#------------------------------------------------------------------------------
# File:         DICOM.pm
#
# Description:  Read DICOM and ACR-NEMA medical images
#
# Revisions:    2005/11/09 - P. Harvey Created
#               2009/11/19 - P. Harvey Added private GE tags from ref 4
#               2009/12/11 - P. Harvey Updated to DICOM 2008 spec
#               2010/04/08 - P. Harvey Updated to DICOM 2009 spec
#
# References:   1) http://medical.nema.org/
#               2) http://www.sph.sc.edu/comd/rorden/dicom.html
#               3) http://www.dclunie.com/
#               4) http://www.gehealthcare.com/usen/interoperability/dicom/docs/2258357r3.pdf
#------------------------------------------------------------------------------

package Image::ExifTool::DICOM;

use strict;
use vars qw($VERSION %uid);
use Image::ExifTool qw(:DataAccess :Utils);

$VERSION = '1.25';

# DICOM VR (Value Representation) format conversions
my %dicomFormat = (
    FD => 'double',
    FL => 'float',
    OB => 'int8u',
    OF => 'float',
    OW => 'int16u',
    SL => 'int32s',
    SS => 'int16s',
    UL => 'int32u',
    US => 'int16u',
);

# VR elements with 32-bit length in explicit VR syntax
my %vr32 = ( OB=>1, OW=>1, OF=>1, SQ=>1, UT=>1, UN=>1 );

# data elements with implicit VR regardless of syntax
my %implicitVR = (
    'FFFE,E000' => 1,
    'FFFE,E00D' => 1,
    'FFFE,E0DD' => 1,
);

# DICOM tags
# Note: "XxxGroupLength" tags are generated automatically if they don't exist
%Image::ExifTool::DICOM::Main = (
    GROUPS => { 2 => 'Image' },
    VARS => { NO_LOOKUP => 1 }, # omit tags from lookup (way too many!)
    NOTES => q{
        The DICOM (Digital Imaging and Communications in Medicine) format is based
        on the ACR-NEMA specification, but adds a file header and a number of new
        tags.  ExifTool will extract information from either type of file.  The Tag
        ID's in the following table are the tag group and element numbers in
        hexadecimal, as given in the DICOM specification (see
        L<https://www.dicomstandard.org/current>).  The table below contains tags
        from the DICOM 2026b and earlier specifications plus some vendor-specific
        private tags.

        Note that DICOM information may be saved in other file formats using the
        L<XMP DICOM Tags|Image::ExifTool::TagNames/XMP DICOM Tags>.
    },
    # file meta information group (names end with VR)
    '0002,0000' => { VR => 'UL', Name => 'FileMetaInfoGroupLength' },
    '0002,0001' => { VR => 'OB', Name => 'FileMetaInfoVersion' },
    '0002,0002' => { VR => 'UI', Name => 'MediaStorageSOPClassUID' },
    '0002,0003' => { VR => 'UI', Name => 'MediaStorageSOPInstanceUID' },
    '0002,0010' => { VR => 'UI', Name => 'TransferSyntaxUID' },
    '0002,0012' => { VR => 'UI', Name => 'ImplementationClassUID' },
    '0002,0013' => { VR => 'SH', Name => 'ImplementationVersionName' },
    '0002,0016' => { VR => 'AE', Name => 'SourceApplicationEntityTitle' },
    '0002,0017' => { VR => 'AE', Name => 'SendingApplicationEntityTitle' },
    '0002,0018' => { VR => 'AE', Name => 'ReceivingApplicationEntityTitle' },
    '0002,0026' => { VR => 'UR', Name => 'SourcePresentationAddress' },
    '0002,0027' => { VR => 'UR', Name => 'SendingPresentationAddress' },
    '0002,0028' => { VR => 'UR', Name => 'ReceivingPresentationAddress' },
    '0002,0031' => { VR => 'OB', Name => 'RTVMetaInformationVersion' },
    '0002,0032' => { VR => 'UI', Name => 'RTVCommunicationSOPClassUID' },
    '0002,0033' => { VR => 'UI', Name => 'RTVCommunicationSOPInstanceUID' },
    '0002,0035' => { VR => 'OB', Name => 'RTVSourceIdentifier' },
    '0002,0036' => { VR => 'OB', Name => 'RTVFlowIdentifier' },
    '0002,0037' => { VR => 'UL', Name => 'RTVFlowRTPSamplingRate' },
    '0002,0038' => { VR => 'FD', Name => 'RTVFlowActualFrameDuration' },
    '0002,0100' => { VR => 'UI', Name => 'PrivateInformationCreatorUID' },
    '0002,0102' => { VR => 'OB', Name => 'PrivateInformation' },
    # directory structuring group
    '0004,1130' => { VR => 'CS', Name => 'FileSetID' },
    '0004,1141' => { VR => 'CS', Name => 'FileSetDescriptorFileID' },
    '0004,1142' => { VR => 'CS', Name => 'SpecificCharacterSetOfFile' },
    '0004,1200' => { VR => 'UL', Name => 'FirstDirectoryRecordOffset' },
    '0004,1202' => { VR => 'UL', Name => 'LastDirectoryRecordOffset' },
    '0004,1212' => { VR => 'US', Name => 'FileSetConsistencyFlag' },
    '0004,1220' => { VR => 'SQ', Name => 'DirectoryRecordSequence' },
    '0004,1400' => { VR => 'UL', Name => 'OffsetOfNextDirectoryRecord' },
    '0004,1410' => { VR => 'US', Name => 'RecordInUseFlag' },
    '0004,1420' => { VR => 'UL', Name => 'LowerLevelDirectoryEntityOffset' },
    '0004,1430' => { VR => 'CS', Name => 'DirectoryRecordType' },
    '0004,1432' => { VR => 'UI', Name => 'PrivateRecordUID' },
    '0004,1500' => { VR => 'CS', Name => 'ReferencedFileID' },
    '0004,1504' => { VR => 'UL', Name => 'MRDRDirectoryRecordOffset' },
    '0004,1510' => { VR => 'UI', Name => 'ReferencedSOPClassUIDInFile' },
    '0004,1511' => { VR => 'UI', Name => 'ReferencedSOPInstanceUIDInFile' },
    '0004,1512' => { VR => 'UI', Name => 'ReferencedTransferSyntaxUIDInFile' },
    '0004,151A' => { VR => 'UI', Name => 'ReferencedRelatedSOPClassUIDInFile' },
    '0004,1600' => { VR => 'UL', Name => 'NumberOfReferences' },
    '0006,0001' => { VR => 'SQ', Name => 'CurrentFrameFunctionalGroupsSeq' },
   # identifying group
    '0008,0000' => { VR => 'UL', Name => 'IdentifyingGroupLength' },
    '0008,0001' => { VR => 'UL', Name => 'LengthToEnd' },
    '0008,0005' => { VR => 'CS', Name => 'SpecificCharacterSet' },
    '0008,0006' => { VR => 'SQ', Name => 'LanguageCodeSequence' },
    '0008,0008' => { VR => 'CS', Name => 'ImageType' },
    '0008,0010' => { VR => 'SH', Name => 'RecognitionCode' },
    '0008,0012' => { VR => 'DA', Name => 'InstanceCreationDate' },
    '0008,0013' => { VR => 'TM', Name => 'InstanceCreationTime' },
    '0008,0014' => { VR => 'UI', Name => 'InstanceCreatorUID' },
    '0008,0015' => { VR => 'DT', Name => 'InstanceCoercionDateTime' },
    '0008,0016' => { VR => 'UI', Name => 'SOPClassUID' },
    '0008,0017' => { VR => 'UI', Name => 'AcquisitionUID' },
    '0008,0018' => { VR => 'UI', Name => 'SOPInstanceUID' },
    '0008,0019' => { VR => 'UI', Name => 'PyramidUID' },
    '0008,001A' => { VR => 'UI', Name => 'RelatedGeneralSOPClassUID' },
    '0008,001B' => { VR => 'UI', Name => 'OriginalSpecializedSOPClassUID' },
    '0008,001C' => { VR => 'CS', Name => 'SyntheticData' },
    '0008,001D' => { VR => 'SQ', Name => 'SensitiveContentCodeSequence' },
    '0008,0020' => { VR => 'DA', Name => 'StudyDate' },
    '0008,0021' => { VR => 'DA', Name => 'SeriesDate' },
    '0008,0022' => { VR => 'DA', Name => 'AcquisitionDate' },
    '0008,0023' => { VR => 'DA', Name => 'ContentDate' },
    '0008,0024' => { VR => 'DA', Name => 'OverlayDate' },
    '0008,0025' => { VR => 'DA', Name => 'CurveDate' },
    '0008,002A' => { VR => 'DT', Name => 'AcquisitionDateTime' },
    '0008,0030' => { VR => 'TM', Name => 'StudyTime' },
    '0008,0031' => { VR => 'TM', Name => 'SeriesTime' },
    '0008,0032' => { VR => 'TM', Name => 'AcquisitionTime' },
    '0008,0033' => { VR => 'TM', Name => 'ContentTime' },
    '0008,0034' => { VR => 'TM', Name => 'OverlayTime' },
    '0008,0035' => { VR => 'TM', Name => 'CurveTime' },
    '0008,0040' => { VR => 'US', Name => 'DataSetType' },
    '0008,0041' => { VR => 'LO', Name => 'DataSetSubtype' },
    '0008,0042' => { VR => 'CS', Name => 'NuclearMedicineSeriesType' },
    '0008,0050' => { VR => 'SH', Name => 'AccessionNumber' },
    '0008,0051' => { VR => 'SQ', Name => 'IssuerOfAccessionNumberSequence' },
    '0008,0052' => { VR => 'CS', Name => 'QueryRetrieveLevel' },
    '0008,0053' => { VR => 'CS', Name => 'QueryRetrieveView' },
    '0008,0054' => { VR => 'AE', Name => 'RetrieveAETitle' },
    '0008,0055' => { VR => 'AE', Name => 'StationAETitle' },
    '0008,0056' => { VR => 'CS', Name => 'InstanceAvailability' },
    '0008,0058' => { VR => 'UI', Name => 'FailedSOPInstanceUIDList' },
    '0008,0060' => { VR => 'CS', Name => 'Modality' },
    '0008,0061' => { VR => 'CS', Name => 'ModalitiesInStudy' },
    '0008,0062' => { VR => 'UI', Name => 'SOPClassesInStudy' },
    '0008,0063' => { VR => 'SQ', Name => 'AnatomicRegionsInStudyCodeSequence' },
    '0008,0064' => { VR => 'CS', Name => 'ConversionType' },
    '0008,0068' => { VR => 'CS', Name => 'PresentationIntentType' },
    '0008,0070' => { VR => 'LO', Name => 'Manufacturer' },
    '0008,0080' => { VR => 'LO', Name => 'InstitutionName' },
    '0008,0081' => { VR => 'ST', Name => 'InstitutionAddress' },
    '0008,0082' => { VR => 'SQ', Name => 'InstitutionCodeSequence' },
    '0008,0090' => { VR => 'PN', Name => 'ReferringPhysicianName' },
    '0008,0092' => { VR => 'ST', Name => 'ReferringPhysicianAddress' },
    '0008,0094' => { VR => 'SH', Name => 'ReferringPhysicianTelephoneNumber' },
    '0008,0096' => { VR => 'SQ', Name => 'ReferringPhysicianIDSequence' },
    '0008,009C' => { VR => 'PN', Name => 'ConsultingPhysicianName' },
    '0008,009D' => { VR => 'SQ', Name => 'ConsultingPhysicianIDSequence' },
    '0008,0100' => { VR => 'SH', Name => 'CodeValue' },
    '0008,0101' => { VR => 'LO', Name => 'ExtendedCodeValue' },
    '0008,0102' => { VR => 'SH', Name => 'CodingSchemeDesignator' },
    '0008,0103' => { VR => 'SH', Name => 'CodingSchemeVersion' },
    '0008,0104' => { VR => 'LO', Name => 'CodeMeaning' },
    '0008,0105' => { VR => 'CS', Name => 'MappingResource' },
    '0008,0106' => { VR => 'DT', Name => 'ContextGroupVersion' },
    '0008,0107' => { VR => 'DT', Name => 'ContextGroupLocalVersion' },
    '0008,0108' => { VR => 'LT', Name => 'ExtendedCodeMeaning' },
    '0008,0109' => { VR => 'SQ', Name => 'CodingSchemeResourcesSequence' },
    '0008,010A' => { VR => 'CS', Name => 'CodingSchemeURLType' },
    '0008,010B' => { VR => 'CS', Name => 'ContextGroupExtensionFlag' },
    '0008,010C' => { VR => 'UI', Name => 'CodingSchemeUID' },
    '0008,010D' => { VR => 'UI', Name => 'ContextGroupExtensionCreatorUID' },
    '0008,010E' => { VR => 'UR', Name => 'CodingSchemeURL' },
    '0008,010F' => { VR => 'CS', Name => 'ContextIdentifier' },
    '0008,0110' => { VR => 'SQ', Name => 'CodingSchemeIDSequence' },
    '0008,0112' => { VR => 'LO', Name => 'CodingSchemeRegistry' },
    '0008,0114' => { VR => 'ST', Name => 'CodingSchemeExternalID' },
    '0008,0115' => { VR => 'ST', Name => 'CodingSchemeName' },
    '0008,0116' => { VR => 'ST', Name => 'ResponsibleOrganization' },
    '0008,0117' => { VR => 'UI', Name => 'ContextUID' },
    '0008,0118' => { VR => 'UI', Name => 'MappingResourceUID' },
    '0008,0119' => { VR => 'UC', Name => 'LongCodeValue' },
    '0008,0120' => { VR => 'UR', Name => 'URNCodeValue' },
    '0008,0121' => { VR => 'SQ', Name => 'EquivalentCodeSequence' },
    '0008,0122' => { VR => 'LO', Name => 'MappingResourceName' },
    '0008,0123' => { VR => 'SQ', Name => 'ContextGroupIDSequence' },
    '0008,0124' => { VR => 'SQ', Name => 'MappingResourceIDSequence' },
    '0008,0201' => { VR => 'SH', Name => 'TimezoneOffsetFromUTC' },
    '0008,0220' => { VR => 'SQ', Name => 'ResponsibleGroupCodeSequence' },
    '0008,0221' => { VR => 'CS', Name => 'EquipmentModality' },
    '0008,0222' => { VR => 'LO', Name => 'ManufacturerRelatedModelGroup' },
    '0008,0300' => { VR => 'SQ', Name => 'PrivateDataElementCharacteristics' },
    '0008,0301' => { VR => 'US', Name => 'PrivateGroupReference' },
    '0008,0302' => { VR => 'LO', Name => 'PrivateCreatorReference' },
    '0008,0303' => { VR => 'CS', Name => 'BlockIdentifyingInformationStatus' },
    '0008,0304' => { VR => 'US', Name => 'NonidentifyingPrivateElements' },
    '0008,0305' => { VR => 'SQ', Name => 'DeidentificationActionSequence' },
    '0008,0306' => { VR => 'US', Name => 'IdentifyingPrivateElements' },
    '0008,0307' => { VR => 'CS', Name => 'DeidentificationAction' },
    '0008,0308' => { VR => 'US', Name => 'PrivateDataElement' },
    '0008,0309' => { VR => 'UL', Name => 'PrivateDataElementValueMult' },
    '0008,030A' => { VR => 'CS', Name => 'PrivateDataElementValueRep' },
    '0008,030B' => { VR => 'UL', Name => 'PrivateDataElementNumberOfItems' },
    '0008,030C' => { VR => 'UC', Name => 'PrivateDataElementName' },
    '0008,030D' => { VR => 'UC', Name => 'PrivateDataElementKeyword' },
    '0008,030E' => { VR => 'UT', Name => 'PrivateDataElementDescription' },
    '0008,030F' => { VR => 'UT', Name => 'PrivateDataElementEncoding' },
    '0008,0310' => { VR => 'SQ', Name => 'PrivateDataElementDefinitionSeq' },
    '0008,0400' => { VR => 'SQ', Name => 'ScopeOfInventorySequence' },
    '0008,0401' => { VR => 'LT', Name => 'InventoryPurpose' },
    '0008,0402' => { VR => 'LT', Name => 'InventoryInstanceDescription' },
    '0008,0403' => { VR => 'CS', Name => 'InventoryLevel' },
    '0008,0404' => { VR => 'DT', Name => 'ItemInventoryDateTime' },
    '0008,0405' => { VR => 'CS', Name => 'RemovedFromOperationalUse' },
    '0008,0406' => { VR => 'SQ', Name => 'ReasonForRemovalCodeSequence' },
    '0008,0407' => { VR => 'UR', Name => 'StoredInstanceBaseURI' },
    '0008,0408' => { VR => 'UR', Name => 'FolderAccessURI' },
    '0008,0409' => { VR => 'UR', Name => 'FileAccessURI' },
    '0008,040A' => { VR => 'CS', Name => 'ContainerFileType' },
    '0008,040B' => { VR => 'UR', Name => 'FilenameInContainer' },
    '0008,040C' => { VR => 'UV', Name => 'FileOffsetInContainer' },
    '0008,040D' => { VR => 'UV', Name => 'FileLengthInContainer' },
    '0008,040E' => { VR => 'UI', Name => 'StoredInstanceTransferSyntaxUID' },
    '0008,040F' => { VR => 'CS', Name => 'ExtendedMatchingMechanisms' },
    '0008,0410' => { VR => 'SQ', Name => 'RangeMatchingSequence' },
    '0008,0411' => { VR => 'SQ', Name => 'ListOfUIDMatchingSequence' },
    '0008,0412' => { VR => 'SQ', Name => 'EmptyValueMatchingSequence' },
    '0008,0413' => { VR => 'SQ', Name => 'GeneralMatchingSequence' },
    '0008,0414' => { VR => 'US', Name => 'RequestedStatusInterval' },
    '0008,0415' => { VR => 'CS', Name => 'RetainInstances' },
    '0008,0416' => { VR => 'DT', Name => 'ExpirationDateTime' },
    '0008,0417' => { VR => 'CS', Name => 'TransactionStatus' },
    '0008,0418' => { VR => 'LT', Name => 'TransactionStatusComment' },
    '0008,0419' => { VR => 'SQ', Name => 'FileSetAccessSequence' },
    '0008,041A' => { VR => 'SQ', Name => 'FileAccessSequence' },
    '0008,041B' => { VR => 'OB', Name => 'RecordKey' },
    '0008,041C' => { VR => 'OB', Name => 'PriorRecordKey' },
    '0008,041D' => { VR => 'SQ', Name => 'MetadataSequence' },
    '0008,041E' => { VR => 'SQ', Name => 'UpdatedMetadataSequence' },
    '0008,041F' => { VR => 'DT', Name => 'StudyUpdateDateTime' },
    '0008,0420' => { VR => 'SQ', Name => 'InventoryAccessEndPointsSequence' },
    '0008,0421' => { VR => 'SQ', Name => 'StudyAccessEndPointsSequence' },
    '0008,0422' => { VR => 'SQ', Name => 'IncorporatedInventoryInstanceSeq' },
    '0008,0423' => { VR => 'SQ', Name => 'InventoriedStudiesSequence' },
    '0008,0424' => { VR => 'SQ', Name => 'InventoriedSeriesSequence' },
    '0008,0425' => { VR => 'SQ', Name => 'InventoriedInstancesSequence' },
    '0008,0426' => { VR => 'CS', Name => 'InventoryCompletionStatus' },
    '0008,0427' => { VR => 'UL', Name => 'NumberOfStudyRecordsInInstance' },
    '0008,0428' => { VR => 'UV', Name => 'TotalNumberOfStudyRecords' },
    '0008,0429' => { VR => 'UV', Name => 'MaximumNumberOfRecords' },
    '0008,1000' => { VR => 'AE', Name => 'NetworkID' },
    '0008,1010' => { VR => 'SH', Name => 'StationName' },
    '0008,1030' => { VR => 'LO', Name => 'StudyDescription' },
    '0008,1032' => { VR => 'SQ', Name => 'ProcedureCodeSequence' },
    '0008,103E' => { VR => 'LO', Name => 'SeriesDescription' },
    '0008,103F' => { VR => 'SQ', Name => 'SeriesDescriptionCodeSequence' },
    '0008,1040' => { VR => 'LO', Name => 'InstitutionalDepartmentName' },
    '0008,1041' => { VR => 'SQ', Name => 'InstitutionalDepartmentTypeCodeSeq' },
    '0008,1048' => { VR => 'PN', Name => 'PhysiciansOfRecord' },
    '0008,1049' => { VR => 'SQ', Name => 'PhysiciansOfRecordIDSequence' },
    '0008,1050' => { VR => 'PN', Name => 'PerformingPhysicianName' },
    '0008,1052' => { VR => 'SQ', Name => 'PerformingPhysicianIDSequence' },
    '0008,1060' => { VR => 'PN', Name => 'NameOfPhysicianReadingStudy' },
    '0008,1062' => { VR => 'SQ', Name => 'PhysicianReadingStudyIDSequence' },
    '0008,1070' => { VR => 'PN', Name => 'OperatorsName' },
    '0008,1072' => { VR => 'SQ', Name => 'OperatorIDSequence' },
    '0008,1080' => { VR => 'LO', Name => 'AdmittingDiagnosesDescription' },
    '0008,1084' => { VR => 'SQ', Name => 'AdmittingDiagnosesCodeSequence' },
    '0008,1088' => { VR => 'LO', Name => 'PyramidDescription' },
    '0008,1090' => { VR => 'LO', Name => 'ManufacturersModelName' },
    '0008,1100' => { VR => 'SQ', Name => 'ReferencedResultsSequence' },
    '0008,1110' => { VR => 'SQ', Name => 'ReferencedStudySequence' },
    '0008,1111' => { VR => 'SQ', Name => 'ReferencedProcedureStepSequence' },
    '0008,1112' => { VR => 'SQ', Name => 'ReferencedInstancesBySOPClassSe' },
    '0008,1115' => { VR => 'SQ', Name => 'ReferencedSeriesSequence' },
    '0008,1120' => { VR => 'SQ', Name => 'ReferencedPatientSequence' },
    '0008,1125' => { VR => 'SQ', Name => 'ReferencedVisitSequence' },
    '0008,1130' => { VR => 'SQ', Name => 'ReferencedOverlaySequence' },
    '0008,1134' => { VR => 'SQ', Name => 'ReferencedStereometricInstanceSeq' },
    '0008,113A' => { VR => 'SQ', Name => 'ReferencedWaveformSequence' },
    '0008,1140' => { VR => 'SQ', Name => 'ReferencedImageSequence' },
    '0008,1145' => { VR => 'SQ', Name => 'ReferencedCurveSequence' },
    '0008,114A' => { VR => 'SQ', Name => 'ReferencedInstanceSequence' },
    '0008,114B' => { VR => 'SQ', Name => 'ReferencedRealWorldValueMappingSeq' },
    '0008,114C' => { VR => 'SQ', Name => 'ReferencedSegmentationSequence' },
    '0008,114D' => { VR => 'SQ', Name => 'ReferencedSurfaceSegmentationSeq' },
    '0008,1150' => { VR => 'UI', Name => 'ReferencedSOPClassUID' },
    '0008,1155' => { VR => 'UI', Name => 'ReferencedSOPInstanceUID' },
    '0008,1156' => { VR => 'SQ', Name => 'DefinitionSourceSequence' },
    '0008,115A' => { VR => 'UI', Name => 'SOPClassesSupported' },
    '0008,1160' => { VR => 'IS', Name => 'ReferencedFrameNumber' },
    '0008,1161' => { VR => 'UL', Name => 'SimpleFrameList' },
    '0008,1162' => { VR => 'UL', Name => 'CalculatedFrameList' },
    '0008,1163' => { VR => 'FD', Name => 'TimeRange' },
    '0008,1164' => { VR => 'SQ', Name => 'FrameExtractionSequence' },
    '0008,1167' => { VR => 'UI', Name => 'MultiFrameSourceSOPInstanceUID' },
    '0008,1190' => { VR => 'UR', Name => 'RetrieveURL' },
    '0008,1195' => { VR => 'UI', Name => 'TransactionUID' },
    '0008,1196' => { VR => 'US', Name => 'WarningReason' },
    '0008,1197' => { VR => 'US', Name => 'FailureReason' },
    '0008,1198' => { VR => 'SQ', Name => 'FailedSOPSequence' },
    '0008,1199' => { VR => 'SQ', Name => 'ReferencedSOPSequence' },
    '0008,119A' => { VR => 'SQ', Name => 'OtherFailuresSequence' },
    '0008,119B' => { VR => 'SQ', Name => 'FailedStudySequence' },
    '0008,1200' => { VR => 'SQ', Name => 'OtherReferencedStudiesSequence' },
    '0008,1250' => { VR => 'SQ', Name => 'RelatedSeriesSequence' },
    '0008,1301' => { VR => 'SQ', Name => 'PrincipalDiagnosisCodeSequence' },
    '0008,1302' => { VR => 'SQ', Name => 'PrimaryDiagnosisCodeSequence' },
    '0008,1303' => { VR => 'SQ', Name => 'SecondaryDiagnosesCodeSequence' },
    '0008,1304' => { VR => 'SQ', Name => 'HistologicalDiagnosesCodeSequence' },
    '0008,2110' => { VR => 'CS', Name => 'LossyImageCompression' },
    '0008,2111' => { VR => 'ST', Name => 'DerivationDescription' },
    '0008,2112' => { VR => 'SQ', Name => 'SourceImageSequence' },
    '0008,2120' => { VR => 'SH', Name => 'StageName' },
    '0008,2122' => { VR => 'IS', Name => 'StageNumber' },
    '0008,2124' => { VR => 'IS', Name => 'NumberOfStages' },
    '0008,2127' => { VR => 'SH', Name => 'ViewName' },
    '0008,2128' => { VR => 'IS', Name => 'ViewNumber' },
    '0008,2129' => { VR => 'IS', Name => 'NumberOfEventTimers' },
    '0008,212A' => { VR => 'IS', Name => 'NumberOfViewsInStage' },
    '0008,2130' => { VR => 'DS', Name => 'EventElapsedTimes' },
    '0008,2132' => { VR => 'LO', Name => 'EventTimerNames' },
    '0008,2133' => { VR => 'SQ', Name => 'EventTimerSequence' },
    '0008,2134' => { VR => 'FD', Name => 'EventTimeOffset' },
    '0008,2135' => { VR => 'SQ', Name => 'EventCodeSequence' },
    '0008,2142' => { VR => 'IS', Name => 'StartTrim' },
    '0008,2143' => { VR => 'IS', Name => 'StopTrim' },
    '0008,2144' => { VR => 'IS', Name => 'RecommendedDisplayFrameRate' },
    '0008,2200' => { VR => 'CS', Name => 'TransducerPosition' },
    '0008,2204' => { VR => 'CS', Name => 'TransducerOrientation' },
    '0008,2208' => { VR => 'CS', Name => 'AnatomicStructure' },
    '0008,2218' => { VR => 'SQ', Name => 'AnatomicRegionSequence' },
    '0008,2220' => { VR => 'SQ', Name => 'AnatomicRegionModifierSequence' },
    '0008,2228' => { VR => 'SQ', Name => 'PrimaryAnatomicStructureSequence' },
    '0008,2229' => { VR => 'SQ', Name => 'AnatomicStructureOrRegionSequence' },
    '0008,2230' => { VR => 'SQ', Name => 'AnatomicStructureModifierSequence' },
    '0008,2240' => { VR => 'SQ', Name => 'TransducerPositionSequence' },
    '0008,2242' => { VR => 'SQ', Name => 'TransducerPositionModifierSequence' },
    '0008,2244' => { VR => 'SQ', Name => 'TransducerOrientationSequence' },
    '0008,2246' => { VR => 'SQ', Name => 'TransducerOrientationModifierSeq' },
    '0008,2251' => { VR => 'SQ', Name => 'AnatomicStructOrRegionCodeSeqTrial' },
    '0008,2253' => { VR => 'SQ', Name => 'AnatomicEntrancePortalCodeSeqTrial' },
    '0008,2255' => { VR => 'SQ', Name => 'AnatomicApproachDirCodeSeqTrial' },
    '0008,2256' => { VR => 'ST', Name => 'AnatomicPerspectiveDescrTrial' },
    '0008,2257' => { VR => 'SQ', Name => 'AnatomicPerspectiveCodeSeqTrial' },
    '0008,2258' => { VR => 'ST', Name => 'AnatomicLocationExamInstrDescTrial' },
    '0008,2259' => { VR => 'SQ', Name => 'AnatomicLocExamInstrCodeSeqTrial' },
    '0008,225A' => { VR => 'SQ', Name => 'AnatomicStructRegnModCodeSeqTrial' },
    '0008,225C' => { VR => 'SQ', Name => 'OnAxisBkgAnatomyStructCodeSeqTrial' },
    '0008,3001' => { VR => 'SQ', Name => 'AlternateRepresentationSequence' },
    '0008,3002' => { VR => 'UI', Name => 'AvailableTransferSyntaxUID' },
    '0008,3010' => { VR => 'UI', Name => 'IrradiationEventUID' },
    '0008,3011' => { VR => 'SQ', Name => 'SourceIrradiationEventSequence' },
    '0008,3012' => { VR => 'UI', Name => 'RadiopharmaceuticalAdminEventUID' },
    '0008,4000' => { VR => 'LT', Name => 'IdentifyingComments' },
    '0008,9007' => { VR => 'CS', Name => 'FrameType' },
    '0008,9092' => { VR => 'SQ', Name => 'ReferencedImageEvidenceSequence' },
    '0008,9121' => { VR => 'SQ', Name => 'ReferencedRawDataSequence' },
    '0008,9123' => { VR => 'UI', Name => 'CreatorVersionUID' },
    '0008,9124' => { VR => 'SQ', Name => 'DerivationImageSequence' },
    '0008,9154' => { VR => 'SQ', Name => 'SourceImageEvidenceSequence' },
    '0008,9205' => { VR => 'CS', Name => 'PixelPresentation' },
    '0008,9206' => { VR => 'CS', Name => 'VolumetricProperties' },
    '0008,9207' => { VR => 'CS', Name => 'VolumeBasedCalculationTechnique' },
    '0008,9208' => { VR => 'CS', Name => 'ComplexImageComponent' },
    '0008,9209' => { VR => 'CS', Name => 'AcquisitionContrast' },
    '0008,9215' => { VR => 'SQ', Name => 'DerivationCodeSequence' },
    '0008,9237' => { VR => 'SQ', Name => 'GrayscalePresentationStateSequence' },
    '0008,9410' => { VR => 'SQ', Name => 'ReferencedOtherPlaneSequence' },
    '0008,9458' => { VR => 'SQ', Name => 'FrameDisplaySequence' },
    '0008,9459' => { VR => 'FL', Name => 'RecommendedDisplayFrameRateInFloat' },
    '0008,9460' => { VR => 'CS', Name => 'SkipFrameRangeFlag' },
    # GEMS_IDEN_01 (ref 4)
    '0009,1001' => { VR => 'LO', Name => 'FullFidelity' },
    '0009,1002' => { VR => 'SH', Name => 'SuiteID' },
    '0009,1004' => { VR => 'SH', Name => 'ProductID' },
    '0009,1027' => { VR => 'SL', Name => 'ImageActualDate' },
    '0009,1030' => { VR => 'SH', Name => 'ServiceID' },
    '0009,1031' => { VR => 'SH', Name => 'MobileLocationNumber' },
    '0009,10E3' => { VR => 'UI', Name => 'EquipmentUID' },
    '0009,10E6' => { VR => 'SH', Name => 'GenesisVersionNow' },
    '0009,10E7' => { VR => 'UL', Name => 'ExamRecordChecksum' },
    '0009,10E9' => { VR => 'SL', Name => 'ActualSeriesDataTimeStamp' },
    # patient group
    '0010,0000' => { VR => 'UL', Name => 'PatientGroupLength' },
    '0010,0010' => { VR => 'PN', Name => 'PatientName' },
    '0010,0011' => { VR => 'SQ', Name => 'PersonNamesToUseSequence' },
    '0010,0012' => { VR => 'LT', Name => 'NameToUse' },
    '0010,0013' => { VR => 'UT', Name => 'NameToUseComment' },
    '0010,0014' => { VR => 'SQ', Name => 'ThirdPersonPronounsSequence' },
    '0010,0015' => { VR => 'SQ', Name => 'PronounCodeSequence' },
    '0010,0016' => { VR => 'UT', Name => 'PronounComment' },
    '0010,0020' => { VR => 'LO', Name => 'PatientID' },
    '0010,0021' => { VR => 'LO', Name => 'IssuerOfPatientID' },
    '0010,0022' => { VR => 'CS', Name => 'TypeOfPatientID' },
    '0010,0024' => { VR => 'SQ', Name => 'IssuerOfPatientIDQualifiersSeq' },
    '0010,0026' => { VR => 'SQ', Name => 'SourcePatientGroupIDSeq' },
    '0010,0027' => { VR => 'SQ', Name => 'GroupOfPatientsIdentificationSeq' },
    '0010,0028' => { VR => 'US', Name => 'SubjectRelativePositionInImage' },
    '0010,0030' => { VR => 'DA', Name => 'PatientBirthDate' },
    '0010,0032' => { VR => 'TM', Name => 'PatientBirthTime' },
    '0010,0033' => { VR => 'LO', Name => 'PatientBirthDateInAltCalendar' },
    '0010,0034' => { VR => 'LO', Name => 'PatientDeathDateInAltCalendar' },
    '0010,0035' => { VR => 'CS', Name => 'PatientAlternativeCalendar' },
    '0010,0040' => { VR => 'CS', Name => 'PatientSex' },
    '0010,0041' => { VR => 'SQ', Name => 'GenderIdentitySequence' },
    '0010,0042' => { VR => 'UT', Name => 'SexParamsForClinicalUseCategoryCom' },
    '0010,0043' => { VR => 'SQ', Name => 'SexParamsForClinicalUseCategorySeq' },
    '0010,0044' => { VR => 'SQ', Name => 'GenderIdentityCodeSequence' },
    '0010,0045' => { VR => 'UT', Name => 'GenderIdentityComment' },
    '0010,0046' => { VR => 'SQ', Name => 'SexParamsForClinicalUseCatCodeSeq' },
    '0010,0047' => { VR => 'UR', Name => 'SexParamsForClinicalUseCategoryRef' },
    '0010,0050' => { VR => 'SQ', Name => 'PatientInsurancePlanCodeSequence' },
    '0010,0101' => { VR => 'SQ', Name => 'PatientPrimaryLanguageCodeSeq' },
    '0010,0102' => { VR => 'SQ', Name => 'PatientPrimaryLanguageCodeModSeq' },
    '0010,0200' => { VR => 'CS', Name => 'QualityControlSubject' },
    '0010,0201' => { VR => 'SQ', Name => 'QualityControlSubjectTypeCodeSeq' },
    '0010,0212' => { VR => 'UC', Name => 'StrainDescription' },
    '0010,0213' => { VR => 'LO', Name => 'StrainNomenclature' },
    '0010,0214' => { VR => 'LO', Name => 'StrainStockNumber' },
    '0010,0215' => { VR => 'SQ', Name => 'StrainSourceRegistryCodeSequence' },
    '0010,0216' => { VR => 'SQ', Name => 'StrainStockSequence' },
    '0010,0217' => { VR => 'LO', Name => 'StrainSource' },
    '0010,0218' => { VR => 'UT', Name => 'StrainAdditionalInformation' },
    '0010,0219' => { VR => 'SQ', Name => 'StrainCodeSequence' },
    '0010,0221' => { VR => 'SQ', Name => 'GeneticModificationsSequence' },
    '0010,0222' => { VR => 'UC', Name => 'GeneticModificationsDescription' },
    '0010,0223' => { VR => 'LO', Name => 'GeneticModificationsNomenclature' },
    '0010,0229' => { VR => 'SQ', Name => 'GeneticModificationsCodeSequence' },
    '0010,1000' => { VR => 'LO', Name => 'OtherPatientIDs' },
    '0010,1001' => { VR => 'PN', Name => 'OtherPatientNames' },
    '0010,1002' => { VR => 'SQ', Name => 'OtherPatientIDsSequence' },
    '0010,1005' => { VR => 'PN', Name => 'PatientBirthName' },
    '0010,1010' => { VR => 'AS', Name => 'PatientAge' },
    '0010,1020' => { VR => 'DS', Name => 'PatientSize' },
    '0010,1021' => { VR => 'SQ', Name => 'PatientSizeCodeSequence' },
    '0010,1022' => { VR => 'DS', Name => 'PatientBodyMassIndex' },
    '0010,1023' => { VR => 'DS', Name => 'MeasuredAPDimension' },
    '0010,1024' => { VR => 'DS', Name => 'MeasuredLateralDimension' },
    '0010,1030' => { VR => 'DS', Name => 'PatientWeight' },
    '0010,1040' => { VR => 'LO', Name => 'PatientAddress' },
    '0010,1050' => { VR => 'LO', Name => 'InsurancePlanIdentification' },
    '0010,1060' => { VR => 'PN', Name => 'PatientMotherBirthName' },
    '0010,1080' => { VR => 'LO', Name => 'MilitaryRank' },
    '0010,1081' => { VR => 'LO', Name => 'BranchOfService' },
    '0010,1090' => { VR => 'LO', Name => 'MedicalRecordLocator' },
    '0010,1100' => { VR => 'SQ', Name => 'ReferencedPatientPhotoSequence' },
    '0010,2000' => { VR => 'LO', Name => 'MedicalAlerts' },
    '0010,2110' => { VR => 'LO', Name => 'Allergies' },
    '0010,2150' => { VR => 'LO', Name => 'CountryOfResidence' },
    '0010,2152' => { VR => 'LO', Name => 'RegionOfResidence' },
    '0010,2154' => { VR => 'SH', Name => 'PatientTelephoneNumbers' },
    '0010,2155' => { VR => 'LT', Name => 'PatientTelecomInformation' },
    '0010,2160' => { VR => 'SH', Name => 'EthnicGroup' },
    '0010,2161' => { VR => 'SQ', Name => 'EthnicGroupCodeSequence' },
    '0010,2162' => { VR => 'UC', Name => 'EthnicGroups' },
    '0010,2180' => { VR => 'SH', Name => 'Occupation' },
    '0010,21A0' => { VR => 'CS', Name => 'SmokingStatus' },
    '0010,21B0' => { VR => 'LT', Name => 'AdditionalPatientHistory' },
    '0010,21C0' => { VR => 'US', Name => 'PregnancyStatus' },
    '0010,21D0' => { VR => 'DA', Name => 'LastMenstrualDate' },
    '0010,21F0' => { VR => 'LO', Name => 'PatientReligiousPreference' },
    '0010,2201' => { VR => 'LO', Name => 'PatientSpeciesDescription' },
    '0010,2202' => { VR => 'SQ', Name => 'PatientSpeciesCodeSequence' },
    '0010,2203' => { VR => 'CS', Name => 'PatientSexNeutered' },
    '0010,2210' => { VR => 'CS', Name => 'AnatomicalOrientationType' },
    '0010,2292' => { VR => 'LO', Name => 'PatientBreedDescription' },
    '0010,2293' => { VR => 'SQ', Name => 'PatientBreedCodeSequence' },
    '0010,2294' => { VR => 'SQ', Name => 'BreedRegistrationSequence' },
    '0010,2295' => { VR => 'LO', Name => 'BreedRegistrationNumber' },
    '0010,2296' => { VR => 'SQ', Name => 'BreedRegistryCodeSequence' },
    '0010,2297' => { VR => 'PN', Name => 'ResponsiblePerson' },
    '0010,2298' => { VR => 'CS', Name => 'ResponsiblePersonRole' },
    '0010,2299' => { VR => 'LO', Name => 'ResponsibleOrganization' },
    '0010,4000' => { VR => 'LT', Name => 'PatientComments' },
    '0010,9431' => { VR => 'FL', Name => 'ExaminedBodyThickness' },
    # GEMS_PATI_01 (ref 4)
    '0011,1010' => { VR => 'SS', Name => 'PatientStatus' },
    # clinical trial group
    '0012,0010' => { VR => 'LO', Name => 'ClinicalTrialSponsorName' },
    '0012,0020' => { VR => 'LO', Name => 'ClinicalTrialProtocolID' },
    '0012,0021' => { VR => 'LO', Name => 'ClinicalTrialProtocolName' },
    '0012,0022' => { VR => 'LO', Name => 'IssuerOfClinicalTrialProtocolID' },
    '0012,0023' => { VR => 'SQ', Name => 'OtherClinicalTrialProtocolIDsSeq' },
    '0012,0030' => { VR => 'LO', Name => 'ClinicalTrialSiteID' },
    '0012,0031' => { VR => 'LO', Name => 'ClinicalTrialSiteName' },
    '0012,0032' => { VR => 'LO', Name => 'IssuerOfClinicalTrialSiteID' },
    '0012,0040' => { VR => 'LO', Name => 'ClinicalTrialSubjectID' },
    '0012,0041' => { VR => 'LO', Name => 'IssuerOfClinicalTrialSubjectID' },
    '0012,0042' => { VR => 'LO', Name => 'ClinicalTrialSubjectReadingID' },
    '0012,0043' => { VR => 'LO', Name => 'ClinicalTrialIssuerSubjReadingID' },
    '0012,0050' => { VR => 'LO', Name => 'ClinicalTrialTimePointID' },
    '0012,0051' => { VR => 'ST', Name => 'ClinicalTrialTimePointDescription' },
    '0012,0052' => { VR => 'FD', Name => 'LongitudTemporalOffsetFromEvent' },
    '0012,0053' => { VR => 'CS', Name => 'LongitudinalTemporalEventType' },
    '0012,0054' => { VR => 'SQ', Name => 'ClinicalTrialTimePointTypeCodeSeq' },
    '0012,0055' => { VR => 'LO', Name => 'IssuerOfClinicalTrialTimePointID' },
    '0012,0060' => { VR => 'LO', Name => 'ClinicalTrialCoordinatingCenter' },
    '0012,0062' => { VR => 'CS', Name => 'PatientIdentityRemoved' },
    '0012,0063' => { VR => 'LO', Name => 'DeidentificationMethod' },
    '0012,0064' => { VR => 'SQ', Name => 'DeidentificationMethodCodeSequence' },
    '0012,0071' => { VR => 'LO', Name => 'ClinicalTrialSeriesID' },
    '0012,0072' => { VR => 'LO', Name => 'ClinicalTrialSeriesDescription' },
    '0012,0073' => { VR => 'LO', Name => 'IssuerOfClinicalTrialSeriesID' },
    '0012,0081' => { VR => 'LO', Name => 'ClinicalTrialProtoEthicsCommittee' },
    '0012,0082' => { VR => 'LO', Name => 'ClinicalTrialEthicsCommApprovalNum' },
    '0012,0083' => { VR => 'SQ', Name => 'ConsentForClinicalTrialUseSequence' },
    '0012,0084' => { VR => 'CS', Name => 'DistributionType' },
    '0012,0085' => { VR => 'CS', Name => 'ConsentForDistributionFlag' },
    '0012,0086' => { VR => 'DA', Name => 'EthicsCommApprovalEffectStartDate' },
    '0012,0087' => { VR => 'DA', Name => 'EthicsCommApprovalEffectEndDate' },
    '0014,0023' => { VR => 'ST', Name => 'CADFileFormat' },
    '0014,0024' => { VR => 'ST', Name => 'ComponentReferenceSystem' },
    '0014,0025' => { VR => 'ST', Name => 'ComponentManufacturingProcedure' },
    '0014,0028' => { VR => 'ST', Name => 'ComponentManufacturer' },
    '0014,0030' => { VR => 'DS', Name => 'MaterialThickness' },
    '0014,0032' => { VR => 'DS', Name => 'MaterialPipeDiameter' },
    '0014,0034' => { VR => 'DS', Name => 'MaterialIsolationDiameter' },
    '0014,0042' => { VR => 'ST', Name => 'MaterialGrade' },
    '0014,0044' => { VR => 'ST', Name => 'MaterialPropertiesDescription' },
    '0014,0045' => { VR => 'ST', Name => 'MaterialPropertiesFileFormat' },
    '0014,0046' => { VR => 'LT', Name => 'MaterialNotes' },
    '0014,0050' => { VR => 'CS', Name => 'ComponentShape' },
    '0014,0052' => { VR => 'CS', Name => 'CurvatureType' },
    '0014,0054' => { VR => 'DS', Name => 'OuterDiameter' },
    '0014,0056' => { VR => 'DS', Name => 'InnerDiameter' },
    '0014,0100' => { VR => 'LO', Name => 'ComponentWelderIDs' },
    '0014,0101' => { VR => 'CS', Name => 'SecondaryApprovalStatus' },
    '0014,0102' => { VR => 'DA', Name => 'SecondaryReviewDate' },
    '0014,0103' => { VR => 'TM', Name => 'SecondaryReviewTime' },
    '0014,0104' => { VR => 'PN', Name => 'SecondaryReviewerName' },
    '0014,0105' => { VR => 'ST', Name => 'RepairID' },
    '0014,0106' => { VR => 'SQ', Name => 'MultipleComponentApprovalSequence' },
    '0014,0107' => { VR => 'CS', Name => 'OtherApprovalStatus' },
    '0014,0108' => { VR => 'CS', Name => 'OtherSecondaryApprovalStatus' },
    '0014,0200' => { VR => 'SQ', Name => 'DataElementLabelSequence' },
    '0014,0201' => { VR => 'SQ', Name => 'DataElementLabelItemSequence' },
    '0014,0202' => { VR => 'AT', Name => 'DataElement' },
    '0014,0203' => { VR => 'LO', Name => 'DataElementName' },
    '0014,0204' => { VR => 'LO', Name => 'DataElementDescription' },
    '0014,0205' => { VR => 'CS', Name => 'DataElementConditionality' },
    '0014,0206' => { VR => 'IS', Name => 'DataElementMinimumCharacters' },
    '0014,0207' => { VR => 'IS', Name => 'DataElementMaximumCharacters' },
    '0014,1010' => { VR => 'ST', Name => 'ActualEnvironmentalConditions' },
    '0014,1020' => { VR => 'DA', Name => 'ExpiryDate' },
    '0014,1040' => { VR => 'ST', Name => 'EnvironmentalConditions' },
    '0014,2002' => { VR => 'SQ', Name => 'EvaluatorSequence' },
    '0014,2004' => { VR => 'IS', Name => 'EvaluatorNumber' },
    '0014,2006' => { VR => 'PN', Name => 'EvaluatorName' },
    '0014,2008' => { VR => 'IS', Name => 'EvaluationAttempt' },
    '0014,2012' => { VR => 'SQ', Name => 'IndicationSequence' },
    '0014,2014' => { VR => 'IS', Name => 'IndicationNumber' },
    '0014,2016' => { VR => 'SH', Name => 'IndicationLabel' },
    '0014,2018' => { VR => 'ST', Name => 'IndicationDescription' },
    '0014,201A' => { VR => 'CS', Name => 'IndicationType' },
    '0014,201C' => { VR => 'CS', Name => 'IndicationDisposition' },
    '0014,201E' => { VR => 'SQ', Name => 'IndicationROISequence' },
    '0014,2030' => { VR => 'SQ', Name => 'IndicationPhysicalPropertySequence' },
    '0014,2032' => { VR => 'SH', Name => 'PropertyLabel' },
    '0014,2202' => { VR => 'IS', Name => 'CoordinateSystemNumberOfAxes' },
    '0014,2204' => { VR => 'SQ', Name => 'CoordinateSystemAxesSequence' },
    '0014,2206' => { VR => 'ST', Name => 'CoordinateSystemAxisDescription' },
    '0014,2208' => { VR => 'CS', Name => 'CoordinateSystemDataSetMapping' },
    '0014,220A' => { VR => 'IS', Name => 'CoordinateSystemAxisNumber' },
    '0014,220C' => { VR => 'CS', Name => 'CoordinateSystemAxisType' },
    '0014,220E' => { VR => 'CS', Name => 'CoordinateSystemAxisUnits' },
    '0014,2210' => { VR => 'OB', Name => 'CoordinateSystemAxisValues' },
    '0014,2220' => { VR => 'SQ', Name => 'CoordinateSystemTransformSequence' },
    '0014,2222' => { VR => 'ST', Name => 'TransformDescription' },
    '0014,2224' => { VR => 'IS', Name => 'TransformNumberOfAxes' },
    '0014,2226' => { VR => 'IS', Name => 'TransformOrderOfAxes' },
    '0014,2228' => { VR => 'CS', Name => 'TransformedAxisUnits' },
    '0014,222A' => { VR => 'DS', Name => 'CoordinateTransformRotationMatrix' },
    '0014,222C' => { VR => 'DS', Name => 'CoordinateTransformTranslateMatrix' },
    '0014,3011' => { VR => 'DS', Name => 'InternalDetectorFrameTime' },
    '0014,3012' => { VR => 'DS', Name => 'NumberOfFramesIntegrated' },
    '0014,3020' => { VR => 'SQ', Name => 'DetectorTemperatureSequence' },
    '0014,3022' => { VR => 'ST', Name => 'SensorName' },
    '0014,3024' => { VR => 'DS', Name => 'HorizontalOffsetOfSensor' },
    '0014,3026' => { VR => 'DS', Name => 'VerticalOffsetOfSensor' },
    '0014,3028' => { VR => 'DS', Name => 'SensorTemperature' },
    '0014,3040' => { VR => 'SQ', Name => 'DarkCurrentSequence' },
    '0014,3050' => { VR => 'OB', Name => 'DarkCurrentCounts' },
    '0014,3060' => { VR => 'SQ', Name => 'GainCorrectionReferenceSequence' },
    '0014,3070' => { VR => 'OB', Name => 'AirCounts' },
    '0014,3071' => { VR => 'DS', Name => 'KVUsedInGainCalibration' },
    '0014,3072' => { VR => 'DS', Name => 'MAUsedInGainCalibration' },
    '0014,3073' => { VR => 'DS', Name => 'NumberOfFramesUsedForIntegration' },
    '0014,3074' => { VR => 'LO', Name => 'FilterMaterialUsedInGainCal' },
    '0014,3075' => { VR => 'DS', Name => 'FilterThicknessUsedInGainCal' },
    '0014,3076' => { VR => 'DA', Name => 'DateOfGainCalibration' },
    '0014,3077' => { VR => 'TM', Name => 'TimeOfGainCalibration' },
    '0014,3080' => { VR => 'OB', Name => 'BadPixelImage' },
    '0014,3099' => { VR => 'LT', Name => 'CalibrationNotes' },
    '0014,3100' => { VR => 'LT', Name => 'LinearityCorrectionTechnique' },
    '0014,3101' => { VR => 'LT', Name => 'BeamHardeningCorrectionTechnique' },
    '0014,4002' => { VR => 'SQ', Name => 'PulserEquipmentSequence' },
    '0014,4004' => { VR => 'CS', Name => 'PulserType' },
    '0014,4006' => { VR => 'LT', Name => 'PulserNotes' },
    '0014,4008' => { VR => 'SQ', Name => 'ReceiverEquipmentSequence' },
    '0014,400A' => { VR => 'CS', Name => 'AmplifierType' },
    '0014,400C' => { VR => 'LT', Name => 'ReceiverNotes' },
    '0014,400E' => { VR => 'SQ', Name => 'PreAmplifierEquipmentSequence' },
    '0014,400F' => { VR => 'LT', Name => 'PreAmplifierNotes' },
    '0014,4010' => { VR => 'SQ', Name => 'TransmitTransducerSequence' },
    '0014,4011' => { VR => 'SQ', Name => 'ReceiveTransducerSequence' },
    '0014,4012' => { VR => 'US', Name => 'NumberOfElements' },
    '0014,4013' => { VR => 'CS', Name => 'ElementShape' },
    '0014,4014' => { VR => 'DS', Name => 'ElementDimensionA' },
    '0014,4015' => { VR => 'DS', Name => 'ElementDimensionB' },
    '0014,4016' => { VR => 'DS', Name => 'ElementPitchA' },
    '0014,4017' => { VR => 'DS', Name => 'MeasuredBeamDimensionA' },
    '0014,4018' => { VR => 'DS', Name => 'MeasuredBeamDimensionB' },
    '0014,4019' => { VR => 'DS', Name => 'LocationOfMeasuredBeamDiameter' },
    '0014,401A' => { VR => 'DS', Name => 'NominalFrequency' },
    '0014,401B' => { VR => 'DS', Name => 'MeasuredCenterFrequency' },
    '0014,401C' => { VR => 'DS', Name => 'MeasuredBandwidth' },
    '0014,401D' => { VR => 'DS', Name => 'ElementPitchB' },
    '0014,4020' => { VR => 'SQ', Name => 'PulserSettingsSequence' },
    '0014,4022' => { VR => 'DS', Name => 'PulseWidth' },
    '0014,4024' => { VR => 'DS', Name => 'ExcitationFrequency' },
    '0014,4026' => { VR => 'CS', Name => 'ModulationType' },
    '0014,4028' => { VR => 'DS', Name => 'Damping' },
    '0014,4030' => { VR => 'SQ', Name => 'ReceiverSettingsSequence' },
    '0014,4031' => { VR => 'DS', Name => 'AcquiredSoundpathLength' },
    '0014,4032' => { VR => 'CS', Name => 'AcquisitionCompressionType' },
    '0014,4033' => { VR => 'IS', Name => 'AcquisitionSampleSize' },
    '0014,4034' => { VR => 'DS', Name => 'RectifierSmoothing' },
    '0014,4035' => { VR => 'SQ', Name => 'DACSequence' },
    '0014,4036' => { VR => 'CS', Name => 'DACType' },
    '0014,4038' => { VR => 'DS', Name => 'DACGainPoints' },
    '0014,403A' => { VR => 'DS', Name => 'DACTimePoints' },
    '0014,403C' => { VR => 'DS', Name => 'DACAmplitude' },
    '0014,4040' => { VR => 'SQ', Name => 'PreAmplifierSettingsSequence' },
    '0014,4050' => { VR => 'SQ', Name => 'TransmitTransducerSettingsSequence' },
    '0014,4051' => { VR => 'SQ', Name => 'ReceiveTransducerSettingsSequence' },
    '0014,4052' => { VR => 'DS', Name => 'IncidentAngle' },
    '0014,4054' => { VR => 'ST', Name => 'CouplingTechnique' },
    '0014,4056' => { VR => 'ST', Name => 'CouplingMedium' },
    '0014,4057' => { VR => 'DS', Name => 'CouplingVelocity' },
    '0014,4058' => { VR => 'DS', Name => 'ProbeCenterLocationX' },
    '0014,4059' => { VR => 'DS', Name => 'ProbeCenterLocationZ' },
    '0014,405A' => { VR => 'DS', Name => 'SoundPathLength' },
    '0014,405C' => { VR => 'ST', Name => 'DelayLawIdentifier' },
    '0014,4060' => { VR => 'SQ', Name => 'GateSettingsSequence' },
    '0014,4062' => { VR => 'DS', Name => 'GateThreshold' },
    '0014,4064' => { VR => 'DS', Name => 'VelocityOfSound' },
    '0014,4070' => { VR => 'SQ', Name => 'CalibrationSettingsSequence' },
    '0014,4072' => { VR => 'ST', Name => 'CalibrationProcedure' },
    '0014,4074' => { VR => 'SH', Name => 'ProcedureVersion' },
    '0014,4076' => { VR => 'DA', Name => 'ProcedureCreationDate' },
    '0014,4078' => { VR => 'DA', Name => 'ProcedureExpirationDate' },
    '0014,407A' => { VR => 'DA', Name => 'ProcedureLastModifiedDate' },
    '0014,407C' => { VR => 'TM', Name => 'CalibrationTime' },
    '0014,407E' => { VR => 'DA', Name => 'CalibrationDate' },
    '0014,4080' => { VR => 'SQ', Name => 'ProbeDriveEquipmentSequence' },
    '0014,4081' => { VR => 'CS', Name => 'DriveType' },
    '0014,4082' => { VR => 'LT', Name => 'ProbeDriveNotes' },
    '0014,4083' => { VR => 'SQ', Name => 'DriveProbeSequence' },
    '0014,4084' => { VR => 'DS', Name => 'ProbeInductance' },
    '0014,4085' => { VR => 'DS', Name => 'ProbeResistance' },
    '0014,4086' => { VR => 'SQ', Name => 'ReceiveProbeSequence' },
    '0014,4087' => { VR => 'SQ', Name => 'ProbeDriveSettingsSequence' },
    '0014,4088' => { VR => 'DS', Name => 'BridgeResistors' },
    '0014,4089' => { VR => 'DS', Name => 'ProbeOrientationAngle' },
    '0014,408B' => { VR => 'DS', Name => 'UserSelectedGainY' },
    '0014,408C' => { VR => 'DS', Name => 'UserSelectedPhase' },
    '0014,408D' => { VR => 'DS', Name => 'UserSelectedOffsetX' },
    '0014,408E' => { VR => 'DS', Name => 'UserSelectedOffsetY' },
    '0014,4091' => { VR => 'SQ', Name => 'ChannelSettingsSequence' },
    '0014,4092' => { VR => 'DS', Name => 'ChannelThreshold' },
    '0014,409A' => { VR => 'SQ', Name => 'ScannerSettingsSequence' },
    '0014,409B' => { VR => 'ST', Name => 'ScanProcedure' },
    '0014,409C' => { VR => 'DS', Name => 'TranslationRateX' },
    '0014,409D' => { VR => 'DS', Name => 'TranslationRateY' },
    '0014,409F' => { VR => 'DS', Name => 'ChannelOverlap' },
    '0014,40A0' => { VR => 'LO', Name => 'ImageQualityIndicatorType' },
    '0014,40A1' => { VR => 'LO', Name => 'ImageQualityIndicatorMaterial' },
    '0014,40A2' => { VR => 'LO', Name => 'ImageQualityIndicatorSize' },
    '0014,4101' => { VR => 'SQ', Name => 'WaveDimensionsDefinitionSequence' },
    '0014,4102' => { VR => 'US', Name => 'WaveDimensionNumber' },
    '0014,4103' => { VR => 'LO', Name => 'WaveDimensionDescription' },
    '0014,4104' => { VR => 'US', Name => 'WaveDimensionUnit' },
    '0014,4105' => { VR => 'CS', Name => 'WaveDimensionValueType' },
    '0014,4106' => { VR => 'SQ', Name => 'WaveDimensionValuesSequence' },
    '0014,4107' => { VR => 'US', Name => 'ReferencedWaveDimension' },
    '0014,4108' => { VR => 'SL', Name => 'IntegerNumericValue' },
    '0014,4109' => { VR => 'OB', Name => 'ByteNumericValue' },
    '0014,410A' => { VR => 'OW', Name => 'ShortNumericValue' },
    '0014,410B' => { VR => 'OF', Name => 'SinglePrecisionFloatingPointValue' },
    '0014,410C' => { VR => 'OD', Name => 'DoublePrecisionFloatingPointValue' },
    '0014,5002' => { VR => 'IS', Name => 'LINACEnergy' },
    '0014,5004' => { VR => 'IS', Name => 'LINACOutput' },
    '0014,5100' => { VR => 'US', Name => 'ActiveAperture' },
    '0014,5101' => { VR => 'DS', Name => 'TotalAperture' },
    '0014,5102' => { VR => 'DS', Name => 'ApertureElevation' },
    '0014,5103' => { VR => 'DS', Name => 'MainLobeAngle' },
    '0014,5104' => { VR => 'DS', Name => 'MainRoofAngle' },
    '0014,5105' => { VR => 'CS', Name => 'ConnectorType' },
    '0014,5106' => { VR => 'SH', Name => 'WedgeModelNumber' },
    '0014,5107' => { VR => 'DS', Name => 'WedgeAngleFloat' },
    '0014,5108' => { VR => 'DS', Name => 'WedgeRoofAngle' },
    '0014,5109' => { VR => 'CS', Name => 'WedgeElement1Position' },
    '0014,510A' => { VR => 'DS', Name => 'WedgeMaterialVelocity' },
    '0014,510B' => { VR => 'SH', Name => 'WedgeMaterial' },
    '0014,510C' => { VR => 'DS', Name => 'WedgeOffsetZ' },
    '0014,510D' => { VR => 'DS', Name => 'WedgeOriginOffsetX' },
    '0014,510E' => { VR => 'DS', Name => 'WedgeTimeDelay' },
    '0014,510F' => { VR => 'SH', Name => 'WedgeName' },
    '0014,5110' => { VR => 'SH', Name => 'WedgeManufacturerName' },
    '0014,5111' => { VR => 'LO', Name => 'WedgeDescription' },
    '0014,5112' => { VR => 'DS', Name => 'NominalBeamAngle' },
    '0014,5113' => { VR => 'DS', Name => 'WedgeOffsetX' },
    '0014,5114' => { VR => 'DS', Name => 'WedgeOffsetY' },
    '0014,5115' => { VR => 'DS', Name => 'WedgeTotalLength' },
    '0014,5116' => { VR => 'DS', Name => 'WedgeInContactLength' },
    '0014,5117' => { VR => 'DS', Name => 'WedgeFrontGap' },
    '0014,5118' => { VR => 'DS', Name => 'WedgeTotalHeight' },
    '0014,5119' => { VR => 'DS', Name => 'WedgeFrontHeight' },
    '0014,511A' => { VR => 'DS', Name => 'WedgeRearHeight' },
    '0014,511B' => { VR => 'DS', Name => 'WedgeTotalWidth' },
    '0014,511C' => { VR => 'DS', Name => 'WedgeInContactWidth' },
    '0014,511D' => { VR => 'DS', Name => 'WedgeChamferHeight' },
    '0014,511E' => { VR => 'CS', Name => 'WedgeCurve' },
    '0014,511F' => { VR => 'DS', Name => 'RadiusAlongWedge' },
    '0014,6001' => { VR => 'SQ', Name => 'ThermalCameraSettingsSequence' },
    '0014,6002' => { VR => 'DS', Name => 'AcquisitionFrameRate' },
    '0014,6003' => { VR => 'DS', Name => 'IntegrationTime' },
    '0014,6004' => { VR => 'DS', Name => 'NumberOfCalibrationFrames' },
    '0014,6005' => { VR => 'DS', Name => 'NumberOfRowsInFullAcquisitionImage' },
    '0014,6006' => { VR => 'DS', Name => 'NumberOfColumnsInFullAcqImage' },
    '0014,6007' => { VR => 'SQ', Name => 'ThermalSourceSettingsSequence' },
    '0014,6008' => { VR => 'DS', Name => 'SourceHorizontalPitch' },
    '0014,6009' => { VR => 'DS', Name => 'SourceVerticalPitch' },
    '0014,600A' => { VR => 'DS', Name => 'SourceHorizontalScanSpeed' },
    '0014,600B' => { VR => 'DS', Name => 'ThermalSourceModulationFrequency' },
    '0014,600C' => { VR => 'SQ', Name => 'InductionSourceSettingSequence' },
    '0014,600D' => { VR => 'DS', Name => 'CoilFrequency' },
    '0014,600E' => { VR => 'DS', Name => 'CurrentAmplitudeAcrossCoil' },
    '0014,600F' => { VR => 'SQ', Name => 'FlashSourceSettingSequence' },
    '0014,6010' => { VR => 'DS', Name => 'FlashDuration' },
    '0014,6011' => { VR => 'DS', Name => 'FlashFrameNumber' },
    '0014,6012' => { VR => 'SQ', Name => 'LaserSourceSettingSequence' },
    '0014,6013' => { VR => 'DS', Name => 'HorizontalLaserSpotDimension' },
    '0014,6014' => { VR => 'DS', Name => 'VerticalLaserSpotDimension' },
    '0014,6015' => { VR => 'DS', Name => 'LaserWavelength' },
    '0014,6016' => { VR => 'DS', Name => 'LaserPower' },
    '0014,6017' => { VR => 'SQ', Name => 'ForcedGasSettingSequence' },
    '0014,6018' => { VR => 'SQ', Name => 'VibrationSourceSettingSequence' },
    '0014,6019' => { VR => 'DS', Name => 'VibrationExcitationFrequency' },
    '0014,601A' => { VR => 'DS', Name => 'VibrationExcitationVoltage' },
    '0014,601B' => { VR => 'CS', Name => 'ThermographyDataCaptureMethod' },
    '0014,601C' => { VR => 'CS', Name => 'ThermalTechnique' },
    '0014,601D' => { VR => 'SQ', Name => 'ThermalCameraCoreSequence' },
    '0014,601E' => { VR => 'CS', Name => 'DetectorWavelengthRange' },
    '0014,601F' => { VR => 'CS', Name => 'ThermalCameraCalibrationType' },
    '0014,6020' => { VR => 'UV', Name => 'AcquisitionImageCounter' },
    '0014,6021' => { VR => 'DS', Name => 'FrontPanelTemperature' },
    '0014,6022' => { VR => 'DS', Name => 'AirGapTemperature' },
    '0014,6023' => { VR => 'DS', Name => 'VerticalPixelSize' },
    '0014,6024' => { VR => 'DS', Name => 'HorizontalPixelSize' },
    '0014,6025' => { VR => 'ST', Name => 'DataStreamingProtocol' },
    '0014,6026' => { VR => 'SQ', Name => 'LensSequence' },
    '0014,6027' => { VR => 'DS', Name => 'FieldOfView' },
    '0014,6028' => { VR => 'LO', Name => 'LensFilterManufacturer' },
    '0014,6029' => { VR => 'CS', Name => 'CutoffFilterType' },
    '0014,602A' => { VR => 'DS', Name => 'LensFilterCutOffWavelength' },
    '0014,602B' => { VR => 'SQ', Name => 'ThermalSourceSequence' },
    '0014,602C' => { VR => 'CS', Name => 'ThermalSourceMotionState' },
    '0014,602D' => { VR => 'CS', Name => 'ThermalSourceMotionType' },
    '0014,602E' => { VR => 'SQ', Name => 'InductionHeatingSequence' },
    '0014,602F' => { VR => 'ST', Name => 'CoilConfigurationID' },
    '0014,6030' => { VR => 'DS', Name => 'NumberOfTurnsInCoil' },
    '0014,6031' => { VR => 'CS', Name => 'ShapeOfIndividualTurn' },
    '0014,6032' => { VR => 'DS', Name => 'SizeOfIndividualTurn' },
    '0014,6033' => { VR => 'DS', Name => 'DistanceBetweenTurns' },
    '0014,6034' => { VR => 'SQ', Name => 'FlashHeatingSequence' },
    '0014,6035' => { VR => 'DS', Name => 'NumberOfLamps' },
    '0014,6036' => { VR => 'ST', Name => 'FlashSynchronizationProtocol' },
    '0014,6037' => { VR => 'CS', Name => 'FlashModificationStatus' },
    '0014,6038' => { VR => 'SQ', Name => 'LaserHeatingSequence' },
    '0014,6039' => { VR => 'LO', Name => 'LaserManufacturer' },
    '0014,603A' => { VR => 'LO', Name => 'LaserModelNumber' },
    '0014,603B' => { VR => 'ST', Name => 'LaserTypeDescription' },
    '0014,603C' => { VR => 'SQ', Name => 'ForcedGasHeatingSequence' },
    '0014,603D' => { VR => 'LO', Name => 'GasUsedForHeatingCoolingPart' },
    '0014,603E' => { VR => 'SQ', Name => 'VibrationSonicHeatingSequence' },
    '0014,603F' => { VR => 'LO', Name => 'ProbeManufacturer' },
    '0014,6040' => { VR => 'LO', Name => 'ProbeModelNumber' },
    '0014,6041' => { VR => 'DS', Name => 'ApertureSize' },
    '0014,6042' => { VR => 'DS', Name => 'ProbeResonantFrequency' },
    '0014,6043' => { VR => 'UT', Name => 'HeatSourceDescription' },
    '0014,6044' => { VR => 'CS', Name => 'SurfacePrepWithOpticalCoating' },
    '0014,6045' => { VR => 'ST', Name => 'OpticalCoatingType' },
    '0014,6046' => { VR => 'DS', Name => 'ThermalConductivityExposedSurface' },
    '0014,6047' => { VR => 'DS', Name => 'MaterialDensity' },
    '0014,6048' => { VR => 'DS', Name => 'SpecificHeatOfInspectionSurface' },
    '0014,6049' => { VR => 'DS', Name => 'EmissivityOfInspectionSurface' },
    '0014,604A' => { VR => 'CS', Name => 'ElectromagneticClassInspectSurface' },
    '0014,604C' => { VR => 'DS', Name => 'MovingWindowSize' },
    '0014,604D' => { VR => 'CS', Name => 'MovingWindowType' },
    '0014,604E' => { VR => 'DS', Name => 'MovingWindowWeights' },
    '0014,604F' => { VR => 'DS', Name => 'MovingWindowPitch' },
    '0014,6050' => { VR => 'CS', Name => 'MovingWindowPaddingScheme' },
    '0014,6051' => { VR => 'DS', Name => 'MovingWindowPaddingLength' },
    '0014,6052' => { VR => 'SQ', Name => 'SpatialFilteringParametersSequence' },
    '0014,6053' => { VR => 'CS', Name => 'SpatialFilteringScheme' },
    '0014,6056' => { VR => 'DS', Name => 'HorizontalMovingWindowSize' },
    '0014,6057' => { VR => 'DS', Name => 'VerticalMovingWindowSize' },
    '0014,6059' => { VR => 'SQ', Name => 'PolynomialFittingSequence' },
    '0014,605A' => { VR => 'CS', Name => 'FittingDataType' },
    '0014,605B' => { VR => 'CS', Name => 'OperationOnTimeAxisBeforeFitting' },
    '0014,605C' => { VR => 'CS', Name => 'OperationPixelIntensBeforeFitting' },
    '0014,605D' => { VR => 'DS', Name => 'OrderOfPolynomial' },
    '0014,605E' => { VR => 'CS', Name => 'IndependentVariableForPolyFit' },
    '0014,605F' => { VR => 'DS', Name => 'PolynomialCoefficients' },
    '0014,6060' => { VR => 'CS', Name => 'ThermographyPixelDataUnit' },
    '0016,0001' => { VR => 'DS', Name => 'WhitePoint' },
    '0016,0002' => { VR => 'DS', Name => 'PrimaryChromaticities' },
    '0016,0003' => { VR => 'UT', Name => 'BatteryLevel' },
    '0016,0004' => { VR => 'DS', Name => 'ExposureTimeInSeconds' },
    '0016,0005' => { VR => 'DS', Name => 'FNumber' },
    '0016,0006' => { VR => 'IS', Name => 'OECFRows' },
    '0016,0007' => { VR => 'IS', Name => 'OECFColumns' },
    '0016,0008' => { VR => 'UC', Name => 'OECFColumnNames' },
    '0016,0009' => { VR => 'DS', Name => 'OECFValues' },
    '0016,000A' => { VR => 'IS', Name => 'SpatialFrequencyResponseRows' },
    '0016,000B' => { VR => 'IS', Name => 'SpatialFrequencyResponseColumns' },
    '0016,000C' => { VR => 'UC', Name => 'SpatialFrequencyRespColumnNames' },
    '0016,000D' => { VR => 'DS', Name => 'SpatialFrequencyResponseValues' },
    '0016,000E' => { VR => 'IS', Name => 'ColorFilterArrayPatternRows' },
    '0016,000F' => { VR => 'IS', Name => 'ColorFilterArrayPatternColumns' },
    '0016,0010' => { VR => 'DS', Name => 'ColorFilterArrayPatternValues' },
    '0016,0011' => { VR => 'US', Name => 'FlashFiringStatus' },
    '0016,0012' => { VR => 'US', Name => 'FlashReturnStatus' },
    '0016,0013' => { VR => 'US', Name => 'FlashMode' },
    '0016,0014' => { VR => 'US', Name => 'FlashFunctionPresent' },
    '0016,0015' => { VR => 'US', Name => 'FlashRedEyeMode' },
    '0016,0016' => { VR => 'US', Name => 'ExposureProgram' },
    '0016,0017' => { VR => 'UT', Name => 'SpectralSensitivity' },
    '0016,0018' => { VR => 'IS', Name => 'PhotographicSensitivity' },
    '0016,0019' => { VR => 'IS', Name => 'SelfTimerMode' },
    '0016,001A' => { VR => 'US', Name => 'SensitivityType' },
    '0016,001B' => { VR => 'IS', Name => 'StandardOutputSensitivity' },
    '0016,001C' => { VR => 'IS', Name => 'RecommendedExposureIndex' },
    '0016,001D' => { VR => 'IS', Name => 'ISOSpeed' },
    '0016,001E' => { VR => 'IS', Name => 'ISOSpeedLatitudeyyy' },
    '0016,001F' => { VR => 'IS', Name => 'ISOSpeedLatitudezzz' },
    '0016,0020' => { VR => 'UT', Name => 'ExifVersion' },
    '0016,0021' => { VR => 'DS', Name => 'ShutterSpeedValue' },
    '0016,0022' => { VR => 'DS', Name => 'ApertureValue' },
    '0016,0023' => { VR => 'DS', Name => 'BrightnessValue' },
    '0016,0024' => { VR => 'DS', Name => 'ExposureBiasValue' },
    '0016,0025' => { VR => 'DS', Name => 'MaxApertureValue' },
    '0016,0026' => { VR => 'DS', Name => 'SubjectDistance' },
    '0016,0027' => { VR => 'US', Name => 'MeteringMode' },
    '0016,0028' => { VR => 'US', Name => 'LightSource' },
    '0016,0029' => { VR => 'DS', Name => 'FocalLength' },
    '0016,002A' => { VR => 'IS', Name => 'SubjectArea' },
    '0016,002B' => { VR => 'OB', Name => 'MakerNote' },
    '0016,0030' => { VR => 'DS', Name => 'Temperature' },
    '0016,0031' => { VR => 'DS', Name => 'Humidity' },
    '0016,0032' => { VR => 'DS', Name => 'Pressure' },
    '0016,0033' => { VR => 'DS', Name => 'WaterDepth' },
    '0016,0034' => { VR => 'DS', Name => 'Acceleration' },
    '0016,0035' => { VR => 'DS', Name => 'CameraElevationAngle' },
    '0016,0036' => { VR => 'DS', Name => 'FlashEnergy' },
    '0016,0037' => { VR => 'IS', Name => 'SubjectLocation' },
    '0016,0038' => { VR => 'DS', Name => 'PhotographicExposureIndex' },
    '0016,0039' => { VR => 'US', Name => 'SensingMethod' },
    '0016,003A' => { VR => 'US', Name => 'FileSource' },
    '0016,003B' => { VR => 'US', Name => 'SceneType' },
    '0016,0041' => { VR => 'US', Name => 'CustomRendered' },
    '0016,0042' => { VR => 'US', Name => 'ExposureMode' },
    '0016,0043' => { VR => 'US', Name => 'WhiteBalance' },
    '0016,0044' => { VR => 'DS', Name => 'DigitalZoomRatio' },
    '0016,0045' => { VR => 'IS', Name => 'FocalLengthIn35mmFilm' },
    '0016,0046' => { VR => 'US', Name => 'SceneCaptureType' },
    '0016,0047' => { VR => 'US', Name => 'GainControl' },
    '0016,0048' => { VR => 'US', Name => 'Contrast' },
    '0016,0049' => { VR => 'US', Name => 'Saturation' },
    '0016,004A' => { VR => 'US', Name => 'Sharpness' },
    '0016,004B' => { VR => 'OB', Name => 'DeviceSettingDescription' },
    '0016,004C' => { VR => 'US', Name => 'SubjectDistanceRange' },
    '0016,004D' => { VR => 'UT', Name => 'CameraOwnerName' },
    '0016,004E' => { VR => 'DS', Name => 'LensSpecification' },
    '0016,004F' => { VR => 'UT', Name => 'LensMake' },
    '0016,0050' => { VR => 'UT', Name => 'LensModel' },
    '0016,0051' => { VR => 'UT', Name => 'LensSerialNumber' },
    '0016,0061' => { VR => 'CS', Name => 'InteroperabilityIndex' },
    '0016,0062' => { VR => 'OB', Name => 'InteroperabilityVersion' },
    '0016,0070' => { VR => 'OB', Name => 'GPSVersionID' },
    '0016,0071' => { VR => 'CS', Name => 'GPSLatitudeRef' },
    '0016,0072' => { VR => 'DS', Name => 'GPSLatitude' },
    '0016,0073' => { VR => 'CS', Name => 'GPSLongitudeRef' },
    '0016,0074' => { VR => 'DS', Name => 'GPSLongitude' },
    '0016,0075' => { VR => 'US', Name => 'GPSAltitudeRef' },
    '0016,0076' => { VR => 'DS', Name => 'GPSAltitude' },
    '0016,0077' => { VR => 'DT', Name => 'GPSTimeStamp' },
    '0016,0078' => { VR => 'UT', Name => 'GPSSatellites' },
    '0016,0079' => { VR => 'CS', Name => 'GPSStatus' },
    '0016,007A' => { VR => 'CS', Name => 'GPSMeasureMode' },
    '0016,007B' => { VR => 'DS', Name => 'GPSDOP' },
    '0016,007C' => { VR => 'CS', Name => 'GPSSpeedRef' },
    '0016,007D' => { VR => 'DS', Name => 'GPSSpeed' },
    '0016,007E' => { VR => 'CS', Name => 'GPSTrackRef' },
    '0016,007F' => { VR => 'DS', Name => 'GPSTrack' },
    '0016,0080' => { VR => 'CS', Name => 'GPSImgDirectionRef' },
    '0016,0081' => { VR => 'DS', Name => 'GPSImgDirection' },
    '0016,0082' => { VR => 'UT', Name => 'GPSMapDatum' },
    '0016,0083' => { VR => 'CS', Name => 'GPSDestLatitudeRef' },
    '0016,0084' => { VR => 'DS', Name => 'GPSDestLatitude' },
    '0016,0085' => { VR => 'CS', Name => 'GPSDestLongitudeRef' },
    '0016,0086' => { VR => 'DS', Name => 'GPSDestLongitude' },
    '0016,0087' => { VR => 'CS', Name => 'GPSDestBearingRef' },
    '0016,0088' => { VR => 'DS', Name => 'GPSDestBearing' },
    '0016,0089' => { VR => 'CS', Name => 'GPSDestDistanceRef' },
    '0016,008A' => { VR => 'DS', Name => 'GPSDestDistance' },
    '0016,008B' => { VR => 'OB', Name => 'GPSProcessingMethod' },
    '0016,008C' => { VR => 'OB', Name => 'GPSAreaInformation' },
    '0016,008D' => { VR => 'DT', Name => 'GPSDateStamp' },
    '0016,008E' => { VR => 'IS', Name => 'GPSDifferential' },
    '0016,1001' => { VR => 'CS', Name => 'LightSourcePolarization' },
    '0016,1002' => { VR => 'DS', Name => 'EmitterColorTemperature' },
    '0016,1003' => { VR => 'CS', Name => 'ContactMethod' },
    '0016,1004' => { VR => 'CS', Name => 'ImmersionMedia' },
    '0016,1005' => { VR => 'DS', Name => 'OpticalMagnificationFactor' },
    # acquisition group
    '0018,0000' => { VR => 'UL', Name => 'AcquisitionGroupLength' },
    '0018,0010' => { VR => 'LO', Name => 'ContrastBolusAgent' },
    '0018,0012' => { VR => 'SQ', Name => 'ContrastBolusAgentSequence' },
    '0018,0013' => { VR => 'FL', Name => 'ContrastBolusT1Relaxivity' },
    '0018,0014' => { VR => 'SQ', Name => 'ContrastBolusAdministrationRoute' },
    '0018,0015' => { VR => 'CS', Name => 'BodyPartExamined' },
    '0018,0020' => { VR => 'CS', Name => 'ScanningSequence' },
    '0018,0021' => { VR => 'CS', Name => 'SequenceVariant' },
    '0018,0022' => { VR => 'CS', Name => 'ScanOptions' },
    '0018,0023' => { VR => 'CS', Name => 'MRAcquisitionType' },
    '0018,0024' => { VR => 'SH', Name => 'SequenceName' },
    '0018,0025' => { VR => 'CS', Name => 'AngioFlag' },
    '0018,0026' => { VR => 'SQ', Name => 'InterventionDrugInformationSeq' },
    '0018,0027' => { VR => 'TM', Name => 'InterventionDrugStopTime' },
    '0018,0028' => { VR => 'DS', Name => 'InterventionDrugDose' },
    '0018,0029' => { VR => 'SQ', Name => 'InterventionDrugSequence' },
    '0018,002A' => { VR => 'SQ', Name => 'AdditionalDrugSequence' },
    '0018,0030' => { VR => 'LO', Name => 'Radionuclide' },
    '0018,0031' => { VR => 'LO', Name => 'Radiopharmaceutical' },
    '0018,0032' => { VR => 'DS', Name => 'EnergyWindowCenterline' },
    '0018,0033' => { VR => 'DS', Name => 'EnergyWindowTotalWidth' },
    '0018,0034' => { VR => 'LO', Name => 'InterventionDrugName' },
    '0018,0035' => { VR => 'TM', Name => 'InterventionDrugStartTime' },
    '0018,0036' => { VR => 'SQ', Name => 'InterventionSequence' },
    '0018,0037' => { VR => 'CS', Name => 'TherapyType' },
    '0018,0038' => { VR => 'CS', Name => 'InterventionStatus' },
    '0018,0039' => { VR => 'CS', Name => 'TherapyDescription' },
    '0018,003A' => { VR => 'ST', Name => 'InterventionDescription' },
    '0018,0040' => { VR => 'IS', Name => 'CineRate' },
    '0018,0042' => { VR => 'CS', Name => 'InitialCineRunState' },
    '0018,0050' => { VR => 'DS', Name => 'SliceThickness' },
    '0018,0060' => { VR => 'DS', Name => 'KVP' },
    '0018,0070' => { VR => 'IS', Name => 'CountsAccumulated' },
    '0018,0071' => { VR => 'CS', Name => 'AcquisitionTerminationCondition' },
    '0018,0072' => { VR => 'DS', Name => 'EffectiveDuration' },
    '0018,0073' => { VR => 'CS', Name => 'AcquisitionStartCondition' },
    '0018,0074' => { VR => 'IS', Name => 'AcquisitionStartConditionData' },
    '0018,0075' => { VR => 'IS', Name => 'AcquisitionEndConditionData' },
    '0018,0080' => { VR => 'DS', Name => 'RepetitionTime' },
    '0018,0081' => { VR => 'DS', Name => 'EchoTime' },
    '0018,0082' => { VR => 'DS', Name => 'InversionTime' },
    '0018,0083' => { VR => 'DS', Name => 'NumberOfAverages' },
    '0018,0084' => { VR => 'DS', Name => 'ImagingFrequency' },
    '0018,0085' => { VR => 'SH', Name => 'ImagedNucleus' },
    '0018,0086' => { VR => 'IS', Name => 'EchoNumber' },
    '0018,0087' => { VR => 'DS', Name => 'MagneticFieldStrength' },
    '0018,0088' => { VR => 'DS', Name => 'SpacingBetweenSlices' },
    '0018,0089' => { VR => 'IS', Name => 'NumberOfPhaseEncodingSteps' },
    '0018,0090' => { VR => 'DS', Name => 'DataCollectionDiameter' },
    '0018,0091' => { VR => 'IS', Name => 'EchoTrainLength' },
    '0018,0093' => { VR => 'DS', Name => 'PercentSampling' },
    '0018,0094' => { VR => 'DS', Name => 'PercentPhaseFieldOfView' },
    '0018,0095' => { VR => 'DS', Name => 'PixelBandwidth' },
    '0018,1000' => { VR => 'LO', Name => 'DeviceSerialNumber' },
    '0018,1002' => { VR => 'UI', Name => 'DeviceUID' },
    '0018,1003' => { VR => 'LO', Name => 'DeviceID' },
    '0018,1004' => { VR => 'LO', Name => 'PlateID' },
    '0018,1005' => { VR => 'LO', Name => 'GeneratorID' },
    '0018,1006' => { VR => 'LO', Name => 'GridID' },
    '0018,1007' => { VR => 'LO', Name => 'CassetteID' },
    '0018,1008' => { VR => 'LO', Name => 'GantryID' },
    '0018,1009' => { VR => 'UT', Name => 'UniqueDeviceIdentifier' },
    '0018,100A' => { VR => 'SQ', Name => 'UDISequence' },
    '0018,100B' => { VR => 'UI', Name => 'ManufacturerDeviceClassUID' },
    '0018,1010' => { VR => 'LO', Name => 'SecondaryCaptureDeviceID' },
    '0018,1011' => { VR => 'LO', Name => 'HardcopyCreationDeviceID' },
    '0018,1012' => { VR => 'DA', Name => 'DateOfSecondaryCapture' },
    '0018,1014' => { VR => 'TM', Name => 'TimeOfSecondaryCapture' },
    '0018,1016' => { VR => 'LO', Name => 'SecondaryCaptureDeviceManufacturer' },
    '0018,1017' => { VR => 'LO', Name => 'HardcopyDeviceManufacturer' },
    '0018,1018' => { VR => 'LO', Name => 'SecondaryCaptureDeviceModelName' },
    '0018,1019' => { VR => 'LO', Name => 'SecondaryCaptureDeviceSoftwareVers' },
    '0018,101A' => { VR => 'LO', Name => 'HardcopyDeviceSoftwareVersion' },
    '0018,101B' => { VR => 'LO', Name => 'HardcopyDeviceModelName' },
    '0018,1020' => { VR => 'LO', Name => 'SoftwareVersion' },
    '0018,1022' => { VR => 'SH', Name => 'VideoImageFormatAcquired' },
    '0018,1023' => { VR => 'LO', Name => 'DigitalImageFormatAcquired' },
    '0018,1030' => { VR => 'LO', Name => 'ProtocolName' },
    '0018,1040' => { VR => 'LO', Name => 'ContrastBolusRoute' },
    '0018,1041' => { VR => 'DS', Name => 'ContrastBolusVolume' },
    '0018,1042' => { VR => 'TM', Name => 'ContrastBolusStartTime' },
    '0018,1043' => { VR => 'TM', Name => 'ContrastBolusStopTime' },
    '0018,1044' => { VR => 'DS', Name => 'ContrastBolusTotalDose' },
    '0018,1045' => { VR => 'IS', Name => 'SyringeCounts' },
    '0018,1046' => { VR => 'DS', Name => 'ContrastFlowRate' },
    '0018,1047' => { VR => 'DS', Name => 'ContrastFlowDuration' },
    '0018,1048' => { VR => 'CS', Name => 'ContrastBolusIngredient' },
    '0018,1049' => { VR => 'DS', Name => 'ContrastBolusConcentration' },
    '0018,1050' => { VR => 'DS', Name => 'SpatialResolution' },
    '0018,1060' => { VR => 'DS', Name => 'TriggerTime' },
    '0018,1061' => { VR => 'LO', Name => 'TriggerSourceOrType' },
    '0018,1062' => { VR => 'IS', Name => 'NominalInterval' },
    '0018,1063' => { VR => 'DS', Name => 'FrameTime' },
    '0018,1064' => { VR => 'LO', Name => 'CardiacFramingType' },
    '0018,1065' => { VR => 'DS', Name => 'FrameTimeVector' },
    '0018,1066' => { VR => 'DS', Name => 'FrameDelay' },
    '0018,1067' => { VR => 'DS', Name => 'ImageTriggerDelay' },
    '0018,1068' => { VR => 'DS', Name => 'MultiplexGroupTimeOffset' },
    '0018,1069' => { VR => 'DS', Name => 'TriggerTimeOffset' },
    '0018,106A' => { VR => 'CS', Name => 'SynchronizationTrigger' },
    '0018,106C' => { VR => 'US', Name => 'SynchronizationChannel' },
    '0018,106E' => { VR => 'UL', Name => 'TriggerSamplePosition' },
    '0018,1070' => { VR => 'LO', Name => 'RadiopharmaceuticalRoute' },
    '0018,1071' => { VR => 'DS', Name => 'RadiopharmaceuticalVolume' },
    '0018,1072' => { VR => 'TM', Name => 'RadiopharmaceuticalStartTime' },
    '0018,1073' => { VR => 'TM', Name => 'RadiopharmaceuticalStopTime' },
    '0018,1074' => { VR => 'DS', Name => 'RadionuclideTotalDose' },
    '0018,1075' => { VR => 'DS', Name => 'RadionuclideHalfLife' },
    '0018,1076' => { VR => 'DS', Name => 'RadionuclidePositronFraction' },
    '0018,1077' => { VR => 'DS', Name => 'RadiopharmaceuticalSpecActivity' },
    '0018,1078' => { VR => 'DT', Name => 'RadiopharmaceuticalStartDateTime' },
    '0018,1079' => { VR => 'DT', Name => 'RadiopharmaceuticalStopDateTime' },
    '0018,1080' => { VR => 'CS', Name => 'BeatRejectionFlag' },
    '0018,1081' => { VR => 'IS', Name => 'LowRRValue' },
    '0018,1082' => { VR => 'IS', Name => 'HighRRValue' },
    '0018,1083' => { VR => 'IS', Name => 'IntervalsAcquired' },
    '0018,1084' => { VR => 'IS', Name => 'IntervalsRejected' },
    '0018,1085' => { VR => 'LO', Name => 'PVCRejection' },
    '0018,1086' => { VR => 'IS', Name => 'SkipBeats' },
    '0018,1088' => { VR => 'IS', Name => 'HeartRate' },
    '0018,1090' => { VR => 'IS', Name => 'CardiacNumberOfImages' },
    '0018,1094' => { VR => 'IS', Name => 'TriggerWindow' },
    '0018,1100' => { VR => 'DS', Name => 'ReconstructionDiameter' },
    '0018,1110' => { VR => 'DS', Name => 'DistanceSourceToDetector' },
    '0018,1111' => { VR => 'DS', Name => 'DistanceSourceToPatient' },
    '0018,1114' => { VR => 'DS', Name => 'EstimatedRadiographicMagnification' },
    '0018,1120' => { VR => 'DS', Name => 'GantryDetectorTilt' },
    '0018,1121' => { VR => 'DS', Name => 'GantryDetectorSlew' },
    '0018,1130' => { VR => 'DS', Name => 'TableHeight' },
    '0018,1131' => { VR => 'DS', Name => 'TableTraverse' },
    '0018,1134' => { VR => 'CS', Name => 'TableMotion' },
    '0018,1135' => { VR => 'DS', Name => 'TableVerticalIncrement' },
    '0018,1136' => { VR => 'DS', Name => 'TableLateralIncrement' },
    '0018,1137' => { VR => 'DS', Name => 'TableLongitudinalIncrement' },
    '0018,1138' => { VR => 'DS', Name => 'TableAngle' },
    '0018,113A' => { VR => 'CS', Name => 'TableType' },
    '0018,1140' => { VR => 'CS', Name => 'RotationDirection' },
    '0018,1141' => { VR => 'DS', Name => 'AngularPosition' },
    '0018,1142' => { VR => 'DS', Name => 'RadialPosition' },
    '0018,1143' => { VR => 'DS', Name => 'ScanArc' },
    '0018,1144' => { VR => 'DS', Name => 'AngularStep' },
    '0018,1145' => { VR => 'DS', Name => 'CenterOfRotationOffset' },
    '0018,1146' => { VR => 'DS', Name => 'RotationOffset' },
    '0018,1147' => { VR => 'CS', Name => 'FieldOfViewShape' },
    '0018,1149' => { VR => 'IS', Name => 'FieldOfViewDimensions' },
    '0018,1150' => { VR => 'IS', Name => 'ExposureTime' },
    '0018,1151' => { VR => 'IS', Name => 'XRayTubeCurrent' },
    '0018,1152' => { VR => 'IS', Name => 'Exposure' },
    '0018,1153' => { VR => 'IS', Name => 'ExposureInMicroAmpSec' },
    '0018,1154' => { VR => 'DS', Name => 'AveragePulseWidth' },
    '0018,1155' => { VR => 'CS', Name => 'RadiationSetting' },
    '0018,1156' => { VR => 'CS', Name => 'RectificationType' },
    '0018,115A' => { VR => 'CS', Name => 'RadiationMode' },
    '0018,115E' => { VR => 'DS', Name => 'ImageAreaDoseProduct' },
    '0018,1160' => { VR => 'SH', Name => 'FilterType' },
    '0018,1161' => { VR => 'LO', Name => 'TypeOfFilters' },
    '0018,1162' => { VR => 'DS', Name => 'IntensifierSize' },
    '0018,1164' => { VR => 'DS', Name => 'ImagerPixelSpacing' },
    '0018,1166' => { VR => 'CS', Name => 'Grid' },
    '0018,1170' => { VR => 'IS', Name => 'GeneratorPower' },
    '0018,1180' => { VR => 'SH', Name => 'CollimatorGridName' },
    '0018,1181' => { VR => 'CS', Name => 'CollimatorType' },
    '0018,1182' => { VR => 'IS', Name => 'FocalDistance' },
    '0018,1183' => { VR => 'DS', Name => 'XFocusCenter' },
    '0018,1184' => { VR => 'DS', Name => 'YFocusCenter' },
    '0018,1190' => { VR => 'DS', Name => 'FocalSpots' },
    '0018,1191' => { VR => 'CS', Name => 'AnodeTargetMaterial' },
    '0018,11A0' => { VR => 'DS', Name => 'BodyPartThickness' },
    '0018,11A2' => { VR => 'DS', Name => 'CompressionForce' },
    '0018,11A3' => { VR => 'DS', Name => 'CompressionPressure' },
    '0018,11A4' => { VR => 'LO', Name => 'PaddleDescription' },
    '0018,11A5' => { VR => 'DS', Name => 'CompressionContactArea' },
    '0018,11B0' => { VR => 'LO', Name => 'AcquisitionMode' },
    '0018,11B1' => { VR => 'LO', Name => 'DoseModeName' },
    '0018,11B2' => { VR => 'CS', Name => 'AcquiredSubtractionMaskFlag' },
    '0018,11B3' => { VR => 'CS', Name => 'FluoroscopyPersistenceFlag' },
    '0018,11B4' => { VR => 'CS', Name => 'FluoroLastImageHoldPersistenceFlag' },
    '0018,11B5' => { VR => 'IS', Name => 'UpperLimitNumPersistFluoroFrames' },
    '0018,11B6' => { VR => 'CS', Name => 'ContrastBolusAutoInjectionTrigFlag' },
    '0018,11B7' => { VR => 'FD', Name => 'ContrastBolusInjectionDelay' },
    '0018,11B8' => { VR => 'SQ', Name => 'XAAcquisitionPhaseDetailsSequence' },
    '0018,11B9' => { VR => 'FD', Name => 'XAAcquisitionFrameRate' },
    '0018,11BA' => { VR => 'SQ', Name => 'XAPlaneDetailsSequence' },
    '0018,11BB' => { VR => 'LO', Name => 'AcquisitionFieldOfViewLabel' },
    '0018,11BC' => { VR => 'SQ', Name => 'XRayFilterDetailsSequence' },
    '0018,11BD' => { VR => 'FD', Name => 'XAAcquisitionDuration' },
    '0018,11BE' => { VR => 'CS', Name => 'ReconstructionPipelineType' },
    '0018,11BF' => { VR => 'SQ', Name => 'ImageFilterDetailsSequence' },
    '0018,11C0' => { VR => 'CS', Name => 'AppliedMaskSubtractionFlag' },
    '0018,11C1' => { VR => 'SQ', Name => 'RequestedSeriesDescriptionCodeSeq' },
    '0018,1200' => { VR => 'DA', Name => 'DateOfLastCalibration' },
    '0018,1201' => { VR => 'TM', Name => 'TimeOfLastCalibration' },
    '0018,1202' => { VR => 'DT', Name => 'DateTimeOfLastCalibration' },
    '0018,1203' => { VR => 'DT', Name => 'CalibrationDateTime' },
    '0018,1204' => { VR => 'DA', Name => 'DateOfManufacture' },
    '0018,1205' => { VR => 'DA', Name => 'DateOfInstallation' },
    '0018,1210' => { VR => 'SH', Name => 'ConvolutionKernel' },
    '0018,1240' => { VR => 'IS', Name => 'UpperLowerPixelValues' },
    '0018,1242' => { VR => 'IS', Name => 'ActualFrameDuration' },
    '0018,1243' => { VR => 'IS', Name => 'CountRate' },
    '0018,1244' => { VR => 'US', Name => 'PreferredPlaybackSequencing' },
    '0018,1250' => { VR => 'SH', Name => 'ReceiveCoilName' },
    '0018,1251' => { VR => 'SH', Name => 'TransmitCoilName' },
    '0018,1260' => { VR => 'SH', Name => 'PlateType' },
    '0018,1261' => { VR => 'LO', Name => 'PhosphorType' },
    '0018,1271' => { VR => 'FD', Name => 'WaterEquivalentDiameter' },
    '0018,1272' => { VR => 'SQ', Name => 'WaterEquiDiameterCalcMethCodeSeq' },
    '0018,1300' => { VR => 'DS', Name => 'ScanVelocity' },
    '0018,1301' => { VR => 'CS', Name => 'WholeBodyTechnique' },
    '0018,1302' => { VR => 'IS', Name => 'ScanLength' },
    '0018,1310' => { VR => 'US', Name => 'AcquisitionMatrix' },
    '0018,1312' => { VR => 'CS', Name => 'InPlanePhaseEncodingDirection' },
    '0018,1314' => { VR => 'DS', Name => 'FlipAngle' },
    '0018,1315' => { VR => 'CS', Name => 'VariableFlipAngleFlag' },
    '0018,1316' => { VR => 'DS', Name => 'SAR' },
    '0018,1318' => { VR => 'DS', Name => 'DB-Dt' },
    '0018,1320' => { VR => 'FL', Name => 'B1rms' },
    '0018,1400' => { VR => 'LO', Name => 'AcquisitionDeviceProcessingDescr' },
    '0018,1401' => { VR => 'LO', Name => 'AcquisitionDeviceProcessingCode' },
    '0018,1402' => { VR => 'CS', Name => 'CassetteOrientation' },
    '0018,1403' => { VR => 'CS', Name => 'CassetteSize' },
    '0018,1404' => { VR => 'US', Name => 'ExposuresOnPlate' },
    '0018,1405' => { VR => 'IS', Name => 'RelativeXRayExposure' },
    '0018,1411' => { VR => 'DS', Name => 'ExposureIndex' },
    '0018,1412' => { VR => 'DS', Name => 'TargetExposureIndex' },
    '0018,1413' => { VR => 'DS', Name => 'DeviationIndex' },
    '0018,1450' => { VR => 'DS', Name => 'ColumnAngulation' },
    '0018,1460' => { VR => 'DS', Name => 'TomoLayerHeight' },
    '0018,1470' => { VR => 'DS', Name => 'TomoAngle' },
    '0018,1480' => { VR => 'DS', Name => 'TomoTime' },
    '0018,1490' => { VR => 'CS', Name => 'TomoType' },
    '0018,1491' => { VR => 'CS', Name => 'TomoClass' },
    '0018,1495' => { VR => 'IS', Name => 'NumberOfTomosynthesisSourceImages' },
    '0018,1500' => { VR => 'CS', Name => 'PositionerMotion' },
    '0018,1508' => { VR => 'CS', Name => 'PositionerType' },
    '0018,1510' => { VR => 'DS', Name => 'PositionerPrimaryAngle' },
    '0018,1511' => { VR => 'DS', Name => 'PositionerSecondaryAngle' },
    '0018,1520' => { VR => 'DS', Name => 'PositionerPrimaryAngleIncrement' },
    '0018,1521' => { VR => 'DS', Name => 'PositionerSecondaryAngleIncrement' },
    '0018,1530' => { VR => 'DS', Name => 'DetectorPrimaryAngle' },
    '0018,1531' => { VR => 'DS', Name => 'DetectorSecondaryAngle' },
    '0018,1600' => { VR => 'CS', Name => 'ShutterShape' },
    '0018,1602' => { VR => 'IS', Name => 'ShutterLeftVerticalEdge' },
    '0018,1604' => { VR => 'IS', Name => 'ShutterRightVerticalEdge' },
    '0018,1606' => { VR => 'IS', Name => 'ShutterUpperHorizontalEdge' },
    '0018,1608' => { VR => 'IS', Name => 'ShutterLowerHorizontalEdge' },
    '0018,1610' => { VR => 'IS', Name => 'CenterOfCircularShutter' },
    '0018,1612' => { VR => 'IS', Name => 'RadiusOfCircularShutter' },
    '0018,1620' => { VR => 'IS', Name => 'VerticesOfPolygonalShutter' },
    '0018,1622' => { VR => 'US', Name => 'ShutterPresentationValue' },
    '0018,1623' => { VR => 'US', Name => 'ShutterOverlayGroup' },
    '0018,1624' => { VR => 'US', Name => 'ShutterPresentationColorCIELabVal' },
    '0018,1630' => { VR => 'CS', Name => 'OutlineShapeType' },
    '0018,1631' => { VR => 'FD', Name => 'OutlineLeftVerticalEdge' },
    '0018,1632' => { VR => 'FD', Name => 'OutlineRightVerticalEdge' },
    '0018,1633' => { VR => 'FD', Name => 'OutlineUpperHorizontalEdge' },
    '0018,1634' => { VR => 'FD', Name => 'OutlineLowerHorizontalEdge' },
    '0018,1635' => { VR => 'FD', Name => 'CenterOfCircularOutline' },
    '0018,1636' => { VR => 'FD', Name => 'DiameterOfCircularOutline' },
    '0018,1637' => { VR => 'UL', Name => 'NumberOfPolygonalVertices' },
    '0018,1638' => { VR => 'OF', Name => 'VerticesOfThePolygonalOutline' },
    '0018,1700' => { VR => 'CS', Name => 'CollimatorShape' },
    '0018,1702' => { VR => 'IS', Name => 'CollimatorLeftVerticalEdge' },
    '0018,1704' => { VR => 'IS', Name => 'CollimatorRightVerticalEdge' },
    '0018,1706' => { VR => 'IS', Name => 'CollimatorUpperHorizontalEdge' },
    '0018,1708' => { VR => 'IS', Name => 'CollimatorLowerHorizontalEdge' },
    '0018,1710' => { VR => 'IS', Name => 'CenterOfCircularCollimator' },
    '0018,1712' => { VR => 'IS', Name => 'RadiusOfCircularCollimator' },
    '0018,1720' => { VR => 'IS', Name => 'VerticesOfPolygonalCollimator' },
    '0018,1800' => { VR => 'CS', Name => 'AcquisitionTimeSynchronized' },
    '0018,1801' => { VR => 'SH', Name => 'TimeSource' },
    '0018,1802' => { VR => 'CS', Name => 'TimeDistributionProtocol' },
    '0018,1803' => { VR => 'LO', Name => 'NTPSourceAddress' },
    '0018,2001' => { VR => 'IS', Name => 'PageNumberVector' },
    '0018,2002' => { VR => 'SH', Name => 'FrameLabelVector' },
    '0018,2003' => { VR => 'DS', Name => 'FramePrimaryAngleVector' },
    '0018,2004' => { VR => 'DS', Name => 'FrameSecondaryAngleVector' },
    '0018,2005' => { VR => 'DS', Name => 'SliceLocationVector' },
    '0018,2006' => { VR => 'SH', Name => 'DisplayWindowLabelVector' },
    '0018,2010' => { VR => 'DS', Name => 'NominalScannedPixelSpacing' },
    '0018,2020' => { VR => 'CS', Name => 'DigitizingDeviceTransportDirection' },
    '0018,2030' => { VR => 'DS', Name => 'RotationOfScannedFilm' },
    '0018,2041' => { VR => 'SQ', Name => 'BiopsyTargetSequence' },
    '0018,2042' => { VR => 'UI', Name => 'TargetUID' },
    '0018,2043' => { VR => 'FL', Name => 'LocalizingCursorPosition' },
    '0018,2044' => { VR => 'FL', Name => 'CalculatedTargetPosition' },
    '0018,2045' => { VR => 'SH', Name => 'TargetLabel' },
    '0018,2046' => { VR => 'FL', Name => 'DisplayedZValue' },
    '0018,3100' => { VR => 'CS', Name => 'IVUSAcquisition' },
    '0018,3101' => { VR => 'DS', Name => 'IVUSPullbackRate' },
    '0018,3102' => { VR => 'DS', Name => 'IVUSGatedRate' },
    '0018,3103' => { VR => 'IS', Name => 'IVUSPullbackStartFrameNumber' },
    '0018,3104' => { VR => 'IS', Name => 'IVUSPullbackStopFrameNumber' },
    '0018,3105' => { VR => 'IS', Name => 'LesionNumber' },
    '0018,4000' => { VR => 'LT', Name => 'AcquisitionComments' },
    '0018,5000' => { VR => 'SH', Name => 'OutputPower' },
    '0018,5010' => { VR => 'LO', Name => 'TransducerData' },
    '0018,5011' => { VR => 'SQ', Name => 'TransducerIdentificationSequence' },
    '0018,5012' => { VR => 'DS', Name => 'FocusDepth' },
    '0018,5020' => { VR => 'LO', Name => 'ProcessingFunction' },
    '0018,5021' => { VR => 'LO', Name => 'PostprocessingFunction' },
    '0018,5022' => { VR => 'DS', Name => 'MechanicalIndex' },
    '0018,5024' => { VR => 'DS', Name => 'BoneThermalIndex' },
    '0018,5026' => { VR => 'DS', Name => 'CranialThermalIndex' },
    '0018,5027' => { VR => 'DS', Name => 'SoftTissueThermalIndex' },
    '0018,5028' => { VR => 'DS', Name => 'SoftTissueFocusThermalIndex' },
    '0018,5029' => { VR => 'DS', Name => 'SoftTissueSurfaceThermalIndex' },
    '0018,5030' => { VR => 'DS', Name => 'DynamicRange' },
    '0018,5040' => { VR => 'DS', Name => 'TotalGain' },
    '0018,5050' => { VR => 'IS', Name => 'DepthOfScanField' },
    '0018,5100' => { VR => 'CS', Name => 'PatientPosition' },
    '0018,5101' => { VR => 'CS', Name => 'ViewPosition' },
    '0018,5104' => { VR => 'SQ', Name => 'ProjectionEponymousNameCodeSeq' },
    '0018,5210' => { VR => 'DS', Name => 'ImageTransformationMatrix' },
    '0018,5212' => { VR => 'DS', Name => 'ImageTranslationVector' },
    '0018,6000' => { VR => 'DS', Name => 'Sensitivity' },
    '0018,6011' => { VR => 'SQ', Name => 'SequenceOfUltrasoundRegions' },
    '0018,6012' => { VR => 'US', Name => 'RegionSpatialFormat' },
    '0018,6014' => { VR => 'US', Name => 'RegionDataType' },
    '0018,6016' => { VR => 'UL', Name => 'RegionFlags' },
    '0018,6018' => { VR => 'UL', Name => 'RegionLocationMinX0' },
    '0018,601A' => { VR => 'UL', Name => 'RegionLocationMinY0' },
    '0018,601C' => { VR => 'UL', Name => 'RegionLocationMaxX1' },
    '0018,601E' => { VR => 'UL', Name => 'RegionLocationMaxY1' },
    '0018,6020' => { VR => 'SL', Name => 'ReferencePixelX0' },
    '0018,6022' => { VR => 'SL', Name => 'ReferencePixelY0' },
    '0018,6024' => { VR => 'US', Name => 'PhysicalUnitsXDirection' },
    '0018,6026' => { VR => 'US', Name => 'PhysicalUnitsYDirection' },
    '0018,6028' => { VR => 'FD', Name => 'ReferencePixelPhysicalValueX' },
    '0018,602A' => { VR => 'FD', Name => 'ReferencePixelPhysicalValueY' },
    '0018,602C' => { VR => 'FD', Name => 'PhysicalDeltaX' },
    '0018,602E' => { VR => 'FD', Name => 'PhysicalDeltaY' },
    '0018,6030' => { VR => 'UL', Name => 'TransducerFrequency' },
    '0018,6031' => { VR => 'CS', Name => 'TransducerType' },
    '0018,6032' => { VR => 'UL', Name => 'PulseRepetitionFrequency' },
    '0018,6034' => { VR => 'FD', Name => 'DopplerCorrectionAngle' },
    '0018,6036' => { VR => 'FD', Name => 'SteeringAngle' },
    '0018,6038' => { VR => 'UL', Name => 'DopplerSampleVolumeXPosRetired' },
    '0018,6039' => { VR => 'SL', Name => 'DopplerSampleVolumeXPosition' },
    '0018,603A' => { VR => 'UL', Name => 'DopplerSampleVolumeYPosRetired' },
    '0018,603B' => { VR => 'SL', Name => 'DopplerSampleVolumeYPosition' },
    '0018,603C' => { VR => 'UL', Name => 'TMLinePositionX0Retired' },
    '0018,603D' => { VR => 'SL', Name => 'TMLinePositionX0' },
    '0018,603E' => { VR => 'UL', Name => 'TMLinePositionY0Retired' },
    '0018,603F' => { VR => 'SL', Name => 'TMLinePositionY0' },
    '0018,6040' => { VR => 'UL', Name => 'TMLinePositionX1Retired' },
    '0018,6041' => { VR => 'SL', Name => 'TMLinePositionX1' },
    '0018,6042' => { VR => 'UL', Name => 'TMLinePositionY1Retired' },
    '0018,6043' => { VR => 'SL', Name => 'TMLinePositionY1' },
    '0018,6044' => { VR => 'US', Name => 'PixelComponentOrganization' },
    '0018,6046' => { VR => 'UL', Name => 'PixelComponentMask' },
    '0018,6048' => { VR => 'UL', Name => 'PixelComponentRangeStart' },
    '0018,604A' => { VR => 'UL', Name => 'PixelComponentRangeStop' },
    '0018,604C' => { VR => 'US', Name => 'PixelComponentPhysicalUnits' },
    '0018,604E' => { VR => 'US', Name => 'PixelComponentDataType' },
    '0018,6050' => { VR => 'UL', Name => 'NumberOfTableBreakPoints' },
    '0018,6052' => { VR => 'UL', Name => 'TableOfXBreakPoints' },
    '0018,6054' => { VR => 'FD', Name => 'TableOfYBreakPoints' },
    '0018,6056' => { VR => 'UL', Name => 'NumberOfTableEntries' },
    '0018,6058' => { VR => 'UL', Name => 'TableOfPixelValues' },
    '0018,605A' => { VR => 'FL', Name => 'TableOfParameterValues' },
    '0018,6060' => { VR => 'FL', Name => 'RWaveTimeVector' },
    '0018,6070' => { VR => 'US', Name => 'ActiveImageAreaOverlayGroup' },
    '0018,7000' => { VR => 'CS', Name => 'DetectorConditionsNominalFlag' },
    '0018,7001' => { VR => 'DS', Name => 'DetectorTemperature' },
    '0018,7004' => { VR => 'CS', Name => 'DetectorType' },
    '0018,7005' => { VR => 'CS', Name => 'DetectorConfiguration' },
    '0018,7006' => { VR => 'LT', Name => 'DetectorDescription' },
    '0018,7008' => { VR => 'LT', Name => 'DetectorMode' },
    '0018,700A' => { VR => 'SH', Name => 'DetectorID' },
    '0018,700C' => { VR => 'DA', Name => 'DateOfLastDetectorCalibration' },
    '0018,700E' => { VR => 'TM', Name => 'TimeOfLastDetectorCalibration' },
    '0018,7010' => { VR => 'IS', Name => 'DetectorExposuresSinceCalibration' },
    '0018,7011' => { VR => 'IS', Name => 'DetectorExposuresSinceManufactured' },
    '0018,7012' => { VR => 'DS', Name => 'DetectorTimeSinceLastExposure' },
    '0018,7014' => { VR => 'DS', Name => 'DetectorActiveTime' },
    '0018,7016' => { VR => 'DS', Name => 'DetectorActiveOffsetFromExposure' },
    '0018,701A' => { VR => 'DS', Name => 'DetectorBinning' },
    '0018,7020' => { VR => 'DS', Name => 'DetectorElementPhysicalSize' },
    '0018,7022' => { VR => 'DS', Name => 'DetectorElementSpacing' },
    '0018,7024' => { VR => 'CS', Name => 'DetectorActiveShape' },
    '0018,7026' => { VR => 'DS', Name => 'DetectorActiveDimensions' },
    '0018,7028' => { VR => 'DS', Name => 'DetectorActiveOrigin' },
    '0018,702A' => { VR => 'LO', Name => 'DetectorManufacturerName' },
    '0018,702B' => { VR => 'LO', Name => 'DetectorManufacturersModelName' },
    '0018,7030' => { VR => 'DS', Name => 'FieldOfViewOrigin' },
    '0018,7032' => { VR => 'DS', Name => 'FieldOfViewRotation' },
    '0018,7034' => { VR => 'CS', Name => 'FieldOfViewHorizontalFlip' },
    '0018,7036' => { VR => 'FL', Name => 'PixelDataAreaOriginRelativeToFOV' },
    '0018,7038' => { VR => 'FL', Name => 'PixelDataAreaRotationAngleRelToFOV' },
    '0018,7040' => { VR => 'LT', Name => 'GridAbsorbingMaterial' },
    '0018,7041' => { VR => 'LT', Name => 'GridSpacingMaterial' },
    '0018,7042' => { VR => 'DS', Name => 'GridThickness' },
    '0018,7044' => { VR => 'DS', Name => 'GridPitch' },
    '0018,7046' => { VR => 'IS', Name => 'GridAspectRatio' },
    '0018,7048' => { VR => 'DS', Name => 'GridPeriod' },
    '0018,704C' => { VR => 'DS', Name => 'GridFocalDistance' },
    '0018,7050' => { VR => 'CS', Name => 'FilterMaterial' },
    '0018,7052' => { VR => 'DS', Name => 'FilterThicknessMinimum' },
    '0018,7054' => { VR => 'DS', Name => 'FilterThicknessMaximum' },
    '0018,7056' => { VR => 'FL', Name => 'FilterBeamPathLengthMinimum' },
    '0018,7058' => { VR => 'FL', Name => 'FilterBeamPathLengthMaximum' },
    '0018,7060' => { VR => 'CS', Name => 'ExposureControlMode' },
    '0018,7062' => { VR => 'LT', Name => 'ExposureControlModeDescription' },
    '0018,7064' => { VR => 'CS', Name => 'ExposureStatus' },
    '0018,7065' => { VR => 'DS', Name => 'PhototimerSetting' },
    '0018,8150' => { VR => 'DS', Name => 'ExposureTimeInMicroSec' },
    '0018,8151' => { VR => 'DS', Name => 'XRayTubeCurrentInMicroAmps' },
    '0018,9004' => { VR => 'CS', Name => 'ContentQualification' },
    '0018,9005' => { VR => 'SH', Name => 'PulseSequenceName' },
    '0018,9006' => { VR => 'SQ', Name => 'MRImagingModifierSequence' },
    '0018,9008' => { VR => 'CS', Name => 'EchoPulseSequence' },
    '0018,9009' => { VR => 'CS', Name => 'InversionRecovery' },
    '0018,9010' => { VR => 'CS', Name => 'FlowCompensation' },
    '0018,9011' => { VR => 'CS', Name => 'MultipleSpinEcho' },
    '0018,9012' => { VR => 'CS', Name => 'MultiPlanarExcitation' },
    '0018,9014' => { VR => 'CS', Name => 'PhaseContrast' },
    '0018,9015' => { VR => 'CS', Name => 'TimeOfFlightContrast' },
    '0018,9016' => { VR => 'CS', Name => 'Spoiling' },
    '0018,9017' => { VR => 'CS', Name => 'SteadyStatePulseSequence' },
    '0018,9018' => { VR => 'CS', Name => 'EchoPlanarPulseSequence' },
    '0018,9019' => { VR => 'FD', Name => 'TagAngleFirstAxis' },
    '0018,9020' => { VR => 'CS', Name => 'MagnetizationTransfer' },
    '0018,9021' => { VR => 'CS', Name => 'T2Preparation' },
    '0018,9022' => { VR => 'CS', Name => 'BloodSignalNulling' },
    '0018,9024' => { VR => 'CS', Name => 'SaturationRecovery' },
    '0018,9025' => { VR => 'CS', Name => 'SpectrallySelectedSuppression' },
    '0018,9026' => { VR => 'CS', Name => 'SpectrallySelectedExcitation' },
    '0018,9027' => { VR => 'CS', Name => 'SpatialPresaturation' },
    '0018,9028' => { VR => 'CS', Name => 'Tagging' },
    '0018,9029' => { VR => 'CS', Name => 'OversamplingPhase' },
    '0018,9030' => { VR => 'FD', Name => 'TagSpacingFirstDimension' },
    '0018,9032' => { VR => 'CS', Name => 'GeometryOfKSpaceTraversal' },
    '0018,9033' => { VR => 'CS', Name => 'SegmentedKSpaceTraversal' },
    '0018,9034' => { VR => 'CS', Name => 'RectilinearPhaseEncodeReordering' },
    '0018,9035' => { VR => 'FD', Name => 'TagThickness' },
    '0018,9036' => { VR => 'CS', Name => 'PartialFourierDirection' },
    '0018,9037' => { VR => 'CS', Name => 'CardiacSynchronizationTechnique' },
    '0018,9041' => { VR => 'LO', Name => 'ReceiveCoilManufacturerName' },
    '0018,9042' => { VR => 'SQ', Name => 'MRReceiveCoilSequence' },
    '0018,9043' => { VR => 'CS', Name => 'ReceiveCoilType' },
    '0018,9044' => { VR => 'CS', Name => 'QuadratureReceiveCoil' },
    '0018,9045' => { VR => 'SQ', Name => 'MultiCoilDefinitionSequence' },
    '0018,9046' => { VR => 'LO', Name => 'MultiCoilConfiguration' },
    '0018,9047' => { VR => 'SH', Name => 'MultiCoilElementName' },
    '0018,9048' => { VR => 'CS', Name => 'MultiCoilElementUsed' },
    '0018,9049' => { VR => 'SQ', Name => 'MRTransmitCoilSequence' },
    '0018,9050' => { VR => 'LO', Name => 'TransmitCoilManufacturerName' },
    '0018,9051' => { VR => 'CS', Name => 'TransmitCoilType' },
    '0018,9052' => { VR => 'FD', Name => 'SpectralWidth' },
    '0018,9053' => { VR => 'FD', Name => 'ChemicalShiftReference' },
    '0018,9054' => { VR => 'CS', Name => 'VolumeLocalizationTechnique' },
    '0018,9058' => { VR => 'US', Name => 'MRAcquisitionFrequencyEncodeSteps' },
    '0018,9059' => { VR => 'CS', Name => 'Decoupling' },
    '0018,9060' => { VR => 'CS', Name => 'DecoupledNucleus' },
    '0018,9061' => { VR => 'FD', Name => 'DecouplingFrequency' },
    '0018,9062' => { VR => 'CS', Name => 'DecouplingMethod' },
    '0018,9063' => { VR => 'FD', Name => 'DecouplingChemicalShiftReference' },
    '0018,9064' => { VR => 'CS', Name => 'KSpaceFiltering' },
    '0018,9065' => { VR => 'CS', Name => 'TimeDomainFiltering' },
    '0018,9066' => { VR => 'US', Name => 'NumberOfZeroFills' },
    '0018,9067' => { VR => 'CS', Name => 'BaselineCorrection' },
    '0018,9069' => { VR => 'FD', Name => 'ParallelReductionFactorInPlane' },
    '0018,9070' => { VR => 'FD', Name => 'CardiacRRIntervalSpecified' },
    '0018,9073' => { VR => 'FD', Name => 'AcquisitionDuration' },
    '0018,9074' => { VR => 'DT', Name => 'FrameAcquisitionDateTime' },
    '0018,9075' => { VR => 'CS', Name => 'DiffusionDirectionality' },
    '0018,9076' => { VR => 'SQ', Name => 'DiffusionGradientDirectionSequence' },
    '0018,9077' => { VR => 'CS', Name => 'ParallelAcquisition' },
    '0018,9078' => { VR => 'CS', Name => 'ParallelAcquisitionTechnique' },
    '0018,9079' => { VR => 'FD', Name => 'InversionTimes' },
    '0018,9080' => { VR => 'ST', Name => 'MetaboliteMapDescription' },
    '0018,9081' => { VR => 'CS', Name => 'PartialFourier' },
    '0018,9082' => { VR => 'FD', Name => 'EffectiveEchoTime' },
    '0018,9083' => { VR => 'SQ', Name => 'MetaboliteMapCodeSequence' },
    '0018,9084' => { VR => 'SQ', Name => 'ChemicalShiftSequence' },
    '0018,9085' => { VR => 'CS', Name => 'CardiacSignalSource' },
    '0018,9087' => { VR => 'FD', Name => 'DiffusionBValue' },
    '0018,9089' => { VR => 'FD', Name => 'DiffusionGradientOrientation' },
    '0018,9090' => { VR => 'FD', Name => 'VelocityEncodingDirection' },
    '0018,9091' => { VR => 'FD', Name => 'VelocityEncodingMinimumValue' },
    '0018,9092' => { VR => 'SQ', Name => 'VelocityEncodingAcquisitionSeq' },
    '0018,9093' => { VR => 'US', Name => 'NumberOfKSpaceTrajectories' },
    '0018,9094' => { VR => 'CS', Name => 'CoverageOfKSpace' },
    '0018,9095' => { VR => 'UL', Name => 'SpectroscopyAcquisitionPhaseRows' },
    '0018,9096' => { VR => 'FD', Name => 'ParallelReductFactorInPlaneRetired' },
    '0018,9098' => { VR => 'FD', Name => 'TransmitterFrequency' },
    '0018,9100' => { VR => 'CS', Name => 'ResonantNucleus' },
    '0018,9101' => { VR => 'CS', Name => 'FrequencyCorrection' },
    '0018,9103' => { VR => 'SQ', Name => 'MRSpectroscopyFOV-GeometrySequence' },
    '0018,9104' => { VR => 'FD', Name => 'SlabThickness' },
    '0018,9105' => { VR => 'FD', Name => 'SlabOrientation' },
    '0018,9106' => { VR => 'FD', Name => 'MidSlabPosition' },
    '0018,9107' => { VR => 'SQ', Name => 'MRSpatialSaturationSequence' },
    '0018,9112' => { VR => 'SQ', Name => 'MRTimingAndRelatedParametersSeq' },
    '0018,9114' => { VR => 'SQ', Name => 'MREchoSequence' },
    '0018,9115' => { VR => 'SQ', Name => 'MRModifierSequence' },
    '0018,9117' => { VR => 'SQ', Name => 'MRDiffusionSequence' },
    '0018,9118' => { VR => 'SQ', Name => 'CardiacTriggerSequence' },
    '0018,9119' => { VR => 'SQ', Name => 'MRAveragesSequence' },
    '0018,9125' => { VR => 'SQ', Name => 'MRFOV-GeometrySequence' },
    '0018,9126' => { VR => 'SQ', Name => 'VolumeLocalizationSequence' },
    '0018,9127' => { VR => 'UL', Name => 'SpectroscopyAcquisitionDataColumns' },
    '0018,9147' => { VR => 'CS', Name => 'DiffusionAnisotropyType' },
    '0018,9151' => { VR => 'DT', Name => 'FrameReferenceDateTime' },
    '0018,9152' => { VR => 'SQ', Name => 'MRMetaboliteMapSequence' },
    '0018,9155' => { VR => 'FD', Name => 'ParallelReductionFactorOutOfPlane' },
    '0018,9159' => { VR => 'UL', Name => 'SpectroscopyOutOfPlanePhaseSteps' },
    '0018,9166' => { VR => 'CS', Name => 'BulkMotionStatus' },
    '0018,9168' => { VR => 'FD', Name => 'ParallelReductionFactSecondInPlane' },
    '0018,9169' => { VR => 'CS', Name => 'CardiacBeatRejectionTechnique' },
    '0018,9170' => { VR => 'CS', Name => 'RespiratoryMotionCompTechnique' },
    '0018,9171' => { VR => 'CS', Name => 'RespiratorySignalSource' },
    '0018,9172' => { VR => 'CS', Name => 'BulkMotionCompensationTechnique' },
    '0018,9173' => { VR => 'CS', Name => 'BulkMotionSignalSource' },
    '0018,9174' => { VR => 'CS', Name => 'ApplicableSafetyStandardAgency' },
    '0018,9175' => { VR => 'LO', Name => 'ApplicableSafetyStandardDescr' },
    '0018,9176' => { VR => 'SQ', Name => 'OperatingModeSequence' },
    '0018,9177' => { VR => 'CS', Name => 'OperatingModeType' },
    '0018,9178' => { VR => 'CS', Name => 'OperatingMode' },
    '0018,9179' => { VR => 'CS', Name => 'SpecificAbsorptionRateDefinition' },
    '0018,9180' => { VR => 'CS', Name => 'GradientOutputType' },
    '0018,9181' => { VR => 'FD', Name => 'SpecificAbsorptionRateValue' },
    '0018,9182' => { VR => 'FD', Name => 'GradientOutput' },
    '0018,9183' => { VR => 'CS', Name => 'FlowCompensationDirection' },
    '0018,9184' => { VR => 'FD', Name => 'TaggingDelay' },
    '0018,9185' => { VR => 'ST', Name => 'RespiratoryMotionCompTechDescr' },
    '0018,9186' => { VR => 'SH', Name => 'RespiratorySignalSourceID' },
    '0018,9195' => { VR => 'FD', Name => 'ChemicalShiftsMinIntegrateLimitHz' },
    '0018,9196' => { VR => 'FD', Name => 'ChemicalShiftsMaxIntegrateLimitHz' },
    '0018,9197' => { VR => 'SQ', Name => 'MRVelocityEncodingSequence' },
    '0018,9198' => { VR => 'CS', Name => 'FirstOrderPhaseCorrection' },
    '0018,9199' => { VR => 'CS', Name => 'WaterReferencedPhaseCorrection' },
    '0018,9200' => { VR => 'CS', Name => 'MRSpectroscopyAcquisitionType' },
    '0018,9214' => { VR => 'CS', Name => 'RespiratoryCyclePosition' },
    '0018,9217' => { VR => 'FD', Name => 'VelocityEncodingMaximumValue' },
    '0018,9218' => { VR => 'FD', Name => 'TagSpacingSecondDimension' },
    '0018,9219' => { VR => 'SS', Name => 'TagAngleSecondAxis' },
    '0018,9220' => { VR => 'FD', Name => 'FrameAcquisitionDuration' },
    '0018,9226' => { VR => 'SQ', Name => 'MRImageFrameTypeSequence' },
    '0018,9227' => { VR => 'SQ', Name => 'MRSpectroscopyFrameTypeSequence' },
    '0018,9231' => { VR => 'US', Name => 'MRAcqPhaseEncodingStepsInPlane' },
    '0018,9232' => { VR => 'US', Name => 'MRAcqPhaseEncodingStepsOutOfPlane' },
    '0018,9234' => { VR => 'UL', Name => 'SpectroscopyAcqPhaseColumns' },
    '0018,9236' => { VR => 'CS', Name => 'CardiacCyclePosition' },
    '0018,9239' => { VR => 'SQ', Name => 'SpecificAbsorptionRateSequence' },
    '0018,9240' => { VR => 'US', Name => 'RFEchoTrainLength' },
    '0018,9241' => { VR => 'US', Name => 'GradientEchoTrainLength' },
    '0018,9250' => { VR => 'CS', Name => 'ArterialSpinLabelingContrast' },
    '0018,9251' => { VR => 'SQ', Name => 'MRArterialSpinLabelingSequence' },
    '0018,9252' => { VR => 'LO', Name => 'ASLTechniqueDescription' },
    '0018,9253' => { VR => 'US', Name => 'ASLSlabNumber' },
    '0018,9254' => { VR => 'FD', Name => 'ASLSlabThickness' },
    '0018,9255' => { VR => 'FD', Name => 'ASLSlabOrientation' },
    '0018,9256' => { VR => 'FD', Name => 'ASLMidSlabPosition' },
    '0018,9257' => { VR => 'CS', Name => 'ASLContext' },
    '0018,9258' => { VR => 'UL', Name => 'ASLPulseTrainDuration' },
    '0018,9259' => { VR => 'CS', Name => 'ASLCrusherFlag' },
    '0018,925A' => { VR => 'FD', Name => 'ASLCrusherFlowLimit' },
    '0018,925B' => { VR => 'LO', Name => 'ASLCrusherDescription' },
    '0018,925C' => { VR => 'CS', Name => 'ASLBolusCutoffFlag' },
    '0018,925D' => { VR => 'SQ', Name => 'ASLBolusCutoffTimingSequence' },
    '0018,925E' => { VR => 'LO', Name => 'ASLBolusCutoffTechnique' },
    '0018,925F' => { VR => 'UL', Name => 'ASLBolusCutoffDelayTime' },
    '0018,9260' => { VR => 'SQ', Name => 'ASLSlabSequence' },
    '0018,9295' => { VR => 'FD', Name => 'ChemicalShiftsMinIntegrateLimitPPM' },
    '0018,9296' => { VR => 'FD', Name => 'ChemicalShiftsMaxIntegrateLimitPPM' },
    '0018,9297' => { VR => 'CS', Name => 'WaterReferenceAcquisition' },
    '0018,9298' => { VR => 'IS', Name => 'EchoPeakPosition' },
    '0018,9301' => { VR => 'SQ', Name => 'CTAcquisitionTypeSequence' },
    '0018,9302' => { VR => 'CS', Name => 'AcquisitionType' },
    '0018,9303' => { VR => 'FD', Name => 'TubeAngle' },
    '0018,9304' => { VR => 'SQ', Name => 'CTAcquisitionDetailsSequence' },
    '0018,9305' => { VR => 'FD', Name => 'RevolutionTime' },
    '0018,9306' => { VR => 'FD', Name => 'SingleCollimationWidth' },
    '0018,9307' => { VR => 'FD', Name => 'TotalCollimationWidth' },
    '0018,9308' => { VR => 'SQ', Name => 'CTTableDynamicsSequence' },
    '0018,9309' => { VR => 'FD', Name => 'TableSpeed' },
    '0018,9310' => { VR => 'FD', Name => 'TableFeedPerRotation' },
    '0018,9311' => { VR => 'FD', Name => 'SpiralPitchFactor' },
    '0018,9312' => { VR => 'SQ', Name => 'CTGeometrySequence' },
    '0018,9313' => { VR => 'FD', Name => 'DataCollectionCenterPatient' },
    '0018,9314' => { VR => 'SQ', Name => 'CTReconstructionSequence' },
    '0018,9315' => { VR => 'CS', Name => 'ReconstructionAlgorithm' },
    '0018,9316' => { VR => 'CS', Name => 'ConvolutionKernelGroup' },
    '0018,9317' => { VR => 'FD', Name => 'ReconstructionFieldOfView' },
    '0018,9318' => { VR => 'FD', Name => 'ReconstructionTargetCenterPatient' },
    '0018,9319' => { VR => 'FD', Name => 'ReconstructionAngle' },
    '0018,9320' => { VR => 'SH', Name => 'ImageFilter' },
    '0018,9321' => { VR => 'SQ', Name => 'CTExposureSequence' },
    '0018,9322' => { VR => 'FD', Name => 'ReconstructionPixelSpacing' },
    '0018,9323' => { VR => 'CS', Name => 'ExposureModulationType' },
    '0018,9324' => { VR => 'FD', Name => 'EstimatedDoseSaving' },
    '0018,9325' => { VR => 'SQ', Name => 'CTXRayDetailsSequence' },
    '0018,9326' => { VR => 'SQ', Name => 'CTPositionSequence' },
    '0018,9327' => { VR => 'FD', Name => 'TablePosition' },
    '0018,9328' => { VR => 'FD', Name => 'ExposureTimeInMilliSec' },
    '0018,9329' => { VR => 'SQ', Name => 'CTImageFrameTypeSequence' },
    '0018,9330' => { VR => 'FD', Name => 'XRayTubeCurrentInMilliAmps' },
    '0018,9332' => { VR => 'FD', Name => 'ExposureInMilliAmpSec' },
    '0018,9333' => { VR => 'CS', Name => 'ConstantVolumeFlag' },
    '0018,9334' => { VR => 'CS', Name => 'FluoroscopyFlag' },
    '0018,9335' => { VR => 'FD', Name => 'SourceToDataCollectionCenterDist' },
    '0018,9337' => { VR => 'US', Name => 'ContrastBolusAgentNumber' },
    '0018,9338' => { VR => 'SQ', Name => 'ContrastBolusIngredientCodeSeq' },
    '0018,9340' => { VR => 'SQ', Name => 'ContrastAdministrationProfileSeq' },
    '0018,9341' => { VR => 'SQ', Name => 'ContrastBolusUsageSequence' },
    '0018,9342' => { VR => 'CS', Name => 'ContrastBolusAgentAdministered' },
    '0018,9343' => { VR => 'CS', Name => 'ContrastBolusAgentDetected' },
    '0018,9344' => { VR => 'CS', Name => 'ContrastBolusAgentPhase' },
    '0018,9345' => { VR => 'FD', Name => 'CTDIvol' },
    '0018,9346' => { VR => 'SQ', Name => 'CTDIPhantomTypeCodeSequence' },
    '0018,9351' => { VR => 'FL', Name => 'CalciumScoringMassFactorPatient' },
    '0018,9352' => { VR => 'FL', Name => 'CalciumScoringMassFactorDevice' },
    '0018,9353' => { VR => 'FL', Name => 'EnergyWeightingFactor' },
    '0018,9360' => { VR => 'SQ', Name => 'CTAdditionalXRaySourceSequence' },
    '0018,9361' => { VR => 'CS', Name => 'MultienergyCTAcquisition' },
    '0018,9362' => { VR => 'SQ', Name => 'MultienergyCTAcquisitionSequence' },
    '0018,9363' => { VR => 'SQ', Name => 'MultienergyCTProcessingSequence' },
    '0018,9364' => { VR => 'SQ', Name => 'MultienergyCTCharacteristicsSeq' },
    '0018,9365' => { VR => 'SQ', Name => 'MultienergyCTXRaySourceSequence' },
    '0018,9366' => { VR => 'US', Name => 'XRaySourceIndex' },
    '0018,9367' => { VR => 'UC', Name => 'XRaySourceID' },
    '0018,9368' => { VR => 'CS', Name => 'MultienergySourceTechnique' },
    '0018,9369' => { VR => 'DT', Name => 'SourceStartDateTime' },
    '0018,936A' => { VR => 'DT', Name => 'SourceEndDateTime' },
    '0018,936B' => { VR => 'US', Name => 'SwitchingPhaseNumber' },
    '0018,936C' => { VR => 'DS', Name => 'SwitchingPhaseNominalDuration' },
    '0018,936D' => { VR => 'DS', Name => 'SwitchingPhaseTransitionDuration' },
    '0018,936E' => { VR => 'DS', Name => 'EffectiveBinEnergy' },
    '0018,936F' => { VR => 'SQ', Name => 'MultienergyCTXRayDetectorSequence' },
    '0018,9370' => { VR => 'US', Name => 'XRayDetectorIndex' },
    '0018,9371' => { VR => 'UC', Name => 'XRayDetectorID' },
    '0018,9372' => { VR => 'CS', Name => 'MultienergyDetectorType' },
    '0018,9373' => { VR => 'ST', Name => 'XRayDetectorLabel' },
    '0018,9374' => { VR => 'DS', Name => 'NominalMaxEnergy' },
    '0018,9375' => { VR => 'DS', Name => 'NominalMinEnergy' },
    '0018,9376' => { VR => 'US', Name => 'ReferencedXRayDetectorIndex' },
    '0018,9377' => { VR => 'US', Name => 'ReferencedXRaySourceIndex' },
    '0018,9378' => { VR => 'US', Name => 'ReferencedPathIndex' },
    '0018,9379' => { VR => 'SQ', Name => 'MultienergyCTPathSequence' },
    '0018,937A' => { VR => 'US', Name => 'MultienergyCTPathIndex' },
    '0018,937B' => { VR => 'UT', Name => 'MultienergyAcquisitionDescription' },
    '0018,937C' => { VR => 'FD', Name => 'MonoenergeticEnergyEquivalent' },
    '0018,937D' => { VR => 'SQ', Name => 'MaterialCodeSequence' },
    '0018,937E' => { VR => 'CS', Name => 'DecompositionMethod' },
    '0018,937F' => { VR => 'UT', Name => 'DecompositionDescription' },
    '0018,9380' => { VR => 'SQ', Name => 'DecompositionAlgorithmIDSeq' },
    '0018,9381' => { VR => 'SQ', Name => 'DecompositionMaterialSequence' },
    '0018,9382' => { VR => 'SQ', Name => 'MaterialAttenuationSequence' },
    '0018,9383' => { VR => 'DS', Name => 'PhotonEnergy' },
    '0018,9384' => { VR => 'DS', Name => 'XRayMassAttenuationCoefficient' },
    '0018,9390' => { VR => 'SQ', Name => 'MetalArtifactReductionSequence' },
    '0018,9391' => { VR => 'CS', Name => 'MetalArtifactReductionApplied' },
    '0018,9392' => { VR => 'SQ', Name => 'MetalArtifactReductAlgorithmIDSeq' },
    '0018,9401' => { VR => 'SQ', Name => 'ProjectionPixelCalibrationSequence' },
    '0018,9402' => { VR => 'FL', Name => 'DistanceSourceToIsocenter' },
    '0018,9403' => { VR => 'FL', Name => 'DistanceObjectToTableTop' },
    '0018,9404' => { VR => 'FL', Name => 'ObjectPixelSpacingInCenterOfBeam' },
    '0018,9405' => { VR => 'SQ', Name => 'PositionerPositionSequence' },
    '0018,9406' => { VR => 'SQ', Name => 'TablePositionSequence' },
    '0018,9407' => { VR => 'SQ', Name => 'CollimatorShapeSequence' },
    '0018,9410' => { VR => 'CS', Name => 'PlanesInAcquisition' },
    '0018,9412' => { VR => 'SQ', Name => 'XA-XRFFrameCharacteristicsSequence' },
    '0018,9417' => { VR => 'SQ', Name => 'FrameAcquisitionSequence' },
    '0018,9420' => { VR => 'CS', Name => 'XRayReceptorType' },
    '0018,9423' => { VR => 'LO', Name => 'AcquisitionProtocolName' },
    '0018,9424' => { VR => 'LT', Name => 'AcquisitionProtocolDescription' },
    '0018,9425' => { VR => 'CS', Name => 'ContrastBolusIngredientOpaque' },
    '0018,9426' => { VR => 'FL', Name => 'DistanceReceptorPlaneToDetHousing' },
    '0018,9427' => { VR => 'CS', Name => 'IntensifierActiveShape' },
    '0018,9428' => { VR => 'FL', Name => 'IntensifierActiveDimensions' },
    '0018,9429' => { VR => 'FL', Name => 'PhysicalDetectorSize' },
    '0018,9430' => { VR => 'US', Name => 'PositionOfIsocenterProjection' },
    '0018,9432' => { VR => 'SQ', Name => 'FieldOfViewSequence' },
    '0018,9433' => { VR => 'LO', Name => 'FieldOfViewDescription' },
    '0018,9434' => { VR => 'SQ', Name => 'ExposureControlSensingRegionsSeq' },
    '0018,9435' => { VR => 'CS', Name => 'ExposureControlSensingRegionShape' },
    '0018,9436' => { VR => 'SS', Name => 'ExposureControlSensRegionLeftEdge' },
    '0018,9437' => { VR => 'SS', Name => 'ExposureControlSensRegionRightEdge' },
    '0018,9438' => { VR => 'SS', Name => 'ExpControlSensingRgnUpperHorizEdge' },
    '0018,9439' => { VR => 'SS', Name => 'ExpControlSensingRgnLowerHorizEdge' },
    '0018,9440' => { VR => 'SS', Name => 'CenterOfCircExposControlSensRegion' },
    '0018,9441' => { VR => 'US', Name => 'RadiusOfCircExposControlSensRegion' },
    '0018,9442' => { VR => 'SS', Name => 'VerticesOfPolyExpControlSensingRgn' },
    '0018,9447' => { VR => 'FL', Name => 'ColumnAngulationPatient' },
    '0018,9449' => { VR => 'FL', Name => 'BeamAngle' },
    '0018,9451' => { VR => 'SQ', Name => 'FrameDetectorParametersSequence' },
    '0018,9452' => { VR => 'FL', Name => 'CalculatedAnatomyThickness' },
    '0018,9455' => { VR => 'SQ', Name => 'CalibrationSequence' },
    '0018,9456' => { VR => 'SQ', Name => 'ObjectThicknessSequence' },
    '0018,9457' => { VR => 'CS', Name => 'PlaneIdentification' },
    '0018,9461' => { VR => 'FL', Name => 'FieldOfViewDimensionsInFloat' },
    '0018,9462' => { VR => 'SQ', Name => 'IsocenterReferenceSystemSequence' },
    '0018,9463' => { VR => 'FL', Name => 'PositionerIsocenterPrimaryAngle' },
    '0018,9464' => { VR => 'FL', Name => 'PositionerIsocenterSecondaryAngle' },
    '0018,9465' => { VR => 'FL', Name => 'PositionerIsocenterDetRotAngle' },
    '0018,9466' => { VR => 'FL', Name => 'TableXPositionToIsocenter' },
    '0018,9467' => { VR => 'FL', Name => 'TableYPositionToIsocenter' },
    '0018,9468' => { VR => 'FL', Name => 'TableZPositionToIsocenter' },
    '0018,9469' => { VR => 'FL', Name => 'TableHorizontalRotationAngle' },
    '0018,9470' => { VR => 'FL', Name => 'TableHeadTiltAngle' },
    '0018,9471' => { VR => 'FL', Name => 'TableCradleTiltAngle' },
    '0018,9472' => { VR => 'SQ', Name => 'FrameDisplayShutterSequence' },
    '0018,9473' => { VR => 'FL', Name => 'AcquiredImageAreaDoseProduct' },
    '0018,9474' => { VR => 'CS', Name => 'CArmPositionerTabletopRelationship' },
    '0018,9476' => { VR => 'SQ', Name => 'XRayGeometrySequence' },
    '0018,9477' => { VR => 'SQ', Name => 'IrradiationEventIDSequence' },
    '0018,9504' => { VR => 'SQ', Name => 'XRay3DFrameTypeSequence' },
    '0018,9506' => { VR => 'SQ', Name => 'ContributingSourcesSequence' },
    '0018,9507' => { VR => 'SQ', Name => 'XRay3DAcquisitionSequence' },
    '0018,9508' => { VR => 'FL', Name => 'PrimaryPositionerScanArc' },
    '0018,9509' => { VR => 'FL', Name => 'SecondaryPositionerScanArc' },
    '0018,9510' => { VR => 'FL', Name => 'PrimaryPositionerScanStartAngle' },
    '0018,9511' => { VR => 'FL', Name => 'SecondaryPositionerScanStartAngle' },
    '0018,9514' => { VR => 'FL', Name => 'PrimaryPositionerIncrement' },
    '0018,9515' => { VR => 'FL', Name => 'SecondaryPositionerIncrement' },
    '0018,9516' => { VR => 'DT', Name => 'StartAcquisitionDateTime' },
    '0018,9517' => { VR => 'DT', Name => 'EndAcquisitionDateTime' },
    '0018,9518' => { VR => 'SS', Name => 'PrimaryPositionerIncrementSign' },
    '0018,9519' => { VR => 'SS', Name => 'SecondaryPositionerIncrementSign' },
    '0018,9524' => { VR => 'LO', Name => 'ApplicationName' },
    '0018,9525' => { VR => 'LO', Name => 'ApplicationVersion' },
    '0018,9526' => { VR => 'LO', Name => 'ApplicationManufacturer' },
    '0018,9527' => { VR => 'CS', Name => 'AlgorithmType' },
    '0018,9528' => { VR => 'LO', Name => 'AlgorithmDescription' },
    '0018,9530' => { VR => 'SQ', Name => 'XRay3DReconstructionSequence' },
    '0018,9531' => { VR => 'LO', Name => 'ReconstructionDescription' },
    '0018,9538' => { VR => 'SQ', Name => 'PerProjectionAcquisitionSequence' },
    '0018,9541' => { VR => 'SQ', Name => 'DetectorPositionSequence' },
    '0018,9542' => { VR => 'SQ', Name => 'XRayAcquisitionDoseSequence' },
    '0018,9543' => { VR => 'FD', Name => 'XRaySourceIsocenterPrimaryAngle' },
    '0018,9544' => { VR => 'FD', Name => 'XRaySourceIsocenterSecondaryAngle' },
    '0018,9545' => { VR => 'FD', Name => 'BreastSupportIsocenterPrimaryAngle' },
    '0018,9546' => { VR => 'FD', Name => 'BreastSupportIsocenterSecondAngle' },
    '0018,9547' => { VR => 'FD', Name => 'BreastSupportXPositionToIsocenter' },
    '0018,9548' => { VR => 'FD', Name => 'BreastSupportYPositionToIsocenter' },
    '0018,9549' => { VR => 'FD', Name => 'BreastSupportZPositionToIsocenter' },
    '0018,9550' => { VR => 'FD', Name => 'DetectorIsocenterPrimaryAngle' },
    '0018,9551' => { VR => 'FD', Name => 'DetectorIsocenterSecondaryAngle' },
    '0018,9552' => { VR => 'FD', Name => 'DetectorXPositionToIsocenter' },
    '0018,9553' => { VR => 'FD', Name => 'DetectorYPositionToIsocenter' },
    '0018,9554' => { VR => 'FD', Name => 'DetectorZPositionToIsocenter' },
    '0018,9555' => { VR => 'SQ', Name => 'XRayGridSequence' },
    '0018,9556' => { VR => 'SQ', Name => 'XRayFilterSequence' },
    '0018,9557' => { VR => 'FD', Name => 'DetectorActiveAreaTLHCPosition' },
    '0018,9558' => { VR => 'FD', Name => 'DetectorActiveAreaOrientation' },
    '0018,9559' => { VR => 'CS', Name => 'PositionerPrimaryAngleDirection' },
    '0018,9601' => { VR => 'SQ', Name => 'DiffusionBMatrixSequence' },
    '0018,9602' => { VR => 'FD', Name => 'DiffusionBValueXX' },
    '0018,9603' => { VR => 'FD', Name => 'DiffusionBValueXY' },
    '0018,9604' => { VR => 'FD', Name => 'DiffusionBValueXZ' },
    '0018,9605' => { VR => 'FD', Name => 'DiffusionBValueYY' },
    '0018,9606' => { VR => 'FD', Name => 'DiffusionBValueYZ' },
    '0018,9607' => { VR => 'FD', Name => 'DiffusionBValueZZ' },
    '0018,9621' => { VR => 'SQ', Name => 'FunctionalMRSequence' },
    '0018,9622' => { VR => 'CS', Name => 'FunctSettlingPhaseFramesPresent' },
    '0018,9623' => { VR => 'DT', Name => 'FunctionalSyncPulse' },
    '0018,9624' => { VR => 'CS', Name => 'SettlingPhaseFrame' },
    '0018,9701' => { VR => 'DT', Name => 'DecayCorrectionDateTime' },
    '0018,9715' => { VR => 'FD', Name => 'StartDensityThreshold' },
    '0018,9716' => { VR => 'FD', Name => 'StartRelativeDensityDiffThresh' },
    '0018,9717' => { VR => 'FD', Name => 'StartCardiacTriggerCountThreshold' },
    '0018,9718' => { VR => 'FD', Name => 'StartRespiratoryTriggerCountThresh' },
    '0018,9719' => { VR => 'FD', Name => 'TerminationCountsThreshold' },
    '0018,9720' => { VR => 'FD', Name => 'TerminationDensityThreshold' },
    '0018,9721' => { VR => 'FD', Name => 'TerminationRelativeDensityThresh' },
    '0018,9722' => { VR => 'FD', Name => 'TerminationTimeThreshold' },
    '0018,9723' => { VR => 'FD', Name => 'TerminationCardiacTrigCountThresh' },
    '0018,9724' => { VR => 'FD', Name => 'TermRespiratoryTrigCountThresh' },
    '0018,9725' => { VR => 'CS', Name => 'DetectorGeometry' },
    '0018,9726' => { VR => 'FD', Name => 'TransverseDetectorSeparation' },
    '0018,9727' => { VR => 'FD', Name => 'AxialDetectorDimension' },
    '0018,9729' => { VR => 'US', Name => 'RadiopharmaceuticalAgentNumber' },
    '0018,9732' => { VR => 'SQ', Name => 'PETFrameAcquisitionSequence' },
    '0018,9733' => { VR => 'SQ', Name => 'PETDetectorMotionDetailsSequence' },
    '0018,9734' => { VR => 'SQ', Name => 'PETTableDynamicsSequence' },
    '0018,9735' => { VR => 'SQ', Name => 'PETPositionSequence' },
    '0018,9736' => { VR => 'SQ', Name => 'PETFrameCorrectionFactorsSequence' },
    '0018,9737' => { VR => 'SQ', Name => 'RadiopharmaceuticalUsageSequence' },
    '0018,9738' => { VR => 'CS', Name => 'AttenuationCorrectionSource' },
    '0018,9739' => { VR => 'US', Name => 'NumberOfIterations' },
    '0018,9740' => { VR => 'US', Name => 'NumberOfSubsets' },
    '0018,9749' => { VR => 'SQ', Name => 'PETReconstructionSequence' },
    '0018,9751' => { VR => 'SQ', Name => 'PETFrameTypeSequence' },
    '0018,9755' => { VR => 'CS', Name => 'TimeOfFlightInformationUsed' },
    '0018,9756' => { VR => 'CS', Name => 'ReconstructionType' },
    '0018,9758' => { VR => 'CS', Name => 'DecayCorrected' },
    '0018,9759' => { VR => 'CS', Name => 'AttenuationCorrected' },
    '0018,9760' => { VR => 'CS', Name => 'ScatterCorrected' },
    '0018,9761' => { VR => 'CS', Name => 'DeadTimeCorrected' },
    '0018,9762' => { VR => 'CS', Name => 'GantryMotionCorrected' },
    '0018,9763' => { VR => 'CS', Name => 'PatientMotionCorrected' },
    '0018,9764' => { VR => 'CS', Name => 'CountLossNormalizationCorrected' },
    '0018,9765' => { VR => 'CS', Name => 'RandomsCorrected' },
    '0018,9766' => { VR => 'CS', Name => 'NonUniformRadialSamplingCorrected' },
    '0018,9767' => { VR => 'CS', Name => 'SensitivityCalibrated' },
    '0018,9768' => { VR => 'CS', Name => 'DetectorNormalizationCorrection' },
    '0018,9769' => { VR => 'CS', Name => 'IterativeReconstructionMethod' },
    '0018,9770' => { VR => 'CS', Name => 'AttenCorrTemporalRelationship' },
    '0018,9771' => { VR => 'SQ', Name => 'PatientPhysiologicalStateSequence' },
    '0018,9772' => { VR => 'SQ', Name => 'PatientPhysiologicalStateCodeSeq' },
    '0018,9801' => { VR => 'FD', Name => 'DepthsOfFocus' },
    '0018,9803' => { VR => 'SQ', Name => 'ExcludedIntervalsSequence' },
    '0018,9804' => { VR => 'DT', Name => 'ExclusionStartDatetime' },
    '0018,9805' => { VR => 'FD', Name => 'ExclusionDuration' },
    '0018,9806' => { VR => 'SQ', Name => 'USImageDescriptionSequence' },
    '0018,9807' => { VR => 'SQ', Name => 'ImageDataTypeSequence' },
    '0018,9808' => { VR => 'CS', Name => 'DataType' },
    '0018,9809' => { VR => 'SQ', Name => 'TransducerScanPatternCodeSequence' },
    '0018,980B' => { VR => 'CS', Name => 'AliasedDataType' },
    '0018,980C' => { VR => 'CS', Name => 'PositionMeasuringDeviceUsed' },
    '0018,980D' => { VR => 'SQ', Name => 'TransducerGeometryCodeSequence' },
    '0018,980E' => { VR => 'SQ', Name => 'TransducerBeamSteeringCodeSequence' },
    '0018,980F' => { VR => 'SQ', Name => 'TransducerApplicationCodeSequence' },
    '0018,9810' => { VR => 'SS', Name => 'ZeroVelocityPixelValue' },
    '0018,9821' => { VR => 'SQ', Name => 'PhotoacousticExcitationCharSeq' },
    '0018,9822' => { VR => 'FD', Name => 'ExcitationSpectralWidth' },
    '0018,9823' => { VR => 'FD', Name => 'ExcitationEnergy' },
    '0018,9824' => { VR => 'FD', Name => 'ExcitationPulseDuration' },
    '0018,9825' => { VR => 'SQ', Name => 'ExcitationWavelengthSequence' },
    '0018,9826' => { VR => 'FD', Name => 'ExcitationWavelength' },
    '0018,9828' => { VR => 'CS', Name => 'IlluminationTranslationFlag' },
    '0018,9829' => { VR => 'CS', Name => 'AcousticCouplingMediumFlag' },
    '0018,982A' => { VR => 'SQ', Name => 'AcousticCouplingMediumCodeSequence' },
    '0018,982B' => { VR => 'FD', Name => 'AcousticCouplingMediumTemperature' },
    '0018,982C' => { VR => 'SQ', Name => 'TransducerResponseSequence' },
    '0018,982D' => { VR => 'FD', Name => 'CenterFrequency' },
    '0018,982E' => { VR => 'FD', Name => 'FractionalBandwidth' },
    '0018,982F' => { VR => 'FD', Name => 'LowerCutoffFrequency' },
    '0018,9830' => { VR => 'FD', Name => 'UpperCutoffFrequency' },
    '0018,9831' => { VR => 'SQ', Name => 'TransducerTechnologySequence' },
    '0018,9832' => { VR => 'SQ', Name => 'SoundSpeedCorrMechanismCodeSeq' },
    '0018,9833' => { VR => 'FD', Name => 'ObjectSoundSpeed' },
    '0018,9834' => { VR => 'FD', Name => 'AcousticCouplingMediumSoundSpeed' },
    '0018,9835' => { VR => 'SQ', Name => 'PhotoacousticImageFrameTypeSeq' },
    '0018,9836' => { VR => 'SQ', Name => 'ImageDataTypeCodeSequence' },
    '0018,9900' => { VR => 'LO', Name => 'ReferenceLocationLabel' },
    '0018,9901' => { VR => 'UT', Name => 'ReferenceLocationDescription' },
    '0018,9902' => { VR => 'SQ', Name => 'ReferenceBasisCodeSequence' },
    '0018,9903' => { VR => 'SQ', Name => 'ReferenceGeometryCodeSequence' },
    '0018,9904' => { VR => 'DS', Name => 'OffsetDistance' },
    '0018,9905' => { VR => 'CS', Name => 'OffsetDirection' },
    '0018,9906' => { VR => 'SQ', Name => 'PotentialScheduledProtocolCodeSeq' },
    '0018,9907' => { VR => 'SQ', Name => 'PotentialRequestedProcedureCodeSeq' },
    '0018,9908' => { VR => 'UC', Name => 'PotentialReasonsForProcedure' },
    '0018,9909' => { VR => 'SQ', Name => 'PotentialReasonsForProcCodeSeq' },
    '0018,990A' => { VR => 'UC', Name => 'PotentialDiagnosticTasks' },
    '0018,990B' => { VR => 'SQ', Name => 'ContraindicationsCodeSequence' },
    '0018,990C' => { VR => 'SQ', Name => 'ReferencedDefinedProtocolSequence' },
    '0018,990D' => { VR => 'SQ', Name => 'ReferencedPerformedProtocolSeq' },
    '0018,990E' => { VR => 'SQ', Name => 'PredecessorProtocolSequence' },
    '0018,990F' => { VR => 'UT', Name => 'ProtocolPlanningInformation' },
    '0018,9910' => { VR => 'UT', Name => 'ProtocolDesignRationale' },
    '0018,9911' => { VR => 'SQ', Name => 'PatientSpecificationSequence' },
    '0018,9912' => { VR => 'SQ', Name => 'ModelSpecificationSequence' },
    '0018,9913' => { VR => 'SQ', Name => 'ParametersSpecificationSequence' },
    '0018,9914' => { VR => 'SQ', Name => 'InstructionSequence' },
    '0018,9915' => { VR => 'US', Name => 'InstructionIndex' },
    '0018,9916' => { VR => 'LO', Name => 'InstructionText' },
    '0018,9917' => { VR => 'UT', Name => 'InstructionDescription' },
    '0018,9918' => { VR => 'CS', Name => 'InstructionPerformedFlag' },
    '0018,9919' => { VR => 'DT', Name => 'InstructionPerformedDateTime' },
    '0018,991A' => { VR => 'UT', Name => 'InstructionPerformanceComment' },
    '0018,991B' => { VR => 'SQ', Name => 'PatientPositioningInstructionSeq' },
    '0018,991C' => { VR => 'SQ', Name => 'PositioningMethodCodeSequence' },
    '0018,991D' => { VR => 'SQ', Name => 'PositioningLandmarkSequence' },
    '0018,991E' => { VR => 'UI', Name => 'TargetFrameOfReferenceUID' },
    '0018,991F' => { VR => 'SQ', Name => 'AcquisitionProtocolElementSpecSeq' },
    '0018,9920' => { VR => 'SQ', Name => 'AcquisitionProtocolElementSequence' },
    '0018,9921' => { VR => 'US', Name => 'ProtocolElementNumber' },
    '0018,9922' => { VR => 'LO', Name => 'ProtocolElementName' },
    '0018,9923' => { VR => 'UT', Name => 'ProtocolElementCharSummary' },
    '0018,9924' => { VR => 'UT', Name => 'ProtocolElementPurpose' },
    '0018,9930' => { VR => 'CS', Name => 'AcquisitionMotion' },
    '0018,9931' => { VR => 'SQ', Name => 'AcquisitionStartLocationSequence' },
    '0018,9932' => { VR => 'SQ', Name => 'AcquisitionEndLocationSequence' },
    '0018,9933' => { VR => 'SQ', Name => 'ReconstructionProtoElementSpecSeq' },
    '0018,9934' => { VR => 'SQ', Name => 'ReconstructionProtocolElementSeq' },
    '0018,9935' => { VR => 'SQ', Name => 'StorageProtocolElementSpecSequence' },
    '0018,9936' => { VR => 'SQ', Name => 'StorageProtocolElementSequence' },
    '0018,9937' => { VR => 'LO', Name => 'RequestedSeriesDescription' },
    '0018,9938' => { VR => 'US', Name => 'SourceAcquisitionProtoElementNum' },
    '0018,9939' => { VR => 'US', Name => 'SourceAcquisitionBeamNumber' },
    '0018,993A' => { VR => 'US', Name => 'SourceReconstructProtoElementNum' },
    '0018,993B' => { VR => 'SQ', Name => 'ReconstructionStartLocationSeq' },
    '0018,993C' => { VR => 'SQ', Name => 'ReconstructionEndLocationSequence' },
    '0018,993D' => { VR => 'SQ', Name => 'ReconstructionAlgorithmSequence' },
    '0018,993E' => { VR => 'SQ', Name => 'ReconstructionTargetCenterLocSeq' },
    '0018,9941' => { VR => 'UT', Name => 'ImageFilterDescription' },
    '0018,9942' => { VR => 'FD', Name => 'CTDIvolNotificationTrigger' },
    '0018,9943' => { VR => 'FD', Name => 'DLPNotificationTrigger' },
    '0018,9944' => { VR => 'CS', Name => 'AutoKVPSelectionType' },
    '0018,9945' => { VR => 'FD', Name => 'AutoKVPUpperBound' },
    '0018,9946' => { VR => 'FD', Name => 'AutoKVPLowerBound' },
    '0018,9947' => { VR => 'CS', Name => 'ProtocolDefinedPatientPosition' },
    '0018,A001' => { VR => 'SQ', Name => 'ContributingEquipmentSequence' },
    '0018,A002' => { VR => 'DT', Name => 'ContributionDateTime' },
    '0018,A003' => { VR => 'ST', Name => 'ContributionDescription' },
    # GEMS_ACQU_01 (ref 4)
    '0019,1002' => { VR => 'SL', Name => 'NumberOfCellsIInDetector' },
    '0019,1003' => { VR => 'DS', Name => 'CellNumberAtTheta' },
    '0019,1004' => { VR => 'DS', Name => 'CellSpacing' },
    '0019,100F' => { VR => 'DS', Name => 'HorizFrameOfRef' },
    '0019,1011' => { VR => 'SS', Name => 'SeriesContrast' },
    '0019,1012' => { VR => 'SS', Name => 'LastPseq' },
    '0019,1013' => { VR => 'SS', Name => 'StartNumberForBaseline' },
    '0019,1014' => { VR => 'SS', Name => 'EndNumberForBaseline' },
    '0019,1015' => { VR => 'SS', Name => 'StartNumberForEnhancedScans' },
    '0019,1016' => { VR => 'SS', Name => 'EndNumberForEnhancedScans' },
    '0019,1017' => { VR => 'SS', Name => 'SeriesPlane' },
    '0019,1018' => { VR => 'LO', Name => 'FirstScanRas' },
    '0019,1019' => { VR => 'DS', Name => 'FirstScanLocation' },
    '0019,101A' => { VR => 'LO', Name => 'LastScanRas' },
    '0019,101B' => { VR => 'DS', Name => 'LastScanLoc' },
    '0019,101E' => { VR => 'DS', Name => 'DisplayFieldOfView' },
    '0019,1023' => { VR => 'DS', Name => 'TableSpeed' },
    '0019,1024' => { VR => 'DS', Name => 'MidScanTime' },
    '0019,1025' => { VR => 'SS', Name => 'MidScanFlag' },
    '0019,1026' => { VR => 'SL', Name => 'DegreesOfAzimuth' },
    '0019,1027' => { VR => 'DS', Name => 'GantryPeriod' },
    '0019,102A' => { VR => 'DS', Name => 'XRayOnPosition' },
    '0019,102B' => { VR => 'DS', Name => 'XRayOffPosition' },
    '0019,102C' => { VR => 'SL', Name => 'NumberOfTriggers' },
    '0019,102E' => { VR => 'DS', Name => 'AngleOfFirstView' },
    '0019,102F' => { VR => 'DS', Name => 'TriggerFrequency' },
    '0019,1039' => { VR => 'SS', Name => 'ScanFOVType' },
    '0019,1040' => { VR => 'SS', Name => 'StatReconFlag' },
    '0019,1041' => { VR => 'SS', Name => 'ComputeType' },
    '0019,1042' => { VR => 'SS', Name => 'SegmentNumber' },
    '0019,1043' => { VR => 'SS', Name => 'TotalSegmentsRequested' },
    '0019,1044' => { VR => 'DS', Name => 'InterscanDelay' },
    '0019,1047' => { VR => 'SS', Name => 'ViewCompressionFactor' },
    '0019,104A' => { VR => 'SS', Name => 'TotalNoOfRefChannels' },
    '0019,104B' => { VR => 'SL', Name => 'DataSizeForScanData' },
    '0019,1052' => { VR => 'SS', Name => 'ReconPostProcflag' },
    '0019,1057' => { VR => 'SS', Name => 'CTWaterNumber' },
    '0019,1058' => { VR => 'SS', Name => 'CTBoneNumber' },
    '0019,105A' => { VR => 'FL', Name => 'AcquisitionDuration' },
    '0019,105E' => { VR => 'SL', Name => 'NumberOfChannels' },
    '0019,105F' => { VR => 'SL', Name => 'IncrementBetweenChannels' },
    '0019,1060' => { VR => 'SL', Name => 'StartingView' },
    '0019,1061' => { VR => 'SL', Name => 'NumberOfViews' },
    '0019,1062' => { VR => 'SL', Name => 'IncrementBetweenViews' },
    '0019,106A' => { VR => 'SS', Name => 'DependantOnNoViewsProcessed' },
    '0019,106B' => { VR => 'SS', Name => 'FieldOfViewInDetectorCells' },
    '0019,1070' => { VR => 'SS', Name => 'ValueOfBackProjectionButton' },
    '0019,1071' => { VR => 'SS', Name => 'SetIfFatqEstimatesWereUsed' },
    '0019,1072' => { VR => 'DS', Name => 'ZChanAvgOverViews' },
    '0019,1073' => { VR => 'DS', Name => 'AvgOfLeftRefChansOverViews' },
    '0019,1074' => { VR => 'DS', Name => 'MaxLeftChanOverViews' },
    '0019,1075' => { VR => 'DS', Name => 'AvgOfRightRefChansOverViews' },
    '0019,1076' => { VR => 'DS', Name => 'MaxRightChanOverViews' },
    '0019,107D' => { VR => 'DS', Name => 'SecondEcho' },
    '0019,107E' => { VR => 'SS', Name => 'NumberOfEchoes' },
    '0019,107F' => { VR => 'DS', Name => 'TableDelta' },
    '0019,1081' => { VR => 'SS', Name => 'Contiguous' },
    '0019,1084' => { VR => 'DS', Name => 'PeakSAR' },
    '0019,1085' => { VR => 'SS', Name => 'MonitorSAR' },
    '0019,1087' => { VR => 'DS', Name => 'CardiacRepetitionTime' },
    '0019,1088' => { VR => 'SS', Name => 'ImagesPerCardiacCycle' },
    '0019,108A' => { VR => 'SS', Name => 'ActualReceiveGainAnalog' },
    '0019,108B' => { VR => 'SS', Name => 'ActualReceiveGainDigital' },
    '0019,108D' => { VR => 'DS', Name => 'DelayAfterTrigger' },
    '0019,108F' => { VR => 'SS', Name => 'Swappf' },
    '0019,1090' => { VR => 'SS', Name => 'PauseInterval' },
    '0019,1091' => { VR => 'DS', Name => 'PulseTime' },
    '0019,1092' => { VR => 'SL', Name => 'SliceOffsetOnFreqAxis' },
    '0019,1093' => { VR => 'DS', Name => 'CenterFrequency' },
    '0019,1094' => { VR => 'SS', Name => 'TransmitGain' },
    '0019,1095' => { VR => 'SS', Name => 'AnalogReceiverGain' },
    '0019,1096' => { VR => 'SS', Name => 'DigitalReceiverGain' },
    '0019,1097' => { VR => 'SL', Name => 'BitmapDefiningCVs' },
    '0019,1098' => { VR => 'SS', Name => 'CenterFreqMethod' },
    '0019,109B' => { VR => 'SS', Name => 'PulseSeqMode' },
    '0019,109C' => { VR => 'LO', Name => 'PulseSeqName' },
    '0019,109D' => { VR => 'DT', Name => 'PulseSeqDate' },
    '0019,109E' => { VR => 'LO', Name => 'InternalPulseSeqName' },
    '0019,109F' => { VR => 'SS', Name => 'TransmittingCoil' },
    '0019,10A0' => { VR => 'SS', Name => 'SurfaceCoilType' },
    '0019,10A1' => { VR => 'SS', Name => 'ExtremityCoilFlag' },
    '0019,10A2' => { VR => 'SL', Name => 'RawDataRunNumber' },
    '0019,10A3' => { VR => 'UL', Name => 'CalibratedFieldStrength' },
    '0019,10A4' => { VR => 'SS', Name => 'SATFatWaterBone' },
    '0019,10A5' => { VR => 'DS', Name => 'ReceiveBandwidth' },
    '0019,10A7' => { VR => 'DS', Name => 'UserData01' },
    '0019,10A8' => { VR => 'DS', Name => 'UserData02' },
    '0019,10A9' => { VR => 'DS', Name => 'UserData03' },
    '0019,10AA' => { VR => 'DS', Name => 'UserData04' },
    '0019,10AB' => { VR => 'DS', Name => 'UserData05' },
    '0019,10AC' => { VR => 'DS', Name => 'UserData06' },
    '0019,10AD' => { VR => 'DS', Name => 'UserData07' },
    '0019,10AE' => { VR => 'DS', Name => 'UserData08' },
    '0019,10AF' => { VR => 'DS', Name => 'UserData09' },
    '0019,10B0' => { VR => 'DS', Name => 'UserData10' },
    '0019,10B1' => { VR => 'DS', Name => 'UserData11' },
    '0019,10B2' => { VR => 'DS', Name => 'UserData12' },
    '0019,10B3' => { VR => 'DS', Name => 'UserData13' },
    '0019,10B4' => { VR => 'DS', Name => 'UserData14' },
    '0019,10B5' => { VR => 'DS', Name => 'UserData15' },
    '0019,10B6' => { VR => 'DS', Name => 'UserData16' },
    '0019,10B7' => { VR => 'DS', Name => 'UserData17' },
    '0019,10B8' => { VR => 'DS', Name => 'UserData18' },
    '0019,10B9' => { VR => 'DS', Name => 'UserData19' },
    '0019,10BA' => { VR => 'DS', Name => 'UserData20' },
    '0019,10BB' => { VR => 'DS', Name => 'UserData21' },
    '0019,10BC' => { VR => 'DS', Name => 'UserData22' },
    '0019,10BD' => { VR => 'DS', Name => 'UserData23' },
    '0019,10BE' => { VR => 'DS', Name => 'ProjectionAngle' },
    '0019,10C0' => { VR => 'SS', Name => 'SaturationPlanes' },
    '0019,10C1' => { VR => 'SS', Name => 'SurfaceCoilIntensity' },
    '0019,10C2' => { VR => 'SS', Name => 'SATLocationR' },
    '0019,10C3' => { VR => 'SS', Name => 'SATLocationL' },
    '0019,10C4' => { VR => 'SS', Name => 'SATLocationA' },
    '0019,10C5' => { VR => 'SS', Name => 'SATLocationP' },
    '0019,10C6' => { VR => 'SS', Name => 'SATLocationH' },
    '0019,10C7' => { VR => 'SS', Name => 'SATLocationF' },
    '0019,10C8' => { VR => 'SS', Name => 'SATThicknessR-L' },
    '0019,10C9' => { VR => 'SS', Name => 'SATThicknessA-P' },
    '0019,10CA' => { VR => 'SS', Name => 'SATThicknessH-F' },
    '0019,10CB' => { VR => 'SS', Name => 'PrescribedFlowAxis' },
    '0019,10CC' => { VR => 'SS', Name => 'VelocityEncoding' },
    '0019,10CD' => { VR => 'SS', Name => 'ThicknessDisclaimer' },
    '0019,10CE' => { VR => 'SS', Name => 'PrescanType' },
    '0019,10CF' => { VR => 'SS', Name => 'PrescanStatus' },
    '0019,10D0' => { VR => 'SH', Name => 'RawDataType' },
    '0019,10D2' => { VR => 'SS', Name => 'ProjectionAlgorithm' },
    '0019,10D3' => { VR => 'SH', Name => 'ProjectionAlgorithm' },
    '0019,10D5' => { VR => 'SS', Name => 'FractionalEcho' },
    '0019,10D6' => { VR => 'SS', Name => 'PrepPulse' },
    '0019,10D7' => { VR => 'SS', Name => 'CardiacPhases' },
    '0019,10D8' => { VR => 'SS', Name => 'VariableEchoflag' },
    '0019,10D9' => { VR => 'DS', Name => 'ConcatenatedSAT' },
    '0019,10DA' => { VR => 'SS', Name => 'ReferenceChannelUsed' },
    '0019,10DB' => { VR => 'DS', Name => 'BackProjectorCoefficient' },
    '0019,10DC' => { VR => 'SS', Name => 'PrimarySpeedCorrectionUsed' },
    '0019,10DD' => { VR => 'SS', Name => 'OverrangeCorrectionUsed' },
    '0019,10DE' => { VR => 'DS', Name => 'DynamicZAlphaValue' },
    '0019,10DF' => { VR => 'DS', Name => 'UserData' },
    '0019,10E0' => { VR => 'DS', Name => 'UserData' },
    '0019,10E2' => { VR => 'DS', Name => 'VelocityEncodeScale' },
    '0019,10F2' => { VR => 'SS', Name => 'FastPhases' },
    '0019,10F9' => { VR => 'DS', Name => 'TransmissionGain' },
    # relationship group
    '0020,0000' => { VR => 'UL', Name => 'RelationshipGroupLength' },
    '0020,000D' => { VR => 'UI', Name => 'StudyInstanceUID' },
    '0020,000E' => { VR => 'UI', Name => 'SeriesInstanceUID' },
    '0020,0010' => { VR => 'SH', Name => 'StudyID' },
    '0020,0011' => { VR => 'IS', Name => 'SeriesNumber' },
    '0020,0012' => { VR => 'IS', Name => 'AcquisitionNumber' },
    '0020,0013' => { VR => 'IS', Name => 'InstanceNumber' },
    '0020,0014' => { VR => 'IS', Name => 'IsotopeNumber' },
    '0020,0015' => { VR => 'IS', Name => 'PhaseNumber' },
    '0020,0016' => { VR => 'IS', Name => 'IntervalNumber' },
    '0020,0017' => { VR => 'IS', Name => 'TimeSlotNumber' },
    '0020,0018' => { VR => 'IS', Name => 'AngleNumber' },
    '0020,0019' => { VR => 'IS', Name => 'ItemNumber' },
    '0020,0020' => { VR => 'CS', Name => 'PatientOrientation' },
    '0020,0022' => { VR => 'IS', Name => 'OverlayNumber' },
    '0020,0024' => { VR => 'IS', Name => 'CurveNumber' },
    '0020,0026' => { VR => 'IS', Name => 'LookupTableNumber' },
    '0020,0027' => { VR => 'LO', Name => 'PyramidLabel' },
    '0020,0030' => { VR => 'DS', Name => 'ImagePosition' },
    '0020,0032' => { VR => 'DS', Name => 'ImagePositionPatient' },
    '0020,0035' => { VR => 'DS', Name => 'ImageOrientation' },
    '0020,0037' => { VR => 'DS', Name => 'ImageOrientationPatient' },
    '0020,0050' => { VR => 'DS', Name => 'Location' },
    '0020,0052' => { VR => 'UI', Name => 'FrameOfReferenceUID' },
    '0020,0060' => { VR => 'CS', Name => 'Laterality' },
    '0020,0062' => { VR => 'CS', Name => 'ImageLaterality' },
    '0020,0070' => { VR => 'LO', Name => 'ImageGeometryType' },
    '0020,0080' => { VR => 'CS', Name => 'MaskingImage' },
    '0020,00AA' => { VR => 'IS', Name => 'ReportNumber' },
    '0020,0100' => { VR => 'IS', Name => 'TemporalPositionIdentifier' },
    '0020,0105' => { VR => 'IS', Name => 'NumberOfTemporalPositions' },
    '0020,0110' => { VR => 'DS', Name => 'TemporalResolution' },
    '0020,0200' => { VR => 'UI', Name => 'SynchronizationFrameOfReferenceUID' },
    '0020,0242' => { VR => 'UI', Name => 'SOPInstanceUIDConcatenationSource' },
    '0020,1000' => { VR => 'IS', Name => 'SeriesInStudy' },
    '0020,1001' => { VR => 'IS', Name => 'AcquisitionsInSeries' },
    '0020,1002' => { VR => 'IS', Name => 'ImagesInAcquisition' },
    '0020,1003' => { VR => 'IS', Name => 'ImagesInSeries' },
    '0020,1004' => { VR => 'IS', Name => 'AcquisitionsInStudy' },
    '0020,1005' => { VR => 'IS', Name => 'ImagesInStudy' },
    '0020,1020' => { VR => 'CS', Name => 'Reference' },
    '0020,103F' => { VR => 'LO', Name => 'TargetPositionReferenceIndicator' },
    '0020,1040' => { VR => 'LO', Name => 'PositionReferenceIndicator' },
    '0020,1041' => { VR => 'DS', Name => 'SliceLocation' },
    '0020,1070' => { VR => 'IS', Name => 'OtherStudyNumbers' },
    '0020,1200' => { VR => 'IS', Name => 'NumberOfPatientRelatedStudies' },
    '0020,1202' => { VR => 'IS', Name => 'NumberOfPatientRelatedSeries' },
    '0020,1204' => { VR => 'IS', Name => 'NumberOfPatientRelatedInstances' },
    '0020,1206' => { VR => 'IS', Name => 'NumberOfStudyRelatedSeries' },
    '0020,1208' => { VR => 'IS', Name => 'NumberOfStudyRelatedInstances' },
    '0020,1209' => { VR => 'IS', Name => 'NumberOfSeriesRelatedInstances' },
    '0020,31xx' => { VR => 'CS', Name => 'SourceImageIDs' },
    '0020,3401' => { VR => 'CS', Name => 'ModifyingDeviceID' },
    '0020,3402' => { VR => 'CS', Name => 'ModifiedImageID' },
    '0020,3403' => { VR => 'DA', Name => 'ModifiedImageDate' },
    '0020,3404' => { VR => 'LO', Name => 'ModifyingDeviceManufacturer' },
    '0020,3405' => { VR => 'TM', Name => 'ModifiedImageTime' },
    '0020,3406' => { VR => 'LO', Name => 'ModifiedImageDescription' },
    '0020,4000' => { VR => 'LT', Name => 'ImageComments' },
    '0020,5000' => { VR => 'AT', Name => 'OriginalImageIdentification' },
    '0020,5002' => { VR => 'CS', Name => 'OriginalImageIdentNomenclature' },
    '0020,9056' => { VR => 'SH', Name => 'StackID' },
    '0020,9057' => { VR => 'UL', Name => 'InStackPositionNumber' },
    '0020,9071' => { VR => 'SQ', Name => 'FrameAnatomySequence' },
    '0020,9072' => { VR => 'CS', Name => 'FrameLaterality' },
    '0020,9111' => { VR => 'SQ', Name => 'FrameContentSequence' },
    '0020,9113' => { VR => 'SQ', Name => 'PlanePositionSequence' },
    '0020,9116' => { VR => 'SQ', Name => 'PlaneOrientationSequence' },
    '0020,9128' => { VR => 'UL', Name => 'TemporalPositionIndex' },
    '0020,9153' => { VR => 'FD', Name => 'TriggerDelayTime' },
    '0020,9154' => { VR => 'FL', Name => 'NominalCardiacTrigTimePriorToRPeak' },
    '0020,9155' => { VR => 'FL', Name => 'ActualCardiacTrigTimePriorToRPeak' },
    '0020,9156' => { VR => 'US', Name => 'FrameAcquisitionNumber' },
    '0020,9157' => { VR => 'UL', Name => 'DimensionIndexValues' },
    '0020,9158' => { VR => 'LT', Name => 'FrameComments' },
    '0020,9161' => { VR => 'UI', Name => 'ConcatenationUID' },
    '0020,9162' => { VR => 'US', Name => 'InConcatenationNumber' },
    '0020,9163' => { VR => 'US', Name => 'InConcatenationTotalNumber' },
    '0020,9164' => { VR => 'UI', Name => 'DimensionOrganizationUID' },
    '0020,9165' => { VR => 'AT', Name => 'DimensionIndexPointer' },
    '0020,9167' => { VR => 'AT', Name => 'FunctionalGroupPointer' },
    '0020,9170' => { VR => 'SQ', Name => 'UnassignedSharedConvertedAttrsSeq' },
    '0020,9171' => { VR => 'SQ', Name => 'UnassignedPerFrameConvertedAttrSeq' },
    '0020,9172' => { VR => 'SQ', Name => 'ConversionSourceAttributesSequence' },
    '0020,9213' => { VR => 'LO', Name => 'DimensionIndexPrivateCreator' },
    '0020,9221' => { VR => 'SQ', Name => 'DimensionOrganizationSequence' },
    '0020,9222' => { VR => 'SQ', Name => 'DimensionIndexSequence' },
    '0020,9228' => { VR => 'UL', Name => 'ConcatenationFrameOffsetNumber' },
    '0020,9238' => { VR => 'LO', Name => 'FunctionalGroupPrivateCreator' },
    '0020,9241' => { VR => 'FL', Name => 'NominalPercentageOfCardiacPhase' },
    '0020,9245' => { VR => 'FL', Name => 'NominalPercentOfRespiratoryPhase' },
    '0020,9246' => { VR => 'FL', Name => 'StartingRespiratoryAmplitude' },
    '0020,9247' => { VR => 'CS', Name => 'StartingRespiratoryPhase' },
    '0020,9248' => { VR => 'FL', Name => 'EndingRespiratoryAmplitude' },
    '0020,9249' => { VR => 'CS', Name => 'EndingRespiratoryPhase' },
    '0020,9250' => { VR => 'CS', Name => 'RespiratoryTriggerType' },
    '0020,9251' => { VR => 'FD', Name => 'RRIntervalTimeNominal' },
    '0020,9252' => { VR => 'FD', Name => 'ActualCardiacTriggerDelayTime' },
    '0020,9253' => { VR => 'SQ', Name => 'RespiratorySynchronizationSequence' },
    '0020,9254' => { VR => 'FD', Name => 'RespiratoryIntervalTime' },
    '0020,9255' => { VR => 'FD', Name => 'NominalRespiratoryTriggerDelayTime' },
    '0020,9256' => { VR => 'FD', Name => 'RespiratoryTriggerDelayThreshold' },
    '0020,9257' => { VR => 'FD', Name => 'ActualRespiratoryTriggerDelayTime' },
    '0020,9301' => { VR => 'FD', Name => 'ImagePositionVolume' },
    '0020,9302' => { VR => 'FD', Name => 'ImageOrientationVolume' },
    '0020,9307' => { VR => 'CS', Name => 'UltrasoundAcquisitionGeometry' },
    '0020,9308' => { VR => 'FD', Name => 'ApexPosition' },
    '0020,9309' => { VR => 'FD', Name => 'VolumeToTransducerMappingMatrix' },
    '0020,930A' => { VR => 'FD', Name => 'VolumeToTableMappingMatrix' },
    '0020,930B' => { VR => 'CS', Name => 'VolumeToTransducerRelationship' },
    '0020,930C' => { VR => 'CS', Name => 'PatientFrameOfReferenceSource' },
    '0020,930D' => { VR => 'FD', Name => 'TemporalPositionTimeOffset' },
    '0020,930E' => { VR => 'SQ', Name => 'PlanePositionVolumeSequence' },
    '0020,930F' => { VR => 'SQ', Name => 'PlaneOrientationVolumeSequence' },
    '0020,9310' => { VR => 'SQ', Name => 'TemporalPositionSequence' },
    '0020,9311' => { VR => 'CS', Name => 'DimensionOrganizationType' },
    '0020,9312' => { VR => 'UI', Name => 'VolumeFrameOfReferenceUID' },
    '0020,9313' => { VR => 'UI', Name => 'TableFrameOfReferenceUID' },
    '0020,9421' => { VR => 'LO', Name => 'DimensionDescriptionLabel' },
    '0020,9450' => { VR => 'SQ', Name => 'PatientOrientationInFrameSequence' },
    '0020,9453' => { VR => 'LO', Name => 'FrameLabel' },
    '0020,9518' => { VR => 'US', Name => 'AcquisitionIndex' },
    '0020,9529' => { VR => 'SQ', Name => 'ContributingSOPInstancesRefSeq' },
    '0020,9536' => { VR => 'US', Name => 'ReconstructionIndex' },
    # GEMS_RELA_01 (ref 4)
    '0021,1003' => { VR => 'SS', Name => 'SeriesFromWhichPrescribed' },
    '0021,1005' => { VR => 'SH', Name => 'GenesisVersionNow' },
    '0021,1005' => { VR => 'SH', Name => 'GenesisVersionNow' },
    '0021,1007' => { VR => 'UL', Name => 'SeriesRecordChecksum' },
    '0021,1018' => { VR => 'SH', Name => 'GenesisVersionNow' },
    '0021,1018' => { VR => 'SH', Name => 'GenesisVersionNow' },
    '0021,1019' => { VR => 'UL', Name => 'AcqReconRecordChecksum' },
    '0021,1019' => { VR => 'UL', Name => 'AcqreconRecordChecksum' },
    '0021,1020' => { VR => 'DS', Name => 'TableStartLocation' },
    '0021,1035' => { VR => 'SS', Name => 'SeriesFromWhichPrescribed' },
    '0021,1036' => { VR => 'SS', Name => 'ImageFromWhichPrescribed' },
    '0021,1037' => { VR => 'SS', Name => 'ScreenFormat' },
    '0021,104A' => { VR => 'LO', Name => 'AnatomicalReferenceForScout' },
    '0021,104F' => { VR => 'SS', Name => 'LocationsInAcquisition' },
    '0021,1050' => { VR => 'SS', Name => 'GraphicallyPrescribed' },
    '0021,1051' => { VR => 'DS', Name => 'RotationFromSourceXRot' },
    '0021,1052' => { VR => 'DS', Name => 'RotationFromSourceYRot' },
    '0021,1053' => { VR => 'DS', Name => 'RotationFromSourceZRot' },
    '0021,1054' => { VR => 'SH', Name => 'ImagePosition' },
    '0021,1055' => { VR => 'SH', Name => 'ImageOrientation' },
    '0021,1056' => { VR => 'SL', Name => 'IntegerSlop' },
    '0021,1057' => { VR => 'SL', Name => 'IntegerSlop' },
    '0021,1058' => { VR => 'SL', Name => 'IntegerSlop' },
    '0021,1059' => { VR => 'SL', Name => 'IntegerSlop' },
    '0021,105A' => { VR => 'SL', Name => 'IntegerSlop' },
    '0021,105B' => { VR => 'DS', Name => 'FloatSlop' },
    '0021,105C' => { VR => 'DS', Name => 'FloatSlop' },
    '0021,105D' => { VR => 'DS', Name => 'FloatSlop' },
    '0021,105E' => { VR => 'DS', Name => 'FloatSlop' },
    '0021,105F' => { VR => 'DS', Name => 'FloatSlop' },
    '0021,1081' => { VR => 'DS', Name => 'AutoWindowLevelAlpha' },
    '0021,1082' => { VR => 'DS', Name => 'AutoWindowLevelBeta' },
    '0021,1083' => { VR => 'DS', Name => 'AutoWindowLevelWindow' },
    '0021,1084' => { VR => 'DS', Name => 'ToWindowLevelLevel' },
    '0021,1090' => { VR => 'SS', Name => 'TubeFocalSpotPosition' },
    '0021,1091' => { VR => 'SS', Name => 'BiopsyPosition' },
    '0021,1092' => { VR => 'FL', Name => 'BiopsyTLocation' },
    '0021,1093' => { VR => 'FL', Name => 'BiopsyRefLocation' },
    # ?
    '0022,0001' => { VR => 'US', Name => 'LightPathFilterPassThroughWavelen' },
    '0022,0002' => { VR => 'US', Name => 'LightPathFilterPassBand' },
    '0022,0003' => { VR => 'US', Name => 'ImagePathFilterPassThroughWavelen' },
    '0022,0004' => { VR => 'US', Name => 'ImagePathFilterPassBand' },
    '0022,0005' => { VR => 'CS', Name => 'PatientEyeMovementCommanded' },
    '0022,0006' => { VR => 'SQ', Name => 'PatientEyeMovementCommandCodeSeq' },
    '0022,0007' => { VR => 'FL', Name => 'SphericalLensPower' },
    '0022,0008' => { VR => 'FL', Name => 'CylinderLensPower' },
    '0022,0009' => { VR => 'FL', Name => 'CylinderAxis' },
    '0022,000A' => { VR => 'FL', Name => 'EmmetropicMagnification' },
    '0022,000B' => { VR => 'FL', Name => 'IntraOcularPressure' },
    '0022,000C' => { VR => 'FL', Name => 'HorizontalFieldOfView' },
    '0022,000D' => { VR => 'CS', Name => 'PupilDilated' },
    '0022,000E' => { VR => 'FL', Name => 'DegreeOfDilation' },
    '0022,000F' => { VR => 'FD', Name => 'VertexDistance' },
    '0022,0010' => { VR => 'FL', Name => 'StereoBaselineAngle' },
    '0022,0011' => { VR => 'FL', Name => 'StereoBaselineDisplacement' },
    '0022,0012' => { VR => 'FL', Name => 'StereoHorizontalPixelOffset' },
    '0022,0013' => { VR => 'FL', Name => 'StereoVerticalPixelOffset' },
    '0022,0014' => { VR => 'FL', Name => 'StereoRotation' },
    '0022,0015' => { VR => 'SQ', Name => 'AcquisitionDeviceTypeCodeSequence' },
    '0022,0016' => { VR => 'SQ', Name => 'IlluminationTypeCodeSequence' },
    '0022,0017' => { VR => 'SQ', Name => 'LightPathFilterTypeStackCodeSeq' },
    '0022,0018' => { VR => 'SQ', Name => 'ImagePathFilterTypeStackCodeSeq' },
    '0022,0019' => { VR => 'SQ', Name => 'LensesCodeSequence' },
    '0022,001A' => { VR => 'SQ', Name => 'ChannelDescriptionCodeSequence' },
    '0022,001B' => { VR => 'SQ', Name => 'RefractiveStateSequence' },
    '0022,001C' => { VR => 'SQ', Name => 'MydriaticAgentCodeSequence' },
    '0022,001D' => { VR => 'SQ', Name => 'RelativeImagePositionCodeSequence' },
    '0022,001E' => { VR => 'FL', Name => 'CameraAngleOfView' },
    '0022,0020' => { VR => 'SQ', Name => 'StereoPairsSequence' },
    '0022,0021' => { VR => 'SQ', Name => 'LeftImageSequence' },
    '0022,0022' => { VR => 'SQ', Name => 'RightImageSequence' },
    '0022,0028' => { VR => 'CS', Name => 'StereoPairsPresent' },
    '0022,0030' => { VR => 'FL', Name => 'AxialLengthOfTheEye' },
    '0022,0031' => { VR => 'SQ', Name => 'OphthalmicFrameLocationSequence' },
    '0022,0032' => { VR => 'FL', Name => 'ReferenceCoordinates' },
    '0022,0035' => { VR => 'FL', Name => 'DepthSpatialResolution' },
    '0022,0036' => { VR => 'FL', Name => 'MaximumDepthDistortion' },
    '0022,0037' => { VR => 'FL', Name => 'AlongScanSpatialResolution' },
    '0022,0038' => { VR => 'FL', Name => 'MaximumAlongScanDistortion' },
    '0022,0039' => { VR => 'CS', Name => 'OphthalmicImageOrientation' },
    '0022,0041' => { VR => 'FL', Name => 'DepthOfTransverseImage' },
    '0022,0042' => { VR => 'SQ', Name => 'MydriaticAgentConcUnitsSeq' },
    '0022,0048' => { VR => 'FL', Name => 'AcrossScanSpatialResolution' },
    '0022,0049' => { VR => 'FL', Name => 'MaximumAcrossScanDistortion' },
    '0022,004E' => { VR => 'DS', Name => 'MydriaticAgentConcentration' },
    '0022,0055' => { VR => 'FL', Name => 'IlluminationWaveLength' },
    '0022,0056' => { VR => 'FL', Name => 'IlluminationPower' },
    '0022,0057' => { VR => 'FL', Name => 'IlluminationBandwidth' },
    '0022,0058' => { VR => 'SQ', Name => 'MydriaticAgentSequence' },
    '0022,1007' => { VR => 'SQ', Name => 'OphthalmicAxialMeasRightEyeSeq' },
    '0022,1008' => { VR => 'SQ', Name => 'OphthalmicAxialMeasLeftEyeSeq' },
    '0022,1009' => { VR => 'CS', Name => 'OphthalmicAxialMeasDeviceType' },
    '0022,1010' => { VR => 'CS', Name => 'OphthalmicAxialLengthMeasType' },
    '0022,1012' => { VR => 'SQ', Name => 'OphthalmicAxialLengthSequence' },
    '0022,1019' => { VR => 'FL', Name => 'OphthalmicAxialLength' },
    '0022,1024' => { VR => 'SQ', Name => 'LensStatusCodeSequence' },
    '0022,1025' => { VR => 'SQ', Name => 'VitreousStatusCodeSequence' },
    '0022,1028' => { VR => 'SQ', Name => 'IOLFormulaCodeSequence' },
    '0022,1029' => { VR => 'LO', Name => 'IOLFormulaDetail' },
    '0022,1033' => { VR => 'FL', Name => 'KeratometerIndex' },
    '0022,1035' => { VR => 'SQ', Name => 'SourceOfOphthalmicAxialLenCodeSeq' },
    '0022,1036' => { VR => 'SQ', Name => 'SourceOfCornealSizeDataCodeSeq' },
    '0022,1037' => { VR => 'FL', Name => 'TargetRefraction' },
    '0022,1039' => { VR => 'CS', Name => 'RefractiveProcedureOccurred' },
    '0022,1040' => { VR => 'SQ', Name => 'RefractiveSurgeryTypeCodeSequence' },
    '0022,1044' => { VR => 'SQ', Name => 'OphthalmicUltrasoundMethodCodeSeq' },
    '0022,1045' => { VR => 'SQ', Name => 'SurgicallyInducedAstigmatismSeq' },
    '0022,1046' => { VR => 'CS', Name => 'TypeOfOpticalCorrection' },
    '0022,1047' => { VR => 'SQ', Name => 'ToricIOLPowerSequence' },
    '0022,1048' => { VR => 'SQ', Name => 'PredictedToricErrorSequence' },
    '0022,1049' => { VR => 'CS', Name => 'PreSelectedForImplantation' },
    '0022,104A' => { VR => 'SQ', Name => 'ToricIOLPowerForExactEmmetropiaSeq' },
    '0022,104B' => { VR => 'SQ', Name => 'ToricIOLPowerExactTargetRefractSeq' },
    '0022,1050' => { VR => 'SQ', Name => 'OphthalmicAxialLengthMeasSequence' },
    '0022,1053' => { VR => 'FL', Name => 'IOLPower' },
    '0022,1054' => { VR => 'FL', Name => 'PredictedRefractiveError' },
    '0022,1059' => { VR => 'FL', Name => 'OphthalmicAxialLengthVelocity' },
    '0022,1065' => { VR => 'LO', Name => 'LensStatusDescription' },
    '0022,1066' => { VR => 'LO', Name => 'VitreousStatusDescription' },
    '0022,1090' => { VR => 'SQ', Name => 'IOLPowerSequence' },
    '0022,1092' => { VR => 'SQ', Name => 'LensConstantSequence' },
    '0022,1093' => { VR => 'LO', Name => 'IOLManufacturer' },
    '0022,1094' => { VR => 'LO', Name => 'LensConstantDescription' },
    '0022,1095' => { VR => 'LO', Name => 'ImplantName' },
    '0022,1096' => { VR => 'SQ', Name => 'KeratometryMeasurementTypeCodeSeq' },
    '0022,1097' => { VR => 'LO', Name => 'ImplantPartNumber' },
    '0022,1100' => { VR => 'SQ', Name => 'ReferencedOphthalmicAxialMeasSeq' },
    '0022,1101' => { VR => 'SQ', Name => 'OphthalmicAxialMeasSegmentCodeSeq' },
    '0022,1103' => { VR => 'SQ', Name => 'RefractErrBeforeRefractSurgCodeSeq' },
    '0022,1121' => { VR => 'FL', Name => 'IOLPowerForExactEmmetropia' },
    '0022,1122' => { VR => 'FL', Name => 'IOLPowerForExactTargetRefraction' },
    '0022,1125' => { VR => 'SQ', Name => 'AnteriorChamberDepthDefCodeSeq' },
    '0022,1127' => { VR => 'SQ', Name => 'LensThicknessSequence' },
    '0022,1128' => { VR => 'SQ', Name => 'AnteriorChamberDepthSequence' },
    '0022,112A' => { VR => 'SQ', Name => 'CalculationCommentSequence' },
    '0022,112B' => { VR => 'CS', Name => 'CalculationCommentType' },
    '0022,112C' => { VR => 'LT', Name => 'CalculationComment' },
    '0022,1130' => { VR => 'FL', Name => 'LensThickness' },
    '0022,1131' => { VR => 'FL', Name => 'AnteriorChamberDepth' },
    '0022,1132' => { VR => 'SQ', Name => 'SourceOfLensThicknessDataCodeSeq' },
    '0022,1133' => { VR => 'SQ', Name => 'SourceAnteriorChamberDepthCodeSeq' },
    '0022,1134' => { VR => 'SQ', Name => 'SourceOfRefractiveMeasurementsSeq' },
    '0022,1135' => { VR => 'SQ', Name => 'SourceOfRefractiveMeasCodeSeq' },
    '0022,1140' => { VR => 'CS', Name => 'OphthalmicAxialLengthMeasModified' },
    '0022,1150' => { VR => 'SQ', Name => 'OphthalmicAxialLenDataSrcCodeSeq' },
    '0022,1153' => { VR => 'SQ', Name => 'OphthalmicAxialLenAcqMethodCodeSeq' },
    '0022,1155' => { VR => 'FL', Name => 'SignalToNoiseRatio' },
    '0022,1159' => { VR => 'LO', Name => 'OphthalmicAxialLenDataSourceDescr' },
    '0022,1210' => { VR => 'SQ', Name => 'OphthalmicAxialLenMeasTotalLenSeq' },
    '0022,1211' => { VR => 'SQ', Name => 'OphthalmicAxialMeasSegmentLenSeq' },
    '0022,1212' => { VR => 'SQ', Name => 'OphthalmicAxialMeasLenSummationSeq' },
    '0022,1220' => { VR => 'SQ', Name => 'UltrasoundOphthalmicAxialMeasSeq' },
    '0022,1225' => { VR => 'SQ', Name => 'OpticalOphthalmicAxialLenMeasSeq' },
    '0022,1230' => { VR => 'SQ', Name => 'UltrasoundSelOphthalmicAxialLenSeq' },
    '0022,1250' => { VR => 'SQ', Name => 'OphthalmicAxialLenSelMethodCodeSeq' },
    '0022,1255' => { VR => 'SQ', Name => 'OpticalSelOphthalmicAxialLenSeq' },
    '0022,1257' => { VR => 'SQ', Name => 'SelSegmentalOphthalmicAxialLenSeq' },
    '0022,1260' => { VR => 'SQ', Name => 'SelTotalOphthalmicAxialLenSeq' },
    '0022,1262' => { VR => 'SQ', Name => 'OphthalmicAxialLenQualityMetricSeq' },
    '0022,1265' => { VR => 'SQ', Name => 'OphthalmicAxialLenQualMetricCode' },
    '0022,1273' => { VR => 'LO', Name => 'OphthalmicAxialLenQualMetricDesc' },
    '0022,1300' => { VR => 'SQ', Name => 'IntraocularLensCalcRightEyeSeq' },
    '0022,1310' => { VR => 'SQ', Name => 'IntraocularLensCalcLeftEyeSeq' },
    '0022,1330' => { VR => 'SQ', Name => 'RefOphthalmicAxialLenMeasQCImgSeq' },
    '0022,1415' => { VR => 'CS', Name => 'OphthalmicMappingDeviceType' },
    '0022,1420' => { VR => 'SQ', Name => 'AcquisitionMethodCodeSequence' },
    '0022,1423' => { VR => 'SQ', Name => 'AcquisitionMethodAlgorithmSequence' },
    '0022,1436' => { VR => 'SQ', Name => 'OphthalmicThicknessMapTypeCodeSeq' },
    '0022,1443' => { VR => 'SQ', Name => 'OphthalmicThickMappingNormalsSeq' },
    '0022,1445' => { VR => 'SQ', Name => 'RetinalThicknessDefinitionCodeSeq' },
    '0022,1450' => { VR => 'SQ', Name => 'PixelValueMappingToCodedConceptSeq' },
    '0022,1452' => { VR => 'SS', Name => 'MappedPixelValue' },
    '0022,1454' => { VR => 'LO', Name => 'PixelValueMappingExplanation' },
    '0022,1458' => { VR => 'SQ', Name => 'OphthalmicThickMapQualityThreshSeq' },
    '0022,1460' => { VR => 'FL', Name => 'OphthalmicThickMapThreshQualRating' },
    '0022,1463' => { VR => 'FL', Name => 'AnatomicStructureReferencePoint' },
    '0022,1465' => { VR => 'SQ', Name => 'RegistrationToLocalizerSequence' },
    '0022,1466' => { VR => 'CS', Name => 'RegisteredLocalizerUnits' },
    '0022,1467' => { VR => 'FL', Name => 'RegisteredLocalizerTopLeftCorner' },
    '0022,1468' => { VR => 'FL', Name => 'RegisteredLocalizerBotRightCorner' },
    '0022,1470' => { VR => 'SQ', Name => 'OphthalmicThickMapQualityRatingSeq' },
    '0022,1472' => { VR => 'SQ', Name => 'RelevantOPTAttributesSequence' },
    '0022,1512' => { VR => 'SQ', Name => 'TransformationMethodCodeSequence' },
    '0022,1513' => { VR => 'SQ', Name => 'TransformationAlgorithmSequence' },
    '0022,1515' => { VR => 'CS', Name => 'OphthalmicAxialLengthMethod' },
    '0022,1517' => { VR => 'FL', Name => 'OphthalmicFOV' },
    '0022,1518' => { VR => 'SQ', Name => 'TwoDimensionalToThreeDimMapSeq' },
    '0022,1525' => { VR => 'SQ', Name => 'WideFieldOphthalmicPhotoQualRating' },
    '0022,1526' => { VR => 'SQ', Name => 'WideFieldOphthalmicPhotoQualThresh' },
    '0022,1527' => { VR => 'FL', Name => 'WideFieldOphthalmicPhotoThreshQual' },
    '0022,1528' => { VR => 'FL', Name => 'XCoordinatesCenterPixelViewAngle' },
    '0022,1529' => { VR => 'FL', Name => 'YCoordinatesCenterPixelViewAngle' },
    '0022,1530' => { VR => 'UL', Name => 'NumberOfMapPoints' },
    '0022,1531' => { VR => 'OF', Name => 'TwoDimensionalToThreeDimMapData' },
    '0022,1612' => { VR => 'SQ', Name => 'DerivationAlgorithmSequence' },
    '0022,1615' => { VR => 'SQ', Name => 'OphthalmicImageTypeCodeSequence' },
    '0022,1616' => { VR => 'LO', Name => 'OphthalmicImageTypeDescription' },
    '0022,1618' => { VR => 'SQ', Name => 'ScanPatternTypeCodeSequence' },
    '0022,1620' => { VR => 'SQ', Name => 'ReferencedSurfaceMeshIDSequence' },
    '0022,1622' => { VR => 'CS', Name => 'OphthalmicVolumetricPropertiesFlag' },
    '0022,1623' => { VR => 'FL', Name => 'OphthalmicAnatomicRefFrameCoord' },
    '0022,1624' => { VR => 'FL', Name => 'OphthalmicAnatomicRefPointXCoord' },
    '0022,1626' => { VR => 'FL', Name => 'OphthalmicAnatomicRefPointYCoord' },
    '0022,1627' => { VR => 'SQ', Name => 'OphthalmicEnFaceVolumeDescrSeq' },
    '0022,1628' => { VR => 'SQ', Name => 'OphthalmicEnFaceImageQualRatingSeq' },
    '0022,1629' => { VR => 'CS', Name => 'OphthalmicEnFaceVolumeDescrScope' },
    '0022,1630' => { VR => 'DS', Name => 'QualityThreshold' },
    '0022,1632' => { VR => 'SQ', Name => 'OphthalmicAnatomicRefPointSeq' },
    '0022,1633' => { VR => 'CS', Name => 'OphthalmicAnatomicRefPointLocaliz' },
    '0022,1634' => { VR => 'IS', Name => 'PrimaryAnatomicStructureItemIndex' },
    '0022,1640' => { VR => 'SQ', Name => 'OCTBscanAnalysisAcqParamSeq' },
    '0022,1642' => { VR => 'UL', Name => 'NumberOfBscansPerFrame' },
    '0022,1643' => { VR => 'FL', Name => 'BscanSlabThickness' },
    '0022,1644' => { VR => 'FL', Name => 'DistanceBetweenBscanSlabs' },
    '0022,1645' => { VR => 'FL', Name => 'BscanCycleTime' },
    '0022,1646' => { VR => 'FL', Name => 'BscanCycleTimeVector' },
    '0022,1649' => { VR => 'FL', Name => 'AscanRate' },
    '0022,1650' => { VR => 'FL', Name => 'BscanRate' },
    '0022,1658' => { VR => 'UL', Name => 'SurfaceMeshZPixelOffset' },
    # GEMS_STDY_01 (ref 4)
    '0023,1001' => { VR => 'SL', Name => 'NumberOfSeriesInStudy' },
    '0023,1002' => { VR => 'SL', Name => 'NumberOfUnarchivedSeries' },
    '0023,1010' => { VR => 'SS', Name => 'ReferenceImageField' },
    '0023,1050' => { VR => 'SS', Name => 'SummaryImage' },
    '0023,1070' => { VR => 'FD', Name => 'StartTimeSecsInFirstAxial' },
    '0023,1074' => { VR => 'SL', Name => 'NoofUpdatesToHeader' },
    '0023,107D' => { VR => 'SS', Name => 'IndicatesIfStudyHasCompleteInfo' },
    '0023,107D' => { VR => 'SS', Name => 'IndicatesIfTheStudyHasCompleteInfo' },
    # DICOM 2026
    '0024,0010' => { VR => 'FL', Name => 'VisualFieldHorizontalExtent' },
    '0024,0011' => { VR => 'FL', Name => 'VisualFieldVerticalExtent' },
    '0024,0012' => { VR => 'CS', Name => 'VisualFieldShape' },
    '0024,0016' => { VR => 'SQ', Name => 'ScreeningTestModeCodeSequence' },
    '0024,0018' => { VR => 'FL', Name => 'MaximumStimulusLuminance' },
    '0024,0020' => { VR => 'FL', Name => 'BackgroundLuminance' },
    '0024,0021' => { VR => 'SQ', Name => 'StimulusColorCodeSequence' },
    '0024,0024' => { VR => 'SQ', Name => 'BackgroundIlluminationColorCodeSeq' },
    '0024,0025' => { VR => 'FL', Name => 'StimulusArea' },
    '0024,0028' => { VR => 'FL', Name => 'StimulusPresentationTime' },
    '0024,0032' => { VR => 'SQ', Name => 'FixationSequence' },
    '0024,0033' => { VR => 'SQ', Name => 'FixationMonitoringCodeSequence' },
    '0024,0034' => { VR => 'SQ', Name => 'VisualFieldCatchTrialSequence' },
    '0024,0035' => { VR => 'US', Name => 'FixationCheckedQuantity' },
    '0024,0036' => { VR => 'US', Name => 'PatientNotProperlyFixatedQuantity' },
    '0024,0037' => { VR => 'CS', Name => 'PresentedVisualStimuliDataFlag' },
    '0024,0038' => { VR => 'US', Name => 'NumberOfVisualStimuli' },
    '0024,0039' => { VR => 'CS', Name => 'ExcessiveFixationLossesDataFlag' },
    '0024,0040' => { VR => 'CS', Name => 'ExcessiveFixationLosses' },
    '0024,0042' => { VR => 'US', Name => 'StimuliRetestingQuantity' },
    '0024,0044' => { VR => 'LT', Name => 'CommentsOnPatientPerfOfVisualField' },
    '0024,0045' => { VR => 'CS', Name => 'FalseNegativesEstimateFlag' },
    '0024,0046' => { VR => 'FL', Name => 'FalseNegativesEstimate' },
    '0024,0048' => { VR => 'US', Name => 'NegativeCatchTrialsQuantity' },
    '0024,0050' => { VR => 'US', Name => 'FalseNegativesQuantity' },
    '0024,0051' => { VR => 'CS', Name => 'ExcessiveFalseNegativesDataFlag' },
    '0024,0052' => { VR => 'CS', Name => 'ExcessiveFalseNegatives' },
    '0024,0053' => { VR => 'CS', Name => 'FalsePositivesEstimateFlag' },
    '0024,0054' => { VR => 'FL', Name => 'FalsePositivesEstimate' },
    '0024,0055' => { VR => 'CS', Name => 'CatchTrialsDataFlag' },
    '0024,0056' => { VR => 'US', Name => 'PositiveCatchTrialsQuantity' },
    '0024,0057' => { VR => 'CS', Name => 'TestPointNormalsDataFlag' },
    '0024,0058' => { VR => 'SQ', Name => 'TestPointNormalsSequence' },
    '0024,0059' => { VR => 'CS', Name => 'GlobalDeviationProbNormalsFlag' },
    '0024,0060' => { VR => 'US', Name => 'FalsePositivesQuantity' },
    '0024,0061' => { VR => 'CS', Name => 'ExcessiveFalsePositivesDataFlag' },
    '0024,0062' => { VR => 'CS', Name => 'ExcessiveFalsePositives' },
    '0024,0063' => { VR => 'CS', Name => 'VisualFieldTestNormalsFlag' },
    '0024,0064' => { VR => 'SQ', Name => 'ResultsNormalsSequence' },
    '0024,0065' => { VR => 'SQ', Name => 'AgeCorSensitivityDeviationAlgoSeq' },
    '0024,0066' => { VR => 'FL', Name => 'GlobalDeviationFromNormal' },
    '0024,0067' => { VR => 'SQ', Name => 'GeneralDefectSensDeviationAlgoSeq' },
    '0024,0068' => { VR => 'FL', Name => 'LocalizedDeviationFromNormal' },
    '0024,0069' => { VR => 'LO', Name => 'PatientReliabilityIndicator' },
    '0024,0070' => { VR => 'FL', Name => 'VisualFieldMeanSensitivity' },
    '0024,0071' => { VR => 'FL', Name => 'GlobalDeviationProbability' },
    '0024,0072' => { VR => 'CS', Name => 'LocalDeviationProbNormalsFlag' },
    '0024,0073' => { VR => 'FL', Name => 'LocalizedDeviationProbability' },
    '0024,0074' => { VR => 'CS', Name => 'ShortTermFluctuationCalculated' },
    '0024,0075' => { VR => 'FL', Name => 'ShortTermFluctuation' },
    '0024,0076' => { VR => 'CS', Name => 'ShortTermFluctuationProbCalculated' },
    '0024,0077' => { VR => 'FL', Name => 'ShortTermFluctuationProbability' },
    '0024,0078' => { VR => 'CS', Name => 'CorrectedLocalDevFromNormalCalc' },
    '0024,0079' => { VR => 'FL', Name => 'CorrectedLocalizedDevFromNormal' },
    '0024,0080' => { VR => 'CS', Name => 'CorrectedLocalDevFrmNormalProbCalc' },
    '0024,0081' => { VR => 'FL', Name => 'CorrectedLocalDevFromNormalProb' },
    '0024,0083' => { VR => 'SQ', Name => 'GlobalDeviationProbabilitySequence' },
    '0024,0085' => { VR => 'SQ', Name => 'LocalizedDeviationProbabilitySeq' },
    '0024,0086' => { VR => 'CS', Name => 'FovealSensitivityMeasured' },
    '0024,0087' => { VR => 'FL', Name => 'FovealSensitivity' },
    '0024,0088' => { VR => 'FL', Name => 'VisualFieldTestDuration' },
    '0024,0089' => { VR => 'SQ', Name => 'VisualFieldTestPointSequence' },
    '0024,0090' => { VR => 'FL', Name => 'VisualFieldTestPointXCoordinate' },
    '0024,0091' => { VR => 'FL', Name => 'VisualFieldTestPointYCoordinate' },
    '0024,0092' => { VR => 'FL', Name => 'AgeCorrectedSensDeviationValue' },
    '0024,0093' => { VR => 'CS', Name => 'StimulusResults' },
    '0024,0094' => { VR => 'FL', Name => 'SensitivityValue' },
    '0024,0095' => { VR => 'CS', Name => 'RetestStimulusSeen' },
    '0024,0096' => { VR => 'FL', Name => 'RetestSensitivityValue' },
    '0024,0097' => { VR => 'SQ', Name => 'VisualFieldTestPointNormalsSeq' },
    '0024,0098' => { VR => 'FL', Name => 'QuantifiedDefect' },
    '0024,0100' => { VR => 'FL', Name => 'AgeCorrectedSensDeviationProbValue' },
    '0024,0102' => { VR => 'CS', Name => 'GeneralizedDefectCorrSensDevFlag' },
    '0024,0103' => { VR => 'FL', Name => 'GeneralizedDefectCorrSensDevValue' },
    '0024,0104' => { VR => 'FL', Name => 'GeneralDefectCorrSensDevProbValue' },
    '0024,0105' => { VR => 'FL', Name => 'MinimumSensitivityValue' },
    '0024,0106' => { VR => 'CS', Name => 'BlindSpotLocalized' },
    '0024,0107' => { VR => 'FL', Name => 'BlindSpotXCoordinate' },
    '0024,0108' => { VR => 'FL', Name => 'BlindSpotYCoordinate' },
    '0024,0110' => { VR => 'SQ', Name => 'VisualAcuityMeasurementSequence' },
    '0024,0112' => { VR => 'SQ', Name => 'RefractiveParamsUsedOnPatientSeq' },
    '0024,0113' => { VR => 'CS', Name => 'MeasurementLaterality' },
    '0024,0114' => { VR => 'SQ', Name => 'OphthalmicPatientClinicalInfoLfEye' },
    '0024,0115' => { VR => 'SQ', Name => 'OphthalmicPatientClinicalInfoRtEye' },
    '0024,0117' => { VR => 'CS', Name => 'FovealPointNormativeDataFlag' },
    '0024,0118' => { VR => 'FL', Name => 'FovealPointProbabilityValue' },
    '0024,0120' => { VR => 'CS', Name => 'ScreeningBaselineMeasured' },
    '0024,0122' => { VR => 'SQ', Name => 'ScreeningBaselineMeasuredSequence' },
    '0024,0124' => { VR => 'CS', Name => 'ScreeningBaselineType' },
    '0024,0126' => { VR => 'FL', Name => 'ScreeningBaselineValue' },
    '0024,0202' => { VR => 'LO', Name => 'AlgorithmSource' },
    '0024,0306' => { VR => 'LO', Name => 'DataSetName' },
    '0024,0307' => { VR => 'LO', Name => 'DataSetVersion' },
    '0024,0308' => { VR => 'LO', Name => 'DataSetSource' },
    '0024,0309' => { VR => 'LO', Name => 'DataSetDescription' },
    '0024,0317' => { VR => 'SQ', Name => 'VisualFieldTestRelyGlobIndexSeq' },
    '0024,0320' => { VR => 'SQ', Name => 'VisualFieldGlobalResultsIndexSeq' },
    '0024,0325' => { VR => 'SQ', Name => 'DataObservationSequence' },
    '0024,0338' => { VR => 'CS', Name => 'IndexNormalsFlag' },
    '0024,0341' => { VR => 'FL', Name => 'IndexProbability' },
    '0024,0344' => { VR => 'SQ', Name => 'IndexProbabilitySequence' },
    # GEMS_SERS_01 (ref 4)
    '0025,1006' => { VR => 'SS', Name => 'LastPulseSequenceUsed' },
    '0025,1007' => { VR => 'SL', Name => 'ImagesInSeries' },
    '0025,1010' => { VR => 'SL', Name => 'LandmarkCounter' },
    '0025,1011' => { VR => 'SS', Name => 'NumberOfAcquisitions' },
    '0025,1014' => { VR => 'SL', Name => 'IndicatesNoofUpdatesToHeader' },
    '0025,1017' => { VR => 'SL', Name => 'SeriesCompleteFlag' },
    '0025,1018' => { VR => 'SL', Name => 'NumberOfImagesArchived' },
    '0025,1019' => { VR => 'SL', Name => 'LastImageNumberUsed' },
    '0025,101A' => { VR => 'SH', Name => 'PrimaryReceiverSuiteAndHost' },
    # GEMS_IMAG_01 (ref 4)
    '0027,1006' => { VR => 'SL', Name => 'ImageArchiveFlag' },
    '0027,1010' => { VR => 'SS', Name => 'ScoutType' },
    '0027,101C' => { VR => 'SL', Name => 'VmaMamp' },
    '0027,101D' => { VR => 'SS', Name => 'VmaPhase' },
    '0027,101E' => { VR => 'SL', Name => 'VmaMod' },
    '0027,101F' => { VR => 'SL', Name => 'VmaClip' },
    '0027,1020' => { VR => 'SS', Name => 'SmartScanOnOffFlag' },
    '0027,1030' => { VR => 'SH', Name => 'ForeignImageRevision' },
    '0027,1031' => { VR => 'SS', Name => 'ImagingMode' },
    '0027,1032' => { VR => 'SS', Name => 'PulseSequence' },
    '0027,1033' => { VR => 'SL', Name => 'ImagingOptions' },
    '0027,1035' => { VR => 'SS', Name => 'PlaneType' },
    '0027,1036' => { VR => 'SL', Name => 'ObliquePlane' },
    '0027,1040' => { VR => 'SH', Name => 'RASLetterOfImageLocation' },
    '0027,1041' => { VR => 'FL', Name => 'ImageLocation' },
    '0027,1042' => { VR => 'FL', Name => 'CenterRCoordOfPlaneImage' },
    '0027,1043' => { VR => 'FL', Name => 'CenterACoordOfPlaneImage' },
    '0027,1044' => { VR => 'FL', Name => 'CenterSCoordOfPlaneImage' },
    '0027,1045' => { VR => 'FL', Name => 'NormalRCoord' },
    '0027,1046' => { VR => 'FL', Name => 'NormalACoord' },
    '0027,1047' => { VR => 'FL', Name => 'NormalSCoord' },
    '0027,1048' => { VR => 'FL', Name => 'RCoordOfTopRightCorner' },
    '0027,1049' => { VR => 'FL', Name => 'ACoordOfTopRightCorner' },
    '0027,104A' => { VR => 'FL', Name => 'SCoordOfTopRightCorner' },
    '0027,104B' => { VR => 'FL', Name => 'RCoordOfBottomRightCorner' },
    '0027,104C' => { VR => 'FL', Name => 'ACoordOfBottomRightCorner' },
    '0027,104D' => { VR => 'FL', Name => 'SCoordOfBottomRightCorner' },
    '0027,1050' => { VR => 'FL', Name => 'TableStartLocation' },
    '0027,1051' => { VR => 'FL', Name => 'TableEndLocation' },
    '0027,1052' => { VR => 'SH', Name => 'RASLetterForSideOfImage' },
    '0027,1053' => { VR => 'SH', Name => 'RASLetterForAnteriorPosterior' },
    '0027,1054' => { VR => 'SH', Name => 'RASLetterForScoutStartLoc' },
    '0027,1055' => { VR => 'SH', Name => 'RASLetterForScoutEndLoc' },
    '0027,1060' => { VR => 'FL', Name => 'ImageDimensionX' },
    '0027,1061' => { VR => 'FL', Name => 'ImageDimensionY' },
    '0027,1062' => { VR => 'FL', Name => 'NumberOfExcitations' },
    # image presentation group
    '0028,0000' => { VR => 'UL', Name => 'ImagePresentationGroupLength' },
    '0028,0002' => { VR => 'US', Name => 'SamplesPerPixel' },
    '0028,0003' => { VR => 'US', Name => 'SamplesPerPixelUsed' },
    '0028,0004' => { VR => 'CS', Name => 'PhotometricInterpretation' },
    '0028,0005' => { VR => 'US', Name => 'ImageDimensions' },
    '0028,0006' => { VR => 'US', Name => 'PlanarConfiguration' },
    '0028,0008' => { VR => 'IS', Name => 'NumberOfFrames' },
    '0028,0009' => { VR => 'AT', Name => 'FrameIncrementPointer' },
    '0028,000A' => { VR => 'AT', Name => 'FrameDimensionPointer' },
    '0028,0010' => { VR => 'US', Name => 'Rows' },
    '0028,0011' => { VR => 'US', Name => 'Columns' },
    '0028,0012' => { VR => 'US', Name => 'Planes' },
    '0028,0014' => { VR => 'US', Name => 'UltrasoundColorDataPresent' },
    '0028,0030' => { VR => 'DS', Name => 'PixelSpacing' },
    '0028,0031' => { VR => 'DS', Name => 'ZoomFactor' },
    '0028,0032' => { VR => 'DS', Name => 'ZoomCenter' },
    '0028,0034' => { VR => 'IS', Name => 'PixelAspectRatio' },
    '0028,0040' => { VR => 'CS', Name => 'ImageFormat' },
    '0028,0050' => { VR => 'LO', Name => 'ManipulatedImage' },
    '0028,0051' => { VR => 'CS', Name => 'CorrectedImage' },
    '0028,005F' => { VR => 'LO', Name => 'CompressionRecognitionCode' },
    '0028,0060' => { VR => 'CS', Name => 'CompressionCode' },
    '0028,0061' => { VR => 'SH', Name => 'CompressionOriginator' },
    '0028,0062' => { VR => 'LO', Name => 'CompressionLabel' },
    '0028,0063' => { VR => 'SH', Name => 'CompressionDescription' },
    '0028,0065' => { VR => 'CS', Name => 'CompressionSequence' },
    '0028,0066' => { VR => 'AT', Name => 'CompressionStepPointers' },
    '0028,0068' => { VR => 'US', Name => 'RepeatInterval' },
    '0028,0069' => { VR => 'US', Name => 'BitsGrouped' },
    '0028,0070' => { VR => 'US', Name => 'PerimeterTable' },
    '0028,0071' => { VR => 'US', Name => 'PerimeterValue' },
    '0028,0080' => { VR => 'US', Name => 'PredictorRows' },
    '0028,0081' => { VR => 'US', Name => 'PredictorColumns' },
    '0028,0082' => { VR => 'US', Name => 'PredictorConstants' },
    '0028,0090' => { VR => 'CS', Name => 'BlockedPixels' },
    '0028,0091' => { VR => 'US', Name => 'BlockRows' },
    '0028,0092' => { VR => 'US', Name => 'BlockColumns' },
    '0028,0093' => { VR => 'US', Name => 'RowOverlap' },
    '0028,0094' => { VR => 'US', Name => 'ColumnOverlap' },
    '0028,0100' => { VR => 'US', Name => 'BitsAllocated' },
    '0028,0101' => { VR => 'US', Name => 'BitsStored' },
    '0028,0102' => { VR => 'US', Name => 'HighBit' },
    '0028,0103' => { VR => 'US', Name => 'PixelRepresentation', PrintConv => { 0 => 'Unsigned', 1 => 'Signed' } },
    '0028,0104' => { VR => 'US', Name => 'SmallestValidPixelValue' },
    '0028,0105' => { VR => 'US', Name => 'LargestValidPixelValue' },
    '0028,0106' => { VR => 'US', Name => 'SmallestImagePixelValue' },
    '0028,0107' => { VR => 'US', Name => 'LargestImagePixelValue' },
    '0028,0108' => { VR => 'US', Name => 'SmallestPixelValueInSeries' },
    '0028,0109' => { VR => 'US', Name => 'LargestPixelValueInSeries' },
    '0028,0110' => { VR => 'US', Name => 'SmallestImagePixelValueInPlane' },
    '0028,0111' => { VR => 'US', Name => 'LargestImagePixelValueInPlane' },
    '0028,0120' => { VR => 'US', Name => 'PixelPaddingValue' },
    '0028,0121' => { VR => 'US', Name => 'PixelPaddingRangeLimit' },
    '0028,0122' => { VR => 'FL', Name => 'FloatPixelPaddingValue' },
    '0028,0123' => { VR => 'FD', Name => 'DoubleFloatPixelPaddingValue' },
    '0028,0124' => { VR => 'FL', Name => 'FloatPixelPaddingRangeLimit' },
    '0028,0125' => { VR => 'FD', Name => 'DoubleFloatPixelPaddingRangeLimit' },
    '0028,0200' => { VR => 'US', Name => 'ImageLocation' },
    '0028,0300' => { VR => 'CS', Name => 'QualityControlImage' },
    '0028,0301' => { VR => 'CS', Name => 'BurnedInAnnotation' },
    '0028,0302' => { VR => 'CS', Name => 'RecognizableVisualFeatures' },
    '0028,0303' => { VR => 'CS', Name => 'LongitudinalTemporalInfoModified' },
    '0028,0304' => { VR => 'UI', Name => 'ReferencedColorPaletteInstanceUID' },
    '0028,0400' => { VR => 'LO', Name => 'TransformLabel' },
    '0028,0401' => { VR => 'LO', Name => 'TransformVersionNumber' },
    '0028,0402' => { VR => 'US', Name => 'NumberOfTransformSteps' },
    '0028,0403' => { VR => 'LO', Name => 'SequenceOfCompressedData' },
    '0028,0404' => { VR => 'AT', Name => 'DetailsOfCoefficients' },
    '0028,04x0' => { VR => 'US', Name => 'RowsForNthOrderCoefficients' },
    '0028,04x1' => { VR => 'US', Name => 'ColumnsForNthOrderCoefficients' },
    '0028,04x2' => { VR => 'LO', Name => 'CoefficientCoding' },
    '0028,04x3' => { VR => 'AT', Name => 'CoefficientCodingPointers' },
    '0028,0700' => { VR => 'LO', Name => 'DCTLabel' },
    '0028,0701' => { VR => 'CS', Name => 'DataBlockDescription' },
    '0028,0702' => { VR => 'AT', Name => 'DataBlock' },
    '0028,0710' => { VR => 'US', Name => 'NormalizationFactorFormat' },
    '0028,0720' => { VR => 'US', Name => 'ZonalMapNumberFormat' },
    '0028,0721' => { VR => 'AT', Name => 'ZonalMapLocation' },
    '0028,0722' => { VR => 'US', Name => 'ZonalMapFormat' },
    '0028,0730' => { VR => 'US', Name => 'AdaptiveMapFormat' },
    '0028,0740' => { VR => 'US', Name => 'CodeNumberFormat' },
    '0028,08x0' => { VR => 'CS', Name => 'CodeLabel' },
    '0028,08x2' => { VR => 'US', Name => 'NumberOfTables' },
    '0028,08x3' => { VR => 'AT', Name => 'CodeTableLocation' },
    '0028,08x4' => { VR => 'US', Name => 'BitsForCodeWord' },
    '0028,08x8' => { VR => 'AT', Name => 'ImageDataLocation' },
    '0028,1040' => { VR => 'CS', Name => 'PixelIntensityRelationship' },
    '0028,0A02' => { VR => 'CS', Name => 'PixelSpacingCalibrationType' },
    '0028,0A04' => { VR => 'LO', Name => 'PixelSpacingCalibrationDescription' },
    '0028,1040' => { VR => 'CS', Name => 'PixelIntensityRelationship' },
    '0028,1041' => { VR => 'SS', Name => 'PixelIntensityRelationshipSign' },
    '0028,1050' => { VR => 'DS', Name => 'WindowCenter' },
    '0028,1051' => { VR => 'DS', Name => 'WindowWidth' },
    '0028,1052' => { VR => 'DS', Name => 'RescaleIntercept' },
    '0028,1053' => { VR => 'DS', Name => 'RescaleSlope' },
    '0028,1054' => { VR => 'LO', Name => 'RescaleType' },
    '0028,1055' => { VR => 'LO', Name => 'WindowCenterAndWidthExplanation' },
    '0028,1056' => { VR => 'CS', Name => 'VOI_LUTFunction' },
    '0028,1080' => { VR => 'CS', Name => 'GrayScale' },
    '0028,1090' => { VR => 'CS', Name => 'RecommendedViewingMode' },
    '0028,1100' => { VR => 'SS', Name => 'GrayLookupTableDescriptor' },
    '0028,1101' => { VR => 'SS', Name => 'RedPaletteColorTableDescriptor' },
    '0028,1102' => { VR => 'SS', Name => 'GreenPaletteColorTableDescriptor' },
    '0028,1103' => { VR => 'SS', Name => 'BluePaletteColorTableDescriptor' },
    '0028,1104' => { VR => 'US', Name => 'AlphaPaletteColorLookupTableDescr' },
    '0028,1111' => { VR => 'SS', Name => 'LargeRedPaletteColorTableDescr' },
    '0028,1112' => { VR => 'SS', Name => 'LargeGreenPaletteColorTableDescr' },
    '0028,1113' => { VR => 'SS', Name => 'LargeBluePaletteColorTableDescr' },
    '0028,1199' => { VR => 'UI', Name => 'PaletteColorTableUID' },
    '0028,1200' => { VR => 'US', Name => 'GrayLookupTableData' },
    '0028,1201' => { VR => 'OW', Name => 'RedPaletteColorTableData' },
    '0028,1202' => { VR => 'OW', Name => 'GreenPaletteColorTableData' },
    '0028,1203' => { VR => 'OW', Name => 'BluePaletteColorTableData' },
    '0028,1204' => { VR => 'OW', Name => 'AlphaPaletteColorLookupTableData' },
    '0028,1211' => { VR => 'OW', Name => 'LargeRedPaletteColorTableData', Binary => 1 },
    '0028,1212' => { VR => 'OW', Name => 'LargeGreenPaletteColorTableData', Binary => 1 },
    '0028,1213' => { VR => 'OW', Name => 'LargeBluePaletteColorTableData', Binary => 1 },
    '0028,1214' => { VR => 'UI', Name => 'LargePaletteColorLookupTableUID' },
    '0028,1221' => { VR => 'OW', Name => 'SegmentedRedColorTableData' },
    '0028,1222' => { VR => 'OW', Name => 'SegmentedGreenColorTableData' },
    '0028,1223' => { VR => 'OW', Name => 'SegmentedBlueColorTableData' },
    '0028,1224' => { VR => 'OW', Name => 'SegmentedAlphaPaletteColLkupTable' },
    '0028,1230' => { VR => 'SQ', Name => 'StoredValueColorRangeSequence' },
    '0028,1231' => { VR => 'FD', Name => 'MinimumStoredValueMapped' },
    '0028,1232' => { VR => 'FD', Name => 'MaximumStoredValueMapped' },
    '0028,1300' => { VR => 'CS', Name => 'BreastImplantPresent' },
    '0028,1350' => { VR => 'CS', Name => 'PartialView' },
    '0028,1351' => { VR => 'ST', Name => 'PartialViewDescription' },
    '0028,1352' => { VR => 'SQ', Name => 'PartialViewCodeSequence' },
    '0028,135A' => { VR => 'CS', Name => 'SpatialLocationsPreserved' },
    '0028,1401' => { VR => 'SQ', Name => 'DataFrameAssignmentSequence' },
    '0028,1402' => { VR => 'CS', Name => 'DataPathAssignment' },
    '0028,1403' => { VR => 'US', Name => 'BitsMappedToColorLookupTable' },
    '0028,1404' => { VR => 'SQ', Name => 'BlendingLUT1Sequence' },
    '0028,1405' => { VR => 'CS', Name => 'BlendingLUT1TransferFunction' },
    '0028,1406' => { VR => 'FD', Name => 'BlendingWeightConstant' },
    '0028,1407' => { VR => 'US', Name => 'BlendingLookupTableDescriptor' },
    '0028,1408' => { VR => 'OW', Name => 'BlendingLookupTableData' },
    '0028,140B' => { VR => 'SQ', Name => 'EnhancedPaletteColorLookupTableSeq' },
    '0028,140C' => { VR => 'SQ', Name => 'BlendingLUT2Sequence' },
    '0028,140D' => { VR => 'CS', Name => 'BlendingLUT2TransferFunction' },
    '0028,140E' => { VR => 'CS', Name => 'DataPathID' },
    '0028,140F' => { VR => 'CS', Name => 'RGBLUTTransferFunction' },
    '0028,1410' => { VR => 'CS', Name => 'AlphaLUTTransferFunction' },
    '0028,2000' => { VR => 'OB', Name => 'ICCProfile' },
    '0028,2002' => { VR => 'CS', Name => 'ColorSpace' },
    '0028,2110' => { VR => 'CS', Name => 'LossyImageCompression' },
    '0028,2112' => { VR => 'DS', Name => 'LossyImageCompressionRatio' },
    '0028,2114' => { VR => 'CS', Name => 'LossyImageCompressionMethod' },
    '0028,3000' => { VR => 'SQ', Name => 'ModalityLUTSequence' },
    '0028,3001' => { VR => 'SQ', Name => 'VariableModalityLUTSequence' },
    '0028,3002' => { VR => 'US', Name => 'LUTDescriptor' },
    '0028,3003' => { VR => 'LO', Name => 'LUTExplanation' },
    '0028,3004' => { VR => 'LO', Name => 'ModalityLUTType' },
    '0028,3006' => { VR => 'SS', Name => 'LUTData' },
    '0028,3010' => { VR => 'SQ', Name => 'VOILUTSequence' },
    '0028,3110' => { VR => 'SQ', Name => 'SoftcopyVOILUTSequence' },
    '0028,4000' => { VR => 'LT', Name => 'ImagePresentationComments' },
    '0028,5000' => { VR => 'SQ', Name => 'BiPlaneAcquisitionSequence' },
    '0028,6010' => { VR => 'US', Name => 'RepresentativeFrameNumber' },
    '0028,6020' => { VR => 'US', Name => 'FrameNumbersOfInterest' },
    '0028,6022' => { VR => 'LO', Name => 'FrameOfInterestDescription' },
    '0028,6023' => { VR => 'CS', Name => 'FrameOfInterestType' },
    '0028,6030' => { VR => 'US', Name => 'MaskPointers' },
    '0028,6040' => { VR => 'US', Name => 'RWavePointer' },
    '0028,6100' => { VR => 'SQ', Name => 'MaskSubtractionSequence' },
    '0028,6101' => { VR => 'CS', Name => 'MaskOperation' },
    '0028,6102' => { VR => 'US', Name => 'ApplicableFrameRange' },
    '0028,6110' => { VR => 'US', Name => 'MaskFrameNumbers' },
    '0028,6112' => { VR => 'US', Name => 'ContrastFrameAveraging' },
    '0028,6114' => { VR => 'FL', Name => 'MaskSubPixelShift' },
    '0028,6120' => { VR => 'SS', Name => 'TIDOffset' },
    '0028,6190' => { VR => 'ST', Name => 'MaskOperationExplanation' },
    '0028,7000' => { VR => 'SQ', Name => 'EquipmentAdministratorSequence' },
    '0028,7001' => { VR => 'US', Name => 'NumberOfDisplaySubsystems' },
    '0028,7002' => { VR => 'US', Name => 'CurrentConfigurationID' },
    '0028,7003' => { VR => 'US', Name => 'DisplaySubsystemID' },
    '0028,7004' => { VR => 'SH', Name => 'DisplaySubsystemName' },
    '0028,7005' => { VR => 'LO', Name => 'DisplaySubsystemDescription' },
    '0028,7006' => { VR => 'CS', Name => 'SystemStatus' },
    '0028,7007' => { VR => 'LO', Name => 'SystemStatusComment' },
    '0028,7008' => { VR => 'SQ', Name => 'TargetLuminanceCharacteristicsSeq' },
    '0028,7009' => { VR => 'US', Name => 'LuminanceCharacteristicsID' },
    '0028,700A' => { VR => 'SQ', Name => 'DisplaySubsystemConfigurationSeq' },
    '0028,700B' => { VR => 'US', Name => 'ConfigurationID' },
    '0028,700C' => { VR => 'SH', Name => 'ConfigurationName' },
    '0028,700D' => { VR => 'LO', Name => 'ConfigurationDescription' },
    '0028,700E' => { VR => 'US', Name => 'ReferencedTargetLuminanceCharID' },
    '0028,700F' => { VR => 'SQ', Name => 'QAResultsSequence' },
    '0028,7010' => { VR => 'SQ', Name => 'DisplaySubsystemQAResultsSequence' },
    '0028,7011' => { VR => 'SQ', Name => 'ConfigurationQAResultsSequence' },
    '0028,7012' => { VR => 'SQ', Name => 'MeasurementEquipmentSequence' },
    '0028,7013' => { VR => 'CS', Name => 'MeasurementFunctions' },
    '0028,7014' => { VR => 'CS', Name => 'MeasurementEquipmentType' },
    '0028,7015' => { VR => 'SQ', Name => 'VisualEvaluationResultSequence' },
    '0028,7016' => { VR => 'SQ', Name => 'DisplayCalibrationResultSequence' },
    '0028,7017' => { VR => 'US', Name => 'DDLValue' },
    '0028,7018' => { VR => 'FL', Name => 'CIExyWhitePoint' },
    '0028,7019' => { VR => 'CS', Name => 'DisplayFunctionType' },
    '0028,701A' => { VR => 'FL', Name => 'GammaValue' },
    '0028,701B' => { VR => 'US', Name => 'NumberOfLuminancePoints' },
    '0028,701C' => { VR => 'SQ', Name => 'LuminanceResponseSequence' },
    '0028,701D' => { VR => 'FL', Name => 'TargetMinimumLuminance' },
    '0028,701E' => { VR => 'FL', Name => 'TargetMaximumLuminance' },
    '0028,701F' => { VR => 'FL', Name => 'LuminanceValue' },
    '0028,7020' => { VR => 'LO', Name => 'LuminanceResponseDescription' },
    '0028,7021' => { VR => 'CS', Name => 'WhitePointFlag' },
    '0028,7022' => { VR => 'SQ', Name => 'DisplayDeviceTypeCodeSequence' },
    '0028,7023' => { VR => 'SQ', Name => 'DisplaySubsystemSequence' },
    '0028,7024' => { VR => 'SQ', Name => 'LuminanceResultSequence' },
    '0028,7025' => { VR => 'CS', Name => 'AmbientLightValueSource' },
    '0028,7026' => { VR => 'CS', Name => 'MeasuredCharacteristics' },
    '0028,7027' => { VR => 'SQ', Name => 'LuminanceUniformityResultSequence' },
    '0028,7028' => { VR => 'SQ', Name => 'VisualEvaluationTestSequence' },
    '0028,7029' => { VR => 'CS', Name => 'TestResult' },
    '0028,702A' => { VR => 'LO', Name => 'TestResultComment' },
    '0028,702B' => { VR => 'CS', Name => 'TestImageValidation' },
    '0028,702C' => { VR => 'SQ', Name => 'TestPatternCodeSequence' },
    '0028,702D' => { VR => 'SQ', Name => 'MeasurementPatternCodeSequence' },
    '0028,702E' => { VR => 'SQ', Name => 'VisualEvaluationMethodCodeSequence' },
    '0028,7FE0' => { VR => 'UT', Name => 'PixelDataProviderURL' },
    '0028,9001' => { VR => 'UL', Name => 'DataPointRows' },
    '0028,9002' => { VR => 'UL', Name => 'DataPointColumns' },
    '0028,9003' => { VR => 'CS', Name => 'SignalDomainColumns' },
    '0028,9099' => { VR => 'US', Name => 'LargestMonochromePixelValue' },
    '0028,9108' => { VR => 'CS', Name => 'DataRepresentation' },
    '0028,9110' => { VR => 'SQ', Name => 'PixelMeasuresSequence' },
    '0028,9132' => { VR => 'SQ', Name => 'FrameVOILUTSequence' },
    '0028,9145' => { VR => 'SQ', Name => 'PixelValueTransformationSequence' },
    '0028,9235' => { VR => 'CS', Name => 'SignalDomainRows' },
    '0028,9411' => { VR => 'FL', Name => 'DisplayFilterPercentage' },
    '0028,9415' => { VR => 'SQ', Name => 'FramePixelShiftSequence' },
    '0028,9416' => { VR => 'US', Name => 'SubtractionItemID' },
    '0028,9422' => { VR => 'SQ', Name => 'PixelIntensityRelationshipLUTSeq' },
    '0028,9443' => { VR => 'SQ', Name => 'FramePixelDataPropertiesSequence' },
    '0028,9444' => { VR => 'CS', Name => 'GeometricalProperties' },
    '0028,9445' => { VR => 'FL', Name => 'GeometricMaximumDistortion' },
    '0028,9446' => { VR => 'CS', Name => 'ImageProcessingApplied' },
    '0028,9454' => { VR => 'CS', Name => 'MaskSelectionMode' },
    '0028,9474' => { VR => 'CS', Name => 'LUTFunction' },
    '0028,9478' => { VR => 'FL', Name => 'MaskVisibilityPercentage' },
    '0028,9501' => { VR => 'SQ', Name => 'PixelShiftSequence' },
    '0028,9502' => { VR => 'SQ', Name => 'RegionPixelShiftSequence' },
    '0028,9503' => { VR => 'SS', Name => 'VerticesOfTheRegion' },
    '0028,9505' => { VR => 'SQ', Name => 'MultiFramePresentationSequence' },
    '0028,9506' => { VR => 'US', Name => 'PixelShiftFrameRange' },
    '0028,9507' => { VR => 'US', Name => 'LUTFrameRange' },
    '0028,9520' => { VR => 'DS', Name => 'ImageToEquipmentMappingMatrix' },
    '0028,9537' => { VR => 'CS', Name => 'EquipmentCoordinateSystemID' },
    # GEMS_IMPS_01 (ref 4)
    '0029,1004' => { VR => 'SL', Name => 'LowerRangeOfPixels1a' },
    '0029,1005' => { VR => 'DS', Name => 'LowerRangeOfPixels1b' },
    '0029,1006' => { VR => 'DS', Name => 'LowerRangeOfPixels1c' },
    '0029,1007' => { VR => 'SL', Name => 'LowerRangeOfPixels1d' },
    '0029,1008' => { VR => 'SH', Name => 'LowerRangeOfPixels1e' },
    '0029,1009' => { VR => 'SH', Name => 'LowerRangeOfPixels1f' },
    '0029,100A' => { VR => 'SS', Name => 'LowerRangeOfPixels1g' },
    '0029,1015' => { VR => 'SL', Name => 'LowerRangeOfPixels1h' },
    '0029,1016' => { VR => 'SL', Name => 'LowerRangeOfPixels1i' },
    '0029,1017' => { VR => 'SL', Name => 'LowerRangeOfPixels2' },
    '0029,1018' => { VR => 'SL', Name => 'UpperRangeOfPixels2' },
    '0029,101A' => { VR => 'SL', Name => 'LenOfTotHdrInBytes' },
    '0029,1026' => { VR => 'SS', Name => 'VersionOfTheHdrStruct' },
    '0029,1034' => { VR => 'SL', Name => 'AdvantageCompOverflow' },
    '0029,1035' => { VR => 'SL', Name => 'AdvantageCompUnderflow' },
    # study group
    '0032,0000' => { VR => 'UL', Name => 'StudyGroupLength' },
    '0032,000A' => { VR => 'CS', Name => 'StudyStatusID' },
    '0032,000C' => { VR => 'CS', Name => 'StudyPriorityID' },
    '0032,0012' => { VR => 'LO', Name => 'StudyIDIssuer' },
    '0032,0032' => { VR => 'DA', Name => 'StudyVerifiedDate' },
    '0032,0033' => { VR => 'TM', Name => 'StudyVerifiedTime' },
    '0032,0034' => { VR => 'DA', Name => 'StudyReadDate' },
    '0032,0035' => { VR => 'TM', Name => 'StudyReadTime' },
    '0032,1000' => { VR => 'DA', Name => 'ScheduledStudyStartDate' },
    '0032,1001' => { VR => 'TM', Name => 'ScheduledStudyStartTime' },
    '0032,1010' => { VR => 'DA', Name => 'ScheduledStudyStopDate' },
    '0032,1011' => { VR => 'TM', Name => 'ScheduledStudyStopTime' },
    '0032,1020' => { VR => 'LO', Name => 'ScheduledStudyLocation' },
    '0032,1021' => { VR => 'AE', Name => 'ScheduledStudyLocationAETitle' },
    '0032,1030' => { VR => 'LO', Name => 'ReasonForStudy' },
    '0032,1031' => { VR => 'SQ', Name => 'RequestingPhysicianIDSequence' },
    '0032,1032' => { VR => 'PN', Name => 'RequestingPhysician' },
    '0032,1033' => { VR => 'LO', Name => 'RequestingService' },
    '0032,1034' => { VR => 'SQ', Name => 'RequestingServiceCodeSequence' },
    '0032,1040' => { VR => 'DA', Name => 'StudyArrivalDate' },
    '0032,1041' => { VR => 'TM', Name => 'StudyArrivalTime' },
    '0032,1050' => { VR => 'DA', Name => 'StudyCompletionDate' },
    '0032,1051' => { VR => 'TM', Name => 'StudyCompletionTime' },
    '0032,1055' => { VR => 'CS', Name => 'StudyComponentStatusID' },
    '0032,1060' => { VR => 'LO', Name => 'RequestedProcedureDescription' },
    '0032,1064' => { VR => 'SQ', Name => 'RequestedProcedureCodeSequence' },
    '0032,1065' => { VR => 'SQ', Name => 'RequestedLateralityCodeSequence' },
    '0032,1066' => { VR => 'UT', Name => 'ReasonForVisit' },
    '0032,1067' => { VR => 'SQ', Name => 'ReasonForVisitCodeSequence' },
    '0032,1070' => { VR => 'LO', Name => 'RequestedContrastAgent' },
    '0032,4000' => { VR => 'LT', Name => 'StudyComments' },
    '0034,0001' => { VR => 'SQ', Name => 'FlowIdentifierSequence' },
    '0034,0002' => { VR => 'OB', Name => 'FlowIdentifier' },
    '0034,0003' => { VR => 'UI', Name => 'FlowTransferSyntaxUID' },
    '0034,0004' => { VR => 'UL', Name => 'FlowRTPSamplingRate' },
    '0034,0005' => { VR => 'OB', Name => 'SourceIdentifier' },
    '0034,0007' => { VR => 'OB', Name => 'FrameOriginTimestamp' },
    '0034,0008' => { VR => 'CS', Name => 'IncludesImagingSubject' },
    '0034,0009' => { VR => 'SQ', Name => 'FrameUsefulnessGroupSequence' },
    '0034,000A' => { VR => 'SQ', Name => 'RealTimeBulkDataFlowSequence' },
    '0034,000B' => { VR => 'SQ', Name => 'CameraPositionGroupSequence' },
    '0034,000C' => { VR => 'CS', Name => 'IncludesInformation' },
    '0034,000D' => { VR => 'SQ', Name => 'TimeOfFrameGroupSequence' },
    # visit group
    '0038,0004' => { VR => 'SQ', Name => 'ReferencedPatientAliasSequence' },
    '0038,0008' => { VR => 'CS', Name => 'VisitStatusID' },
    '0038,0010' => { VR => 'LO', Name => 'AdmissionID' },
    '0038,0011' => { VR => 'LO', Name => 'IssuerOfAdmissionID' },
    '0038,0014' => { VR => 'SQ', Name => 'IssuerOfAdmissionIDSequence' },
    '0038,0016' => { VR => 'LO', Name => 'RouteOfAdmissions' },
    '0038,001A' => { VR => 'DA', Name => 'ScheduledAdmissionDate' },
    '0038,001B' => { VR => 'TM', Name => 'ScheduledAdmissionTime' },
    '0038,001C' => { VR => 'DA', Name => 'ScheduledDischargeDate' },
    '0038,001D' => { VR => 'TM', Name => 'ScheduledDischargeTime' },
    '0038,001E' => { VR => 'LO', Name => 'ScheduledPatientInstitResidence' },
    '0038,0020' => { VR => 'DA', Name => 'AdmittingDate' },
    '0038,0021' => { VR => 'TM', Name => 'AdmittingTime' },
    '0038,0030' => { VR => 'DA', Name => 'DischargeDate' },
    '0038,0032' => { VR => 'TM', Name => 'DischargeTime' },
    '0038,0040' => { VR => 'LO', Name => 'DischargeDiagnosisDescription' },
    '0038,0044' => { VR => 'SQ', Name => 'DischargeDiagnosisCodeSequence' },
    '0038,0050' => { VR => 'LO', Name => 'SpecialNeeds' },
    '0038,0060' => { VR => 'LO', Name => 'ServiceEpisodeID' },
    '0038,0061' => { VR => 'LO', Name => 'IssuerOfServiceEpisodeID' },
    '0038,0062' => { VR => 'LO', Name => 'ServiceEpisodeDescription' },
    '0038,0064' => { VR => 'SQ', Name => 'IssuerOfServiceEpisodeIDSequence' },
    '0038,0100' => { VR => 'SQ', Name => 'PertinentDocumentsSequence' },
    '0038,0101' => { VR => 'SQ', Name => 'PertinentResourcesSequence' },
    '0038,0102' => { VR => 'LO', Name => 'ResourceDescription' },
    '0038,0300' => { VR => 'LO', Name => 'CurrentPatientLocation' },
    '0038,0400' => { VR => 'LO', Name => 'PatientInstitutionResidence' },
    '0038,0500' => { VR => 'LO', Name => 'PatientState' },
    '0038,0502' => { VR => 'SQ', Name => 'PatientClinicalTrialParticipSeq' },
    '0038,4000' => { VR => 'LT', Name => 'VisitComments' },
    '003A,0004' => { VR => 'CS', Name => 'WaveformOriginality' },
    '003A,0005' => { VR => 'US', Name => 'NumberOfWaveformChannels' },
    '003A,0010' => { VR => 'UL', Name => 'NumberOfWaveformSamples' },
    '003A,001A' => { VR => 'DS', Name => 'SamplingFrequency' },
    '003A,0020' => { VR => 'SH', Name => 'MultiplexGroupLabel' },
    '003A,0200' => { VR => 'SQ', Name => 'ChannelDefinitionSequence' },
    '003A,0202' => { VR => 'IS', Name => 'WaveformChannelNumber' },
    '003A,0203' => { VR => 'SH', Name => 'ChannelLabel' },
    '003A,0205' => { VR => 'CS', Name => 'ChannelStatus' },
    '003A,0208' => { VR => 'SQ', Name => 'ChannelSourceSequence' },
    '003A,0209' => { VR => 'SQ', Name => 'ChannelSourceModifiersSequence' },
    '003A,020A' => { VR => 'SQ', Name => 'SourceWaveformSequence' },
    '003A,020C' => { VR => 'LO', Name => 'ChannelDerivationDescription' },
    '003A,0210' => { VR => 'DS', Name => 'ChannelSensitivity' },
    '003A,0211' => { VR => 'SQ', Name => 'ChannelSensitivityUnitsSequence' },
    '003A,0212' => { VR => 'DS', Name => 'ChannelSensitivityCorrectionFactor' },
    '003A,0213' => { VR => 'DS', Name => 'ChannelBaseline' },
    '003A,0214' => { VR => 'DS', Name => 'ChannelTimeSkew' },
    '003A,0215' => { VR => 'DS', Name => 'ChannelSampleSkew' },
    '003A,0218' => { VR => 'DS', Name => 'ChannelOffset' },
    '003A,021A' => { VR => 'US', Name => 'WaveformBitsStored' },
    '003A,0220' => { VR => 'DS', Name => 'FilterLowFrequency' },
    '003A,0221' => { VR => 'DS', Name => 'FilterHighFrequency' },
    '003A,0222' => { VR => 'DS', Name => 'NotchFilterFrequency' },
    '003A,0223' => { VR => 'DS', Name => 'NotchFilterBandwidth' },
    '003A,0230' => { VR => 'FL', Name => 'WaveformDataDisplayScale' },
    '003A,0231' => { VR => 'US', Name => 'WaveformDisplayBkgCIELabValue' },
    '003A,0240' => { VR => 'SQ', Name => 'WaveformPresentationGroupSequence' },
    '003A,0241' => { VR => 'US', Name => 'PresentationGroupNumber' },
    '003A,0242' => { VR => 'SQ', Name => 'ChannelDisplaySequence' },
    '003A,0244' => { VR => 'US', Name => 'ChannelRecommendDisplayCIELabValue' },
    '003A,0245' => { VR => 'FL', Name => 'ChannelPosition' },
    '003A,0246' => { VR => 'CS', Name => 'DisplayShadingFlag' },
    '003A,0247' => { VR => 'FL', Name => 'FractionalChannelDisplayScale' },
    '003A,0248' => { VR => 'FL', Name => 'AbsoluteChannelDisplayScale' },
    '003A,0300' => { VR => 'SQ', Name => 'MultiplexAudioChannelsDescrCodeSeq' },
    '003A,0301' => { VR => 'IS', Name => 'ChannelIdentificationCode' },
    '003A,0302' => { VR => 'CS', Name => 'ChannelMode' },
    '003A,0310' => { VR => 'UI', Name => 'MultiplexGroupUID' },
    '003A,0311' => { VR => 'DS', Name => 'PowerlineFrequency' },
    '003A,0312' => { VR => 'SQ', Name => 'ChannelImpedanceSequence' },
    '003A,0313' => { VR => 'DS', Name => 'ImpedanceValue' },
    '003A,0314' => { VR => 'DT', Name => 'ImpedanceMeasurementDateTime' },
    '003A,0315' => { VR => 'DS', Name => 'ImpedanceMeasurementFrequency' },
    '003A,0316' => { VR => 'CS', Name => 'ImpedanceMeasurementCurrentType' },
    '003A,0317' => { VR => 'CS', Name => 'WaveformAmplifierType' },
    '003A,0318' => { VR => 'SQ', Name => 'FilterLowFrequencyCharSeq' },
    '003A,0319' => { VR => 'SQ', Name => 'FilterHighFrequencyCharSeq' },
    '003A,0320' => { VR => 'SQ', Name => 'SummarizedFilterLookupTableSeq' },
    '003A,0321' => { VR => 'SQ', Name => 'NotchFilterCharacteristicsSequence' },
    '003A,0322' => { VR => 'CS', Name => 'WaveformFilterType' },
    '003A,0323' => { VR => 'SQ', Name => 'AnalogFilterCharacteristicsSeq' },
    '003A,0324' => { VR => 'DS', Name => 'AnalogFilterRollOff' },
    '003A,0325' => { VR => 'SQ', Name => 'AnalogFilterTypeCodeSequence' },
    '003A,0326' => { VR => 'SQ', Name => 'DigitalFilterCharacteristicsSeq' },
    '003A,0327' => { VR => 'IS', Name => 'DigitalFilterOrder' },
    '003A,0328' => { VR => 'SQ', Name => 'DigitalFilterTypeCodeSequence' },
    '003A,0329' => { VR => 'ST', Name => 'WaveformFilterDescription' },
    '003A,032A' => { VR => 'SQ', Name => 'FilterLookupTableSequence' },
    '003A,032B' => { VR => 'ST', Name => 'FilterLookupTableDescription' },
    '003A,032C' => { VR => 'SQ', Name => 'FrequencyEncodingCodeSequence' },
    '003A,032D' => { VR => 'SQ', Name => 'MagnitudeEncodingCodeSequence' },
    '003A,032E' => { VR => 'OD', Name => 'FilterLookupTableData' },
    '0040,0001' => { VR => 'AE', Name => 'ScheduledStationAETitle' },
    '0040,0002' => { VR => 'DA', Name => 'ScheduledProcedureStepStartDate' },
    '0040,0003' => { VR => 'TM', Name => 'ScheduledProcedureStepStartTime' },
    '0040,0004' => { VR => 'DA', Name => 'ScheduledProcedureStepEndDate' },
    '0040,0005' => { VR => 'TM', Name => 'ScheduledProcedureStepEndTime' },
    '0040,0006' => { VR => 'PN', Name => 'ScheduledPerformingPhysiciansName' },
    '0040,0007' => { VR => 'LO', Name => 'ScheduledProcedureStepDescription' },
    '0040,0008' => { VR => 'SQ', Name => 'ScheduledProtocolCodeSequence' },
    '0040,0009' => { VR => 'SH', Name => 'ScheduledProcedureStepID' },
    '0040,000A' => { VR => 'SQ', Name => 'StageCodeSequence' },
    '0040,000B' => { VR => 'SQ', Name => 'ScheduledPerformingPhysicianIDSeq' },
    '0040,0010' => { VR => 'SH', Name => 'ScheduledStationName' },
    '0040,0011' => { VR => 'SH', Name => 'ScheduledProcedureStepLocation' },
    '0040,0012' => { VR => 'LO', Name => 'PreMedication' },
    '0040,0020' => { VR => 'CS', Name => 'ScheduledProcedureStepStatus' },
    '0040,0026' => { VR => 'SQ', Name => 'OrderPlacerIdentifierSequence' },
    '0040,0027' => { VR => 'SQ', Name => 'OrderFillerIdentifierSequence' },
    '0040,0031' => { VR => 'UT', Name => 'LocalNamespaceEntityID' },
    '0040,0032' => { VR => 'UT', Name => 'UniversalEntityID' },
    '0040,0033' => { VR => 'CS', Name => 'UniversalEntityIDType' },
    '0040,0035' => { VR => 'CS', Name => 'IdentifierTypeCode' },
    '0040,0036' => { VR => 'SQ', Name => 'AssigningFacilitySequence' },
    '0040,0039' => { VR => 'SQ', Name => 'AssigningJurisdictionCodeSequence' },
    '0040,003A' => { VR => 'SQ', Name => 'AssigningAgencyOrDepartmentCodeSeq' },
    '0040,0100' => { VR => 'SQ', Name => 'ScheduledProcedureStepSequence' },
    '0040,0220' => { VR => 'SQ', Name => 'ReferencedNonImageCompositeSOPSeq' },
    '0040,0241' => { VR => 'AE', Name => 'PerformedStationAETitle' },
    '0040,0242' => { VR => 'SH', Name => 'PerformedStationName' },
    '0040,0243' => { VR => 'SH', Name => 'PerformedLocation' },
    '0040,0244' => { VR => 'DA', Name => 'PerformedProcedureStepStartDate' },
    '0040,0245' => { VR => 'TM', Name => 'PerformedProcedureStepStartTime' },
    '0040,0250' => { VR => 'DA', Name => 'PerformedProcedureStepEndDate' },
    '0040,0251' => { VR => 'TM', Name => 'PerformedProcedureStepEndTime' },
    '0040,0252' => { VR => 'CS', Name => 'PerformedProcedureStepStatus' },
    '0040,0253' => { VR => 'SH', Name => 'PerformedProcedureStepID' },
    '0040,0254' => { VR => 'LO', Name => 'PerformedProcedureStepDescription' },
    '0040,0255' => { VR => 'LO', Name => 'PerformedProcedureTypeDescription' },
    '0040,0260' => { VR => 'SQ', Name => 'PerformedProtocolCodeSequence' },
    '0040,0261' => { VR => 'CS', Name => 'PerformedProtocolType' },
    '0040,0270' => { VR => 'SQ', Name => 'ScheduledStepAttributesSequence' },
    '0040,0275' => { VR => 'SQ', Name => 'RequestAttributesSequence' },
    '0040,0280' => { VR => 'ST', Name => 'CommentsOnPerformedProcedureStep' },
    '0040,0281' => { VR => 'SQ', Name => 'ProcStepDiscontinueReasonCodeSeq' },
    '0040,0293' => { VR => 'SQ', Name => 'QuantitySequence' },
    '0040,0294' => { VR => 'DS', Name => 'Quantity' },
    '0040,0295' => { VR => 'SQ', Name => 'MeasuringUnitsSequence' },
    '0040,0296' => { VR => 'SQ', Name => 'BillingItemSequence' },
    '0040,0300' => { VR => 'US', Name => 'TotalTimeOfFluoroscopy' },
    '0040,0301' => { VR => 'US', Name => 'TotalNumberOfExposures' },
    '0040,0302' => { VR => 'US', Name => 'EntranceDose' },
    '0040,0303' => { VR => 'US', Name => 'ExposedArea' },
    '0040,0306' => { VR => 'DS', Name => 'DistanceSourceToEntrance' },
    '0040,0307' => { VR => 'DS', Name => 'DistanceSourceToSupport' },
    '0040,030E' => { VR => 'SQ', Name => 'ExposureDoseSequence' },
    '0040,0310' => { VR => 'ST', Name => 'CommentsOnRadiationDose' },
    '0040,0312' => { VR => 'DS', Name => 'XRayOutput' },
    '0040,0314' => { VR => 'DS', Name => 'HalfValueLayer' },
    '0040,0316' => { VR => 'DS', Name => 'OrganDose' },
    '0040,0318' => { VR => 'CS', Name => 'OrganExposed' },
    '0040,0320' => { VR => 'SQ', Name => 'BillingProcedureStepSequence' },
    '0040,0321' => { VR => 'SQ', Name => 'FilmConsumptionSequence' },
    '0040,0324' => { VR => 'SQ', Name => 'BillingSuppliesAndDevicesSequence' },
    '0040,0330' => { VR => 'SQ', Name => 'ReferencedProcedureStepSequence' },
    '0040,0340' => { VR => 'SQ', Name => 'PerformedSeriesSequence' },
    '0040,0400' => { VR => 'LT', Name => 'CommentsOnScheduledProcedureStep' },
    '0040,0440' => { VR => 'SQ', Name => 'ProtocolContextSequence' },
    '0040,0441' => { VR => 'SQ', Name => 'ContentItemModifierSequence' },
    '0040,0500' => { VR => 'SQ', Name => 'ScheduledSpecimenSequence' },
    '0040,050A' => { VR => 'LO', Name => 'SpecimenAccessionNumber' },
    '0040,0512' => { VR => 'LO', Name => 'ContainerIdentifier' },
    '0040,0513' => { VR => 'SQ', Name => 'IssuerOfTheContainerIdentifierSeq' },
    '0040,0515' => { VR => 'SQ', Name => 'AlternateContainerIdentifierSeq' },
    '0040,0518' => { VR => 'SQ', Name => 'ContainerTypeCodeSequence' },
    '0040,051A' => { VR => 'LO', Name => 'ContainerDescription' },
    '0040,0520' => { VR => 'SQ', Name => 'ContainerComponentSequence' },
    '0040,0550' => { VR => 'SQ', Name => 'SpecimenSequence' },
    '0040,0551' => { VR => 'LO', Name => 'SpecimenIdentifier' },
    '0040,0552' => { VR => 'SQ', Name => 'SpecimenDescriptionSequenceTrial' },
    '0040,0553' => { VR => 'ST', Name => 'SpecimenDescriptionTrial' },
    '0040,0554' => { VR => 'UI', Name => 'SpecimenUID' },
    '0040,0555' => { VR => 'SQ', Name => 'AcquisitionContextSequence' },
    '0040,0556' => { VR => 'ST', Name => 'AcquisitionContextDescription' },
    '0040,0560' => { VR => 'SQ', Name => 'SpecimenDescriptionSequence' },
    '0040,0562' => { VR => 'SQ', Name => 'IssuerOfTheSpecimenIdentifierSeq' },
    '0040,059A' => { VR => 'SQ', Name => 'SpecimenTypeCodeSequence' },
    '0040,0600' => { VR => 'LO', Name => 'SpecimenShortDescription' },
    '0040,0602' => { VR => 'UT', Name => 'SpecimenDetailedDescription' },
    '0040,0610' => { VR => 'SQ', Name => 'SpecimenPreparationSequence' },
    '0040,0612' => { VR => 'SQ', Name => 'SpecimenPrepStepContentItemSeq' },
    '0040,0620' => { VR => 'SQ', Name => 'SpecimenLocalizationContentItemSeq' },
    '0040,06FA' => { VR => 'LO', Name => 'SlideIdentifier' },
    '0040,0710' => { VR => 'SQ', Name => 'WholeSlideMicroscopyImageFrameSeq' },
    '0040,071A' => { VR => 'SQ', Name => 'ImageCenterPointCoordinatesSeq' },
    '0040,072A' => { VR => 'DS', Name => 'XOffsetInSlideCoordinateSystem' },
    '0040,073A' => { VR => 'DS', Name => 'YOffsetInSlideCoordinateSystem' },
    '0040,074A' => { VR => 'DS', Name => 'ZOffsetInSlideCoordinateSystem' },
    '0040,08D8' => { VR => 'SQ', Name => 'PixelSpacingSequence' },
    '0040,08DA' => { VR => 'SQ', Name => 'CoordinateSystemAxisCodeSequence' },
    '0040,08EA' => { VR => 'SQ', Name => 'MeasurementUnitsCodeSequence' },
    '0040,09F8' => { VR => 'SQ', Name => 'VitalStainCodeSequenceTrial' },
    '0040,1001' => { VR => 'SH', Name => 'RequestedProcedureID' },
    '0040,1002' => { VR => 'LO', Name => 'ReasonForRequestedProcedure' },
    '0040,1003' => { VR => 'SH', Name => 'RequestedProcedurePriority' },
    '0040,1004' => { VR => 'LO', Name => 'PatientTransportArrangements' },
    '0040,1005' => { VR => 'LO', Name => 'RequestedProcedureLocation' },
    '0040,1006' => { VR => 'SH', Name => 'PlacerOrderNumber-Procedure' },
    '0040,1007' => { VR => 'SH', Name => 'FillerOrderNumber-Procedure' },
    '0040,1008' => { VR => 'LO', Name => 'ConfidentialityCode' },
    '0040,1009' => { VR => 'SH', Name => 'ReportingPriority' },
    '0040,100A' => { VR => 'SQ', Name => 'ReasonForRequestedProcedureCodeSeq' },
    '0040,1010' => { VR => 'PN', Name => 'NamesOfIntendedRecipientsOfResults' },
    '0040,1011' => { VR => 'SQ', Name => 'IntendedRecipientsOfResultsIDSeq' },
    '0040,1012' => { VR => 'SQ', Name => 'ReasonForPerformedProcedureCodeSeq' },
    '0040,1060' => { VR => 'LO', Name => 'RequestedProcedureDescriptionTrial' },
    '0040,1101' => { VR => 'SQ', Name => 'PersonIdentificationCodeSequence' },
    '0040,1102' => { VR => 'ST', Name => 'PersonAddress' },
    '0040,1103' => { VR => 'LO', Name => 'PersonTelephoneNumbers' },
    '0040,1104' => { VR => 'LT', Name => 'PersonTelecomInformation' },
    '0040,1400' => { VR => 'LT', Name => 'RequestedProcedureComments' },
    '0040,2001' => { VR => 'LO', Name => 'ReasonForImagingServiceRequest' },
    '0040,2004' => { VR => 'DA', Name => 'IssueDateOfImagingServiceRequest' },
    '0040,2005' => { VR => 'TM', Name => 'IssueTimeOfImagingServiceRequest' },
    '0040,2006' => { VR => 'SH', Name => 'PlacerOrderNum-ImagingServiceReq' },
    '0040,2007' => { VR => 'SH', Name => 'FillerOrderNum-ImagingServiceReq' },
    '0040,2008' => { VR => 'PN', Name => 'OrderEnteredBy' },
    '0040,2009' => { VR => 'SH', Name => 'OrderEntererLocation' },
    '0040,2010' => { VR => 'SH', Name => 'OrderCallbackPhoneNumber' },
    '0040,2011' => { VR => 'LT', Name => 'OrderCallbackTelecomInformation' },
    '0040,2016' => { VR => 'LO', Name => 'PlacerOrderNum-ImagingServiceReq' },
    '0040,2017' => { VR => 'LO', Name => 'FillerOrderNum-ImagingServiceReq' },
    '0040,2400' => { VR => 'LT', Name => 'ImagingServiceRequestComments' },
    '0040,3001' => { VR => 'LO', Name => 'ConfidentialityOnPatientDataDescr' },
    '0040,4001' => { VR => 'CS', Name => 'GenPurposeScheduledProcStepStatus' },
    '0040,4002' => { VR => 'CS', Name => 'GenPurposePerformedProcStepStatus' },
    '0040,4003' => { VR => 'CS', Name => 'GenPurposeSchedProcStepPriority' },
    '0040,4004' => { VR => 'SQ', Name => 'SchedProcessingApplicationsCodeSeq' },
    '0040,4005' => { VR => 'DT', Name => 'SchedProcedureStepStartDateAndTime' },
    '0040,4006' => { VR => 'CS', Name => 'MultipleCopiesFlag' },
    '0040,4007' => { VR => 'SQ', Name => 'PerformedProcessingAppsCodeSeq' },
    '0040,4008' => { VR => 'DT', Name => 'ScheduledProcedureStepExpDateTime' },
    '0040,4009' => { VR => 'SQ', Name => 'HumanPerformerCodeSequence' },
    '0040,4010' => { VR => 'DT', Name => 'SchedProcStepModificationDateTime' },
    '0040,4011' => { VR => 'DT', Name => 'ExpectedCompletionDateAndTime' },
    '0040,4015' => { VR => 'SQ', Name => 'ResultingGenPurposePerfProcStepSeq' },
    '0040,4016' => { VR => 'SQ', Name => 'RefGenPurposeSchedProcStepSeq' },
    '0040,4018' => { VR => 'SQ', Name => 'ScheduledWorkitemCodeSequence' },
    '0040,4019' => { VR => 'SQ', Name => 'PerformedWorkitemCodeSequence' },
    '0040,4020' => { VR => 'CS', Name => 'InputAvailabilityFlag' },
    '0040,4021' => { VR => 'SQ', Name => 'InputInformationSequence' },
    '0040,4022' => { VR => 'SQ', Name => 'RelevantInformationSequence' },
    '0040,4023' => { VR => 'UI', Name => 'RefGenPurSchedProcStepTransUID' },
    '0040,4025' => { VR => 'SQ', Name => 'ScheduledStationNameCodeSequence' },
    '0040,4026' => { VR => 'SQ', Name => 'ScheduledStationClassCodeSequence' },
    '0040,4027' => { VR => 'SQ', Name => 'SchedStationGeographicLocCodeSeq' },
    '0040,4028' => { VR => 'SQ', Name => 'PerformedStationNameCodeSequence' },
    '0040,4029' => { VR => 'SQ', Name => 'PerformedStationClassCodeSequence' },
    '0040,4030' => { VR => 'SQ', Name => 'PerformedStationGeogLocCodeSeq' },
    '0040,4031' => { VR => 'SQ', Name => 'RequestedSubsequentWorkItemCodeSeq' },
    '0040,4032' => { VR => 'SQ', Name => 'NonDICOMOutputCodeSequence' },
    '0040,4033' => { VR => 'SQ', Name => 'OutputInformationSequence' },
    '0040,4034' => { VR => 'SQ', Name => 'ScheduledHumanPerformersSequence' },
    '0040,4035' => { VR => 'SQ', Name => 'ActualHumanPerformersSequence' },
    '0040,4036' => { VR => 'LO', Name => 'HumanPerformersOrganization' },
    '0040,4037' => { VR => 'PN', Name => 'HumanPerformerName' },
    '0040,4040' => { VR => 'CS', Name => 'RawDataHandling' },
    '0040,4041' => { VR => 'CS', Name => 'InputReadinessState' },
    '0040,4050' => { VR => 'DT', Name => 'PerformedProcedureStepStartTime' },
    '0040,4051' => { VR => 'DT', Name => 'PerformedProcedureStepEndDateTime' },
    '0040,4052' => { VR => 'DT', Name => 'ProcedureStepCancellationDateTime' },
    '0040,4070' => { VR => 'SQ', Name => 'OutputDestinationSequence' },
    '0040,4071' => { VR => 'SQ', Name => 'DICOMStorageSequence' },
    '0040,4072' => { VR => 'SQ', Name => 'STOWRSStorageSequence' },
    '0040,4073' => { VR => 'UR', Name => 'StorageURL' },
    '0040,4074' => { VR => 'SQ', Name => 'XDSStorageSequence' },
    '0040,8302' => { VR => 'DS', Name => 'EntranceDoseInMilliGy' },
    '0040,8303' => { VR => 'CS', Name => 'EntranceDoseDerivation' },
    '0040,9092' => { VR => 'SQ', Name => 'ParametricMapFrameTypeSequence' },
    '0040,9094' => { VR => 'SQ', Name => 'RefImageRealWorldValueMappingSeq' },
    '0040,9096' => { VR => 'SQ', Name => 'RealWorldValueMappingSequence' },
    '0040,9098' => { VR => 'SQ', Name => 'PixelValueMappingCodeSequence' },
    '0040,9210' => { VR => 'SH', Name => 'LUTLabel' },
    '0040,9211' => { VR => 'SS', Name => 'RealWorldValueLastValueMapped' },
    '0040,9212' => { VR => 'FD', Name => 'RealWorldValueLUTData' },
    '0040,9213' => { VR => 'FD', Name => 'DoubleRealWorldLastValueMapped' },
    '0040,9214' => { VR => 'FD', Name => 'DoubleRealWorldFirstValueMapped' },
    '0040,9216' => { VR => 'SS', Name => 'RealWorldValueFirstValueMapped' },
    '0040,9220' => { VR => 'SQ', Name => 'QuantityDefinitionSequence' },
    '0040,9224' => { VR => 'FD', Name => 'RealWorldValueIntercept' },
    '0040,9225' => { VR => 'FD', Name => 'RealWorldValueSlope' },
    '0040,A007' => { VR => 'CS', Name => 'FindingsFlagTrial' },
    '0040,A010' => { VR => 'CS', Name => 'RelationshipType' },
    '0040,A020' => { VR => 'SQ', Name => 'FindingsSequenceTrial' },
    '0040,A021' => { VR => 'UI', Name => 'FindingsGroupUIDTrial' },
    '0040,A022' => { VR => 'UI', Name => 'ReferencedFindingsGroupUIDTrial' },
    '0040,A023' => { VR => 'DA', Name => 'FindingsGroupRecordingDateTrial' },
    '0040,A024' => { VR => 'TM', Name => 'FindingsGroupRecordingTimeTrial' },
    '0040,A026' => { VR => 'SQ', Name => 'FindingsSourceCategoryCodeSeqTrial' },
    '0040,A027' => { VR => 'LO', Name => 'VerifyingOrganization' },
    '0040,A028' => { VR => 'SQ', Name => 'DocumentingOrgIDCodeSeqTrial' },
    '0040,A030' => { VR => 'DT', Name => 'VerificationDateTime' },
    '0040,A032' => { VR => 'DT', Name => 'ObservationDateTime' },
    '0040,A033' => { VR => 'DT', Name => 'ObservationStartDateTime' },
    '0040,A034' => { VR => 'DT', Name => 'EffectiveStartDateTime' },
    '0040,A035' => { VR => 'DT', Name => 'EffectiveStopDateTime' },
    '0040,A040' => { VR => 'CS', Name => 'ValueType' },
    '0040,A043' => { VR => 'SQ', Name => 'ConceptNameCodeSequence' },
    '0040,A047' => { VR => 'LO', Name => 'MeasurementPrecisionDescrTrial' },
    '0040,A050' => { VR => 'CS', Name => 'ContinuityOfContent' },
    '0040,A057' => { VR => 'CS', Name => 'UrgencyOrPriorityAlertsTrial' },
    '0040,A060' => { VR => 'LO', Name => 'SequencingIndicatorTrial' },
    '0040,A066' => { VR => 'SQ', Name => 'DocumentIdentifierCodeSeqTrial' },
    '0040,A067' => { VR => 'PN', Name => 'DocumentAuthorTrial' },
    '0040,A068' => { VR => 'SQ', Name => 'DocumentAuthorIDCodeSeqTrial' },
    '0040,A070' => { VR => 'SQ', Name => 'IdentifierCodeSequenceTrial' },
    '0040,A073' => { VR => 'SQ', Name => 'VerifyingObserverSequence' },
    '0040,A074' => { VR => 'OB', Name => 'ObjectBinaryIdentifierTrial' },
    '0040,A075' => { VR => 'PN', Name => 'VerifyingObserverName' },
    '0040,A076' => { VR => 'SQ', Name => 'DocumentingObserverIDCodeSeqTrial' },
    '0040,A078' => { VR => 'SQ', Name => 'AuthorObserverSequence' },
    '0040,A07A' => { VR => 'SQ', Name => 'ParticipantSequence' },
    '0040,A07C' => { VR => 'SQ', Name => 'CustodialOrganizationSequence' },
    '0040,A080' => { VR => 'CS', Name => 'ParticipationType' },
    '0040,A082' => { VR => 'DT', Name => 'ParticipationDateTime' },
    '0040,A084' => { VR => 'CS', Name => 'ObserverType' },
    '0040,A085' => { VR => 'SQ', Name => 'ProcedureIdentifierCodeSeqTrial' },
    '0040,A088' => { VR => 'SQ', Name => 'VerifyingObserverIdentCodeSequence' },
    '0040,A089' => { VR => 'OB', Name => 'ObjectDirectoryBinaryIDTrial' },
    '0040,A090' => { VR => 'SQ', Name => 'EquivalentCDADocumentSequence' },
    '0040,A0B0' => { VR => 'US', Name => 'ReferencedWaveformChannels' },
    '0040,A110' => { VR => 'DA', Name => 'DateOfDocOrVerbalTransactionTrial' },
    '0040,A112' => { VR => 'TM', Name => 'TimeOfDocCreateOrVerbalTransaction' },
    '0040,A120' => { VR => 'DT', Name => 'DateTime' },
    '0040,A121' => { VR => 'DA', Name => 'Date' },
    '0040,A122' => { VR => 'TM', Name => 'Time' },
    '0040,A123' => { VR => 'PN', Name => 'PersonName' },
    '0040,A124' => { VR => 'UI', Name => 'UID' },
    '0040,A125' => { VR => 'CS', Name => 'ReportStatusIDTrial' },
    '0040,A130' => { VR => 'CS', Name => 'TemporalRangeType' },
    '0040,A132' => { VR => 'UL', Name => 'ReferencedSamplePositions' },
    '0040,A136' => { VR => 'US', Name => 'ReferencedFrameNumbers' },
    '0040,A138' => { VR => 'DS', Name => 'ReferencedTimeOffsets' },
    '0040,A13A' => { VR => 'DT', Name => 'ReferencedDateTime' },
    '0040,A160' => { VR => 'UT', Name => 'TextValue' },
    '0040,A161' => { VR => 'FD', Name => 'FloatingPointValue' },
    '0040,A162' => { VR => 'SL', Name => 'RationalNumeratorValue' },
    '0040,A163' => { VR => 'UL', Name => 'RationalDenominatorValue' },
    '0040,A167' => { VR => 'SQ', Name => 'ObservationCategoryCodeSeqTrial' },
    '0040,A168' => { VR => 'SQ', Name => 'ConceptCodeSequence' },
    '0040,A16A' => { VR => 'ST', Name => 'BibliographicCitationTrial' },
    '0040,A170' => { VR => 'SQ', Name => 'PurposeOfReferenceCodeSequence' },
    '0040,A171' => { VR => 'UI', Name => 'ObservationUID' },
    '0040,A172' => { VR => 'UI', Name => 'ReferencedObservationUIDTrial' },
    '0040,A173' => { VR => 'CS', Name => 'ReferencedObservationClassTrial' },
    '0040,A174' => { VR => 'CS', Name => 'ReferencedObjectObsClassTrial' },
    '0040,A180' => { VR => 'US', Name => 'AnnotationGroupNumber' },
    '0040,A192' => { VR => 'DA', Name => 'ObservationDateTrial' },
    '0040,A193' => { VR => 'TM', Name => 'ObservationTimeTrial' },
    '0040,A194' => { VR => 'CS', Name => 'MeasurementAutomationTrial' },
    '0040,A195' => { VR => 'SQ', Name => 'ModifierCodeSequence' },
    '0040,A224' => { VR => 'ST', Name => 'IdentificationDescriptionTrial' },
    '0040,A290' => { VR => 'CS', Name => 'CoordinatesSetGeometricTypeTrial' },
    '0040,A296' => { VR => 'SQ', Name => 'AlgorithmCodeSequenceTrial' },
    '0040,A297' => { VR => 'ST', Name => 'AlgorithmDescriptionTrial' },
    '0040,A29A' => { VR => 'SL', Name => 'PixelCoordinatesSetTrial' },
    '0040,A300' => { VR => 'SQ', Name => 'MeasuredValueSequence' },
    '0040,A301' => { VR => 'SQ', Name => 'NumericValueQualifierCodeSequence' },
    '0040,A307' => { VR => 'PN', Name => 'CurrentObserverTrial' },
    '0040,A30A' => { VR => 'DS', Name => 'NumericValue' },
    '0040,A313' => { VR => 'SQ', Name => 'ReferencedAccessionSequenceTrial' },
    '0040,A33A' => { VR => 'ST', Name => 'ReportStatusCommentTrial' },
    '0040,A340' => { VR => 'SQ', Name => 'ProcedureContextSequenceTrial' },
    '0040,A352' => { VR => 'PN', Name => 'VerbalSourceTrial' },
    '0040,A353' => { VR => 'ST', Name => 'AddressTrial' },
    '0040,A354' => { VR => 'LO', Name => 'TelephoneNumberTrial' },
    '0040,A358' => { VR => 'SQ', Name => 'VerbalSourceIdentifierCodeSeqTrial' },
    '0040,A360' => { VR => 'SQ', Name => 'PredecessorDocumentsSequence' },
    '0040,A370' => { VR => 'SQ', Name => 'ReferencedRequestSequence' },
    '0040,A372' => { VR => 'SQ', Name => 'PerformedProcedureCodeSequence' },
    '0040,A375' => { VR => 'SQ', Name => 'CurrentRequestedProcEvidenceSeq' },
    '0040,A380' => { VR => 'SQ', Name => 'ReportDetailSequenceTrial' },
    '0040,A385' => { VR => 'SQ', Name => 'PertinentOtherEvidenceSequence' },
    '0040,A390' => { VR => 'SQ', Name => 'HL7StructuredDocumentRefSeq' },
    '0040,A402' => { VR => 'UI', Name => 'ObservationSubjectUIDTrial' },
    '0040,A403' => { VR => 'CS', Name => 'ObservationSubjectClassTrial' },
    '0040,A404' => { VR => 'SQ', Name => 'ObservationSubjectTypeCodeSeqTrial' },
    '0040,A491' => { VR => 'CS', Name => 'CompletionFlag' },
    '0040,A492' => { VR => 'LO', Name => 'CompletionFlagDescription' },
    '0040,A493' => { VR => 'CS', Name => 'VerificationFlag' },
    '0040,A494' => { VR => 'CS', Name => 'ArchiveRequested' },
    '0040,A496' => { VR => 'CS', Name => 'PreliminaryFlag' },
    '0040,A504' => { VR => 'SQ', Name => 'ContentTemplateSequence' },
    '0040,A525' => { VR => 'SQ', Name => 'IdenticalDocumentsSequence' },
    '0040,A600' => { VR => 'CS', Name => 'ObservationSubjectContextFlagTrial' },
    '0040,A601' => { VR => 'CS', Name => 'ObserverContextFlagTrial' },
    '0040,A603' => { VR => 'CS', Name => 'ProcedureContextFlagTrial' },
    '0040,A730' => { VR => 'SQ', Name => 'ContentSequence' },
    '0040,A731' => { VR => 'SQ', Name => 'RelationshipSequenceTrial' },
    '0040,A732' => { VR => 'SQ', Name => 'RelationshipTypeCodeSequenceTrial' },
    '0040,A744' => { VR => 'SQ', Name => 'LanguageCodeSequenceTrial' },
    '0040,A801' => { VR => 'SQ', Name => 'TabulatedValuesSequence' },
    '0040,A802' => { VR => 'UL', Name => 'NumberOfTableRows' },
    '0040,A803' => { VR => 'UL', Name => 'NumberOfTableColumns' },
    '0040,A804' => { VR => 'UL', Name => 'TableRowNumber' },
    '0040,A805' => { VR => 'UL', Name => 'TableColumnNumber' },
    '0040,A806' => { VR => 'SQ', Name => 'TableRowDefinitionSequence' },
    '0040,A807' => { VR => 'SQ', Name => 'TableColumnDefinitionSequence' },
    '0040,A808' => { VR => 'SQ', Name => 'CellValuesSequence' },
    '0040,A992' => { VR => 'ST', Name => 'UniformResourceLocatorTrial' },
    '0040,B020' => { VR => 'SQ', Name => 'AnnotationSequence' },
    '0040,B030' => { VR => 'SQ', Name => 'StructuredWaveformAnnotationSeq' },
    '0040,B031' => { VR => 'SQ', Name => 'WaveformAnnotationDisplaySelSeq' },
    '0040,B032' => { VR => 'US', Name => 'ReferencedMontageIndex' },
    '0040,B033' => { VR => 'SQ', Name => 'WaveformTextualAnnotationSequence' },
    '0040,B034' => { VR => 'DT', Name => 'AnnotationDateTime' },
    '0040,B035' => { VR => 'SQ', Name => 'DisplayedWaveformSegmentSequence' },
    '0040,B036' => { VR => 'DT', Name => 'SegmentDefinitionDateTime' },
    '0040,B037' => { VR => 'SQ', Name => 'MontageActivationSequence' },
    '0040,B038' => { VR => 'DS', Name => 'MontageActivationTimeOffset' },
    '0040,B039' => { VR => 'SQ', Name => 'WaveformMontageSequence' },
    '0040,B03A' => { VR => 'IS', Name => 'ReferencedMontageChannelNumber' },
    '0040,B03B' => { VR => 'LT', Name => 'MontageName' },
    '0040,B03C' => { VR => 'SQ', Name => 'MontageChannelSequence' },
    '0040,B03D' => { VR => 'US', Name => 'MontageIndex' },
    '0040,B03E' => { VR => 'IS', Name => 'MontageChannelNumber' },
    '0040,B03F' => { VR => 'LO', Name => 'MontageChannelLabel' },
    '0040,B040' => { VR => 'SQ', Name => 'MontageChannelSourceCodeSequence' },
    '0040,B041' => { VR => 'SQ', Name => 'ContributingChannelSourcesSequence' },
    '0040,B042' => { VR => 'FL', Name => 'ChannelWeight' },
    '0040,DB00' => { VR => 'CS', Name => 'TemplateIdentifier' },
    '0040,DB06' => { VR => 'DT', Name => 'TemplateVersion' },
    '0040,DB07' => { VR => 'DT', Name => 'TemplateLocalVersion' },
    '0040,DB0B' => { VR => 'CS', Name => 'TemplateExtensionFlag' },
    '0040,DB0C' => { VR => 'UI', Name => 'TemplateExtensionOrganizationUID' },
    '0040,DB0D' => { VR => 'UI', Name => 'TemplateExtensionCreatorUID' },
    '0040,DB73' => { VR => 'UL', Name => 'ReferencedContentItemIdentifier' },
    '0040,E001' => { VR => 'ST', Name => 'HL7InstanceIdentifier' },
    '0040,E004' => { VR => 'DT', Name => 'HL7DocumentEffectiveTime' },
    '0040,E006' => { VR => 'SQ', Name => 'HL7DocumentTypeCodeSequence' },
    '0040,E008' => { VR => 'SQ', Name => 'DocumentClassCodeSequence' },
    '0040,E010' => { VR => 'UT', Name => 'RetrieveURI' },
    '0040,E011' => { VR => 'UI', Name => 'RetrieveLocationUID' },
    '0040,E020' => { VR => 'CS', Name => 'TypeOfInstances' },
    '0040,E021' => { VR => 'SQ', Name => 'DICOMRetrievalSequence' },
    '0040,E022' => { VR => 'SQ', Name => 'DICOMMediaRetrievalSequence' },
    '0040,E023' => { VR => 'SQ', Name => 'WADORetrievalSequence' },
    '0040,E024' => { VR => 'SQ', Name => 'XDSRetrievalSequence' },
    '0040,E025' => { VR => 'SQ', Name => 'WADORSRetrievalSequence' },
    '0040,E030' => { VR => 'UI', Name => 'RepositoryUniqueID' },
    '0040,E031' => { VR => 'UI', Name => 'HomeCommunityID' },
    '0042,0010' => { VR => 'ST', Name => 'DocumentTitle' },
    '0042,0011' => { VR => 'OB', Name => 'EncapsulatedDocument' },
    '0042,0012' => { VR => 'LO', Name => 'MIMETypeOfEncapsulatedDocument' },
    '0042,0013' => { VR => 'SQ', Name => 'SourceInstanceSequence' },
    '0042,0014' => { VR => 'LO', Name => 'ListOfMIMETypes' },
    '0042,0015' => { VR => 'UL', Name => 'EncapsulatedDocumentLength' },
    # GEMS_PARM_01 (ref 4)
    '0043,1001' => { VR => 'SS', Name => 'BitmapOfPrescanOptions' },
    '0043,1002' => { VR => 'SS', Name => 'GradientOffsetInX' },
    '0043,1003' => { VR => 'SS', Name => 'GradientOffsetInY' },
    '0043,1004' => { VR => 'SS', Name => 'GradientOffsetInZ' },
    '0043,1005' => { VR => 'SS', Name => 'ImgIsOriginalOrUnoriginal' },
    '0043,1006' => { VR => 'SS', Name => 'NumberOfEPIShots' },
    '0043,1007' => { VR => 'SS', Name => 'ViewsPerSegment' },
    '0043,1008' => { VR => 'SS', Name => 'RespiratoryRateBpm' },
    '0043,1009' => { VR => 'SS', Name => 'RespiratoryTriggerPoint' },
    '0043,100A' => { VR => 'SS', Name => 'TypeOfReceiverUsed' },
    '0043,100B' => { VR => 'DS', Name => 'PeakRateOfChangeOfGradientField' },
    '0043,100C' => { VR => 'DS', Name => 'LimitsInUnitsOfPercent' },
    '0043,100D' => { VR => 'DS', Name => 'PSDEstimatedLimit' },
    '0043,100E' => { VR => 'DS', Name => 'PSDEstimatedLimitInTeslaPerSecond' },
    '0043,100F' => { VR => 'DS', Name => 'Saravghead' },
    '0043,1010' => { VR => 'US', Name => 'WindowValue' },
    '0043,1011' => { VR => 'US', Name => 'TotalInputViews' },
    '0043,1012' => { VR => 'SS', Name => 'X-RayChain' },
    '0043,1013' => { VR => 'SS', Name => 'DeconKernelParameters' },
    '0043,1014' => { VR => 'SS', Name => 'CalibrationParameters' },
    '0043,1015' => { VR => 'SS', Name => 'TotalOutputViews' },
    '0043,1016' => { VR => 'SS', Name => 'NumberOfOverranges' },
    '0043,1017' => { VR => 'DS', Name => 'IBHImageScaleFactors' },
    '0043,1018' => { VR => 'DS', Name => 'BBHCoefficients' },
    '0043,1019' => { VR => 'SS', Name => 'NumberOfBBHChainsToBlend' },
    '0043,101A' => { VR => 'SL', Name => 'StartingChannelNumber' },
    '0043,101B' => { VR => 'SS', Name => 'PpscanParameters' },
    '0043,101C' => { VR => 'SS', Name => 'GEImageIntegrity' },
    '0043,101D' => { VR => 'SS', Name => 'LevelValue' },
    '0043,101E' => { VR => 'DS', Name => 'DeltaStartTime' },
    '0043,101F' => { VR => 'SL', Name => 'MaxOverrangesInAView' },
    '0043,1020' => { VR => 'DS', Name => 'AvgOverrangesAllViews' },
    '0043,1021' => { VR => 'SS', Name => 'CorrectedAfterGlowTerms' },
    '0043,1025' => { VR => 'SS', Name => 'ReferenceChannels' },
    '0043,1026' => { VR => 'US', Name => 'NoViewsRefChansBlocked' },
    '0043,1027' => { VR => 'SH', Name => 'ScanPitchRatio' },
    '0043,1028' => { VR => 'OB', Name => 'UniqueImageIden' },
    '0043,1029' => { VR => 'OB', Name => 'HistogramTables' },
    '0043,102A' => { VR => 'OB', Name => 'UserDefinedData' },
    '0043,102B' => { VR => 'SS', Name => 'PrivateScanOptions' },
    '0043,102C' => { VR => 'SS', Name => 'EffectiveEchoSpacing' },
    '0043,102D' => { VR => 'SH', Name => 'StringSlopField1' },
    '0043,102E' => { VR => 'SH', Name => 'StringSlopField2' },
    '0043,102F' => { VR => 'SS', Name => 'RawDataType' },
    '0043,1030' => { VR => 'SS', Name => 'RawDataType' },
    '0043,1031' => { VR => 'DS', Name => 'RACordOfTargetReconCenter' },
    '0043,1032' => { VR => 'SS', Name => 'RawDataType' },
    '0043,1033' => { VR => 'FL', Name => 'NegScanspacing' },
    '0043,1034' => { VR => 'IS', Name => 'OffsetFrequency' },
    '0043,1035' => { VR => 'UL', Name => 'UserUsageTag' },
    '0043,1036' => { VR => 'UL', Name => 'UserFillMapMSW' },
    '0043,1037' => { VR => 'UL', Name => 'UserFillMapLSW' },
    '0043,1038' => { VR => 'FL', Name => 'User25-48' },
    '0043,1039' => { VR => 'IS', Name => 'SlopInt6-9' },
    '0043,1040' => { VR => 'FL', Name => 'TriggerOnPosition' },
    '0043,1041' => { VR => 'FL', Name => 'DegreeOfRotation' },
    '0043,1042' => { VR => 'SL', Name => 'DASTriggerSource' },
    '0043,1043' => { VR => 'SL', Name => 'DASFpaGain' },
    '0043,1044' => { VR => 'SL', Name => 'DASOutputSource' },
    '0043,1045' => { VR => 'SL', Name => 'DASAdInput' },
    '0043,1046' => { VR => 'SL', Name => 'DASCalMode' },
    '0043,1047' => { VR => 'SL', Name => 'DASCalFrequency' },
    '0043,1048' => { VR => 'SL', Name => 'DASRegXm' },
    '0043,1049' => { VR => 'SL', Name => 'DASAutoZero' },
    '0043,104A' => { VR => 'SS', Name => 'StartingChannelOfView' },
    '0043,104B' => { VR => 'SL', Name => 'DASXmPattern' },
    '0043,104C' => { VR => 'SS', Name => 'TGGCTriggerMode' },
    '0043,104D' => { VR => 'FL', Name => 'StartScanToXrayOnDelay' },
    '0043,104E' => { VR => 'FL', Name => 'DurationOfXrayOn' },
    '0043,1060' => { VR => 'IS', Name => 'SlopInt10-17' },
    '0043,1061' => { VR => 'UI', Name => 'ScannerStudyEntityUID' },
    '0043,1062' => { VR => 'SH', Name => 'ScannerStudyID' },
    '0043,106f' => { VR => 'DS', Name => 'ScannerTableEntry' },
    # ?
    '0044,0001' => { VR => 'ST', Name => 'ProductPackageIdentifier' },
    '0044,0002' => { VR => 'CS', Name => 'SubstanceAdministrationApproval' },
    '0044,0003' => { VR => 'LT', Name => 'ApprovalStatusFurtherDescription' },
    '0044,0004' => { VR => 'DT', Name => 'ApprovalStatusDateTime' },
    '0044,0007' => { VR => 'SQ', Name => 'ProductTypeCodeSequence' },
    '0044,0008' => { VR => 'LO', Name => 'ProductName' },
    '0044,0009' => { VR => 'LT', Name => 'ProductDescription' },
    '0044,000A' => { VR => 'LO', Name => 'ProductLotIdentifier' },
    '0044,000B' => { VR => 'DT', Name => 'ProductExpirationDateTime' },
    '0044,0010' => { VR => 'DT', Name => 'SubstanceAdministrationDateTime' },
    '0044,0011' => { VR => 'LO', Name => 'SubstanceAdministrationNotes' },
    '0044,0012' => { VR => 'LO', Name => 'SubstanceAdministrationDeviceID' },
    '0044,0013' => { VR => 'SQ', Name => 'ProductParameterSequence' },
    '0044,0019' => { VR => 'SQ', Name => 'SubstanceAdminParameterSeq' },
    # DICOM 2026
    '0044,0100' => { VR => 'SQ', Name => 'ApprovalSequence' },
    '0044,0101' => { VR => 'SQ', Name => 'AssertionCodeSequence' },
    '0044,0102' => { VR => 'UI', Name => 'AssertionUID' },
    '0044,0103' => { VR => 'SQ', Name => 'AsserterIdentificationSequence' },
    '0044,0104' => { VR => 'DT', Name => 'AssertionDateTime' },
    '0044,0105' => { VR => 'DT', Name => 'AssertionExpirationDateTime' },
    '0044,0106' => { VR => 'UT', Name => 'AssertionComments' },
    '0044,0107' => { VR => 'SQ', Name => 'RelatedAssertionSequence' },
    '0044,0108' => { VR => 'UI', Name => 'ReferencedAssertionUID' },
    '0044,0109' => { VR => 'SQ', Name => 'ApprovalSubjectSequence' },
    '0044,010A' => { VR => 'SQ', Name => 'OrganizationalRoleCodeSequence' },
    '0044,0110' => { VR => 'SQ', Name => 'RTAssertionsSequence' },
    # GEMS_HELIOS_01 (ref 4)
    '0045,1001' => { VR => 'LO', Name => 'NumberOfMacroRowsInDetector' },
    '0045,1002' => { VR => 'FL', Name => 'MacroWidthAtISOCenter' },
    '0045,1003' => { VR => 'SS', Name => 'DASType' },
    '0045,1004' => { VR => 'SS', Name => 'DASGain' },
    '0045,1005' => { VR => 'SS', Name => 'DASTemperature' },
    '0045,1006' => { VR => 'CS', Name => 'TableDirectionInOrOut' },
    '0045,1007' => { VR => 'FL', Name => 'ZSmoothingFactor' },
    '0045,1008' => { VR => 'SS', Name => 'ViewWeightingMode' },
    '0045,1009' => { VR => 'SS', Name => 'SigmaRowNumberWhichRowsWereUsed' },
    '0045,100A' => { VR => 'FL', Name => 'MinimumDasValueFoundInTheScanData' },
    '0045,100B' => { VR => 'FL', Name => 'MaximumOffsetShiftValueUsed' },
    '0045,100C' => { VR => 'SS', Name => 'NumberOfViewsShifted' },
    '0045,100D' => { VR => 'SS', Name => 'ZTrackingFlag' },
    '0045,100E' => { VR => 'FL', Name => 'MeanZError' },
    '0045,100F' => { VR => 'FL', Name => 'ZTrackingMaximumError' },
    '0045,1010' => { VR => 'SS', Name => 'StartingViewForRow2a' },
    '0045,1011' => { VR => 'SS', Name => 'NumberOfViewsInRow2a' },
    '0045,1012' => { VR => 'SS', Name => 'StartingViewForRow1a' },
    '0045,1013' => { VR => 'SS', Name => 'SigmaMode' },
    '0045,1014' => { VR => 'SS', Name => 'NumberOfViewsInRow1a' },
    '0045,1015' => { VR => 'SS', Name => 'StartingViewForRow2b' },
    '0045,1016' => { VR => 'SS', Name => 'NumberOfViewsInRow2b' },
    '0045,1017' => { VR => 'SS', Name => 'StartingViewForRow1b' },
    '0045,1018' => { VR => 'SS', Name => 'NumberOfViewsInRow1b' },
    '0045,1019' => { VR => 'SS', Name => 'AirFilterCalibrationDate' },
    '0045,101A' => { VR => 'SS', Name => 'AirFilterCalibrationTime' },
    '0045,101B' => { VR => 'SS', Name => 'PhantomCalibrationDate' },
    '0045,101C' => { VR => 'SS', Name => 'PhantomCalibrationTime' },
    '0045,101D' => { VR => 'SS', Name => 'ZSlopeCalibrationDate' },
    '0045,101E' => { VR => 'SS', Name => 'ZSlopeCalibrationTime' },
    '0045,101F' => { VR => 'SS', Name => 'CrosstalkCalibrationDate' },
    '0045,1020' => { VR => 'SS', Name => 'CrosstalkCalibrationTime' },
    '0045,1021' => { VR => 'SS', Name => 'IterboneOptionFlag' },
    '0045,1022' => { VR => 'SS', Name => 'PeristalticFlagOption' },
    '0046,0012' => { VR => 'LO', Name => 'LensDescription' },
    '0046,0014' => { VR => 'SQ', Name => 'RightLensSequence' },
    '0046,0015' => { VR => 'SQ', Name => 'LeftLensSequence' },
    '0046,0016' => { VR => 'SQ', Name => 'UnspecifiedLateralityLensSequence' },
    '0046,0018' => { VR => 'SQ', Name => 'CylinderSequence' },
    '0046,0028' => { VR => 'SQ', Name => 'PrismSequence' },
    '0046,0030' => { VR => 'FD', Name => 'HorizontalPrismPower' },
    '0046,0032' => { VR => 'CS', Name => 'HorizontalPrismBase' },
    '0046,0034' => { VR => 'FD', Name => 'VerticalPrismPower' },
    '0046,0036' => { VR => 'CS', Name => 'VerticalPrismBase' },
    '0046,0038' => { VR => 'CS', Name => 'LensSegmentType' },
    '0046,0040' => { VR => 'FD', Name => 'OpticalTransmittance' },
    '0046,0042' => { VR => 'FD', Name => 'ChannelWidth' },
    '0046,0044' => { VR => 'FD', Name => 'PupilSize' },
    '0046,0046' => { VR => 'FD', Name => 'CornealSize' },
    '0046,0047' => { VR => 'SQ', Name => 'CornealSizeSequence' },
    '0046,0050' => { VR => 'SQ', Name => 'AutorefractionRightEyeSequence' },
    '0046,0052' => { VR => 'SQ', Name => 'AutorefractionLeftEyeSequence' },
    '0046,0060' => { VR => 'FD', Name => 'DistancePupillaryDistance' },
    '0046,0062' => { VR => 'FD', Name => 'NearPupillaryDistance' },
    '0046,0063' => { VR => 'FD', Name => 'IntermediatePupillaryDistance' },
    '0046,0064' => { VR => 'FD', Name => 'OtherPupillaryDistance' },
    '0046,0070' => { VR => 'SQ', Name => 'KeratometryRightEyeSequence' },
    '0046,0071' => { VR => 'SQ', Name => 'KeratometryLeftEyeSequence' },
    '0046,0074' => { VR => 'SQ', Name => 'SteepKeratometricAxisSequence' },
    '0046,0075' => { VR => 'FD', Name => 'RadiusOfCurvature' },
    '0046,0076' => { VR => 'FD', Name => 'KeratometricPower' },
    '0046,0077' => { VR => 'FD', Name => 'KeratometricAxis' },
    '0046,0080' => { VR => 'SQ', Name => 'FlatKeratometricAxisSequence' },
    '0046,0092' => { VR => 'CS', Name => 'BackgroundColor' },
    '0046,0094' => { VR => 'CS', Name => 'Optotype' },
    '0046,0095' => { VR => 'CS', Name => 'OptotypePresentation' },
    '0046,0097' => { VR => 'SQ', Name => 'SubjectiveRefractionRightEyeSeq' },
    '0046,0098' => { VR => 'SQ', Name => 'SubjectiveRefractionLeftEyeSeq' },
    '0046,0100' => { VR => 'SQ', Name => 'AddNearSequence' },
    '0046,0101' => { VR => 'SQ', Name => 'AddIntermediateSequence' },
    '0046,0102' => { VR => 'SQ', Name => 'AddOtherSequence' },
    '0046,0104' => { VR => 'FD', Name => 'AddPower' },
    '0046,0106' => { VR => 'FD', Name => 'ViewingDistance' },
    '0046,0110' => { VR => 'SQ', Name => 'CorneaMeasurementsSequence' },
    '0046,0111' => { VR => 'SQ', Name => 'SourceOfCorneaMeasDataCodeSeq' },
    '0046,0112' => { VR => 'SQ', Name => 'SteepCornealAxisSequence' },
    '0046,0113' => { VR => 'SQ', Name => 'FlatCornealAxisSequence' },
    '0046,0114' => { VR => 'FD', Name => 'CornealPower' },
    '0046,0115' => { VR => 'FD', Name => 'CornealAxis' },
    '0046,0116' => { VR => 'SQ', Name => 'CorneaMeasurementMethodCodeSeq' },
    '0046,0117' => { VR => 'FL', Name => 'RefractiveIndexOfCornea' },
    '0046,0118' => { VR => 'FL', Name => 'RefractiveIndexOfAqueousHumor' },
    '0046,0121' => { VR => 'SQ', Name => 'VisualAcuityTypeCodeSequence' },
    '0046,0122' => { VR => 'SQ', Name => 'VisualAcuityRightEyeSequence' },
    '0046,0123' => { VR => 'SQ', Name => 'VisualAcuityLeftEyeSequence' },
    '0046,0124' => { VR => 'SQ', Name => 'VisualAcuityBothEyesOpenSequence' },
    '0046,0125' => { VR => 'CS', Name => 'ViewingDistanceType' },
    '0046,0135' => { VR => 'SS', Name => 'VisualAcuityModifiers' },
    '0046,0137' => { VR => 'FD', Name => 'DecimalVisualAcuity' },
    '0046,0139' => { VR => 'LO', Name => 'OptotypeDetailedDefinition' },
    '0046,0145' => { VR => 'SQ', Name => 'ReferencedRefractiveMeasSeq' },
    '0046,0146' => { VR => 'FD', Name => 'SpherePower' },
    '0046,0147' => { VR => 'FD', Name => 'CylinderPower' },
    '0046,0201' => { VR => 'CS', Name => 'CornealTopographySurface' },
    '0046,0202' => { VR => 'FL', Name => 'CornealVertexLocation' },
    '0046,0203' => { VR => 'FL', Name => 'PupilCentroidXCoordinate' },
    '0046,0204' => { VR => 'FL', Name => 'PupilCentroidYCoordinate' },
    '0046,0205' => { VR => 'FL', Name => 'EquivalentPupilRadius' },
    '0046,0207' => { VR => 'SQ', Name => 'CornealTopographyMapTypeCodeSeq' },
    '0046,0208' => { VR => 'IS', Name => 'VerticesOfTheOutlineOfPupil' },
    '0046,0210' => { VR => 'SQ', Name => 'CornealTopographyMappingNormalsSeq' },
    '0046,0211' => { VR => 'SQ', Name => 'MaximumCornealCurvatureSequence' },
    '0046,0212' => { VR => 'FL', Name => 'MaximumCornealCurvature' },
    '0046,0213' => { VR => 'FL', Name => 'MaximumCornealCurvatureLocation' },
    '0046,0215' => { VR => 'SQ', Name => 'MinimumKeratometricSequence' },
    '0046,0218' => { VR => 'SQ', Name => 'SimulatedKeratometricCylinderSeq' },
    '0046,0220' => { VR => 'FL', Name => 'AverageCornealPower' },
    '0046,0224' => { VR => 'FL', Name => 'CornealISValue' },
    '0046,0227' => { VR => 'FL', Name => 'AnalyzedArea' },
    '0046,0230' => { VR => 'FL', Name => 'SurfaceRegularityIndex' },
    '0046,0232' => { VR => 'FL', Name => 'SurfaceAsymmetryIndex' },
    '0046,0234' => { VR => 'FL', Name => 'CornealEccentricityIndex' },
    '0046,0236' => { VR => 'FL', Name => 'KeratoconusPredictionIndex' },
    '0046,0238' => { VR => 'FL', Name => 'DecimalPotentialVisualAcuity' },
    '0046,0242' => { VR => 'CS', Name => 'CornealTopoMapQualityEvaluation' },
    '0046,0244' => { VR => 'SQ', Name => 'SourceImageCornealProcessedDataSeq' },
    '0046,0247' => { VR => 'FL', Name => 'CornealPointLocation' },
    '0046,0248' => { VR => 'CS', Name => 'CornealPointEstimated' },
    '0046,0249' => { VR => 'FL', Name => 'AxialPower' },
    '0046,0250' => { VR => 'FL', Name => 'TangentialPower' },
    '0046,0251' => { VR => 'FL', Name => 'RefractivePower' },
    '0046,0252' => { VR => 'FL', Name => 'RelativeElevation' },
    '0046,0253' => { VR => 'FL', Name => 'CornealWavefront' },
    '0048,0001' => { VR => 'FL', Name => 'ImagedVolumeWidth' },
    '0048,0002' => { VR => 'FL', Name => 'ImagedVolumeHeight' },
    '0048,0003' => { VR => 'FL', Name => 'ImagedVolumeDepth' },
    '0048,0006' => { VR => 'UL', Name => 'TotalPixelMatrixColumns' },
    '0048,0007' => { VR => 'UL', Name => 'TotalPixelMatrixRows' },
    '0048,0008' => { VR => 'SQ', Name => 'TotalPixelMatrixOriginSequence' },
    '0048,0010' => { VR => 'CS', Name => 'SpecimenLabelInImage' },
    '0048,0011' => { VR => 'CS', Name => 'FocusMethod' },
    '0048,0012' => { VR => 'CS', Name => 'ExtendedDepthOfField' },
    '0048,0013' => { VR => 'US', Name => 'NumberOfFocalPlanes' },
    '0048,0014' => { VR => 'FL', Name => 'DistanceBetweenFocalPlanes' },
    '0048,0015' => { VR => 'US', Name => 'RecommendedAbsentPixelCIELabValue' },
    '0048,0100' => { VR => 'SQ', Name => 'IlluminatorTypeCodeSequence' },
    '0048,0102' => { VR => 'DS', Name => 'ImageOrientationSlide' },
    '0048,0105' => { VR => 'SQ', Name => 'OpticalPathSequence' },
    '0048,0106' => { VR => 'SH', Name => 'OpticalPathIdentifier' },
    '0048,0107' => { VR => 'ST', Name => 'OpticalPathDescription' },
    '0048,0108' => { VR => 'SQ', Name => 'IlluminationColorCodeSequence' },
    '0048,0110' => { VR => 'SQ', Name => 'SpecimenReferenceSequence' },
    '0048,0111' => { VR => 'DS', Name => 'CondenserLensPower' },
    '0048,0112' => { VR => 'DS', Name => 'ObjectiveLensPower' },
    '0048,0113' => { VR => 'DS', Name => 'ObjectiveLensNumericalAperture' },
    '0048,0114' => { VR => 'CS', Name => 'ConfocalMode' },
    '0048,0115' => { VR => 'CS', Name => 'TissueLocation' },
    '0048,0116' => { VR => 'SQ', Name => 'ConfocalMicroscopyImageFrameSeq' },
    '0048,0117' => { VR => 'FD', Name => 'ImageAcquisitionDepth' },
    '0048,0120' => { VR => 'SQ', Name => 'PaletteColorLookupTableSequence' },
    '0048,0200' => { VR => 'SQ', Name => 'ReferencedImageNavigationSequence' },
    '0048,0201' => { VR => 'US', Name => 'TopLeftHandCornerOfLocalizerArea' },
    '0048,0202' => { VR => 'US', Name => 'BottomRightCornerOfLocalizerArea' },
    '0048,0207' => { VR => 'SQ', Name => 'OpticalPathIdentificationSequence' },
    '0048,021A' => { VR => 'SQ', Name => 'PlanePositionSlideSequence' },
    '0048,021E' => { VR => 'SL', Name => 'ColumnPosInTotalImagePixelMatrix' },
    '0048,021F' => { VR => 'SL', Name => 'RowPositionInTotalImagePixelMatrix' },
    '0048,0301' => { VR => 'CS', Name => 'PixelOriginInterpretation' },
    '0048,0302' => { VR => 'UL', Name => 'NumberOfOpticalPaths' },
    '0048,0303' => { VR => 'UL', Name => 'TotalPixelMatrixFocalPlanes' },
    '0048,0304' => { VR => 'CS', Name => 'TilesOverlap' },
    # calibration group
    '0050,0004' => { VR => 'CS', Name => 'CalibrationImage' },
    '0050,0010' => { VR => 'SQ', Name => 'DeviceSequence' },
    '0050,0012' => { VR => 'SQ', Name => 'ContainerComponentTypeCodeSequence' },
    '0050,0013' => { VR => 'FD', Name => 'ContainerComponentThickness' },
    '0050,0014' => { VR => 'DS', Name => 'DeviceLength' },
    '0050,0015' => { VR => 'FD', Name => 'ContainerComponentWidth' },
    '0050,0016' => { VR => 'DS', Name => 'DeviceDiameter' },
    '0050,0017' => { VR => 'CS', Name => 'DeviceDiameterUnits' },
    '0050,0018' => { VR => 'DS', Name => 'DeviceVolume' },
    '0050,0019' => { VR => 'DS', Name => 'InterMarkerDistance' },
    '0050,001A' => { VR => 'CS', Name => 'ContainerComponentMaterial' },
    '0050,001B' => { VR => 'LO', Name => 'ContainerComponentID' },
    '0050,001C' => { VR => 'FD', Name => 'ContainerComponentLength' },
    '0050,001D' => { VR => 'FD', Name => 'ContainerComponentDiameter' },
    '0050,001E' => { VR => 'LO', Name => 'ContainerComponentDescription' },
    '0050,0020' => { VR => 'LO', Name => 'DeviceDescription' },
    '0050,0021' => { VR => 'ST', Name => 'LongDeviceDescription' },
    '0052,0001' => { VR => 'FL', Name => 'ContrastBolusIngredientVolPercent' },
    '0052,0002' => { VR => 'FD', Name => 'OCTFocalDistance' },
    '0052,0003' => { VR => 'FD', Name => 'BeamSpotSize' },
    '0052,0004' => { VR => 'FD', Name => 'EffectiveRefractiveIndex' },
    '0052,0006' => { VR => 'CS', Name => 'OCTAcquisitionDomain' },
    '0052,0007' => { VR => 'FD', Name => 'OCTOpticalCenterWavelength' },
    '0052,0008' => { VR => 'FD', Name => 'AxialResolution' },
    '0052,0009' => { VR => 'FD', Name => 'RangingDepth' },
    '0052,0011' => { VR => 'FD', Name => 'ALineRate' },
    '0052,0012' => { VR => 'US', Name => 'ALinesPerFrame' },
    '0052,0013' => { VR => 'FD', Name => 'CatheterRotationalRate' },
    '0052,0014' => { VR => 'FD', Name => 'ALinePixelSpacing' },
    '0052,0016' => { VR => 'SQ', Name => 'ModeOfPercutaneousAccessSequence' },
    '0052,0025' => { VR => 'SQ', Name => 'IntravascularOCTFrameTypeSequence' },
    '0052,0026' => { VR => 'CS', Name => 'OCTZOffsetApplied' },
    '0052,0027' => { VR => 'SQ', Name => 'IntravascularFrameContentSequence' },
    '0052,0028' => { VR => 'FD', Name => 'IntravascularLongitudinalDistance' },
    '0052,0029' => { VR => 'SQ', Name => 'IntravascularOCTFrameContentSeq' },
    '0052,0030' => { VR => 'SS', Name => 'OCTZOffsetCorrection' },
    '0052,0031' => { VR => 'CS', Name => 'CatheterDirectionOfRotation' },
    '0052,0033' => { VR => 'FD', Name => 'SeamLineLocation' },
    '0052,0034' => { VR => 'FD', Name => 'FirstALineLocation' },
    '0052,0036' => { VR => 'US', Name => 'SeamLineIndex' },
    '0052,0038' => { VR => 'US', Name => 'NumberOfPaddedALines' },
    '0052,0039' => { VR => 'CS', Name => 'InterpolationType' },
    '0052,003A' => { VR => 'CS', Name => 'RefractiveIndexApplied' },
    # nuclear acquisition group
    '0054,0010' => { VR => 'US', Name => 'EnergyWindowVector' },
    '0054,0011' => { VR => 'US', Name => 'NumberOfEnergyWindows' },
    '0054,0012' => { VR => 'SQ', Name => 'EnergyWindowInformationSequence' },
    '0054,0013' => { VR => 'SQ', Name => 'EnergyWindowRangeSequence' },
    '0054,0014' => { VR => 'DS', Name => 'EnergyWindowLowerLimit' },
    '0054,0015' => { VR => 'DS', Name => 'EnergyWindowUpperLimit' },
    '0054,0016' => { VR => 'SQ', Name => 'RadiopharmaceuticalInformationSeq' },
    '0054,0017' => { VR => 'IS', Name => 'ResidualSyringeCounts' },
    '0054,0018' => { VR => 'SH', Name => 'EnergyWindowName' },
    '0054,0020' => { VR => 'US', Name => 'DetectorVector' },
    '0054,0021' => { VR => 'US', Name => 'NumberOfDetectors' },
    '0054,0022' => { VR => 'SQ', Name => 'DetectorInformationSequence' },
    '0054,0030' => { VR => 'US', Name => 'PhaseVector' },
    '0054,0031' => { VR => 'US', Name => 'NumberOfPhases' },
    '0054,0032' => { VR => 'SQ', Name => 'PhaseInformationSequence' },
    '0054,0033' => { VR => 'US', Name => 'NumberOfFramesInPhase' },
    '0054,0036' => { VR => 'IS', Name => 'PhaseDelay' },
    '0054,0038' => { VR => 'IS', Name => 'PauseBetweenFrames' },
    '0054,0039' => { VR => 'CS', Name => 'PhaseDescription' },
    '0054,0050' => { VR => 'US', Name => 'RotationVector' },
    '0054,0051' => { VR => 'US', Name => 'NumberOfRotations' },
    '0054,0052' => { VR => 'SQ', Name => 'RotationInformationSequence' },
    '0054,0053' => { VR => 'US', Name => 'NumberOfFramesInRotation' },
    '0054,0060' => { VR => 'US', Name => 'RRIntervalVector' },
    '0054,0061' => { VR => 'US', Name => 'NumberOfRRIntervals' },
    '0054,0062' => { VR => 'SQ', Name => 'GatedInformationSequence' },
    '0054,0063' => { VR => 'SQ', Name => 'DataInformationSequence' },
    '0054,0070' => { VR => 'US', Name => 'TimeSlotVector' },
    '0054,0071' => { VR => 'US', Name => 'NumberOfTimeSlots' },
    '0054,0072' => { VR => 'SQ', Name => 'TimeSlotInformationSequence' },
    '0054,0073' => { VR => 'DS', Name => 'TimeSlotTime' },
    '0054,0080' => { VR => 'US', Name => 'SliceVector' },
    '0054,0081' => { VR => 'US', Name => 'NumberOfSlices' },
    '0054,0090' => { VR => 'US', Name => 'AngularViewVector' },
    '0054,0100' => { VR => 'US', Name => 'TimeSliceVector' },
    '0054,0101' => { VR => 'US', Name => 'NumberOfTimeSlices' },
    '0054,0200' => { VR => 'DS', Name => 'StartAngle' },
    '0054,0202' => { VR => 'CS', Name => 'TypeOfDetectorMotion' },
    '0054,0210' => { VR => 'IS', Name => 'TriggerVector' },
    '0054,0211' => { VR => 'US', Name => 'NumberOfTriggersInPhase' },
    '0054,0220' => { VR => 'SQ', Name => 'ViewCodeSequence' },
    '0054,0222' => { VR => 'SQ', Name => 'ViewModifierCodeSequence' },
    '0054,0300' => { VR => 'SQ', Name => 'RadionuclideCodeSequence' },
    '0054,0302' => { VR => 'SQ', Name => 'AdministrationRouteCodeSequence' },
    '0054,0304' => { VR => 'SQ', Name => 'RadiopharmaceuticalCodeSequence' },
    '0054,0306' => { VR => 'SQ', Name => 'CalibrationDataSequence' },
    '0054,0308' => { VR => 'US', Name => 'EnergyWindowNumber' },
    '0054,0400' => { VR => 'SH', Name => 'ImageID' },
    '0054,0410' => { VR => 'SQ', Name => 'PatientOrientationCodeSequence' },
    '0054,0412' => { VR => 'SQ', Name => 'PatientOrientationModifierCodeSeq' },
    '0054,0414' => { VR => 'SQ', Name => 'PatientGantryRelationshipCodeSeq' },
    '0054,0500' => { VR => 'CS', Name => 'SliceProgressionDirection' },
    '0054,0501' => { VR => 'CS', Name => 'ScanProgressionDirection' },
    '0054,1000' => { VR => 'CS', Name => 'SeriesType' },
    '0054,1001' => { VR => 'CS', Name => 'Units' },
    '0054,1002' => { VR => 'CS', Name => 'CountsSource' },
    '0054,1004' => { VR => 'CS', Name => 'ReprojectionMethod' },
    '0054,1006' => { VR => 'CS', Name => 'SUVType' },
    '0054,1100' => { VR => 'CS', Name => 'RandomsCorrectionMethod' },
    '0054,1101' => { VR => 'LO', Name => 'AttenuationCorrectionMethod' },
    '0054,1102' => { VR => 'CS', Name => 'DecayCorrection' },
    '0054,1103' => { VR => 'LO', Name => 'ReconstructionMethod' },
    '0054,1104' => { VR => 'LO', Name => 'DetectorLinesOfResponseUsed' },
    '0054,1105' => { VR => 'LO', Name => 'ScatterCorrectionMethod' },
    '0054,1200' => { VR => 'DS', Name => 'AxialAcceptance' },
    '0054,1201' => { VR => 'IS', Name => 'AxialMash' },
    '0054,1202' => { VR => 'IS', Name => 'TransverseMash' },
    '0054,1203' => { VR => 'DS', Name => 'DetectorElementSize' },
    '0054,1210' => { VR => 'DS', Name => 'CoincidenceWindowWidth' },
    '0054,1220' => { VR => 'CS', Name => 'SecondaryCountsType' },
    '0054,1300' => { VR => 'DS', Name => 'FrameReferenceTime' },
    '0054,1310' => { VR => 'IS', Name => 'PrimaryCountsAccumulated' },
    '0054,1311' => { VR => 'IS', Name => 'SecondaryCountsAccumulated' },
    '0054,1320' => { VR => 'DS', Name => 'SliceSensitivityFactor' },
    '0054,1321' => { VR => 'DS', Name => 'DecayFactor' },
    '0054,1322' => { VR => 'DS', Name => 'DoseCalibrationFactor' },
    '0054,1323' => { VR => 'DS', Name => 'ScatterFractionFactor' },
    '0054,1324' => { VR => 'DS', Name => 'DeadTimeFactor' },
    '0054,1330' => { VR => 'US', Name => 'ImageIndex' },
    '0054,1400' => { VR => 'CS', Name => 'CountsIncluded' },
    '0054,1401' => { VR => 'CS', Name => 'DeadTimeCorrectionFlag' },
    '0060,3000' => { VR => 'SQ', Name => 'HistogramSequence' },
    '0060,3002' => { VR => 'US', Name => 'HistogramNumberOfBins' },
    '0060,3004' => { VR => 'US', Name => 'HistogramFirstBinValue' },
    '0060,3006' => { VR => 'US', Name => 'HistogramLastBinValue' },
    '0060,3008' => { VR => 'US', Name => 'HistogramBinWidth' },
    '0060,3010' => { VR => 'LO', Name => 'HistogramExplanation' },
    '0060,3020' => { VR => 'UL', Name => 'HistogramData' },
    '0062,0001' => { VR => 'CS', Name => 'SegmentationType' },
    '0062,0002' => { VR => 'SQ', Name => 'SegmentSequence' },
    '0062,0003' => { VR => 'SQ', Name => 'SegmentedPropertyCategoryCodeSeq' },
    '0062,0004' => { VR => 'US', Name => 'SegmentNumber' },
    '0062,0005' => { VR => 'LO', Name => 'SegmentLabel' },
    '0062,0006' => { VR => 'ST', Name => 'SegmentDescription' },
    '0062,0007' => { VR => 'SQ', Name => 'SegmentationAlgorithmIDSequence' },
    '0062,0008' => { VR => 'CS', Name => 'SegmentAlgorithmType' },
    '0062,0009' => { VR => 'LO', Name => 'SegmentAlgorithmName' },
    '0062,000A' => { VR => 'SQ', Name => 'SegmentIdentificationSequence' },
    '0062,000B' => { VR => 'US', Name => 'ReferencedSegmentNumber' },
    '0062,000C' => { VR => 'US', Name => 'RecommendedDisplayGrayscaleValue' },
    '0062,000D' => { VR => 'US', Name => 'RecommendedDisplayCIELabValue' },
    '0062,000E' => { VR => 'US', Name => 'MaximumFractionalValue' },
    '0062,000F' => { VR => 'SQ', Name => 'SegmentedPropertyTypeCodeSequence' },
    '0062,0010' => { VR => 'CS', Name => 'SegmentationFractionalType' },
    '0062,0011' => { VR => 'SQ', Name => 'SegmentedPropertyTypeModCodeSeq' },
    '0062,0012' => { VR => 'SQ', Name => 'UsedSegmentsSequence' },
    '0062,0013' => { VR => 'CS', Name => 'SegmentsOverlap' },
    '0062,0020' => { VR => 'UT', Name => 'TrackingID' },
    '0062,0021' => { VR => 'UI', Name => 'TrackingUID' },
    '0064,0002' => { VR => 'SQ', Name => 'DeformableRegistrationSequence' },
    '0064,0003' => { VR => 'UI', Name => 'SourceFrameOfReferenceUID' },
    '0064,0005' => { VR => 'SQ', Name => 'DeformableRegistrationGridSequence' },
    '0064,0007' => { VR => 'UL', Name => 'GridDimensions' },
    '0064,0008' => { VR => 'FD', Name => 'GridResolution' },
    '0064,0009' => { VR => 'OF', Name => 'VectorGridData' },
    '0064,000F' => { VR => 'SQ', Name => 'PreDeformationMatrixRegistSeq' },
    '0064,0010' => { VR => 'SQ', Name => 'PostDeformationMatrixRegistSeq' },
    '0066,0001' => { VR => 'UL', Name => 'NumberOfSurfaces' },
    '0066,0002' => { VR => 'SQ', Name => 'SurfaceSequence' },
    '0066,0003' => { VR => 'UL', Name => 'SurfaceNumber' },
    '0066,0004' => { VR => 'LT', Name => 'SurfaceComments' },
    '0066,0005' => { VR => 'FL', Name => 'SurfaceOffset' },
    '0066,0009' => { VR => 'CS', Name => 'SurfaceProcessing' },
    '0066,000A' => { VR => 'FL', Name => 'SurfaceProcessingRatio' },
    '0066,000B' => { VR => 'LO', Name => 'SurfaceProcessingDescription' },
    '0066,000C' => { VR => 'FL', Name => 'RecommendedPresentationOpacity' },
    '0066,000D' => { VR => 'CS', Name => 'RecommendedPresentationType' },
    '0066,000E' => { VR => 'CS', Name => 'FiniteVolume' },
    '0066,0010' => { VR => 'CS', Name => 'Manifold' },
    '0066,0011' => { VR => 'SQ', Name => 'SurfacePointsSequence' },
    '0066,0012' => { VR => 'SQ', Name => 'SurfacePointsNormalsSequence' },
    '0066,0013' => { VR => 'SQ', Name => 'SurfaceMeshPrimitivesSequence' },
    '0066,0015' => { VR => 'UL', Name => 'NumberOfSurfacePoints' },
    '0066,0016' => { VR => 'OF', Name => 'PointCoordinatesData' },
    '0066,0017' => { VR => 'FL', Name => 'PointPositionAccuracy' },
    '0066,0018' => { VR => 'FL', Name => 'MeanPointDistance' },
    '0066,0019' => { VR => 'FL', Name => 'MaximumPointDistance' },
    '0066,001A' => { VR => 'FL', Name => 'PointsBoundingBoxCoordinates' },
    '0066,001B' => { VR => 'FL', Name => 'AxisOfRotation' },
    '0066,001C' => { VR => 'FL', Name => 'CenterOfRotation' },
    '0066,001E' => { VR => 'UL', Name => 'NumberOfVectors' },
    '0066,001F' => { VR => 'US', Name => 'VectorDimensionality' },
    '0066,0020' => { VR => 'FL', Name => 'VectorAccuracy' },
    '0066,0021' => { VR => 'OF', Name => 'VectorCoordinateData' },
    '0066,0022' => { VR => 'OD', Name => 'DoublePointCoordinatesData' },
    '0066,0023' => { VR => 'OW', Name => 'TrianglePointIndexList' },
    '0066,0024' => { VR => 'OW', Name => 'EdgePointIndexList' },
    '0066,0025' => { VR => 'OW', Name => 'VertexPointIndexList' },
    '0066,0026' => { VR => 'SQ', Name => 'TriangleStripSequence' },
    '0066,0027' => { VR => 'SQ', Name => 'TriangleFanSequence' },
    '0066,0028' => { VR => 'SQ', Name => 'LineSequence' },
    '0066,0029' => { VR => 'OW', Name => 'PrimitivePointIndexList' },
    '0066,002A' => { VR => 'UL', Name => 'SurfaceCount' },
    '0066,002B' => { VR => 'SQ', Name => 'ReferencedSurfaceSequence' },
    '0066,002C' => { VR => 'UL', Name => 'ReferencedSurfaceNumber' },
    '0066,002D' => { VR => 'SQ', Name => 'SegmentSurfaceGenAlgorithmIDSeq' },
    '0066,002E' => { VR => 'SQ', Name => 'SegmentSurfaceSourceInstanceSeq' },
    '0066,002F' => { VR => 'SQ', Name => 'AlgorithmFamilyCodeSequ' },
    '0066,0030' => { VR => 'SQ', Name => 'AlgorithmNameCodeSequence' },
    '0066,0031' => { VR => 'LO', Name => 'AlgorithmVersion' },
    '0066,0032' => { VR => 'LT', Name => 'AlgorithmParameters' },
    '0066,0034' => { VR => 'SQ', Name => 'FacetSequence' },
    '0066,0035' => { VR => 'SQ', Name => 'SurfaceProcessingAlgorithmIDSeq' },
    '0066,0036' => { VR => 'LO', Name => 'AlgorithmName' },
    '0066,0037' => { VR => 'FL', Name => 'RecommendedPointRadius' },
    '0066,0038' => { VR => 'FL', Name => 'RecommendedLineThickness' },
    '0066,0040' => { VR => 'OL', Name => 'LongPrimitivePointIndexList' },
    '0066,0041' => { VR => 'OL', Name => 'LongTrianglePointIndexList' },
    '0066,0042' => { VR => 'OL', Name => 'LongEdgePointIndexList' },
    '0066,0043' => { VR => 'OL', Name => 'LongVertexPointIndexList' },
    '0066,0101' => { VR => 'SQ', Name => 'TrackSetSequence' },
    '0066,0102' => { VR => 'SQ', Name => 'TrackSequence' },
    '0066,0103' => { VR => 'OW', Name => 'RecommendedDisplayCIELabValueList' },
    '0066,0104' => { VR => 'SQ', Name => 'TrackingAlgorithmIDSequence' },
    '0066,0105' => { VR => 'UL', Name => 'TrackSetNumber' },
    '0066,0106' => { VR => 'LO', Name => 'TrackSetLabel' },
    '0066,0107' => { VR => 'UT', Name => 'TrackSetDescription' },
    '0066,0108' => { VR => 'SQ', Name => 'TrackSetAnatomicalTypeCodeSequence' },
    '0066,0121' => { VR => 'SQ', Name => 'MeasurementsSequence' },
    '0066,0124' => { VR => 'SQ', Name => 'TrackSetStatisticsSequence' },
    '0066,0125' => { VR => 'OF', Name => 'FloatingPointValues' },
    '0066,0129' => { VR => 'OL', Name => 'TrackPointIndexList' },
    '0066,0130' => { VR => 'SQ', Name => 'TrackStatisticsSequence' },
    '0066,0132' => { VR => 'SQ', Name => 'MeasurementValuesSequence' },
    '0066,0133' => { VR => 'SQ', Name => 'DiffusionAcquisitionCodeSequence' },
    '0066,0134' => { VR => 'SQ', Name => 'DiffusionModelCodeSequence' },
    '0068,6210' => { VR => 'LO', Name => 'ImplantSize' },
    '0068,6221' => { VR => 'LO', Name => 'ImplantTemplateVersion' },
    '0068,6222' => { VR => 'SQ', Name => 'ReplacedImplantTemplateSequence' },
    '0068,6223' => { VR => 'CS', Name => 'ImplantType' },
    '0068,6224' => { VR => 'SQ', Name => 'DerivationImplantTemplateSequence' },
    '0068,6225' => { VR => 'SQ', Name => 'OriginalImplantTemplateSequence' },
    '0068,6226' => { VR => 'DT', Name => 'EffectiveDateTime' },
    '0068,6230' => { VR => 'SQ', Name => 'ImplantTargetAnatomySequence' },
    '0068,6260' => { VR => 'SQ', Name => 'InformationFromManufacturerSeq' },
    '0068,6265' => { VR => 'SQ', Name => 'NotificationFromManufacturerSeq' },
    '0068,6270' => { VR => 'DT', Name => 'InformationIssueDateTime' },
    '0068,6280' => { VR => 'ST', Name => 'InformationSummary' },
    '0068,62A0' => { VR => 'SQ', Name => 'ImplantRegulatoryDisapprovalCode' },
    '0068,62A5' => { VR => 'FD', Name => 'OverallTemplateSpatialTolerance' },
    '0068,62C0' => { VR => 'SQ', Name => 'HPGLDocumentSequence' },
    '0068,62D0' => { VR => 'US', Name => 'HPGLDocumentID' },
    '0068,62D5' => { VR => 'LO', Name => 'HPGLDocumentLabel' },
    '0068,62E0' => { VR => 'SQ', Name => 'ViewOrientationCodeSequence' },
    '0068,62F0' => { VR => 'SQ', Name => 'ViewOrientationModifierCodeSeq' },
    '0068,62F2' => { VR => 'FD', Name => 'HPGLDocumentScaling' },
    '0068,6300' => { VR => 'OB', Name => 'HPGLDocument' },
    '0068,6310' => { VR => 'US', Name => 'HPGLContourPenNumber' },
    '0068,6320' => { VR => 'SQ', Name => 'HPGLPenSequence' },
    '0068,6330' => { VR => 'US', Name => 'HPGLPenNumber' },
    '0068,6340' => { VR => 'LO', Name => 'HPGLPenLabel' },
    '0068,6345' => { VR => 'ST', Name => 'HPGLPenDescription' },
    '0068,6346' => { VR => 'FD', Name => 'RecommendedRotationPoint' },
    '0068,6347' => { VR => 'FD', Name => 'BoundingRectangle' },
    '0068,6350' => { VR => 'US', Name => 'ImplantTemplate3DModelSurfaceNum' },
    '0068,6360' => { VR => 'SQ', Name => 'SurfaceModelDescriptionSequence' },
    '0068,6380' => { VR => 'LO', Name => 'SurfaceModelLabel' },
    '0068,6390' => { VR => 'FD', Name => 'SurfaceModelScalingFactor' },
    '0068,63A0' => { VR => 'SQ', Name => 'MaterialsCodeSequence' },
    '0068,63A4' => { VR => 'SQ', Name => 'CoatingMaterialsCodeSequence' },
    '0068,63A8' => { VR => 'SQ', Name => 'ImplantTypeCodeSequence' },
    '0068,63AC' => { VR => 'SQ', Name => 'FixationMethodCodeSequence' },
    '0068,63B0' => { VR => 'SQ', Name => 'MatingFeatureSetsSequence' },
    '0068,63C0' => { VR => 'US', Name => 'MatingFeatureSetID' },
    '0068,63D0' => { VR => 'LO', Name => 'MatingFeatureSetLabel' },
    '0068,63E0' => { VR => 'SQ', Name => 'MatingFeatureSequence' },
    '0068,63F0' => { VR => 'US', Name => 'MatingFeatureID' },
    '0068,6400' => { VR => 'SQ', Name => 'MatingFeatureDegreeOfFreedomSeq' },
    '0068,6410' => { VR => 'US', Name => 'DegreeOfFreedomID' },
    '0068,6420' => { VR => 'CS', Name => 'DegreeOfFreedomType' },
    '0068,6430' => { VR => 'SQ', Name => 'TwoDMatingFeatureCoordinatesSeq' },
    '0068,6440' => { VR => 'US', Name => 'ReferencedHPGLDocumentID' },
    '0068,6450' => { VR => 'FD', Name => 'TwoDMatingPoint' },
    '0068,6460' => { VR => 'FD', Name => 'TwoDMatingAxes' },
    '0068,6470' => { VR => 'SQ', Name => 'TwoDDegreeOfFreedomSequence' },
    '0068,6490' => { VR => 'FD', Name => 'ThreeDDegreeOfFreedomAxis' },
    '0068,64A0' => { VR => 'FD', Name => 'RangeOfFreedom' },
    '0068,64C0' => { VR => 'FD', Name => 'ThreeDMatingPoint' },
    '0068,64D0' => { VR => 'FD', Name => 'ThreeDMatingAxes' },
    '0068,64F0' => { VR => 'FD', Name => 'TwoDDegreeOfFreedomAxis' },
    '0068,6500' => { VR => 'SQ', Name => 'PlanningLandmarkPointSequence' },
    '0068,6510' => { VR => 'SQ', Name => 'PlanningLandmarkLineSequence' },
    '0068,6520' => { VR => 'SQ', Name => 'PlanningLandmarkPlaneSequence' },
    '0068,6530' => { VR => 'US', Name => 'PlanningLandmarkID' },
    '0068,6540' => { VR => 'LO', Name => 'PlanningLandmarkDescription' },
    '0068,6545' => { VR => 'SQ', Name => 'PlanningLandmarkIDCodeSeq' },
    '0068,6550' => { VR => 'SQ', Name => 'TwoDPointCoordinatesSequence' },
    '0068,6560' => { VR => 'FD', Name => 'TwoDPointCoordinates' },
    '0068,6590' => { VR => 'FD', Name => 'ThreeDPointCoordinates' },
    '0068,65A0' => { VR => 'SQ', Name => 'TwoDLineCoordinatesSequence' },
    '0068,65B0' => { VR => 'FD', Name => 'TwoDLineCoordinates' },
    '0068,65D0' => { VR => 'FD', Name => 'ThreeDLineCoordinates' },
    '0068,65E0' => { VR => 'SQ', Name => 'TwoDPlaneCoordinatesSequence' },
    '0068,65F0' => { VR => 'FD', Name => 'TwoDPlaneIntersection' },
    '0068,6610' => { VR => 'FD', Name => 'ThreeDPlaneOrigin' },
    '0068,6620' => { VR => 'FD', Name => 'ThreeDPlaneNormal' },
    '0068,7001' => { VR => 'CS', Name => 'ModelModification' },
    '0068,7002' => { VR => 'CS', Name => 'ModelMirroring' },
    '0068,7003' => { VR => 'SQ', Name => 'ModelUsageCodeSequence' },
    '0068,7004' => { VR => 'UI', Name => 'ModelGroupUID' },
    '0068,7005' => { VR => 'UR', Name => 'RelativeURIRefWithinEncapsulDoc' },
    '006A,0001' => { VR => 'CS', Name => 'AnnotationCoordinateType' },
    '006A,0002' => { VR => 'SQ', Name => 'AnnotationGroupSequence' },
    '006A,0003' => { VR => 'UI', Name => 'AnnotationGroupUID' },
    '006A,0005' => { VR => 'LO', Name => 'AnnotationGroupLabel' },
    '006A,0006' => { VR => 'UT', Name => 'AnnotationGroupDescription' },
    '006A,0007' => { VR => 'CS', Name => 'AnnotationGroupGenerationType' },
    '006A,0008' => { VR => 'SQ', Name => 'AnnotationGroupAlgorithmIDSeq' },
    '006A,0009' => { VR => 'SQ', Name => 'AnnotationPropertyCategoryCodeSeq' },
    '006A,000A' => { VR => 'SQ', Name => 'AnnotationPropertyTypeCodeSequence' },
    '006A,000B' => { VR => 'SQ', Name => 'AnnotationPropertyTypeModCodeSeq' },
    '006A,000C' => { VR => 'UL', Name => 'NumberOfAnnotations' },
    '006A,000D' => { VR => 'CS', Name => 'AnnotationAppliesToAllOpticalPaths' },
    '006A,000E' => { VR => 'SH', Name => 'ReferencedOpticalPathIdentifier' },
    '006A,000F' => { VR => 'CS', Name => 'AnnotationAppliesToAllZPlanes' },
    '006A,0010' => { VR => 'FD', Name => 'CommonZCoordinateValue' },
    '006A,0011' => { VR => 'OL', Name => 'AnnotationIndexList' },
    '0070,0001' => { VR => 'SQ', Name => 'GraphicAnnotationSequence' },
    '0070,0002' => { VR => 'CS', Name => 'GraphicLayer' },
    '0070,0003' => { VR => 'CS', Name => 'BoundingBoxAnnotationUnits' },
    '0070,0004' => { VR => 'CS', Name => 'AnchorPointAnnotationUnits' },
    '0070,0005' => { VR => 'CS', Name => 'GraphicAnnotationUnits' },
    '0070,0006' => { VR => 'ST', Name => 'UnformattedTextValue' },
    '0070,0008' => { VR => 'SQ', Name => 'TextObjectSequence' },
    '0070,0009' => { VR => 'SQ', Name => 'GraphicObjectSequence' },
    '0070,0010' => { VR => 'FL', Name => 'BoundingBoxTopLeftHandCorner' },
    '0070,0011' => { VR => 'FL', Name => 'BoundingBoxBottomRightHandCorner' },
    '0070,0012' => { VR => 'CS', Name => 'BoundingBoxTextHorizJustification' },
    '0070,0014' => { VR => 'FL', Name => 'AnchorPoint' },
    '0070,0015' => { VR => 'CS', Name => 'AnchorPointVisibility' },
    '0070,0020' => { VR => 'US', Name => 'GraphicDimensions' },
    '0070,0021' => { VR => 'US', Name => 'NumberOfGraphicPoints' },
    '0070,0022' => { VR => 'FL', Name => 'GraphicData' },
    '0070,0023' => { VR => 'CS', Name => 'GraphicType' },
    '0070,0024' => { VR => 'CS', Name => 'GraphicFilled' },
    '0070,0040' => { VR => 'IS', Name => 'ImageRotationRetired' },
    '0070,0041' => { VR => 'CS', Name => 'ImageHorizontalFlip' },
    '0070,0042' => { VR => 'US', Name => 'ImageRotation' },
    '0070,0050' => { VR => 'US', Name => 'DisplayedAreaTopLeftTrial' },
    '0070,0051' => { VR => 'US', Name => 'DisplayedAreaBottomRightTrial' },
    '0070,0052' => { VR => 'SL', Name => 'DisplayedAreaTopLeft' },
    '0070,0053' => { VR => 'SL', Name => 'DisplayedAreaBottomRight' },
    '0070,005A' => { VR => 'SQ', Name => 'DisplayedAreaSelectionSequence' },
    '0070,0060' => { VR => 'SQ', Name => 'GraphicLayerSequence' },
    '0070,0062' => { VR => 'IS', Name => 'GraphicLayerOrder' },
    '0070,0066' => { VR => 'US', Name => 'GraphicLayerRecDisplayGraysclValue' },
    '0070,0067' => { VR => 'US', Name => 'GraphicLayerRecDisplayRGBValue' },
    '0070,0068' => { VR => 'LO', Name => 'GraphicLayerDescription' },
    '0070,0080' => { VR => 'CS', Name => 'ContentLabel' },
    '0070,0081' => { VR => 'LO', Name => 'ContentDescription' },
    '0070,0082' => { VR => 'DA', Name => 'PresentationCreationDate' },
    '0070,0083' => { VR => 'TM', Name => 'PresentationCreationTime' },
    '0070,0084' => { VR => 'PN', Name => 'ContentCreatorName' },
    '0070,0086' => { VR => 'SQ', Name => 'ContentCreatorIDCodeSequence' },
    '0070,0087' => { VR => 'SQ', Name => 'AlternateContentDescriptionSeq' },
    '0070,0100' => { VR => 'CS', Name => 'PresentationSizeMode' },
    '0070,0101' => { VR => 'DS', Name => 'PresentationPixelSpacing' },
    '0070,0102' => { VR => 'IS', Name => 'PresentationPixelAspectRatio' },
    '0070,0103' => { VR => 'FL', Name => 'PresentationPixelMagRatio' },
    '0070,0207' => { VR => 'LO', Name => 'GraphicGroupLabel' },
    '0070,0208' => { VR => 'ST', Name => 'GraphicGroupDescription' },
    '0070,0209' => { VR => 'SQ', Name => 'CompoundGraphicSequence' },
    '0070,0226' => { VR => 'UL', Name => 'CompoundGraphicInstanceID' },
    '0070,0227' => { VR => 'LO', Name => 'FontName' },
    '0070,0228' => { VR => 'CS', Name => 'FontNameType' },
    '0070,0229' => { VR => 'LO', Name => 'CSSFontName' },
    '0070,0230' => { VR => 'FD', Name => 'RotationAngle' },
    '0070,0231' => { VR => 'SQ', Name => 'TextStyleSequence' },
    '0070,0232' => { VR => 'SQ', Name => 'LineStyleSequence' },
    '0070,0233' => { VR => 'SQ', Name => 'FillStyleSequence' },
    '0070,0234' => { VR => 'SQ', Name => 'GraphicGroupSequence' },
    '0070,0241' => { VR => 'US', Name => 'TextColorCIELabValue' },
    '0070,0242' => { VR => 'CS', Name => 'HorizontalAlignment' },
    '0070,0243' => { VR => 'CS', Name => 'VerticalAlignment' },
    '0070,0244' => { VR => 'CS', Name => 'ShadowStyle' },
    '0070,0245' => { VR => 'FL', Name => 'ShadowOffsetX' },
    '0070,0246' => { VR => 'FL', Name => 'ShadowOffsetY' },
    '0070,0247' => { VR => 'US', Name => 'ShadowColorCIELabValue' },
    '0070,0248' => { VR => 'CS', Name => 'Underlined' },
    '0070,0249' => { VR => 'CS', Name => 'Bold' },
    '0070,0250' => { VR => 'CS', Name => 'Italic' },
    '0070,0251' => { VR => 'US', Name => 'PatternOnColorCIELabValue' },
    '0070,0252' => { VR => 'US', Name => 'PatternOffColorCIELabValue' },
    '0070,0253' => { VR => 'FL', Name => 'LineThickness' },
    '0070,0254' => { VR => 'CS', Name => 'LineDashingStyle' },
    '0070,0255' => { VR => 'UL', Name => 'LinePattern' },
    '0070,0256' => { VR => 'OB', Name => 'FillPattern' },
    '0070,0257' => { VR => 'CS', Name => 'FillMode' },
    '0070,0258' => { VR => 'FL', Name => 'ShadowOpacity' },
    '0070,0261' => { VR => 'FL', Name => 'GapLength' },
    '0070,0262' => { VR => 'FL', Name => 'DiameterOfVisibility' },
    '0070,0273' => { VR => 'FL', Name => 'RotationPoint' },
    '0070,0274' => { VR => 'CS', Name => 'TickAlignment' },
    '0070,0278' => { VR => 'CS', Name => 'ShowTickLabel' },
    '0070,0279' => { VR => 'CS', Name => 'TickLabelAlignment' },
    '0070,0282' => { VR => 'CS', Name => 'CompoundGraphicUnits' },
    '0070,0284' => { VR => 'FL', Name => 'PatternOnOpacity' },
    '0070,0285' => { VR => 'FL', Name => 'PatternOffOpacity' },
    '0070,0287' => { VR => 'SQ', Name => 'MajorTicksSequence' },
    '0070,0288' => { VR => 'FL', Name => 'TickPosition' },
    '0070,0289' => { VR => 'SH', Name => 'TickLabel' },
    '0070,0294' => { VR => 'CS', Name => 'CompoundGraphicType' },
    '0070,0295' => { VR => 'UL', Name => 'GraphicGroupID' },
    '0070,0306' => { VR => 'CS', Name => 'ShapeType' },
    '0070,0308' => { VR => 'SQ', Name => 'RegistrationSequence' },
    '0070,0309' => { VR => 'SQ', Name => 'MatrixRegistrationSequence' },
    '0070,030A' => { VR => 'SQ', Name => 'MatrixSequence' },
    '0070,030B' => { VR => 'FD', Name => 'FrameOfRef2DisplayCoordTransMatrix' },
    '0070,030C' => { VR => 'CS', Name => 'FrameOfRefTransformationMatrixType' },
    '0070,030D' => { VR => 'SQ', Name => 'RegistrationTypeCodeSequence' },
    '0070,030F' => { VR => 'ST', Name => 'FiducialDescription' },
    '0070,0310' => { VR => 'SH', Name => 'FiducialIdentifier' },
    '0070,0311' => { VR => 'SQ', Name => 'FiducialIdentifierCodeSequence' },
    '0070,0312' => { VR => 'FD', Name => 'ContourUncertaintyRadius' },
    '0070,0314' => { VR => 'SQ', Name => 'UsedFiducialsSequence' },
    '0070,0315' => { VR => 'SQ', Name => 'UsedRTStructureSetROISequence' },
    '0070,0318' => { VR => 'SQ', Name => 'GraphicCoordinatesDataSequence' },
    '0070,031A' => { VR => 'UI', Name => 'FiducialUID' },
    '0070,031B' => { VR => 'UI', Name => 'ReferencedFiducialUID' },
    '0070,031C' => { VR => 'SQ', Name => 'FiducialSetSequence' },
    '0070,031E' => { VR => 'SQ', Name => 'FiducialSequence' },
    '0070,031F' => { VR => 'SQ', Name => 'FiducialsPropertyCategoryCodeSeq' },
    '0070,0401' => { VR => 'US', Name => 'GraphicLayerRecomDisplayCIELabVal' },
    '0070,0402' => { VR => 'SQ', Name => 'BlendingSequence' },
    '0070,0403' => { VR => 'FL', Name => 'RelativeOpacity' },
    '0070,0404' => { VR => 'SQ', Name => 'ReferencedSpatialRegistrationSeq' },
    '0070,0405' => { VR => 'CS', Name => 'BlendingPosition' },
    '0070,1101' => { VR => 'UI', Name => 'PresentationDisplayCollectionUID' },
    '0070,1102' => { VR => 'UI', Name => 'PresentationSequenceCollectionUID' },
    '0070,1103' => { VR => 'US', Name => 'PresentationSequencePositionIndex' },
    '0070,1104' => { VR => 'SQ', Name => 'RenderedImageReferenceSequence' },
    '0070,1201' => { VR => 'SQ', Name => 'VolumetricPresentStateInputSeq' },
    '0070,1202' => { VR => 'CS', Name => 'PresentationInputType' },
    '0070,1203' => { VR => 'US', Name => 'InputSequencePositionIndex' },
    '0070,1204' => { VR => 'CS', Name => 'Crop' },
    '0070,1205' => { VR => 'US', Name => 'CroppingSpecificationIndex' },
    '0070,1206' => { VR => 'CS', Name => 'CompositingMethod' },
    '0070,1207' => { VR => 'US', Name => 'VolumetricPresentationInputNumber' },
    '0070,1208' => { VR => 'CS', Name => 'ImageVolumeGeometry' },
    '0070,1209' => { VR => 'UI', Name => 'VolumetricPresentationInputSetUID' },
    '0070,120A' => { VR => 'SQ', Name => 'VolumetricPresentationInputSetSeq' },
    '0070,120B' => { VR => 'CS', Name => 'GlobalCrop' },
    '0070,120C' => { VR => 'US', Name => 'GlobalCroppingSpecificationIndex' },
    '0070,120D' => { VR => 'CS', Name => 'RenderingMethod' },
    '0070,1301' => { VR => 'SQ', Name => 'VolumeCroppingSequence' },
    '0070,1302' => { VR => 'CS', Name => 'VolumeCroppingMethod' },
    '0070,1303' => { VR => 'FD', Name => 'BoundingBoxCrop' },
    '0070,1304' => { VR => 'SQ', Name => 'ObliqueCroppingPlaneSequence' },
    '0070,1305' => { VR => 'FD', Name => 'Plane' },
    '0070,1306' => { VR => 'FD', Name => 'PlaneNormal' },
    '0070,1309' => { VR => 'US', Name => 'CroppingSpecificationNumber' },
    '0070,1501' => { VR => 'CS', Name => 'MultiPlanarReconstructionStyle' },
    '0070,1502' => { VR => 'CS', Name => 'MPRThicknessType' },
    '0070,1503' => { VR => 'FD', Name => 'MPRSlabThickness' },
    '0070,1505' => { VR => 'FD', Name => 'MPRTopLeftHandCorner' },
    '0070,1507' => { VR => 'FD', Name => 'MPRViewWidthDirection' },
    '0070,1508' => { VR => 'FD', Name => 'MPRViewWidth' },
    '0070,150C' => { VR => 'UL', Name => 'NumberOfVolumetricCurvePoints' },
    '0070,150D' => { VR => 'OD', Name => 'VolumetricCurvePoints' },
    '0070,1511' => { VR => 'FD', Name => 'MPRViewHeightDirection' },
    '0070,1512' => { VR => 'FD', Name => 'MPRViewHeight' },
    '0070,1602' => { VR => 'CS', Name => 'RenderProjection' },
    '0070,1603' => { VR => 'FD', Name => 'ViewpointPosition' },
    '0070,1604' => { VR => 'FD', Name => 'ViewpointLookAtPoint' },
    '0070,1605' => { VR => 'FD', Name => 'ViewpointUpDirection' },
    '0070,1606' => { VR => 'FD', Name => 'RenderFieldOfView' },
    '0070,1607' => { VR => 'FD', Name => 'SamplingStepSize' },
    '0070,1701' => { VR => 'CS', Name => 'ShadingStyle' },
    '0070,1702' => { VR => 'FD', Name => 'AmbientReflectionIntensity' },
    '0070,1703' => { VR => 'FD', Name => 'LightDirection' },
    '0070,1704' => { VR => 'FD', Name => 'DiffuseReflectionIntensity' },
    '0070,1705' => { VR => 'FD', Name => 'SpecularReflectionIntensity' },
    '0070,1706' => { VR => 'FD', Name => 'Shininess' },
    '0070,1801' => { VR => 'SQ', Name => 'PresentationStateClassComponentSeq' },
    '0070,1802' => { VR => 'CS', Name => 'ComponentType' },
    '0070,1803' => { VR => 'SQ', Name => 'ComponentInputSequence' },
    '0070,1804' => { VR => 'US', Name => 'VolumetricPresentationInputIndex' },
    '0070,1805' => { VR => 'SQ', Name => 'PresentationStateCompositorCompSeq' },
    '0070,1806' => { VR => 'SQ', Name => 'WeightingTransferFunctionSequence' },
    '0070,1807' => { VR => 'US', Name => 'WeightingLookupTableDescriptor' },
    '0070,1808' => { VR => 'OB', Name => 'WeightingLookupTableData' },
    '0070,1901' => { VR => 'SQ', Name => 'VolumetricAnnotationSequence' },
    '0070,1903' => { VR => 'SQ', Name => 'ReferencedStructuredContextSeq' },
    '0070,1904' => { VR => 'UI', Name => 'ReferencedContentItem' },
    '0070,1905' => { VR => 'SQ', Name => 'VolumetricPresentInputAnnotSeq' },
    '0070,1907' => { VR => 'CS', Name => 'AnnotationClipping' },
    '0070,1A01' => { VR => 'CS', Name => 'PresentationAnimationStyle' },
    '0070,1A03' => { VR => 'FD', Name => 'RecommendedAnimationRate' },
    '0070,1A04' => { VR => 'SQ', Name => 'AnimationCurveSequence' },
    '0070,1A05' => { VR => 'FD', Name => 'AnimationStepSize' },
    '0070,1A06' => { VR => 'FD', Name => 'SwivelRange' },
    '0070,1A07' => { VR => 'OD', Name => 'VolumetricCurveUpDirections' },
    '0070,1A08' => { VR => 'SQ', Name => 'VolumeStreamSequence' },
    '0070,1A09' => { VR => 'LO', Name => 'RGBATransferFunctionDescription' },
    '0070,1B01' => { VR => 'SQ', Name => 'AdvancedBlendingSequence' },
    '0070,1B02' => { VR => 'US', Name => 'BlendingInputNumber' },
    '0070,1B03' => { VR => 'SQ', Name => 'BlendingDisplayInputSequence' },
    '0070,1B04' => { VR => 'SQ', Name => 'BlendingDisplaySequence' },
    '0070,1B06' => { VR => 'CS', Name => 'BlendingMode' },
    '0070,1B07' => { VR => 'CS', Name => 'TimeSeriesBlending' },
    '0070,1B08' => { VR => 'CS', Name => 'GeometryForDisplay' },
    '0070,1B11' => { VR => 'SQ', Name => 'ThresholdSequence' },
    '0070,1B12' => { VR => 'SQ', Name => 'ThresholdValueSequence' },
    '0070,1B13' => { VR => 'CS', Name => 'ThresholdType' },
    '0070,1B14' => { VR => 'FD', Name => 'ThresholdValue' },
    '0072,0002' => { VR => 'SH', Name => 'HangingProtocolName' },
    '0072,0004' => { VR => 'LO', Name => 'HangingProtocolDescription' },
    '0072,0006' => { VR => 'CS', Name => 'HangingProtocolLevel' },
    '0072,0008' => { VR => 'LO', Name => 'HangingProtocolCreator' },
    '0072,000A' => { VR => 'DT', Name => 'HangingProtocolCreationDateTime' },
    '0072,000C' => { VR => 'SQ', Name => 'HangingProtocolDefinitionSequence' },
    '0072,000E' => { VR => 'SQ', Name => 'HangingProtocolUserIDCodeSequence' },
    '0072,0010' => { VR => 'LO', Name => 'HangingProtocolUserGroupName' },
    '0072,0012' => { VR => 'SQ', Name => 'SourceHangingProtocolSequence' },
    '0072,0014' => { VR => 'US', Name => 'NumberOfPriorsReferenced' },
    '0072,0020' => { VR => 'SQ', Name => 'ImageSetsSequence' },
    '0072,0022' => { VR => 'SQ', Name => 'ImageSetSelectorSequence' },
    '0072,0024' => { VR => 'CS', Name => 'ImageSetSelectorUsageFlag' },
    '0072,0026' => { VR => 'AT', Name => 'SelectorAttribute' },
    '0072,0028' => { VR => 'US', Name => 'SelectorValueNumber' },
    '0072,0030' => { VR => 'SQ', Name => 'TimeBasedImageSetsSequence' },
    '0072,0032' => { VR => 'US', Name => 'ImageSetNumber' },
    '0072,0034' => { VR => 'CS', Name => 'ImageSetSelectorCategory' },
    '0072,0038' => { VR => 'US', Name => 'RelativeTime' },
    '0072,003A' => { VR => 'CS', Name => 'RelativeTimeUnits' },
    '0072,003C' => { VR => 'SS', Name => 'AbstractPriorValue' },
    '0072,003E' => { VR => 'SQ', Name => 'AbstractPriorCodeSequence' },
    '0072,0040' => { VR => 'LO', Name => 'ImageSetLabel' },
    '0072,0050' => { VR => 'CS', Name => 'SelectorAttributeVR' },
    '0072,0052' => { VR => 'AT', Name => 'SelectorSequencePointer' },
    '0072,0054' => { VR => 'LO', Name => 'SelectorSeqPointerPrivateCreator' },
    '0072,0056' => { VR => 'LO', Name => 'SelectorAttributePrivateCreator' },
    '0072,005E' => { VR => 'AE', Name => 'SelectorAEValue' },
    '0072,005F' => { VR => 'AS', Name => 'SelectorASValue' },
    '0072,0060' => { VR => 'AT', Name => 'SelectorATValue' },
    '0072,0061' => { VR => 'DA', Name => 'SelectorDAValue' },
    '0072,0062' => { VR => 'CS', Name => 'SelectorCSValue' },
    '0072,0063' => { VR => 'DT', Name => 'SelectorDTValue' },
    '0072,0064' => { VR => 'IS', Name => 'SelectorISValue' },
    '0072,0065' => { VR => 'OB', Name => 'SelectorOBValue' },
    '0072,0066' => { VR => 'LO', Name => 'SelectorLOValue' },
    '0072,0067' => { VR => 'OF', Name => 'SelectorOFValue' },
    '0072,0068' => { VR => 'LT', Name => 'SelectorLTValue' },
    '0072,0069' => { VR => 'OW', Name => 'SelectorOWValue' },
    '0072,006A' => { VR => 'PN', Name => 'SelectorPNValue' },
    '0072,006B' => { VR => 'TM', Name => 'SelectorTMValue' },
    '0072,006C' => { VR => 'SH', Name => 'SelectorSHValue' },
    '0072,006D' => { VR => 'UN', Name => 'SelectorUNValue' },
    '0072,006E' => { VR => 'ST', Name => 'SelectorSTValue' },
    '0072,006F' => { VR => 'UC', Name => 'SelectorUCValue' },
    '0072,0070' => { VR => 'UT', Name => 'SelectorUTValue' },
    '0072,0071' => { VR => 'UR', Name => 'SelectorURValue' },
    '0072,0072' => { VR => 'DS', Name => 'SelectorDSValue' },
    '0072,0073' => { VR => 'OD', Name => 'SelectorODValue' },
    '0072,0074' => { VR => 'FD', Name => 'SelectorFDValue' },
    '0072,0075' => { VR => 'OL', Name => 'SelectorOLValue' },
    '0072,0076' => { VR => 'FL', Name => 'SelectorFLValue' },
    '0072,0078' => { VR => 'UL', Name => 'SelectorULValue' },
    '0072,007A' => { VR => 'US', Name => 'SelectorUSValue' },
    '0072,007C' => { VR => 'SL', Name => 'SelectorSLValue' },
    '0072,007E' => { VR => 'SS', Name => 'SelectorSSValue' },
    '0072,007F' => { VR => 'UI', Name => 'SelectorUIValue' },
    '0072,0080' => { VR => 'SQ', Name => 'SelectorCodeSequenceValue' },
    '0072,0081' => { VR => 'OV', Name => 'SelectorOVValue' },
    '0072,0082' => { VR => 'SV', Name => 'SelectorSVValue' },
    '0072,0083' => { VR => 'UV', Name => 'SelectorUVValue' },
    '0072,0100' => { VR => 'US', Name => 'NumberOfScreens' },
    '0072,0102' => { VR => 'SQ', Name => 'NominalScreenDefinitionSequence' },
    '0072,0104' => { VR => 'US', Name => 'NumberOfVerticalPixels' },
    '0072,0106' => { VR => 'US', Name => 'NumberOfHorizontalPixels' },
    '0072,0108' => { VR => 'FD', Name => 'DisplayEnvironmentSpatialPosition' },
    '0072,010A' => { VR => 'US', Name => 'ScreenMinimumGrayscaleBitDepth' },
    '0072,010C' => { VR => 'US', Name => 'ScreenMinimumColorBitDepth' },
    '0072,010E' => { VR => 'US', Name => 'ApplicationMaximumRepaintTime' },
    '0072,0200' => { VR => 'SQ', Name => 'DisplaySetsSequence' },
    '0072,0202' => { VR => 'US', Name => 'DisplaySetNumber' },
    '0072,0203' => { VR => 'LO', Name => 'DisplaySetLabel' },
    '0072,0204' => { VR => 'US', Name => 'DisplaySetPresentationGroup' },
    '0072,0206' => { VR => 'LO', Name => 'DisplaySetPresentationGroupDescr' },
    '0072,0208' => { VR => 'CS', Name => 'PartialDataDisplayHandling' },
    '0072,0210' => { VR => 'SQ', Name => 'SynchronizedScrollingSequence' },
    '0072,0212' => { VR => 'US', Name => 'DisplaySetScrollingGroup' },
    '0072,0214' => { VR => 'SQ', Name => 'NavigationIndicatorSequence' },
    '0072,0216' => { VR => 'US', Name => 'NavigationDisplaySet' },
    '0072,0218' => { VR => 'US', Name => 'ReferenceDisplaySets' },
    '0072,0300' => { VR => 'SQ', Name => 'ImageBoxesSequence' },
    '0072,0302' => { VR => 'US', Name => 'ImageBoxNumber' },
    '0072,0304' => { VR => 'CS', Name => 'ImageBoxLayoutType' },
    '0072,0306' => { VR => 'US', Name => 'ImageBoxTileHorizontalDimension' },
    '0072,0308' => { VR => 'US', Name => 'ImageBoxTileVerticalDimension' },
    '0072,0310' => { VR => 'CS', Name => 'ImageBoxScrollDirection' },
    '0072,0312' => { VR => 'CS', Name => 'ImageBoxSmallScrollType' },
    '0072,0314' => { VR => 'US', Name => 'ImageBoxSmallScrollAmount' },
    '0072,0316' => { VR => 'CS', Name => 'ImageBoxLargeScrollType' },
    '0072,0318' => { VR => 'US', Name => 'ImageBoxLargeScrollAmount' },
    '0072,0320' => { VR => 'US', Name => 'ImageBoxOverlapPriority' },
    '0072,0330' => { VR => 'FD', Name => 'CineRelativeToRealTime' },
    '0072,0400' => { VR => 'SQ', Name => 'FilterOperationsSequence' },
    '0072,0402' => { VR => 'CS', Name => 'FilterByCategory' },
    '0072,0404' => { VR => 'CS', Name => 'FilterByAttributePresence' },
    '0072,0406' => { VR => 'CS', Name => 'FilterByOperator' },
    '0072,0420' => { VR => 'US', Name => 'StructuredDisplayBkgCIELabValue' },
    '0072,0421' => { VR => 'US', Name => 'EmptyImageBoxCIELabValue' },
    '0072,0422' => { VR => 'SQ', Name => 'StructuredDisplayImageBoxSequence' },
    '0072,0424' => { VR => 'SQ', Name => 'StructuredDisplayTextBoxSequence' },
    '0072,0427' => { VR => 'SQ', Name => 'ReferencedFirstFrameSequence' },
    '0072,0430' => { VR => 'SQ', Name => 'ImageBoxSynchronizationSequence' },
    '0072,0432' => { VR => 'US', Name => 'SynchronizedImageBoxList' },
    '0072,0434' => { VR => 'CS', Name => 'TypeOfSynchronization' },
    '0072,0500' => { VR => 'CS', Name => 'BlendingOperationType' },
    '0072,0510' => { VR => 'CS', Name => 'ReformattingOperationType' },
    '0072,0512' => { VR => 'FD', Name => 'ReformattingThickness' },
    '0072,0514' => { VR => 'FD', Name => 'ReformattingInterval' },
    '0072,0516' => { VR => 'CS', Name => 'ReformattingOpInitialViewDir' },
    '0072,0520' => { VR => 'CS', Name => 'RenderingType3D' },
    '0072,0600' => { VR => 'SQ', Name => 'SortingOperationsSequence' },
    '0072,0602' => { VR => 'CS', Name => 'SortByCategory' },
    '0072,0604' => { VR => 'CS', Name => 'SortingDirection' },
    '0072,0700' => { VR => 'CS', Name => 'DisplaySetPatientOrientation' },
    '0072,0702' => { VR => 'CS', Name => 'VOIType' },
    '0072,0704' => { VR => 'CS', Name => 'PseudoColorType' },
    '0072,0705' => { VR => 'SQ', Name => 'PseudoColorPaletteInstanceRefSeq' },
    '0072,0706' => { VR => 'CS', Name => 'ShowGrayscaleInverted' },
    '0072,0710' => { VR => 'CS', Name => 'ShowImageTrueSizeFlag' },
    '0072,0712' => { VR => 'CS', Name => 'ShowGraphicAnnotationFlag' },
    '0072,0714' => { VR => 'CS', Name => 'ShowPatientDemographicsFlag' },
    '0072,0716' => { VR => 'CS', Name => 'ShowAcquisitionTechniquesFlag' },
    '0072,0717' => { VR => 'CS', Name => 'DisplaySetHorizontalJustification' },
    '0072,0718' => { VR => 'CS', Name => 'DisplaySetVerticalJustification' },
    '0074,0120' => { VR => 'FD', Name => 'ContinuationStartMeterset' },
    '0074,0121' => { VR => 'FD', Name => 'ContinuationEndMeterset' },
    '0074,1000' => { VR => 'CS', Name => 'UnifiedProcedureStepState' },
    '0074,1002' => { VR => 'SQ', Name => 'UPSProgressInformationSequence' },
    '0074,1004' => { VR => 'DS', Name => 'UnifiedProcedureStepProgress' },
    '0074,1006' => { VR => 'ST', Name => 'UnifiedProcedureStepProgressDescr' },
    '0074,1007' => { VR => 'SQ', Name => 'ProcedureStepProgressParametersSeq' },
    '0074,1008' => { VR => 'SQ', Name => 'UnifiedProcedureStepComURISeq' },
    '0074,100a' => { VR => 'ST', Name => 'ContactURI' },
    '0074,100c' => { VR => 'LO', Name => 'ContactDisplayName' },
    '0074,100E' => { VR => 'SQ', Name => 'ProcedureStepDiscontinuationReason' },
    '0074,1020' => { VR => 'SQ', Name => 'BeamTaskSequence' },
    '0074,1022' => { VR => 'CS', Name => 'BeamTaskType' },
    '0074,1024' => { VR => 'IS', Name => 'BeamOrderIndex' },
    '0074,1025' => { VR => 'CS', Name => 'AutosequenceFlag' },
    '0074,1026' => { VR => 'FD', Name => 'TableTopVerticalAdjustedPosition' },
    '0074,1027' => { VR => 'FD', Name => 'TableTopLongitudinalAdjustedPos' },
    '0074,1028' => { VR => 'FD', Name => 'TableTopLateralAdjustedPosition' },
    '0074,102A' => { VR => 'FD', Name => 'PatientSupportAdjustedAngle' },
    '0074,102B' => { VR => 'FD', Name => 'TableTopEccentricAdjustedAngle' },
    '0074,102C' => { VR => 'FD', Name => 'TableTopPitchAdjustedAngle' },
    '0074,102D' => { VR => 'FD', Name => 'TableTopRollAdjustedAngle' },
    '0074,1030' => { VR => 'SQ', Name => 'DeliveryVerificationImageSequence' },
    '0074,1032' => { VR => 'CS', Name => 'VerificationImageTiming' },
    '0074,1034' => { VR => 'CS', Name => 'DoubleExposureFlag' },
    '0074,1036' => { VR => 'CS', Name => 'DoubleExposureOrdering' },
    '0074,1038' => { VR => 'DS', Name => 'DoubleExposureMeterset' },
    '0074,103A' => { VR => 'DS', Name => 'DoubleExposureFieldDelta' },
    '0074,1040' => { VR => 'SQ', Name => 'RelatedReferenceRTImageSequence' },
    '0074,1042' => { VR => 'SQ', Name => 'GeneralMachineVerificationSequence' },
    '0074,1044' => { VR => 'SQ', Name => 'ConventionalMachineVerificationSeq' },
    '0074,1046' => { VR => 'SQ', Name => 'IonMachineVerificationSequence' },
    '0074,1048' => { VR => 'SQ', Name => 'FailedAttributesSequence' },
    '0074,104A' => { VR => 'SQ', Name => 'OverriddenAttributesSequence' },
    '0074,104C' => { VR => 'SQ', Name => 'ConventionalControlPointVerifySeq' },
    '0074,104E' => { VR => 'SQ', Name => 'IonControlPointVerificationSeq' },
    '0074,1050' => { VR => 'SQ', Name => 'AttributeOccurrenceSequence' },
    '0074,1052' => { VR => 'AT', Name => 'AttributeOccurrencePointer' },
    '0074,1054' => { VR => 'UL', Name => 'AttributeItemSelector' },
    '0074,1056' => { VR => 'LO', Name => 'AttributeOccurrencePrivateCreator' },
    '0074,1057' => { VR => 'IS', Name => 'SelectorSequencePointerItems' },
    '0074,1200' => { VR => 'CS', Name => 'ScheduledProcedureStepPriority' },
    '0074,1202' => { VR => 'LO', Name => 'WorklistLabel' },
    '0074,1204' => { VR => 'LO', Name => 'ProcedureStepLabel' },
    '0074,1210' => { VR => 'SQ', Name => 'ScheduledProcessingParametersSeq' },
    '0074,1212' => { VR => 'SQ', Name => 'PerformedProcessingParametersSeq' },
    '0074,1216' => { VR => 'SQ', Name => 'UPSPerformedProcedureSequence' },
    '0074,1220' => { VR => 'SQ', Name => 'RelatedProcedureStepSequence' },
    '0074,1222' => { VR => 'LO', Name => 'ProcedureStepRelationshipType' },
    '0074,1224' => { VR => 'SQ', Name => 'ReplacedProcedureStepSequence' },
    '0074,1230' => { VR => 'LO', Name => 'DeletionLock' },
    '0074,1234' => { VR => 'AE', Name => 'ReceivingAE' },
    '0074,1236' => { VR => 'AE', Name => 'RequestingAE' },
    '0074,1238' => { VR => 'LT', Name => 'ReasonForCancellation' },
    '0074,1242' => { VR => 'CS', Name => 'SCPStatus' },
    '0074,1244' => { VR => 'CS', Name => 'SubscriptionListStatus' },
    '0074,1246' => { VR => 'CS', Name => 'UPSListStatus' },
    '0074,1324' => { VR => 'UL', Name => 'BeamOrderIndex' },
    '0074,1338' => { VR => 'FD', Name => 'DoubleExposureMeterset' },
    '0074,133A' => { VR => 'FD', Name => 'DoubleExposureFieldDelta' },
    '0074,1401' => { VR => 'SQ', Name => 'BrachyTaskSequence' },
    '0074,1402' => { VR => 'DS', Name => 'ContinuationStartTotalRefAirKerma' },
    '0074,1403' => { VR => 'DS', Name => 'ContinuationEndTotalRefAirKerma' },
    '0074,1404' => { VR => 'IS', Name => 'ContinuationPulseNumber' },
    '0074,1405' => { VR => 'SQ', Name => 'ChannelDeliveryOrderSequence' },
    '0074,1406' => { VR => 'IS', Name => 'ReferencedChannelNumber' },
    '0074,1407' => { VR => 'DS', Name => 'StartCumulativeTimeWeight' },
    '0074,1408' => { VR => 'DS', Name => 'EndCumulativeTimeWeight' },
    '0074,1409' => { VR => 'SQ', Name => 'OmittedChannelSequence' },
    '0074,140A' => { VR => 'CS', Name => 'ReasonForChannelOmission' },
    '0074,140B' => { VR => 'LO', Name => 'ReasonForChannelOmissionDescr' },
    '0074,140C' => { VR => 'IS', Name => 'ChannelDeliveryOrderIndex' },
    '0074,140D' => { VR => 'SQ', Name => 'ChannelDeliveryContinuationSeq' },
    '0074,140E' => { VR => 'SQ', Name => 'OmittedApplicationSetupSequence' },
    '0076,0001' => { VR => 'LO', Name => 'ImplantAssemblyTemplateName' },
    '0076,0003' => { VR => 'LO', Name => 'ImplantAssemblyTemplateIssuer' },
    '0076,0006' => { VR => 'LO', Name => 'ImplantAssemblyTemplateVersion' },
    '0076,0008' => { VR => 'SQ', Name => 'ReplacedImplantAssemblyTemplateSeq' },
    '0076,000A' => { VR => 'CS', Name => 'ImplantAssemblyTemplateType' },
    '0076,000C' => { VR => 'SQ', Name => 'OriginalImplantAssemblyTemplateSeq' },
    '0076,000E' => { VR => 'SQ', Name => 'DerivationImplantAssyTemplateSeq' },
    '0076,0010' => { VR => 'SQ', Name => 'ImplantAssyTemplateTgtAnatomySeq' },
    '0076,0020' => { VR => 'SQ', Name => 'ProcedureTypeCodeSequence' },
    '0076,0030' => { VR => 'LO', Name => 'SurgicalTechnique' },
    '0076,0032' => { VR => 'SQ', Name => 'ComponentTypesSequence' },
    '0076,0034' => { VR => 'SQ', Name => 'ComponentTypeCodeSequence' },
    '0076,0036' => { VR => 'CS', Name => 'ExclusiveComponentType' },
    '0076,0038' => { VR => 'CS', Name => 'MandatoryComponentType' },
    '0076,0040' => { VR => 'SQ', Name => 'ComponentSequence' },
    '0076,0055' => { VR => 'US', Name => 'ComponentID' },
    '0076,0060' => { VR => 'SQ', Name => 'ComponentAssemblySequence' },
    '0076,0070' => { VR => 'US', Name => 'Component1ReferencedID' },
    '0076,0080' => { VR => 'US', Name => 'Component1RefMatingFeatureSetID' },
    '0076,0090' => { VR => 'US', Name => 'Component1RefMatingFeatureID' },
    '0076,00A0' => { VR => 'US', Name => 'Component2ReferencedID' },
    '0076,00B0' => { VR => 'US', Name => 'Component2RefMatingFeatureSetID' },
    '0076,00C0' => { VR => 'US', Name => 'Component2RefMatingFeatureID' },
    '0078,0001' => { VR => 'LO', Name => 'ImplantTemplateGroupName' },
    '0078,0010' => { VR => 'ST', Name => 'ImplantTemplateGroupDescription' },
    '0078,0020' => { VR => 'LO', Name => 'ImplantTemplateGroupIssuer' },
    '0078,0024' => { VR => 'LO', Name => 'ImplantTemplateGroupVersion' },
    '0078,0026' => { VR => 'SQ', Name => 'ReplacedImplantTemplateGroupSeq' },
    '0078,0028' => { VR => 'SQ', Name => 'ImplantTemplateGroupTgtAnatomySeq' },
    '0078,002A' => { VR => 'SQ', Name => 'ImplantTemplateGroupMembersSeq' },
    '0078,002E' => { VR => 'US', Name => 'ImplantTemplateGroupMemberID' },
    '0078,0050' => { VR => 'FD', Name => 'ThreeDImplantTemplGrpMemMatchPoint' },
    '0078,0060' => { VR => 'FD', Name => 'ThreeDImplantTemplGrpMemMatchAxes' },
    '0078,0070' => { VR => 'SQ', Name => 'ImplantTemplGrpMemMatch2DCoordSeq' },
    '0078,0090' => { VR => 'FD', Name => 'TwoDImplantTemplGrpMemMatchPoint' },
    '0078,00A0' => { VR => 'FD', Name => 'TwoDImplantTemplGrpMemberMatchAxes' },
    '0078,00B0' => { VR => 'SQ', Name => 'ImplantTemplGrpVariationDimSeq' },
    '0078,00B2' => { VR => 'LO', Name => 'ImplantTemplGrpVariationDimName' },
    '0078,00B4' => { VR => 'SQ', Name => 'ImplantTemplGrpVariationDimRankSeq' },
    '0078,00B6' => { VR => 'US', Name => 'RefImplantTemplateGroupMemberID' },
    '0078,00B8' => { VR => 'US', Name => 'ImplantTemplGrpVariationDimRank' },
    '0080,0001' => { VR => 'SQ', Name => 'SurfaceScanAcquisitionTypeCodeSeq' },
    '0080,0002' => { VR => 'SQ', Name => 'SurfaceScanModeCodeSequence' },
    '0080,0003' => { VR => 'SQ', Name => 'RegistrationMethodCodeSequence' },
    '0080,0004' => { VR => 'FD', Name => 'ShotDurationTime' },
    '0080,0005' => { VR => 'FD', Name => 'ShotOffsetTime' },
    '0080,0006' => { VR => 'US', Name => 'SurfacePointPresentationValueData' },
    '0080,0007' => { VR => 'US', Name => 'SurfacePointColorCIELabValueData' },
    '0080,0008' => { VR => 'SQ', Name => 'UVMappingSequence' },
    '0080,0009' => { VR => 'SH', Name => 'TextureLabel' },
    '0080,0010' => { VR => 'OF', Name => 'UValueData' },
    '0080,0011' => { VR => 'OF', Name => 'VValueData' },
    '0080,0012' => { VR => 'SQ', Name => 'ReferencedTextureSequence' },
    '0080,0013' => { VR => 'SQ', Name => 'ReferencedSurfaceDataSequence' },
    '0082,0001' => { VR => 'CS', Name => 'AssessmentSummary' },
    '0082,0003' => { VR => 'UT', Name => 'AssessmentSummaryDescription' },
    '0082,0004' => { VR => 'SQ', Name => 'AssessedSOPInstanceSequence' },
    '0082,0005' => { VR => 'SQ', Name => 'ReferencedComparisonSOPInstanceSeq' },
    '0082,0006' => { VR => 'UL', Name => 'NumberOfAssessmentObservations' },
    '0082,0007' => { VR => 'SQ', Name => 'AssessmentObservationsSequence' },
    '0082,0008' => { VR => 'CS', Name => 'ObservationSignificance' },
    '0082,000A' => { VR => 'UT', Name => 'ObservationDescription' },
    '0082,000C' => { VR => 'SQ', Name => 'StructuredConstraintObservationSeq' },
    '0082,0010' => { VR => 'SQ', Name => 'AssessedAttributeValueSequence' },
    '0082,0016' => { VR => 'LO', Name => 'AssessmentSetID' },
    '0082,0017' => { VR => 'SQ', Name => 'AssessmentRequesterSequence' },
    '0082,0018' => { VR => 'LO', Name => 'SelectorAttributeName' },
    '0082,0019' => { VR => 'LO', Name => 'SelectorAttributeKeyword' },
    '0082,0021' => { VR => 'SQ', Name => 'AssessmentTypeCodeSequence' },
    '0082,0022' => { VR => 'SQ', Name => 'ObservationBasisCodeSequence' },
    '0082,0023' => { VR => 'LO', Name => 'AssessmentLabel' },
    '0082,0032' => { VR => 'CS', Name => 'ConstraintType' },
    '0082,0033' => { VR => 'UT', Name => 'SpecificationSelectionGuidance' },
    '0082,0034' => { VR => 'SQ', Name => 'ConstraintValueSequence' },
    '0082,0035' => { VR => 'SQ', Name => 'RecommendedDefaultValueSequence' },
    '0082,0036' => { VR => 'CS', Name => 'ConstraintViolationSignificance' },
    '0082,0037' => { VR => 'UT', Name => 'ConstraintViolationCondition' },
    '0082,0038' => { VR => 'CS', Name => 'ModifiableConstraintFlag' },
    # storage group
    '0088,0130' => { VR => 'SH', Name => 'StorageMediaFileSetID' },
    '0088,0140' => { VR => 'UI', Name => 'StorageMediaFileSetUID' },
    '0088,0200' => { VR => 'SQ', Name => 'IconImageSequence' },
    '0088,0904' => { VR => 'LO', Name => 'TopicTitle' },
    '0088,0906' => { VR => 'ST', Name => 'TopicSubject' },
    '0088,0910' => { VR => 'LO', Name => 'TopicAuthor' },
    '0088,0912' => { VR => 'LO', Name => 'TopicKeywords' },
    '0100,0410' => { VR => 'CS', Name => 'SOPInstanceStatus' },
    '0100,0420' => { VR => 'DT', Name => 'SOPAuthorizationDateAndTime' },
    '0100,0424' => { VR => 'LT', Name => 'SOPAuthorizationComment' },
    '0100,0426' => { VR => 'LO', Name => 'AuthorizationEquipmentCertNumber' },
    '0400,0005' => { VR => 'US', Name => 'MACIDNumber' },
    '0400,0010' => { VR => 'UI', Name => 'MACCalculationTransferSyntaxUID' },
    '0400,0015' => { VR => 'CS', Name => 'MACAlgorithm' },
    '0400,0020' => { VR => 'AT', Name => 'DataElementsSigned' },
    '0400,0100' => { VR => 'UI', Name => 'DigitalSignatureUID' },
    '0400,0105' => { VR => 'DT', Name => 'DigitalSignatureDateTime' },
    '0400,0110' => { VR => 'CS', Name => 'CertificateType' },
    '0400,0115' => { VR => 'OB', Name => 'CertificateOfSigner' },
    '0400,0120' => { VR => 'OB', Name => 'Signature' },
    '0400,0305' => { VR => 'CS', Name => 'CertifiedTimestampType' },
    '0400,0310' => { VR => 'OB', Name => 'CertifiedTimestamp' },
    '0400,0401' => { VR => 'SQ', Name => 'DigitalSignaturePurposeCodeSeq' },
    '0400,0402' => { VR => 'SQ', Name => 'ReferencedDigitalSignatureSeq' },
    '0400,0403' => { VR => 'SQ', Name => 'ReferencedSOPInstanceMACSeq' },
    '0400,0404' => { VR => 'OB', Name => 'MAC' },
    '0400,0500' => { VR => 'SQ', Name => 'EncryptedAttributesSequence' },
    '0400,0510' => { VR => 'UI', Name => 'EncryptedContentTransferSyntaxUID' },
    '0400,0520' => { VR => 'OB', Name => 'EncryptedContent' },
    '0400,0550' => { VR => 'SQ', Name => 'ModifiedAttributesSequence' },
    '0400,0551' => { VR => 'SQ', Name => 'NonconformingModifiedAttributesSeq' },
    '0400,0552' => { VR => 'OB', Name => 'NonconformingDataElementValue' },
    '0400,0561' => { VR => 'SQ', Name => 'OriginalAttributesSequence' },
    '0400,0562' => { VR => 'DT', Name => 'AttributeModificationDateTime' },
    '0400,0563' => { VR => 'LO', Name => 'ModifyingSystem' },
    '0400,0564' => { VR => 'LO', Name => 'SourceOfPreviousValues' },
    '0400,0565' => { VR => 'CS', Name => 'ReasonForTheAttributeModification' },
    '0400,0600' => { VR => 'CS', Name => 'InstanceOriginStatus' },
    '1000,xxx0' => { VR => 'US', Name => 'EscapeTriplet' },
    '1000,xxx1' => { VR => 'US', Name => 'RunLengthTriplet' },
    '1000,xxx2' => { VR => 'US', Name => 'HuffmanTableSize' },
    '1000,xxx3' => { VR => 'US', Name => 'HuffmanTableTriplet' },
    '1000,xxx4' => { VR => 'US', Name => 'ShiftTableSize' },
    '1000,xxx5' => { VR => 'US', Name => 'ShiftTableTriplet' },
    '1010,xxxx' => { VR => 'US', Name => 'ZonalMap' },
    '2000,0010' => { VR => 'IS', Name => 'NumberOfCopies' },
    '2000,001E' => { VR => 'SQ', Name => 'PrinterConfigurationSequence' },
    '2000,0020' => { VR => 'CS', Name => 'PrintPriority' },
    '2000,0030' => { VR => 'CS', Name => 'MediumType' },
    '2000,0040' => { VR => 'CS', Name => 'FilmDestination' },
    '2000,0050' => { VR => 'LO', Name => 'FilmSessionLabel' },
    '2000,0060' => { VR => 'IS', Name => 'MemoryAllocation' },
    '2000,0061' => { VR => 'IS', Name => 'MaximumMemoryAllocation' },
    '2000,0062' => { VR => 'CS', Name => 'ColorImagePrintingFlag' },
    '2000,0063' => { VR => 'CS', Name => 'CollationFlag' },
    '2000,0065' => { VR => 'CS', Name => 'AnnotationFlag' },
    '2000,0067' => { VR => 'CS', Name => 'ImageOverlayFlag' },
    '2000,0069' => { VR => 'CS', Name => 'PresentationLUTFlag' },
    '2000,006A' => { VR => 'CS', Name => 'ImageBoxPresentationLUTFlag' },
    '2000,00A0' => { VR => 'US', Name => 'MemoryBitDepth' },
    '2000,00A1' => { VR => 'US', Name => 'PrintingBitDepth' },
    '2000,00A2' => { VR => 'SQ', Name => 'MediaInstalledSequence' },
    '2000,00A4' => { VR => 'SQ', Name => 'OtherMediaAvailableSequence' },
    '2000,00A8' => { VR => 'SQ', Name => 'SupportedImageDisplayFormatSeq' },
    '2000,0500' => { VR => 'SQ', Name => 'ReferencedFilmBoxSequence' },
    '2000,0510' => { VR => 'SQ', Name => 'ReferencedStoredPrintSequence' },
    # film box group
    '2010,0010' => { VR => 'ST', Name => 'ImageDisplayFormat' },
    '2010,0030' => { VR => 'CS', Name => 'AnnotationDisplayFormatID' },
    '2010,0040' => { VR => 'CS', Name => 'FilmOrientation' },
    '2010,0050' => { VR => 'CS', Name => 'FilmSizeID' },
    '2010,0052' => { VR => 'CS', Name => 'PrinterResolutionID' },
    '2010,0054' => { VR => 'CS', Name => 'DefaultPrinterResolutionID' },
    '2010,0060' => { VR => 'CS', Name => 'MagnificationType' },
    '2010,0080' => { VR => 'CS', Name => 'SmoothingType' },
    '2010,00A6' => { VR => 'CS', Name => 'DefaultMagnificationType' },
    '2010,00A7' => { VR => 'CS', Name => 'OtherMagnificationTypesAvailable' },
    '2010,00A8' => { VR => 'CS', Name => 'DefaultSmoothingType' },
    '2010,00A9' => { VR => 'CS', Name => 'OtherSmoothingTypesAvailable' },
    '2010,0100' => { VR => 'CS', Name => 'BorderDensity' },
    '2010,0110' => { VR => 'CS', Name => 'EmptyImageDensity' },
    '2010,0120' => { VR => 'US', Name => 'MinDensity' },
    '2010,0130' => { VR => 'US', Name => 'MaxDensity' },
    '2010,0140' => { VR => 'CS', Name => 'Trim' },
    '2010,0150' => { VR => 'ST', Name => 'ConfigurationInformation' },
    '2010,0152' => { VR => 'LT', Name => 'ConfigurationInformationDescr' },
    '2010,0154' => { VR => 'IS', Name => 'MaximumCollatedFilms' },
    '2010,015E' => { VR => 'US', Name => 'Illumination' },
    '2010,0160' => { VR => 'US', Name => 'ReflectedAmbientLight' },
    '2010,0376' => { VR => 'DS', Name => 'PrinterPixelSpacing' },
    '2010,0500' => { VR => 'SQ', Name => 'ReferencedFilmSessionSequence' },
    '2010,0510' => { VR => 'SQ', Name => 'ReferencedImageBoxSequence' },
    '2010,0520' => { VR => 'SQ', Name => 'ReferencedBasicAnnotationBoxSeq' },
    # image box group
    '2020,0010' => { VR => 'US', Name => 'ImageBoxPosition' },
    '2020,0020' => { VR => 'CS', Name => 'Polarity' },
    '2020,0030' => { VR => 'DS', Name => 'RequestedImageSize' },
    '2020,0040' => { VR => 'CS', Name => 'RequestedDecimate-CropBehavior' },
    '2020,0050' => { VR => 'CS', Name => 'RequestedResolutionID' },
    '2020,00A0' => { VR => 'CS', Name => 'RequestedImageSizeFlag' },
    '2020,00A2' => { VR => 'CS', Name => 'DecimateCropResult' },
    '2020,0110' => { VR => 'SQ', Name => 'BasicGrayscaleImageSequence' },
    '2020,0111' => { VR => 'SQ', Name => 'BasicColorImageSequence' },
    '2020,0130' => { VR => 'SQ', Name => 'ReferencedImageOverlayBoxSequence' },
    '2020,0140' => { VR => 'SQ', Name => 'ReferencedVOILUTBoxSequence' },
    # annotation group
    '2030,0010' => { VR => 'US', Name => 'AnnotationPosition' },
    '2030,0020' => { VR => 'LO', Name => 'TextString' },
    # overlay box group
    '2040,0010' => { VR => 'SQ', Name => 'ReferencedOverlayPlaneSequence' },
    '2040,0011' => { VR => 'US', Name => 'ReferencedOverlayPlaneGroups' },
    '2040,0020' => { VR => 'SQ', Name => 'OverlayPixelDataSequence' },
    '2040,0060' => { VR => 'CS', Name => 'OverlayMagnificationType' },
    '2040,0070' => { VR => 'CS', Name => 'OverlaySmoothingType' },
    '2040,0072' => { VR => 'CS', Name => 'OverlayOrImageMagnification' },
    '2040,0074' => { VR => 'US', Name => 'MagnifyToNumberOfColumns' },
    '2040,0080' => { VR => 'CS', Name => 'OverlayForegroundDensity' },
    '2040,0082' => { VR => 'CS', Name => 'OverlayBackgroundDensity' },
    '2040,0090' => { VR => 'CS', Name => 'OverlayMode' },
    '2040,0100' => { VR => 'CS', Name => 'ThresholdDensity' },
    '2040,0500' => { VR => 'SQ', Name => 'ReferencedImageBoxSequence' },
    '2050,0010' => { VR => 'SQ', Name => 'PresentationLUTSequence' },
    '2050,0020' => { VR => 'CS', Name => 'PresentationLUTShape' },
    '2050,0500' => { VR => 'SQ', Name => 'ReferencedPresentationLUTSequence' },
    '2100,0010' => { VR => 'SH', Name => 'PrintJobID' },
    '2100,0020' => { VR => 'CS', Name => 'ExecutionStatus' },
    '2100,0030' => { VR => 'CS', Name => 'ExecutionStatusInfo' },
    '2100,0040' => { VR => 'DA', Name => 'CreationDate' },
    '2100,0050' => { VR => 'TM', Name => 'CreationTime' },
    '2100,0070' => { VR => 'AE', Name => 'Originator' },
    '2100,0140' => { VR => 'AE', Name => 'DestinationAE' },
    '2100,0160' => { VR => 'SH', Name => 'OwnerID' },
    '2100,0170' => { VR => 'IS', Name => 'NumberOfFilms' },
    '2100,0500' => { VR => 'SQ', Name => 'ReferencedPrintJobSequence' },
    # printer group
    '2110,0010' => { VR => 'CS', Name => 'PrinterStatus' },
    '2110,0020' => { VR => 'CS', Name => 'PrinterStatusInfo' },
    '2110,0030' => { VR => 'LO', Name => 'PrinterName' },
    '2110,0099' => { VR => 'SH', Name => 'PrintQueueID' },
    '2120,0010' => { VR => 'CS', Name => 'QueueStatus' },
    # print job group
    '2120,0050' => { VR => 'SQ', Name => 'PrintJobDescriptionSequence' },
    '2120,0070' => { VR => 'SQ', Name => 'ReferencedPrintJobSequence' },
    '2130,0010' => { VR => 'SQ', Name => 'PrintManagementCapabilitiesSeq' },
    '2130,0015' => { VR => 'SQ', Name => 'PrinterCharacteristicsSequence' },
    '2130,0030' => { VR => 'SQ', Name => 'FilmBoxContentSequence' },
    '2130,0040' => { VR => 'SQ', Name => 'ImageBoxContentSequence' },
    '2130,0050' => { VR => 'SQ', Name => 'AnnotationContentSequence' },
    '2130,0060' => { VR => 'SQ', Name => 'ImageOverlayBoxContentSequence' },
    '2130,0080' => { VR => 'SQ', Name => 'PresentationLUTContentSequence' },
    '2130,00A0' => { VR => 'SQ', Name => 'ProposedStudySequence' },
    '2130,00C0' => { VR => 'SQ', Name => 'OriginalImageSequence' },
    '2200,0001' => { VR => 'CS', Name => 'LabelFromInfoExtractedFromInstance' },
    '2200,0002' => { VR => 'UT', Name => 'LabelText' },
    '2200,0003' => { VR => 'CS', Name => 'LabelStyleSelection' },
    '2200,0004' => { VR => 'LT', Name => 'MediaDisposition' },
    '2200,0005' => { VR => 'LT', Name => 'BarcodeValue' },
    '2200,0006' => { VR => 'CS', Name => 'BarcodeSymbology' },
    '2200,0007' => { VR => 'CS', Name => 'AllowMediaSplitting' },
    '2200,0008' => { VR => 'CS', Name => 'IncludeNonDICOMObjects' },
    '2200,0009' => { VR => 'CS', Name => 'IncludeDisplayApplication' },
    '2200,000A' => { VR => 'CS', Name => 'SaveCompInstancesAfterMediaCreate' },
    '2200,000B' => { VR => 'US', Name => 'TotalNumberMediaPiecesCreated' },
    '2200,000C' => { VR => 'LO', Name => 'RequestedMediaApplicationProfile' },
    '2200,000D' => { VR => 'SQ', Name => 'ReferencedStorageMediaSequence' },
    '2200,000E' => { VR => 'AT', Name => 'FailureAttributes' },
    '2200,000F' => { VR => 'CS', Name => 'AllowLossyCompression' },
    '2200,0020' => { VR => 'CS', Name => 'RequestPriority' },
    '3002,0002' => { VR => 'SH', Name => 'RTImageLabel' },
    '3002,0003' => { VR => 'LO', Name => 'RTImageName' },
    '3002,0004' => { VR => 'ST', Name => 'RTImageDescription' },
    '3002,000A' => { VR => 'CS', Name => 'ReportedValuesOrigin' },
    '3002,000C' => { VR => 'CS', Name => 'RTImagePlane' },
    '3002,000D' => { VR => 'DS', Name => 'XRayImageReceptorTranslation' },
    '3002,000E' => { VR => 'DS', Name => 'XRayImageReceptorAngle' },
    '3002,0010' => { VR => 'DS', Name => 'RTImageOrientation' },
    '3002,0011' => { VR => 'DS', Name => 'ImagePlanePixelSpacing' },
    '3002,0012' => { VR => 'DS', Name => 'RTImagePosition' },
    '3002,0020' => { VR => 'SH', Name => 'RadiationMachineName' },
    '3002,0022' => { VR => 'DS', Name => 'RadiationMachineSAD' },
    '3002,0024' => { VR => 'DS', Name => 'RadiationMachineSSD' },
    '3002,0026' => { VR => 'DS', Name => 'RTImageSID' },
    '3002,0028' => { VR => 'DS', Name => 'SourceToReferenceObjectDistance' },
    '3002,0029' => { VR => 'IS', Name => 'FractionNumber' },
    '3002,0030' => { VR => 'SQ', Name => 'ExposureSequence' },
    '3002,0032' => { VR => 'DS', Name => 'MetersetExposure' },
    '3002,0034' => { VR => 'DS', Name => 'DiaphragmPosition' },
    '3002,0040' => { VR => 'SQ', Name => 'FluenceMapSequence' },
    '3002,0041' => { VR => 'CS', Name => 'FluenceDataSource' },
    '3002,0042' => { VR => 'DS', Name => 'FluenceDataScale' },
    '3002,0050' => { VR => 'SQ', Name => 'PrimaryFluenceModeSequence' },
    '3002,0051' => { VR => 'CS', Name => 'FluenceMode' },
    '3002,0052' => { VR => 'SH', Name => 'FluenceModeID' },
    '3002,0100' => { VR => 'IS', Name => 'SelectedFrameNumber' },
    '3002,0101' => { VR => 'SQ', Name => 'SelectedFrameFunctionalGroupsSeq' },
    '3002,0102' => { VR => 'SQ', Name => 'RTImageFrameGeneralContentSequence' },
    '3002,0103' => { VR => 'SQ', Name => 'RTImageFrameContextSequence' },
    '3002,0104' => { VR => 'SQ', Name => 'RTImageScopeSequence' },
    '3002,0105' => { VR => 'CS', Name => 'BeamModifierCoordPresenceFlag' },
    '3002,0106' => { VR => 'FD', Name => 'StartCumulativeMeterset' },
    '3002,0107' => { VR => 'FD', Name => 'StopCumulativeMeterset' },
    '3002,0108' => { VR => 'SQ', Name => 'RTAcquisitionPatientPosSeq' },
    '3002,0109' => { VR => 'SQ', Name => 'RTImageFrameImagingDevicePosSeq' },
    '3002,010A' => { VR => 'SQ', Name => 'RTImageFramekVRadiationAcqSeq' },
    '3002,010B' => { VR => 'SQ', Name => 'RTImageFrameMVRadiationAcqSeq' },
    '3002,010C' => { VR => 'SQ', Name => 'RTImageFrameRadiationAcqSeq' },
    '3002,010D' => { VR => 'SQ', Name => 'ImagingSourcePositionSequence' },
    '3002,010E' => { VR => 'SQ', Name => 'ImageReceptorPositionSequence' },
    '3002,010F' => { VR => 'FD', Name => 'DevicePosToEquipmentMappingMatrix' },
    '3002,0110' => { VR => 'SQ', Name => 'DevicePositionParameterSequence' },
    '3002,0111' => { VR => 'CS', Name => 'ImagingSourceLocSpecificationType' },
    '3002,0112' => { VR => 'SQ', Name => 'ImagingDeviceLocationMatrixSeq' },
    '3002,0113' => { VR => 'SQ', Name => 'ImagingDeviceLocationParameterSeq' },
    '3002,0114' => { VR => 'SQ', Name => 'ImagingApertureSequence' },
    '3002,0115' => { VR => 'CS', Name => 'ImagingApertureSpecificationType' },
    '3002,0116' => { VR => 'US', Name => 'NumberOfAcquisitionDevices' },
    '3002,0117' => { VR => 'SQ', Name => 'AcquisitionDeviceSequence' },
    '3002,0118' => { VR => 'SQ', Name => 'AcquisitionTaskSequence' },
    '3002,0119' => { VR => 'SQ', Name => 'AcquisitionTaskWorkitemCodeSeq' },
    '3002,011A' => { VR => 'SQ', Name => 'AcquisitionSubtaskSequence' },
    '3002,011B' => { VR => 'SQ', Name => 'SubtaskWorkitemCodeSequence' },
    '3002,011C' => { VR => 'US', Name => 'AcquisitionTaskIndex' },
    '3002,011D' => { VR => 'US', Name => 'AcquisitionSubtaskIndex' },
    '3002,011E' => { VR => 'SQ', Name => 'RefBaseParamsRTRadiationInstanceSq' },
    '3002,011F' => { VR => 'SQ', Name => 'PositionAcquisitionTemplateIDSeq' },
    '3002,0120' => { VR => 'ST', Name => 'PositionAcquisitionTemplateID' },
    '3002,0121' => { VR => 'LO', Name => 'PositionAcquisitionTemplateName' },
    '3002,0122' => { VR => 'SQ', Name => 'PositionAcqTemplateCodeSequence' },
    '3002,0123' => { VR => 'LT', Name => 'PositionAcqTemplateDescription' },
    '3002,0124' => { VR => 'SQ', Name => 'AcquisitionTaskApplicabilitySeq' },
    '3002,0125' => { VR => 'SQ', Name => 'ProjectionImagingAcqParamSeq' },
    '3002,0126' => { VR => 'SQ', Name => 'CTImagingAcquisitionParameterSeq' },
    '3002,0127' => { VR => 'SQ', Name => 'KVImagingGenerationParametersSeq' },
    '3002,0128' => { VR => 'SQ', Name => 'MVImagingGenerationParametersSeq' },
    '3002,0129' => { VR => 'CS', Name => 'AcquisitionSignalType' },
    '3002,012A' => { VR => 'CS', Name => 'AcquisitionMethod' },
    '3002,012B' => { VR => 'SQ', Name => 'ScanStartPositionSequence' },
    '3002,012C' => { VR => 'SQ', Name => 'ScanStopPositionSequence' },
    '3002,012D' => { VR => 'FD', Name => 'ImagingSourceToBeamModDefPlaneDist' },
    '3002,012E' => { VR => 'CS', Name => 'ScanArcType' },
    '3002,012F' => { VR => 'CS', Name => 'DetectorPositioningType' },
    '3002,0130' => { VR => 'SQ', Name => 'AdditionalRTAccessoryDeviceSeq' },
    '3002,0131' => { VR => 'SQ', Name => 'DeviceSpecificAcquisitionParamSeq' },
    '3002,0132' => { VR => 'SQ', Name => 'ReferencedPositionRefInstanceSeq' },
    '3002,0133' => { VR => 'SQ', Name => 'EnergyDerivationCodeSequence' },
    '3002,0134' => { VR => 'FD', Name => 'MaximumCumulativeMetersetExposure' },
    '3002,0135' => { VR => 'SQ', Name => 'AcquisitionInitiationSequence' },
    '3002,0136' => { VR => 'SQ', Name => 'RTConeBeamImagingGeometrySequence' },
    '3004,0001' => { VR => 'CS', Name => 'DVHType' },
    '3004,0002' => { VR => 'CS', Name => 'DoseUnits' },
    '3004,0004' => { VR => 'CS', Name => 'DoseType' },
    '3004,0005' => { VR => 'CS', Name => 'SpatialTransformOfDose' },
    '3004,0006' => { VR => 'LO', Name => 'DoseComment' },
    '3004,0008' => { VR => 'DS', Name => 'NormalizationPoint' },
    '3004,000A' => { VR => 'CS', Name => 'DoseSummationType' },
    '3004,000C' => { VR => 'DS', Name => 'GridFrameOffsetVector' },
    '3004,000E' => { VR => 'DS', Name => 'DoseGridScaling' },
    '3004,0010' => { VR => 'SQ', Name => 'RTDoseROISequence' },
    '3004,0012' => { VR => 'DS', Name => 'DoseValue' },
    '3004,0014' => { VR => 'CS', Name => 'TissueHeterogeneityCorrection' },
    '3004,0016' => { VR => 'SQ', Name => 'RecommendedIsodoseLevelSequence' },
    '3004,0020' => { VR => 'SQ', Name => 'DoseUnitCodeSequence' },
    '3004,0021' => { VR => 'SQ', Name => 'RTDoseInterpretedTypeCodeSequence' },
    '3004,0022' => { VR => 'SQ', Name => 'RTDoseInterpretedTypeCodeModSeq' },
    '3004,0023' => { VR => 'SQ', Name => 'DoseRadiobiologicalInterpretSeq' },
    '3004,0024' => { VR => 'SQ', Name => 'RTDoseIntentCodeSequence' },
    '3004,0040' => { VR => 'DS', Name => 'DVHNormalizationPoint' },
    '3004,0042' => { VR => 'DS', Name => 'DVHNormalizationDoseValue' },
    '3004,0050' => { VR => 'SQ', Name => 'DVHSequence' },
    '3004,0052' => { VR => 'DS', Name => 'DVHDoseScaling' },
    '3004,0054' => { VR => 'CS', Name => 'DVHVolumeUnits' },
    '3004,0056' => { VR => 'IS', Name => 'DVHNumberOfBins' },
    '3004,0058' => { VR => 'DS', Name => 'DVHData' },
    '3004,0060' => { VR => 'SQ', Name => 'DVHReferencedROISequence' },
    '3004,0062' => { VR => 'CS', Name => 'DVHROIContributionType' },
    '3004,0070' => { VR => 'DS', Name => 'DVHMinimumDose' },
    '3004,0072' => { VR => 'DS', Name => 'DVHMaximumDose' },
    '3004,0074' => { VR => 'DS', Name => 'DVHMeanDose' },
    '3004,0080' => { VR => 'SQ', Name => 'DoseCalculationModelSequence' },
    '3004,0081' => { VR => 'SQ', Name => 'DoseCalculationAlgorithmSequence' },
    '3004,0082' => { VR => 'CS', Name => 'CommissioningStatus' },
    '3004,0083' => { VR => 'SQ', Name => 'DoseCalculationModelParameterSeq' },
    '3004,0084' => { VR => 'CS', Name => 'DoseDepositionCalculationMedium' },
    '3006,0002' => { VR => 'SH', Name => 'StructureSetLabel' },
    '3006,0004' => { VR => 'LO', Name => 'StructureSetName' },
    '3006,0006' => { VR => 'ST', Name => 'StructureSetDescription' },
    '3006,0008' => { VR => 'DA', Name => 'StructureSetDate' },
    '3006,0009' => { VR => 'TM', Name => 'StructureSetTime' },
    '3006,0010' => { VR => 'SQ', Name => 'ReferencedFrameOfReferenceSequence' },
    '3006,0012' => { VR => 'SQ', Name => 'RTReferencedStudySequence' },
    '3006,0014' => { VR => 'SQ', Name => 'RTReferencedSeriesSequence' },
    '3006,0016' => { VR => 'SQ', Name => 'ContourImageSequence' },
    '3006,0018' => { VR => 'SQ', Name => 'PredecessorStructureSetSequence' },
    '3006,0020' => { VR => 'SQ', Name => 'StructureSetROISequence' },
    '3006,0022' => { VR => 'IS', Name => 'ROINumber' },
    '3006,0024' => { VR => 'UI', Name => 'ReferencedFrameOfReferenceUID' },
    '3006,0026' => { VR => 'LO', Name => 'ROIName' },
    '3006,0028' => { VR => 'ST', Name => 'ROIDescription' },
    '3006,002A' => { VR => 'IS', Name => 'ROIDisplayColor' },
    '3006,002C' => { VR => 'DS', Name => 'ROIVolume' },
    '3006,002D' => { VR => 'DT', Name => 'ROIDateTime' },
    '3006,002E' => { VR => 'DT', Name => 'ROIObservationDateTime' },
    '3006,0030' => { VR => 'SQ', Name => 'RTRelatedROISequence' },
    '3006,0033' => { VR => 'CS', Name => 'RTROIRelationship' },
    '3006,0036' => { VR => 'CS', Name => 'ROIGenerationAlgorithm' },
    '3006,0037' => { VR => 'SQ', Name => 'ROIDerivationAlgorithmIDSeq' },
    '3006,0038' => { VR => 'LO', Name => 'ROIGenerationDescription' },
    '3006,0039' => { VR => 'SQ', Name => 'ROIContourSequence' },
    '3006,0040' => { VR => 'SQ', Name => 'ContourSequence' },
    '3006,0042' => { VR => 'CS', Name => 'ContourGeometricType' },
    '3006,0044' => { VR => 'DS', Name => 'ContourSlabThickness' },
    '3006,0045' => { VR => 'DS', Name => 'ContourOffsetVector' },
    '3006,0046' => { VR => 'IS', Name => 'NumberOfContourPoints' },
    '3006,0048' => { VR => 'IS', Name => 'ContourNumber' },
    '3006,0049' => { VR => 'IS', Name => 'AttachedContours' },
    '3006,004A' => { VR => 'SQ', Name => 'SourcePixelPlanesCharSeq' },
    '3006,004B' => { VR => 'SQ', Name => 'SourceSeriesSequence' },
    '3006,004C' => { VR => 'SQ', Name => 'SourceSeriesInformationSequence' },
    '3006,004D' => { VR => 'SQ', Name => 'ROICreatorSequence' },
    '3006,004E' => { VR => 'SQ', Name => 'ROIInterpreterSequence' },
    '3006,004F' => { VR => 'SQ', Name => 'ROIObservationContextCodeSequence' },
    '3006,0050' => { VR => 'DS', Name => 'ContourData' },
    '3006,0080' => { VR => 'SQ', Name => 'RTROIObservationsSequence' },
    '3006,0082' => { VR => 'IS', Name => 'ObservationNumber' },
    '3006,0084' => { VR => 'IS', Name => 'ReferencedROINumber' },
    '3006,0085' => { VR => 'SH', Name => 'ROIObservationLabel' },
    '3006,0086' => { VR => 'SQ', Name => 'RTROIIdentificationCodeSequence' },
    '3006,0088' => { VR => 'ST', Name => 'ROIObservationDescription' },
    '3006,00A0' => { VR => 'SQ', Name => 'RelatedRTROIObservationsSequence' },
    '3006,00A4' => { VR => 'CS', Name => 'RTROIInterpretedType' },
    '3006,00A6' => { VR => 'PN', Name => 'ROIInterpreter' },
    '3006,00B0' => { VR => 'SQ', Name => 'ROIPhysicalPropertiesSequence' },
    '3006,00B2' => { VR => 'CS', Name => 'ROIPhysicalProperty' },
    '3006,00B4' => { VR => 'DS', Name => 'ROIPhysicalPropertyValue' },
    '3006,00B6' => { VR => 'SQ', Name => 'ROIElementalCompositionSequence' },
    '3006,00B7' => { VR => 'US', Name => 'ROIElementalCompAtomicNumber' },
    '3006,00B8' => { VR => 'FL', Name => 'ROIElementalCompAtomicMassFraction' },
    '3006,00B9' => { VR => 'SQ', Name => 'AdditionalRTROIIDCodeSeq' },
    '3006,00C0' => { VR => 'SQ', Name => 'FrameOfReferenceRelationshipSeq' },
    '3006,00C2' => { VR => 'UI', Name => 'RelatedFrameOfReferenceUID' },
    '3006,00C4' => { VR => 'CS', Name => 'FrameOfReferenceTransformType' },
    '3006,00C6' => { VR => 'DS', Name => 'FrameOfReferenceTransformMatrix' },
    '3006,00C8' => { VR => 'LO', Name => 'FrameOfReferenceTransformComment' },
    '3006,00C9' => { VR => 'SQ', Name => 'PatientLocationCoordinatesSequence' },
    '3006,00CA' => { VR => 'SQ', Name => 'PatientLocationCoordinatesCodeSeq' },
    '3006,00CB' => { VR => 'SQ', Name => 'PatientSupportPositionSequence' },
    '3008,0010' => { VR => 'SQ', Name => 'MeasuredDoseReferenceSequence' },
    '3008,0012' => { VR => 'ST', Name => 'MeasuredDoseDescription' },
    '3008,0014' => { VR => 'CS', Name => 'MeasuredDoseType' },
    '3008,0016' => { VR => 'DS', Name => 'MeasuredDoseValue' },
    '3008,0020' => { VR => 'SQ', Name => 'TreatmentSessionBeamSequence' },
    '3008,0021' => { VR => 'SQ', Name => 'TreatmentSessionIonBeamSequence' },
    '3008,0022' => { VR => 'IS', Name => 'CurrentFractionNumber' },
    '3008,0024' => { VR => 'DA', Name => 'TreatmentControlPointDate' },
    '3008,0025' => { VR => 'TM', Name => 'TreatmentControlPointTime' },
    '3008,002A' => { VR => 'CS', Name => 'TreatmentTerminationStatus' },
    '3008,002B' => { VR => 'SH', Name => 'TreatmentTerminationCode' },
    '3008,002C' => { VR => 'CS', Name => 'TreatmentVerificationStatus' },
    '3008,0030' => { VR => 'SQ', Name => 'ReferencedTreatmentRecordSequence' },
    '3008,0032' => { VR => 'DS', Name => 'SpecifiedPrimaryMeterset' },
    '3008,0033' => { VR => 'DS', Name => 'SpecifiedSecondaryMeterset' },
    '3008,0036' => { VR => 'DS', Name => 'DeliveredPrimaryMeterset' },
    '3008,0037' => { VR => 'DS', Name => 'DeliveredSecondaryMeterset' },
    '3008,003A' => { VR => 'DS', Name => 'SpecifiedTreatmentTime' },
    '3008,003B' => { VR => 'DS', Name => 'DeliveredTreatmentTime' },
    '3008,0040' => { VR => 'SQ', Name => 'ControlPointDeliverySequence' },
    '3008,0041' => { VR => 'SQ', Name => 'IonControlPointDeliverySequence' },
    '3008,0042' => { VR => 'DS', Name => 'SpecifiedMeterset' },
    '3008,0044' => { VR => 'DS', Name => 'DeliveredMeterset' },
    '3008,0045' => { VR => 'FL', Name => 'MetersetRateSet' },
    '3008,0046' => { VR => 'FL', Name => 'MetersetRateDelivered' },
    '3008,0047' => { VR => 'FL', Name => 'ScanSpotMetersetsDelivered' },
    '3008,0048' => { VR => 'DS', Name => 'DoseRateDelivered' },
    '3008,0050' => { VR => 'SQ', Name => 'TreatmentSummaryCalcDoseRefSeq' },
    '3008,0052' => { VR => 'DS', Name => 'CumulativeDoseToDoseReference' },
    '3008,0054' => { VR => 'DA', Name => 'FirstTreatmentDate' },
    '3008,0056' => { VR => 'DA', Name => 'MostRecentTreatmentDate' },
    '3008,005A' => { VR => 'IS', Name => 'NumberOfFractionsDelivered' },
    '3008,0060' => { VR => 'SQ', Name => 'OverrideSequence' },
    '3008,0061' => { VR => 'AT', Name => 'ParameterSequencePointer' },
    '3008,0062' => { VR => 'AT', Name => 'OverrideParameterPointer' },
    '3008,0063' => { VR => 'IS', Name => 'ParameterItemIndex' },
    '3008,0064' => { VR => 'IS', Name => 'MeasuredDoseReferenceNumber' },
    '3008,0065' => { VR => 'AT', Name => 'ParameterPointer' },
    '3008,0066' => { VR => 'ST', Name => 'OverrideReason' },
    '3008,0067' => { VR => 'US', Name => 'ParameterValueNumber' },
    '3008,0068' => { VR => 'SQ', Name => 'CorrectedParameterSequence' },
    '3008,006A' => { VR => 'FL', Name => 'CorrectionValue' },
    '3008,0070' => { VR => 'SQ', Name => 'CalculatedDoseReferenceSequence' },
    '3008,0072' => { VR => 'IS', Name => 'CalculatedDoseReferenceNumber' },
    '3008,0074' => { VR => 'ST', Name => 'CalculatedDoseReferenceDescription' },
    '3008,0076' => { VR => 'DS', Name => 'CalculatedDoseReferenceDoseValue' },
    '3008,0078' => { VR => 'DS', Name => 'StartMeterset' },
    '3008,007A' => { VR => 'DS', Name => 'EndMeterset' },
    '3008,0080' => { VR => 'SQ', Name => 'ReferencedMeasuredDoseReferenceSeq' },
    '3008,0082' => { VR => 'IS', Name => 'ReferencedMeasuredDoseReferenceNum' },
    '3008,0090' => { VR => 'SQ', Name => 'ReferencedCalculatedDoseRefSeq' },
    '3008,0092' => { VR => 'IS', Name => 'ReferencedCalculatedDoseRefNumber' },
    '3008,00A0' => { VR => 'SQ', Name => 'BeamLimitingDeviceLeafPairsSeq' },
    '3008,00A1' => { VR => 'SQ', Name => 'EnhancedRTBeamLimitingDeviceSeq' },
    '3008,00A2' => { VR => 'SQ', Name => 'EnhancedRTBeamLimitingOpeningSeq' },
    '3008,00A3' => { VR => 'CS', Name => 'EnhancedRTBeamLimitingDevDefFlag' },
    '3008,00A4' => { VR => 'FD', Name => 'ParallelRTBeamDelimiterOpenExtents' },
    '3008,00B0' => { VR => 'SQ', Name => 'RecordedWedgeSequence' },
    '3008,00C0' => { VR => 'SQ', Name => 'RecordedCompensatorSequence' },
    '3008,00D0' => { VR => 'SQ', Name => 'RecordedBlockSequence' },
    '3008,00D1' => { VR => 'SQ', Name => 'RecordedBlockSlabSequence' },
    '3008,00E0' => { VR => 'SQ', Name => 'TreatmentSummaryMeasuredDoseRefSeq' },
    '3008,00F0' => { VR => 'SQ', Name => 'RecordedSnoutSequence' },
    '3008,00F2' => { VR => 'SQ', Name => 'RecordedRangeShifterSequence' },
    '3008,00F4' => { VR => 'SQ', Name => 'RecordedLateralSpreadingDeviceSeq' },
    '3008,00F6' => { VR => 'SQ', Name => 'RecordedRangeModulatorSequence' },
    '3008,0100' => { VR => 'SQ', Name => 'RecordedSourceSequence' },
    '3008,0105' => { VR => 'LO', Name => 'SourceSerialNumber' },
    '3008,0110' => { VR => 'SQ', Name => 'TreatmentSessionAppSetupSeq' },
    '3008,0116' => { VR => 'CS', Name => 'ApplicationSetupCheck' },
    '3008,0120' => { VR => 'SQ', Name => 'RecordedBrachyAccessoryDeviceSeq' },
    '3008,0122' => { VR => 'IS', Name => 'ReferencedBrachyAccessoryDeviceNum' },
    '3008,0130' => { VR => 'SQ', Name => 'RecordedChannelSequence' },
    '3008,0132' => { VR => 'DS', Name => 'SpecifiedChannelTotalTime' },
    '3008,0134' => { VR => 'DS', Name => 'DeliveredChannelTotalTime' },
    '3008,0136' => { VR => 'IS', Name => 'SpecifiedNumberOfPulses' },
    '3008,0138' => { VR => 'IS', Name => 'DeliveredNumberOfPulses' },
    '3008,013A' => { VR => 'DS', Name => 'SpecifiedPulseRepetitionInterval' },
    '3008,013C' => { VR => 'DS', Name => 'DeliveredPulseRepetitionInterval' },
    '3008,0140' => { VR => 'SQ', Name => 'RecordedSourceApplicatorSequence' },
    '3008,0142' => { VR => 'IS', Name => 'ReferencedSourceApplicatorNumber' },
    '3008,0150' => { VR => 'SQ', Name => 'RecordedChannelShieldSequence' },
    '3008,0152' => { VR => 'IS', Name => 'ReferencedChannelShieldNumber' },
    '3008,0160' => { VR => 'SQ', Name => 'BrachyControlPointDeliveredSeq' },
    '3008,0162' => { VR => 'DA', Name => 'SafePositionExitDate' },
    '3008,0164' => { VR => 'TM', Name => 'SafePositionExitTime' },
    '3008,0166' => { VR => 'DA', Name => 'SafePositionReturnDate' },
    '3008,0168' => { VR => 'TM', Name => 'SafePositionReturnTime' },
    '3008,0171' => { VR => 'SQ', Name => 'PulseSpecBrachyCntrlPtDeliveredSeq' },
    '3008,0172' => { VR => 'US', Name => 'PulseNumber' },
    '3008,0173' => { VR => 'SQ', Name => 'BrachyPulseCntrlPointDeliveredSeq' },
    '3008,0200' => { VR => 'CS', Name => 'CurrentTreatmentStatus' },
    '3008,0202' => { VR => 'ST', Name => 'TreatmentStatusComment' },
    '3008,0220' => { VR => 'SQ', Name => 'FractionGroupSummarySequence' },
    '3008,0223' => { VR => 'IS', Name => 'ReferencedFractionNumber' },
    '3008,0224' => { VR => 'CS', Name => 'FractionGroupType' },
    '3008,0230' => { VR => 'CS', Name => 'BeamStopperPosition' },
    '3008,0240' => { VR => 'SQ', Name => 'FractionStatusSummarySequence' },
    '3008,0250' => { VR => 'DA', Name => 'TreatmentDate' },
    '3008,0251' => { VR => 'TM', Name => 'TreatmentTime' },
    '300A,0002' => { VR => 'SH', Name => 'RTPlanLabel' },
    '300A,0003' => { VR => 'LO', Name => 'RTPlanName' },
    '300A,0004' => { VR => 'ST', Name => 'RTPlanDescription' },
    '300A,0006' => { VR => 'DA', Name => 'RTPlanDate' },
    '300A,0007' => { VR => 'TM', Name => 'RTPlanTime' },
    '300A,0009' => { VR => 'LO', Name => 'TreatmentProtocols' },
    '300A,000A' => { VR => 'CS', Name => 'PlanIntent' },
    '300A,000B' => { VR => 'LO', Name => 'TreatmentSites' },
    '300A,000C' => { VR => 'CS', Name => 'RTPlanGeometry' },
    '300A,000E' => { VR => 'ST', Name => 'PrescriptionDescription' },
    '300A,0010' => { VR => 'SQ', Name => 'DoseReferenceSequence' },
    '300A,0012' => { VR => 'IS', Name => 'DoseReferenceNumber' },
    '300A,0013' => { VR => 'UI', Name => 'DoseReferenceUID' },
    '300A,0014' => { VR => 'CS', Name => 'DoseReferenceStructureType' },
    '300A,0015' => { VR => 'CS', Name => 'NominalBeamEnergyUnit' },
    '300A,0016' => { VR => 'LO', Name => 'DoseReferenceDescription' },
    '300A,0018' => { VR => 'DS', Name => 'DoseReferencePointCoordinates' },
    '300A,001A' => { VR => 'DS', Name => 'NominalPriorDose' },
    '300A,0020' => { VR => 'CS', Name => 'DoseReferenceType' },
    '300A,0021' => { VR => 'DS', Name => 'ConstraintWeight' },
    '300A,0022' => { VR => 'DS', Name => 'DeliveryWarningDose' },
    '300A,0023' => { VR => 'DS', Name => 'DeliveryMaximumDose' },
    '300A,0025' => { VR => 'DS', Name => 'TargetMinimumDose' },
    '300A,0026' => { VR => 'DS', Name => 'TargetPrescriptionDose' },
    '300A,0027' => { VR => 'DS', Name => 'TargetMaximumDose' },
    '300A,0028' => { VR => 'DS', Name => 'TargetUnderdoseVolumeFraction' },
    '300A,002A' => { VR => 'DS', Name => 'OrganAtRiskFullVolumeDose' },
    '300A,002B' => { VR => 'DS', Name => 'OrganAtRiskLimitDose' },
    '300A,002C' => { VR => 'DS', Name => 'OrganAtRiskMaximumDose' },
    '300A,002D' => { VR => 'DS', Name => 'OrganAtRiskOverdoseVolumeFraction' },
    '300A,0040' => { VR => 'SQ', Name => 'ToleranceTableSequence' },
    '300A,0042' => { VR => 'IS', Name => 'ToleranceTableNumber' },
    '300A,0043' => { VR => 'SH', Name => 'ToleranceTableLabel' },
    '300A,0044' => { VR => 'DS', Name => 'GantryAngleTolerance' },
    '300A,0046' => { VR => 'DS', Name => 'BeamLimitingDeviceAngleTolerance' },
    '300A,0048' => { VR => 'SQ', Name => 'BeamLimitingDeviceToleranceSeq' },
    '300A,004A' => { VR => 'DS', Name => 'BeamLimitingDevicePositionTol' },
    '300A,004B' => { VR => 'FL', Name => 'SnoutPositionTolerance' },
    '300A,004C' => { VR => 'DS', Name => 'PatientSupportAngleTolerance' },
    '300A,004E' => { VR => 'DS', Name => 'TableTopEccentricAngleTolerance' },
    '300A,004F' => { VR => 'FL', Name => 'TableTopPitchAngleTolerance' },
    '300A,0050' => { VR => 'FL', Name => 'TableTopRollAngleTolerance' },
    '300A,0051' => { VR => 'DS', Name => 'TableTopVerticalPositionTolerance' },
    '300A,0052' => { VR => 'DS', Name => 'TableTopLongitudinalPositionTol' },
    '300A,0053' => { VR => 'DS', Name => 'TableTopLateralPositionTolerance' },
    '300A,0054' => { VR => 'UI', Name => 'TableTopPositionAlignmentUID' },
    '300A,0055' => { VR => 'CS', Name => 'RTPlanRelationship' },
    '300A,0070' => { VR => 'SQ', Name => 'FractionGroupSequence' },
    '300A,0071' => { VR => 'IS', Name => 'FractionGroupNumber' },
    '300A,0072' => { VR => 'LO', Name => 'FractionGroupDescription' },
    '300A,0078' => { VR => 'IS', Name => 'NumberOfFractionsPlanned' },
    '300A,0079' => { VR => 'IS', Name => 'NumberFractionPatternDigitsPerDay' },
    '300A,007A' => { VR => 'IS', Name => 'RepeatFractionCycleLength' },
    '300A,007B' => { VR => 'LT', Name => 'FractionPattern' },
    '300A,0080' => { VR => 'IS', Name => 'NumberOfBeams' },
    '300A,0082' => { VR => 'DS', Name => 'BeamDoseSpecificationPoint' },
    '300A,0083' => { VR => 'UI', Name => 'ReferencedDoseReferenceUID' },
    '300A,0084' => { VR => 'DS', Name => 'BeamDose' },
    '300A,0086' => { VR => 'DS', Name => 'BeamMeterset' },
    '300A,0088' => { VR => 'FL', Name => 'BeamDosePointDepth' },
    '300A,0089' => { VR => 'FL', Name => 'BeamDosePointEquivalentDepth' },
    '300A,008A' => { VR => 'FL', Name => 'BeamDosePointSSD' },
    '300A,008B' => { VR => 'CS', Name => 'BeamDoseMeaning' },
    '300A,008C' => { VR => 'SQ', Name => 'BeamDoseVerificationCntrlPointSeq' },
    '300A,008D' => { VR => 'FL', Name => 'AverageBeamDosePointDepth' },
    '300A,008E' => { VR => 'FL', Name => 'AverageBeamDosePointEquivDepth' },
    '300A,008F' => { VR => 'FL', Name => 'AverageBeamDosePointSSD' },
    '300A,0090' => { VR => 'CS', Name => 'BeamDoseType' },
    '300A,0091' => { VR => 'DS', Name => 'AlternateBeamDose' },
    '300A,0092' => { VR => 'CS', Name => 'AlternateBeamDoseType' },
    '300A,0093' => { VR => 'CS', Name => 'DepthValueAveragingFlag' },
    '300A,0094' => { VR => 'DS', Name => 'BeamDosePointSrcToExtContourDist' },
    '300A,00A0' => { VR => 'IS', Name => 'NumberOfBrachyApplicationSetups' },
    '300A,00A2' => { VR => 'DS', Name => 'BrachyAppSetupDoseSpecPoint' },
    '300A,00A4' => { VR => 'DS', Name => 'BrachyApplicationSetupDose' },
    '300A,00B0' => { VR => 'SQ', Name => 'BeamSequence' },
    '300A,00B2' => { VR => 'SH', Name => 'TreatmentMachineName' },
    '300A,00B3' => { VR => 'CS', Name => 'PrimaryDosimeterUnit' },
    '300A,00B4' => { VR => 'DS', Name => 'SourceAxisDistance' },
    '300A,00B6' => { VR => 'SQ', Name => 'BeamLimitingDeviceSequence' },
    '300A,00B8' => { VR => 'CS', Name => 'RTBeamLimitingDeviceType' },
    '300A,00BA' => { VR => 'DS', Name => 'SourceToBeamLimitingDeviceDistance' },
    '300A,00BB' => { VR => 'FL', Name => 'IsocenterToBeamLimitingDeviceDist' },
    '300A,00BC' => { VR => 'IS', Name => 'NumberOfLeafJawPairs' },
    '300A,00BE' => { VR => 'DS', Name => 'LeafPositionBoundaries' },
    '300A,00C0' => { VR => 'IS', Name => 'BeamNumber' },
    '300A,00C2' => { VR => 'LO', Name => 'BeamName' },
    '300A,00C3' => { VR => 'ST', Name => 'BeamDescription' },
    '300A,00C4' => { VR => 'CS', Name => 'BeamType' },
    '300A,00C5' => { VR => 'FD', Name => 'BeamDeliveryDurationLimit' },
    '300A,00C6' => { VR => 'CS', Name => 'RadiationType' },
    '300A,00C7' => { VR => 'CS', Name => 'HighDoseTechniqueType' },
    '300A,00C8' => { VR => 'IS', Name => 'ReferenceImageNumber' },
    '300A,00CA' => { VR => 'SQ', Name => 'PlannedVerificationImageSequence' },
    '300A,00CC' => { VR => 'LO', Name => 'ImagingDeviceSpecificAcqParams' },
    '300A,00CE' => { VR => 'CS', Name => 'TreatmentDeliveryType' },
    '300A,00D0' => { VR => 'IS', Name => 'NumberOfWedges' },
    '300A,00D1' => { VR => 'SQ', Name => 'WedgeSequence' },
    '300A,00D2' => { VR => 'IS', Name => 'WedgeNumber' },
    '300A,00D3' => { VR => 'CS', Name => 'WedgeType' },
    '300A,00D4' => { VR => 'SH', Name => 'WedgeID' },
    '300A,00D5' => { VR => 'IS', Name => 'WedgeAngle' },
    '300A,00D6' => { VR => 'DS', Name => 'WedgeFactor' },
    '300A,00D7' => { VR => 'FL', Name => 'TotalWedgeTrayWaterEquivThickness' },
    '300A,00D8' => { VR => 'DS', Name => 'WedgeOrientation' },
    '300A,00D9' => { VR => 'FL', Name => 'IsocenterToWedgeTrayDistance' },
    '300A,00DA' => { VR => 'DS', Name => 'SourceToWedgeTrayDistance' },
    '300A,00DB' => { VR => 'FL', Name => 'WedgeThinEdgePosition' },
    '300A,00DC' => { VR => 'SH', Name => 'BolusID' },
    '300A,00DD' => { VR => 'ST', Name => 'BolusDescription' },
    '300A,00DE' => { VR => 'DS', Name => 'EffectiveWedgeAngle' },
    '300A,00E0' => { VR => 'IS', Name => 'NumberOfCompensators' },
    '300A,00E1' => { VR => 'SH', Name => 'MaterialID' },
    '300A,00E2' => { VR => 'DS', Name => 'TotalCompensatorTrayFactor' },
    '300A,00E3' => { VR => 'SQ', Name => 'CompensatorSequence' },
    '300A,00E4' => { VR => 'IS', Name => 'CompensatorNumber' },
    '300A,00E5' => { VR => 'SH', Name => 'CompensatorID' },
    '300A,00E6' => { VR => 'DS', Name => 'SourceToCompensatorTrayDistance' },
    '300A,00E7' => { VR => 'IS', Name => 'CompensatorRows' },
    '300A,00E8' => { VR => 'IS', Name => 'CompensatorColumns' },
    '300A,00E9' => { VR => 'DS', Name => 'CompensatorPixelSpacing' },
    '300A,00EA' => { VR => 'DS', Name => 'CompensatorPosition' },
    '300A,00EB' => { VR => 'DS', Name => 'CompensatorTransmissionData' },
    '300A,00EC' => { VR => 'DS', Name => 'CompensatorThicknessData' },
    '300A,00ED' => { VR => 'IS', Name => 'NumberOfBoli' },
    '300A,00EE' => { VR => 'CS', Name => 'CompensatorType' },
    '300A,00EF' => { VR => 'SH', Name => 'CompensatorTrayID' },
    '300A,00F0' => { VR => 'IS', Name => 'NumberOfBlocks' },
    '300A,00F2' => { VR => 'DS', Name => 'TotalBlockTrayFactor' },
    '300A,00F3' => { VR => 'FL', Name => 'TotalBlockTrayWaterEquivThickness' },
    '300A,00F4' => { VR => 'SQ', Name => 'BlockSequence' },
    '300A,00F5' => { VR => 'SH', Name => 'BlockTrayID' },
    '300A,00F6' => { VR => 'DS', Name => 'SourceToBlockTrayDistance' },
    '300A,00F7' => { VR => 'FL', Name => 'IsocenterToBlockTrayDistance' },
    '300A,00F8' => { VR => 'CS', Name => 'BlockType' },
    '300A,00F9' => { VR => 'LO', Name => 'AccessoryCode' },
    '300A,00FA' => { VR => 'CS', Name => 'BlockDivergence' },
    '300A,00FB' => { VR => 'CS', Name => 'BlockMountingPosition' },
    '300A,00FC' => { VR => 'IS', Name => 'BlockNumber' },
    '300A,00FE' => { VR => 'LO', Name => 'BlockName' },
    '300A,0100' => { VR => 'DS', Name => 'BlockThickness' },
    '300A,0102' => { VR => 'DS', Name => 'BlockTransmission' },
    '300A,0104' => { VR => 'IS', Name => 'BlockNumberOfPoints' },
    '300A,0106' => { VR => 'DS', Name => 'BlockData' },
    '300A,0107' => { VR => 'SQ', Name => 'ApplicatorSequence' },
    '300A,0108' => { VR => 'SH', Name => 'ApplicatorID' },
    '300A,0109' => { VR => 'CS', Name => 'ApplicatorType' },
    '300A,010A' => { VR => 'LO', Name => 'ApplicatorDescription' },
    '300A,010C' => { VR => 'DS', Name => 'CumulativeDoseReferenceCoefficient' },
    '300A,010E' => { VR => 'DS', Name => 'FinalCumulativeMetersetWeight' },
    '300A,0110' => { VR => 'IS', Name => 'NumberOfControlPoints' },
    '300A,0111' => { VR => 'SQ', Name => 'ControlPointSequence' },
    '300A,0112' => { VR => 'IS', Name => 'ControlPointIndex' },
    '300A,0114' => { VR => 'DS', Name => 'NominalBeamEnergy' },
    '300A,0115' => { VR => 'DS', Name => 'DoseRateSet' },
    '300A,0116' => { VR => 'SQ', Name => 'WedgePositionSequence' },
    '300A,0118' => { VR => 'CS', Name => 'WedgePosition' },
    '300A,011A' => { VR => 'SQ', Name => 'BeamLimitingDevicePositionSequence' },
    '300A,011C' => { VR => 'DS', Name => 'LeafJawPositions' },
    '300A,011E' => { VR => 'DS', Name => 'GantryAngle' },
    '300A,011F' => { VR => 'CS', Name => 'GantryRotationDirection' },
    '300A,0120' => { VR => 'DS', Name => 'BeamLimitingDeviceAngle' },
    '300A,0121' => { VR => 'CS', Name => 'BeamLimitingDeviceRotateDirection' },
    '300A,0122' => { VR => 'DS', Name => 'PatientSupportAngle' },
    '300A,0123' => { VR => 'CS', Name => 'PatientSupportRotationDirection' },
    '300A,0124' => { VR => 'DS', Name => 'TableTopEccentricAxisDistance' },
    '300A,0125' => { VR => 'DS', Name => 'TableTopEccentricAngle' },
    '300A,0126' => { VR => 'CS', Name => 'TableTopEccentricRotateDirection' },
    '300A,0128' => { VR => 'DS', Name => 'TableTopVerticalPosition' },
    '300A,0129' => { VR => 'DS', Name => 'TableTopLongitudinalPosition' },
    '300A,012A' => { VR => 'DS', Name => 'TableTopLateralPosition' },
    '300A,012C' => { VR => 'DS', Name => 'IsocenterPosition' },
    '300A,012E' => { VR => 'DS', Name => 'SurfaceEntryPoint' },
    '300A,0130' => { VR => 'DS', Name => 'SourceToSurfaceDistance' },
    '300A,0131' => { VR => 'FL', Name => 'AvgBeamDosePtSrcToExtContourDist' },
    '300A,0132' => { VR => 'FL', Name => 'SourceToExternalContourDistance' },
    '300A,0133' => { VR => 'FL', Name => 'ExternalContourEntryPoint' },
    '300A,0134' => { VR => 'DS', Name => 'CumulativeMetersetWeight' },
    '300A,0140' => { VR => 'FL', Name => 'TableTopPitchAngle' },
    '300A,0142' => { VR => 'CS', Name => 'TableTopPitchRotationDirection' },
    '300A,0144' => { VR => 'FL', Name => 'TableTopRollAngle' },
    '300A,0146' => { VR => 'CS', Name => 'TableTopRollRotationDirection' },
    '300A,0148' => { VR => 'FL', Name => 'HeadFixationAngle' },
    '300A,014A' => { VR => 'FL', Name => 'GantryPitchAngle' },
    '300A,014C' => { VR => 'CS', Name => 'GantryPitchRotationDirection' },
    '300A,014E' => { VR => 'FL', Name => 'GantryPitchAngleTolerance' },
    '300A,0150' => { VR => 'CS', Name => 'FixationEye' },
    '300A,0151' => { VR => 'DS', Name => 'ChairHeadFramePosition' },
    '300A,0152' => { VR => 'DS', Name => 'HeadFixationAngleTolerance' },
    '300A,0153' => { VR => 'DS', Name => 'ChairHeadFramePositionTolerance' },
    '300A,0154' => { VR => 'DS', Name => 'FixationLightAzimuthalAngleTol' },
    '300A,0155' => { VR => 'DS', Name => 'FixationLightPolarAngleTolerance' },
    '300A,0180' => { VR => 'SQ', Name => 'PatientSetupSequence' },
    '300A,0182' => { VR => 'IS', Name => 'PatientSetupNumber' },
    '300A,0183' => { VR => 'LO', Name => 'PatientSetupLabel' },
    '300A,0184' => { VR => 'LO', Name => 'PatientAdditionalPosition' },
    '300A,0190' => { VR => 'SQ', Name => 'FixationDeviceSequence' },
    '300A,0192' => { VR => 'CS', Name => 'FixationDeviceType' },
    '300A,0194' => { VR => 'SH', Name => 'FixationDeviceLabel' },
    '300A,0196' => { VR => 'ST', Name => 'FixationDeviceDescription' },
    '300A,0198' => { VR => 'SH', Name => 'FixationDevicePosition' },
    '300A,0199' => { VR => 'FL', Name => 'FixationDevicePitchAngle' },
    '300A,019A' => { VR => 'FL', Name => 'FixationDeviceRollAngle' },
    '300A,01A0' => { VR => 'SQ', Name => 'ShieldingDeviceSequence' },
    '300A,01A2' => { VR => 'CS', Name => 'ShieldingDeviceType' },
    '300A,01A4' => { VR => 'SH', Name => 'ShieldingDeviceLabel' },
    '300A,01A6' => { VR => 'ST', Name => 'ShieldingDeviceDescription' },
    '300A,01A8' => { VR => 'SH', Name => 'ShieldingDevicePosition' },
    '300A,01B0' => { VR => 'CS', Name => 'SetupTechnique' },
    '300A,01B2' => { VR => 'ST', Name => 'SetupTechniqueDescription' },
    '300A,01B4' => { VR => 'SQ', Name => 'SetupDeviceSequence' },
    '300A,01B6' => { VR => 'CS', Name => 'SetupDeviceType' },
    '300A,01B8' => { VR => 'SH', Name => 'SetupDeviceLabel' },
    '300A,01BA' => { VR => 'ST', Name => 'SetupDeviceDescription' },
    '300A,01BC' => { VR => 'DS', Name => 'SetupDeviceParameter' },
    '300A,01D0' => { VR => 'ST', Name => 'SetupReferenceDescription' },
    '300A,01D2' => { VR => 'DS', Name => 'TableTopVerticalSetupDisplacement' },
    '300A,01D4' => { VR => 'DS', Name => 'TableTopLongitudinalSetupDisplace' },
    '300A,01D6' => { VR => 'DS', Name => 'TableTopLateralSetupDisplacement' },
    '300A,0200' => { VR => 'CS', Name => 'BrachyTreatmentTechnique' },
    '300A,0202' => { VR => 'CS', Name => 'BrachyTreatmentType' },
    '300A,0206' => { VR => 'SQ', Name => 'TreatmentMachineSequence' },
    '300A,0210' => { VR => 'SQ', Name => 'SourceSequence' },
    '300A,0212' => { VR => 'IS', Name => 'SourceNumber' },
    '300A,0214' => { VR => 'CS', Name => 'SourceType' },
    '300A,0216' => { VR => 'LO', Name => 'SourceManufacturer' },
    '300A,0218' => { VR => 'DS', Name => 'ActiveSourceDiameter' },
    '300A,021A' => { VR => 'DS', Name => 'ActiveSourceLength' },
    '300A,021B' => { VR => 'SH', Name => 'SourceModelID' },
    '300A,021C' => { VR => 'LO', Name => 'SourceDescription' },
    '300A,0222' => { VR => 'DS', Name => 'SourceEncapsulationNomThickness' },
    '300A,0224' => { VR => 'DS', Name => 'SourceEncapsulationNomTransmission' },
    '300A,0226' => { VR => 'LO', Name => 'SourceIsotopeName' },
    '300A,0228' => { VR => 'DS', Name => 'SourceIsotopeHalfLife' },
    '300A,0229' => { VR => 'CS', Name => 'SourceStrengthUnits' },
    '300A,022A' => { VR => 'DS', Name => 'ReferenceAirKermaRate' },
    '300A,022B' => { VR => 'DS', Name => 'SourceStrength' },
    '300A,022C' => { VR => 'DA', Name => 'SourceStrengthReferenceDate' },
    '300A,022E' => { VR => 'TM', Name => 'SourceStrengthReferenceTime' },
    '300A,0230' => { VR => 'SQ', Name => 'ApplicationSetupSequence' },
    '300A,0232' => { VR => 'CS', Name => 'ApplicationSetupType' },
    '300A,0234' => { VR => 'IS', Name => 'ApplicationSetupNumber' },
    '300A,0236' => { VR => 'LO', Name => 'ApplicationSetupName' },
    '300A,0238' => { VR => 'LO', Name => 'ApplicationSetupManufacturer' },
    '300A,0240' => { VR => 'IS', Name => 'TemplateNumber' },
    '300A,0242' => { VR => 'SH', Name => 'TemplateType' },
    '300A,0244' => { VR => 'LO', Name => 'TemplateName' },
    '300A,0250' => { VR => 'DS', Name => 'TotalReferenceAirKerma' },
    '300A,0260' => { VR => 'SQ', Name => 'BrachyAccessoryDeviceSequence' },
    '300A,0262' => { VR => 'IS', Name => 'BrachyAccessoryDeviceNumber' },
    '300A,0263' => { VR => 'SH', Name => 'BrachyAccessoryDeviceID' },
    '300A,0264' => { VR => 'CS', Name => 'BrachyAccessoryDeviceType' },
    '300A,0266' => { VR => 'LO', Name => 'BrachyAccessoryDeviceName' },
    '300A,026A' => { VR => 'DS', Name => 'BrachyAccessoryDeviceNomThickness' },
    '300A,026C' => { VR => 'DS', Name => 'BrachyAccessoryDevNomTransmission' },
    '300A,0271' => { VR => 'DS', Name => 'ChannelEffectiveLength' },
    '300A,0272' => { VR => 'DS', Name => 'ChannelInnerLength' },
    '300A,0273' => { VR => 'SH', Name => 'AfterloaderChannelID' },
    '300A,0274' => { VR => 'DS', Name => 'SourceApplicatorTipLength' },
    '300A,0280' => { VR => 'SQ', Name => 'ChannelSequence' },
    '300A,0282' => { VR => 'IS', Name => 'ChannelNumber' },
    '300A,0284' => { VR => 'DS', Name => 'ChannelLength' },
    '300A,0286' => { VR => 'DS', Name => 'ChannelTotalTime' },
    '300A,0288' => { VR => 'CS', Name => 'SourceMovementType' },
    '300A,028A' => { VR => 'IS', Name => 'NumberOfPulses' },
    '300A,028C' => { VR => 'DS', Name => 'PulseRepetitionInterval' },
    '300A,0290' => { VR => 'IS', Name => 'SourceApplicatorNumber' },
    '300A,0291' => { VR => 'SH', Name => 'SourceApplicatorID' },
    '300A,0292' => { VR => 'CS', Name => 'SourceApplicatorType' },
    '300A,0294' => { VR => 'LO', Name => 'SourceApplicatorName' },
    '300A,0296' => { VR => 'DS', Name => 'SourceApplicatorLength' },
    '300A,0298' => { VR => 'LO', Name => 'SourceApplicatorManufacturer' },
    '300A,029C' => { VR => 'DS', Name => 'SourceApplicatorWallNomThickness' },
    '300A,029E' => { VR => 'DS', Name => 'SourceApplicatorWallNomTrans' },
    '300A,02A0' => { VR => 'DS', Name => 'SourceApplicatorStepSize' },
    '300A,02A1' => { VR => 'IS', Name => 'ApplicatorShapeReferencedROINumber' },
    '300A,02A2' => { VR => 'IS', Name => 'TransferTubeNumber' },
    '300A,02A4' => { VR => 'DS', Name => 'TransferTubeLength' },
    '300A,02B0' => { VR => 'SQ', Name => 'ChannelShieldSequence' },
    '300A,02B2' => { VR => 'IS', Name => 'ChannelShieldNumber' },
    '300A,02B3' => { VR => 'SH', Name => 'ChannelShieldID' },
    '300A,02B4' => { VR => 'LO', Name => 'ChannelShieldName' },
    '300A,02B8' => { VR => 'DS', Name => 'ChannelShieldNominalThickness' },
    '300A,02BA' => { VR => 'DS', Name => 'ChannelShieldNominalTransmission' },
    '300A,02C8' => { VR => 'DS', Name => 'FinalCumulativeTimeWeight' },
    '300A,02D0' => { VR => 'SQ', Name => 'BrachyControlPointSequence' },
    '300A,02D2' => { VR => 'DS', Name => 'ControlPointRelativePosition' },
    '300A,02D4' => { VR => 'DS', Name => 'ControlPoint3DPosition' },
    '300A,02D6' => { VR => 'DS', Name => 'CumulativeTimeWeight' },
    '300A,02E0' => { VR => 'CS', Name => 'CompensatorDivergence' },
    '300A,02E1' => { VR => 'CS', Name => 'CompensatorMountingPosition' },
    '300A,02E2' => { VR => 'DS', Name => 'SourceToCompensatorDistance' },
    '300A,02E3' => { VR => 'FL', Name => 'TotalCompTrayWaterEquivThickness' },
    '300A,02E4' => { VR => 'FL', Name => 'IsocenterToCompensatorTrayDistance' },
    '300A,02E5' => { VR => 'FL', Name => 'CompensatorColumnOffset' },
    '300A,02E6' => { VR => 'FL', Name => 'IsocenterToCompensatorDistances' },
    '300A,02E7' => { VR => 'FL', Name => 'CompensatorRelStoppingPowerRatio' },
    '300A,02E8' => { VR => 'FL', Name => 'CompensatorMillingToolDiameter' },
    '300A,02EA' => { VR => 'SQ', Name => 'IonRangeCompensatorSequence' },
    '300A,02EB' => { VR => 'LT', Name => 'CompensatorDescription' },
    '300A,02EC' => { VR => 'CS', Name => 'CompensatorSurfaceRepFlag' },
    '300A,0302' => { VR => 'IS', Name => 'RadiationMassNumber' },
    '300A,0304' => { VR => 'IS', Name => 'RadiationAtomicNumber' },
    '300A,0306' => { VR => 'SS', Name => 'RadiationChargeState' },
    '300A,0308' => { VR => 'CS', Name => 'ScanMode' },
    '300A,0309' => { VR => 'CS', Name => 'ModulatedScanModeType' },
    '300A,030A' => { VR => 'FL', Name => 'VirtualSourceAxisDistances' },
    '300A,030C' => { VR => 'SQ', Name => 'SnoutSequence' },
    '300A,030D' => { VR => 'FL', Name => 'SnoutPosition' },
    '300A,030F' => { VR => 'SH', Name => 'SnoutID' },
    '300A,0312' => { VR => 'IS', Name => 'NumberOfRangeShifters' },
    '300A,0314' => { VR => 'SQ', Name => 'RangeShifterSequence' },
    '300A,0316' => { VR => 'IS', Name => 'RangeShifterNumber' },
    '300A,0318' => { VR => 'SH', Name => 'RangeShifterID' },
    '300A,0320' => { VR => 'CS', Name => 'RangeShifterType' },
    '300A,0322' => { VR => 'LO', Name => 'RangeShifterDescription' },
    '300A,0330' => { VR => 'IS', Name => 'NumberOfLateralSpreadingDevices' },
    '300A,0332' => { VR => 'SQ', Name => 'LateralSpreadingDeviceSequence' },
    '300A,0334' => { VR => 'IS', Name => 'LateralSpreadingDeviceNumber' },
    '300A,0336' => { VR => 'SH', Name => 'LateralSpreadingDeviceID' },
    '300A,0338' => { VR => 'CS', Name => 'LateralSpreadingDeviceType' },
    '300A,033A' => { VR => 'LO', Name => 'LateralSpreadingDeviceDescription' },
    '300A,033C' => { VR => 'FL', Name => 'LateralSpreadingDevWaterEquivThick' },
    '300A,0340' => { VR => 'IS', Name => 'NumberOfRangeModulators' },
    '300A,0342' => { VR => 'SQ', Name => 'RangeModulatorSequence' },
    '300A,0344' => { VR => 'IS', Name => 'RangeModulatorNumber' },
    '300A,0346' => { VR => 'SH', Name => 'RangeModulatorID' },
    '300A,0348' => { VR => 'CS', Name => 'RangeModulatorType' },
    '300A,034A' => { VR => 'LO', Name => 'RangeModulatorDescription' },
    '300A,034C' => { VR => 'SH', Name => 'BeamCurrentModulationID' },
    '300A,0350' => { VR => 'CS', Name => 'PatientSupportType' },
    '300A,0352' => { VR => 'SH', Name => 'PatientSupportID' },
    '300A,0354' => { VR => 'LO', Name => 'PatientSupportAccessoryCode' },
    '300A,0355' => { VR => 'LO', Name => 'TrayAccessoryCode' },
    '300A,0356' => { VR => 'FL', Name => 'FixationLightAzimuthalAngle' },
    '300A,0358' => { VR => 'FL', Name => 'FixationLightPolarAngle' },
    '300A,035A' => { VR => 'FL', Name => 'MetersetRate' },
    '300A,0360' => { VR => 'SQ', Name => 'RangeShifterSettingsSequence' },
    '300A,0362' => { VR => 'LO', Name => 'RangeShifterSetting' },
    '300A,0364' => { VR => 'FL', Name => 'IsocenterToRangeShifterDistance' },
    '300A,0366' => { VR => 'FL', Name => 'RangeShifterWaterEquivThickness' },
    '300A,0370' => { VR => 'SQ', Name => 'LateralSpreadingDeviceSettingsSeq' },
    '300A,0372' => { VR => 'LO', Name => 'LateralSpreadingDeviceSetting' },
    '300A,0374' => { VR => 'FL', Name => 'IsocenterToLateralSpreadingDevDist' },
    '300A,0380' => { VR => 'SQ', Name => 'RangeModulatorSettingsSequence' },
    '300A,0382' => { VR => 'FL', Name => 'RangeModulatorGatingStartValue' },
    '300A,0384' => { VR => 'FL', Name => 'RangeModulatorGatingStopValue' },
    '300A,0386' => { VR => 'FL', Name => 'RangeModuGatingStartWaterEquiThick' },
    '300A,0388' => { VR => 'FL', Name => 'RangeModuGatingStopWaterEquivThick' },
    '300A,038A' => { VR => 'FL', Name => 'IsocenterToRangeModulatorDistance' },
    '300A,038F' => { VR => 'FL', Name => 'ScanSpotTimeOffset' },
    '300A,0390' => { VR => 'SH', Name => 'ScanSpotTuneID' },
    '300A,0391' => { VR => 'IS', Name => 'ScanSpotPrescribedIndices' },
    '300A,0392' => { VR => 'IS', Name => 'NumberOfScanSpotPositions' },
    '300A,0393' => { VR => 'CS', Name => 'ScanSpotReordered' },
    '300A,0394' => { VR => 'FL', Name => 'ScanSpotPositionMap' },
    '300A,0395' => { VR => 'CS', Name => 'ScanSpotReorderingAllowed' },
    '300A,0396' => { VR => 'FL', Name => 'ScanSpotMetersetWeights' },
    '300A,0398' => { VR => 'FL', Name => 'ScanningSpotSize' },
    '300A,0399' => { VR => 'FL', Name => 'ScanSpotSizesDelivered' },
    '300A,039A' => { VR => 'IS', Name => 'NumberOfPaintings' },
    '300A,039B' => { VR => 'FL', Name => 'ScanSpotGantryAngles' },
    '300A,039C' => { VR => 'FL', Name => 'ScanSpotPatientSupportAngles' },
    '300A,03A0' => { VR => 'SQ', Name => 'IonToleranceTableSequence' },
    '300A,03A2' => { VR => 'SQ', Name => 'IonBeamSequence' },
    '300A,03A4' => { VR => 'SQ', Name => 'IonBeamLimitingDeviceSequence' },
    '300A,03A6' => { VR => 'SQ', Name => 'IonBlockSequence' },
    '300A,03A8' => { VR => 'SQ', Name => 'IonControlPointSequence' },
    '300A,03AA' => { VR => 'SQ', Name => 'IonWedgeSequence' },
    '300A,03AC' => { VR => 'SQ', Name => 'IonWedgePositionSequence' },
    '300A,0401' => { VR => 'SQ', Name => 'ReferencedSetupImageSequence' },
    '300A,0402' => { VR => 'ST', Name => 'SetupImageComment' },
    '300A,0410' => { VR => 'SQ', Name => 'MotionSynchronizationSequence' },
    '300A,0412' => { VR => 'FL', Name => 'ControlPointOrientation' },
    '300A,0420' => { VR => 'SQ', Name => 'GeneralAccessorySequence' },
    '300A,0421' => { VR => 'SH', Name => 'GeneralAccessoryID' },
    '300A,0422' => { VR => 'ST', Name => 'GeneralAccessoryDescription' },
    '300A,0423' => { VR => 'CS', Name => 'GeneralAccessoryType' },
    '300A,0424' => { VR => 'IS', Name => 'GeneralAccessoryNumber' },
    '300A,0425' => { VR => 'FL', Name => 'SourceToGeneralAccessoryDistance' },
    '300A,0426' => { VR => 'DS', Name => 'IsocenterToGeneralAccessoryDist' },
    '300A,0431' => { VR => 'SQ', Name => 'ApplicatorGeometrySequence' },
    '300A,0432' => { VR => 'CS', Name => 'ApplicatorApertureShape' },
    '300A,0433' => { VR => 'FL', Name => 'ApplicatorOpening' },
    '300A,0434' => { VR => 'FL', Name => 'ApplicatorOpeningX' },
    '300A,0435' => { VR => 'FL', Name => 'ApplicatorOpeningY' },
    '300A,0436' => { VR => 'FL', Name => 'SourceToApplicatorMountingPosDist' },
    '300A,0440' => { VR => 'IS', Name => 'NumberOfBlockSlabItems' },
    '300A,0441' => { VR => 'SQ', Name => 'BlockSlabSequence' },
    '300A,0442' => { VR => 'DS', Name => 'BlockSlabThickness' },
    '300A,0443' => { VR => 'US', Name => 'BlockSlabNumber' },
    '300A,0450' => { VR => 'SQ', Name => 'DeviceMotionControlSequence' },
    '300A,0451' => { VR => 'CS', Name => 'DeviceMotionExecutionMode' },
    '300A,0452' => { VR => 'CS', Name => 'DeviceMotionObservationMode' },
    '300A,0453' => { VR => 'SQ', Name => 'DeviceMotionParameterCodeSequence' },
    '300A,0501' => { VR => 'FL', Name => 'DistalDepthFraction' },
    '300A,0502' => { VR => 'FL', Name => 'DistalDepth' },
    '300A,0503' => { VR => 'FL', Name => 'NominalRangeModulationFractions' },
    '300A,0504' => { VR => 'FL', Name => 'NominalRangeModulatedRegionDepths' },
    '300A,0505' => { VR => 'SQ', Name => 'DepthDoseParametersSequence' },
    '300A,0506' => { VR => 'SQ', Name => 'DeliveredDepthDoseParametersSeq' },
    '300A,0507' => { VR => 'FL', Name => 'DeliveredDistalDepthFraction' },
    '300A,0508' => { VR => 'FL', Name => 'DeliveredDistalDepth' },
    '300A,0509' => { VR => 'FL', Name => 'DeliveredNomRngModulationFractions' },
    '300A,0510' => { VR => 'FL', Name => 'DeliveredNomRngModulatedRgnDepths' },
    '300A,0511' => { VR => 'CS', Name => 'DeliveredReferenceDoseDefinition' },
    '300A,0512' => { VR => 'CS', Name => 'ReferenceDoseDefinition' },
    '300A,0600' => { VR => 'US', Name => 'RTControlPointIndex' },
    '300A,0601' => { VR => 'US', Name => 'RadiationGenerationModeIndex' },
    '300A,0602' => { VR => 'US', Name => 'ReferencedDefinedDeviceIndex' },
    '300A,0603' => { VR => 'US', Name => 'RadiationDoseIdentificationIndex' },
    '300A,0604' => { VR => 'US', Name => 'NumberOfRTControlPoints' },
    '300A,0605' => { VR => 'US', Name => 'RefRadiationGenerationModeIndex' },
    '300A,0606' => { VR => 'US', Name => 'TreatmentPositionIndex' },
    '300A,0607' => { VR => 'US', Name => 'ReferencedDeviceIndex' },
    '300A,0608' => { VR => 'LO', Name => 'TreatmentPositionGroupLabel' },
    '300A,0609' => { VR => 'UI', Name => 'TreatmentPositionGroupUID' },
    '300A,060A' => { VR => 'SQ', Name => 'TreatmentPositionGroupSequence' },
    '300A,060B' => { VR => 'US', Name => 'ReferencedTreatmentPositionIndex' },
    '300A,060C' => { VR => 'US', Name => 'ReferencedRadiationDoseIDIndex' },
    '300A,060D' => { VR => 'FD', Name => 'RTAccHolderWaterEquivThickness' },
    '300A,060E' => { VR => 'US', Name => 'ReferencedRTAccHolderDeviceIndex' },
    '300A,060F' => { VR => 'CS', Name => 'RTAccessoryHolderSlotExistenceFlag' },
    '300A,0610' => { VR => 'SQ', Name => 'RTAccessoryHolderSlotSequence' },
    '300A,0611' => { VR => 'LO', Name => 'RTAccessoryHolderSlotID' },
    '300A,0612' => { VR => 'FD', Name => 'RTAccessoryHolderSlotDistance' },
    '300A,0613' => { VR => 'FD', Name => 'RTAccessorySlotDistance' },
    '300A,0614' => { VR => 'SQ', Name => 'RTAccessoryHolderDefinitionSeq' },
    '300A,0615' => { VR => 'LO', Name => 'RTAccessoryDeviceSlotID' },
    '300A,0616' => { VR => 'SQ', Name => 'RTRadiationSequence' },
    '300A,0617' => { VR => 'SQ', Name => 'RadiationDoseSequence' },
    '300A,0618' => { VR => 'SQ', Name => 'RadiationDoseIdentificationSeq' },
    '300A,0619' => { VR => 'LO', Name => 'RadiationDoseIdentificationLabel' },
    '300A,061A' => { VR => 'CS', Name => 'ReferenceDoseType' },
    '300A,061B' => { VR => 'CS', Name => 'PrimaryDoseValueIndicator' },
    '300A,061C' => { VR => 'SQ', Name => 'DoseValuesSequence' },
    '300A,061D' => { VR => 'CS', Name => 'DoseValuePurpose' },
    '300A,061E' => { VR => 'FD', Name => 'ReferenceDosePointCoordinates' },
    '300A,061F' => { VR => 'SQ', Name => 'RadiationDoseValuesParametersSeq' },
    '300A,0620' => { VR => 'SQ', Name => 'MetersetToDoseMappingSequence' },
    '300A,0621' => { VR => 'SQ', Name => 'ExpectedInVivoMeasurementValuesSeq' },
    '300A,0622' => { VR => 'US', Name => 'ExpectedInVivoMeasurementValueIdx' },
    '300A,0623' => { VR => 'LO', Name => 'RadiationDoseInVivoMeasurementLbl' },
    '300A,0624' => { VR => 'FD', Name => 'RadiationDoseCentralAxisDispl' },
    '300A,0625' => { VR => 'FD', Name => 'RadiationDoseValue' },
    '300A,0626' => { VR => 'FD', Name => 'RadiationDoseSourceToSkinDistance' },
    '300A,0627' => { VR => 'FD', Name => 'RadiationDoseMeasPointCoordinates' },
    '300A,0628' => { VR => 'FD', Name => 'RadiationDoseSrcToExtContourDist' },
    '300A,0629' => { VR => 'SQ', Name => 'RTToleranceSetSequence' },
    '300A,062A' => { VR => 'LO', Name => 'RTToleranceSetLabel' },
    '300A,062B' => { VR => 'SQ', Name => 'AttributeToleranceValuesSequence' },
    '300A,062C' => { VR => 'FD', Name => 'ToleranceValue' },
    '300A,062D' => { VR => 'SQ', Name => 'PatientSupportPositionToleranceSeq' },
    '300A,062E' => { VR => 'FD', Name => 'TreatmentTimeLimit' },
    '300A,062F' => { VR => 'SQ', Name => 'CArmPhotonElectronControlPointSeq' },
    '300A,0630' => { VR => 'SQ', Name => 'ReferencedRTRadiationSequence' },
    '300A,0631' => { VR => 'SQ', Name => 'ReferencedRTInstanceSequence' },
    '300A,0632' => { VR => 'SQ', Name => 'ReferencedRTPatientSetupSequence' },
    '300A,0634' => { VR => 'FD', Name => 'SourceToPatientSurfaceDistance' },
    '300A,0635' => { VR => 'SQ', Name => 'TreatmentMachineSpecialModeCodeSeq' },
    '300A,0636' => { VR => 'US', Name => 'IntendedNumberOfFractions' },
    '300A,0637' => { VR => 'CS', Name => 'RTRadiationSetIntent' },
    '300A,0638' => { VR => 'CS', Name => 'RTRadPhysAndGeomContentDetailFlag' },
    '300A,0639' => { VR => 'CS', Name => 'RTRecordFlag' },
    '300A,063A' => { VR => 'SQ', Name => 'TreatmentDeviceIdentificationSeq' },
    '300A,063B' => { VR => 'SQ', Name => 'ReferencedRTPhysicianIntentSeq' },
    '300A,063C' => { VR => 'FD', Name => 'CumulativeMeterset' },
    '300A,063D' => { VR => 'FD', Name => 'DeliveryRate' },
    '300A,063E' => { VR => 'SQ', Name => 'DeliveryRateUnitSequence' },
    '300A,063F' => { VR => 'SQ', Name => 'TreatmentPositionSequence' },
    '300A,0640' => { VR => 'FD', Name => 'RadiationSourceAxisDistance' },
    '300A,0641' => { VR => 'US', Name => 'NumberOfRTBeamLimitingDevices' },
    '300A,0642' => { VR => 'FD', Name => 'RTBeamLimitingDeviceProximalDist' },
    '300A,0643' => { VR => 'FD', Name => 'RTBeamLimitingDeviceDistalDistance' },
    '300A,0644' => { VR => 'SQ', Name => 'ParallelRTBeamDelimDevOrientLabel' },
    '300A,0645' => { VR => 'FD', Name => 'BeamModifierOrientationAngle' },
    '300A,0646' => { VR => 'SQ', Name => 'FixedRTBeamDelimiterDeviceSequence' },
    '300A,0647' => { VR => 'SQ', Name => 'ParallelRTBeamDelimiterDeviceSeq' },
    '300A,0648' => { VR => 'US', Name => 'NumberOfParallelRTBeamDelimiters' },
    '300A,0649' => { VR => 'FD', Name => 'ParallelRTBeamDelimiterBoundaries' },
    '300A,064A' => { VR => 'FD', Name => 'ParallelRTBeamDelimiterPositions' },
    '300A,064B' => { VR => 'FD', Name => 'RTBeamLimitingDeviceOffset' },
    '300A,064C' => { VR => 'SQ', Name => 'RTBeamDelimiterGeometrySequence' },
    '300A,064D' => { VR => 'SQ', Name => 'RTBeamLimitingDeviceDefinitionSeq' },
    '300A,064E' => { VR => 'CS', Name => 'ParallelRTBeamDelimiterOpeningMode' },
    '300A,064F' => { VR => 'CS', Name => 'ParallelRTBeamDelimLeafMountSide' },
    '300A,0650' => { VR => 'UI', Name => 'PatientSetupUID' },
    '300A,0651' => { VR => 'SQ', Name => 'WedgeDefinitionSequence' },
    '300A,0652' => { VR => 'FD', Name => 'RadiationBeamWedgeAngle' },
    '300A,0653' => { VR => 'FD', Name => 'RadiationBeamWedgeThinEdgeDistance' },
    '300A,0654' => { VR => 'FD', Name => 'RadiationBeamEffectiveWedgeAngle' },
    '300A,0655' => { VR => 'US', Name => 'NumberOfWedgePositions' },
    '300A,0656' => { VR => 'SQ', Name => 'RTBeamLimitingDeviceOpeningSeq' },
    '300A,0657' => { VR => 'US', Name => 'NumberOfRTBeamLimitingDevOpenings' },
    '300A,0658' => { VR => 'SQ', Name => 'RadiationDosimeterUnitSequence' },
    '300A,0659' => { VR => 'SQ', Name => 'RTDeviceDistanceRefLocCodeSeq' },
    '300A,065A' => { VR => 'SQ', Name => 'RadiationDevConfigAndCommissKeySeq' },
    '300A,065B' => { VR => 'SQ', Name => 'PatientSupportPositionParameterSeq' },
    '300A,065C' => { VR => 'CS', Name => 'PatientSupportPositionSpecMethod' },
    '300A,065D' => { VR => 'SQ', Name => 'PatientSupportPositionDevParamSeq' },
    '300A,065E' => { VR => 'US', Name => 'DeviceOrderIndex' },
    '300A,065F' => { VR => 'US', Name => 'PatientSupportPosParamOrderIndex' },
    '300A,0660' => { VR => 'SQ', Name => 'PatientSupportPosDeviceTolSeq' },
    '300A,0661' => { VR => 'US', Name => 'PatientSupportPosTolOrderIndex' },
    '300A,0662' => { VR => 'SQ', Name => 'CompensatorDefinitionSequence' },
    '300A,0663' => { VR => 'CS', Name => 'CompensatorMapOrientation' },
    '300A,0664' => { VR => 'OF', Name => 'CompensatorProximalThicknessMap' },
    '300A,0665' => { VR => 'OF', Name => 'CompensatorDistalThicknessMap' },
    '300A,0666' => { VR => 'FD', Name => 'CompensatorBasePlaneOffset' },
    '300A,0667' => { VR => 'SQ', Name => 'CompensatorShapeFabricationCodeSeq' },
    '300A,0668' => { VR => 'SQ', Name => 'CompensatorShapeSequence' },
    '300A,0669' => { VR => 'FD', Name => 'RadiationBeamCompMillingToolDia' },
    '300A,066A' => { VR => 'SQ', Name => 'BlockDefinitionSequence' },
    '300A,066B' => { VR => 'OF', Name => 'BlockEdgeData' },
    '300A,066C' => { VR => 'CS', Name => 'BlockOrientation' },
    '300A,066D' => { VR => 'FD', Name => 'RadiationBeamBlockThickness' },
    '300A,066E' => { VR => 'FD', Name => 'RadiationBeamBlockSlabThickness' },
    '300A,066F' => { VR => 'SQ', Name => 'BlockEdgeDataSequence' },
    '300A,0670' => { VR => 'US', Name => 'NumberOfRTAccessoryHolders' },
    '300A,0671' => { VR => 'SQ', Name => 'GeneralAccessoryDefinitionSequence' },
    '300A,0672' => { VR => 'US', Name => 'NumberOfGeneralAccessories' },
    '300A,0673' => { VR => 'SQ', Name => 'BolusDefinitionSequence' },
    '300A,0674' => { VR => 'US', Name => 'NumberOfBoluses' },
    '300A,0675' => { VR => 'UI', Name => 'EquipmentFrameOfReferenceUID' },
    '300A,0676' => { VR => 'ST', Name => 'EquipmentFrameOfReferenceDescr' },
    '300A,0677' => { VR => 'SQ', Name => 'EquipmentReferencePointCoordsSeq' },
    '300A,0678' => { VR => 'SQ', Name => 'EquipmentReferencePointCodeSeq' },
    '300A,0679' => { VR => 'FD', Name => 'RTBeamLimitingDeviceAngle' },
    '300A,067A' => { VR => 'FD', Name => 'SourceRollAngle' },
    '300A,067B' => { VR => 'SQ', Name => 'RadiationGenerationModeSequence' },
    '300A,067C' => { VR => 'SH', Name => 'RadiationGenerationModeLabel' },
    '300A,067D' => { VR => 'ST', Name => 'RadiationGenerationModeDescription' },
    '300A,067E' => { VR => 'SQ', Name => 'RadiationGenModeMachineCodeSeq' },
    '300A,067F' => { VR => 'SQ', Name => 'RadiationTypeCodeSequence' },
    '300A,0680' => { VR => 'DS', Name => 'NominalEnergy' },
    '300A,0681' => { VR => 'DS', Name => 'MinimumNominalEnergy' },
    '300A,0682' => { VR => 'DS', Name => 'MaximumNominalEnergy' },
    '300A,0683' => { VR => 'SQ', Name => 'RadiationFluenceModifierCodeSeq' },
    '300A,0684' => { VR => 'SQ', Name => 'EnergyUnitCodeSequence' },
    '300A,0685' => { VR => 'US', Name => 'NumberOfRadiationGenerationModes' },
    '300A,0686' => { VR => 'SQ', Name => 'PatientSupportDevicesSequence' },
    '300A,0687' => { VR => 'US', Name => 'NumberOfPatientSupportDevices' },
    '300A,0688' => { VR => 'FD', Name => 'RTBeamModifierDefinitionDistance' },
    '300A,0689' => { VR => 'SQ', Name => 'BeamAreaLimitSequence' },
    '300A,068A' => { VR => 'SQ', Name => 'ReferencedRTPrescriptionSequence' },
    '300A,068B' => { VR => 'CS', Name => 'DoseValueInterpretation' },
    '300A,0700' => { VR => 'UI', Name => 'TreatmentSessionUID' },
    '300A,0701' => { VR => 'CS', Name => 'RTRadiationUsage' },
    '300A,0702' => { VR => 'SQ', Name => 'ReferencedRTRadiationSetSequence' },
    '300A,0703' => { VR => 'SQ', Name => 'ReferencedRTRadiationRecordSeq' },
    '300A,0704' => { VR => 'US', Name => 'RTRadiationSetDeliveryNumber' },
    '300A,0705' => { VR => 'US', Name => 'ClinicalFractionNumber' },
    '300A,0706' => { VR => 'CS', Name => 'RTTreatmentFractionCompletionStat' },
    '300A,0707' => { VR => 'CS', Name => 'RTRadiationSetUsage' },
    '300A,0708' => { VR => 'CS', Name => 'TreatmentDeliveryContinuationFlag' },
    '300A,0709' => { VR => 'CS', Name => 'TreatmentRecordContentOrigin' },
    '300A,0714' => { VR => 'CS', Name => 'RTTreatmentTerminationStatus' },
    '300A,0715' => { VR => 'SQ', Name => 'RTTreatmentTermReasonCodeSeq' },
    '300A,0716' => { VR => 'SQ', Name => 'MachineSpecTreatmentTermCodeSeq' },
    '300A,0722' => { VR => 'SQ', Name => 'RTRadiationSalvageRecordCtrlPtSeq' },
    '300A,0723' => { VR => 'CS', Name => 'StartingMetersetValueKnownFlag' },
    '300A,0730' => { VR => 'ST', Name => 'TreatmentTerminationDescription' },
    '300A,0731' => { VR => 'SQ', Name => 'TreatmentToleranceViolationSeq' },
    '300A,0732' => { VR => 'CS', Name => 'TreatmentTolViolationCategory' },
    '300A,0733' => { VR => 'SQ', Name => 'TreatmentTolViolationAttributeSeq' },
    '300A,0734' => { VR => 'ST', Name => 'TreatmentTolViolationDescription' },
    '300A,0735' => { VR => 'ST', Name => 'TreatmentTolViolationID' },
    '300A,0736' => { VR => 'DT', Name => 'TreatmentTolViolationDateTime' },
    '300A,073A' => { VR => 'DT', Name => 'RecordedRTControlPointDateTime' },
    '300A,073B' => { VR => 'US', Name => 'ReferencedRadiationRTCtrlPtIndex' },
    '300A,073E' => { VR => 'SQ', Name => 'AlternateValueSequence' },
    '300A,073F' => { VR => 'SQ', Name => 'ConfirmationSequence' },
    '300A,0740' => { VR => 'SQ', Name => 'InterlockSequence' },
    '300A,0741' => { VR => 'DT', Name => 'InterlockDateTime' },
    '300A,0742' => { VR => 'ST', Name => 'InterlockDescription' },
    '300A,0743' => { VR => 'SQ', Name => 'InterlockOriginatingDeviceSequence' },
    '300A,0744' => { VR => 'SQ', Name => 'InterlockCodeSequence' },
    '300A,0745' => { VR => 'SQ', Name => 'InterlockResolutionCodeSequence' },
    '300A,0746' => { VR => 'SQ', Name => 'InterlockResolutionUserSequence' },
    '300A,0760' => { VR => 'DT', Name => 'OverrideDateTime' },
    '300A,0761' => { VR => 'SQ', Name => 'TreatmentTolViolationTypeCodeSeq' },
    '300A,0762' => { VR => 'SQ', Name => 'TreatmentTolViolationCauseCodeSeq' },
    '300A,0772' => { VR => 'SQ', Name => 'MeasuredMetersetToDoseMappingSeq' },
    '300A,0773' => { VR => 'US', Name => 'RefExpectedInVivoMeasValueIndex' },
    '300A,0774' => { VR => 'SQ', Name => 'DoseMeasurementDeviceCodeSequence' },
    '300A,0780' => { VR => 'SQ', Name => 'AdditionalParamRecordInstanceSeq' },
    '300A,0783' => { VR => 'ST', Name => 'InterlockOriginDescription' },
    '300A,0784' => { VR => 'SQ', Name => 'RTPatientPositionScopeSequence' },
    '300A,0785' => { VR => 'UI', Name => 'ReferencedTreatmentPosGroupUID' },
    '300A,0786' => { VR => 'US', Name => 'RadiationOrderIndex' },
    '300A,0787' => { VR => 'SQ', Name => 'OmittedRadiationSequence' },
    '300A,0788' => { VR => 'SQ', Name => 'ReasonForOmissionCodeSequence' },
    '300A,0789' => { VR => 'SQ', Name => 'RTDeliveryStartPatientPositionSeq' },
    '300A,078A' => { VR => 'SQ', Name => 'RTTreatmentPrepPatientPosSeq' },
    '300A,078B' => { VR => 'SQ', Name => 'ReferencedRTTreatmentPrepSeq' },
    '300A,078C' => { VR => 'SQ', Name => 'ReferencedPatientSetupPhotoSeq' },
    '300A,078D' => { VR => 'SQ', Name => 'PatientTreatmentPrepMethodCodeSeq' },
    '300A,078E' => { VR => 'LT', Name => 'PatientTreatmentPrepProcParamDesc' },
    '300A,078F' => { VR => 'SQ', Name => 'PatientTreatmentPrepDeviceSeq' },
    '300A,0790' => { VR => 'SQ', Name => 'PatientTreatmentPrepProcedureSeq' },
    '300A,0791' => { VR => 'SQ', Name => 'PatientTreatmentPrepProcCodeSeq' },
    '300A,0792' => { VR => 'LT', Name => 'PatientTreatmentPrepMethodDesc' },
    '300A,0793' => { VR => 'SQ', Name => 'PatientTreatmentPrepProcParamSeq' },
    '300A,0794' => { VR => 'LT', Name => 'PatientSetupPhotoDescription' },
    '300A,0795' => { VR => 'US', Name => 'PatientTreatmentPrepProcIndex' },
    '300A,0796' => { VR => 'US', Name => 'ReferencedPatientSetupProcedureIdx' },
    '300A,0797' => { VR => 'SQ', Name => 'RTRadiationTaskSequence' },
    '300A,0798' => { VR => 'SQ', Name => 'RTPatientPositionDisplacementSeq' },
    '300A,0799' => { VR => 'SQ', Name => 'RTPatientPositionSequence' },
    '300A,079A' => { VR => 'LO', Name => 'DisplacementReferenceLabel' },
    '300A,079B' => { VR => 'FD', Name => 'DisplacementMatrix' },
    '300A,079C' => { VR => 'SQ', Name => 'PatientSupportDisplacementSequence' },
    '300A,079D' => { VR => 'SQ', Name => 'DisplacementReferenceLocCodeSeq' },
    '300A,079E' => { VR => 'CS', Name => 'RTRadiationSetDeliveryUsage' },
    '300A,079F' => { VR => 'SQ', Name => 'PatientTreatmentPreparationSeq' },
    '300A,07A0' => { VR => 'SQ', Name => 'PatientToEquipmentRelationshipSeq' },
    '300A,07A1' => { VR => 'SQ', Name => 'ImagingEquipToTreatDelivDevRelSeq' },
    '300C,0002' => { VR => 'SQ', Name => 'ReferencedRTPlanSequence' },
    '300C,0004' => { VR => 'SQ', Name => 'ReferencedBeamSequence' },
    '300C,0006' => { VR => 'IS', Name => 'ReferencedBeamNumber' },
    '300C,0007' => { VR => 'IS', Name => 'ReferencedReferenceImageNumber' },
    '300C,0008' => { VR => 'DS', Name => 'StartCumulativeMetersetWeight' },
    '300C,0009' => { VR => 'DS', Name => 'EndCumulativeMetersetWeight' },
    '300C,000A' => { VR => 'SQ', Name => 'ReferencedBrachyAppSetupSeq' },
    '300C,000C' => { VR => 'IS', Name => 'ReferencedBrachyAppSetupNumber' },
    '300C,000E' => { VR => 'IS', Name => 'ReferencedSourceNumber' },
    '300C,0020' => { VR => 'SQ', Name => 'ReferencedFractionGroupSequence' },
    '300C,0022' => { VR => 'IS', Name => 'ReferencedFractionGroupNumber' },
    '300C,0040' => { VR => 'SQ', Name => 'ReferencedVerificationImageSeq' },
    '300C,0042' => { VR => 'SQ', Name => 'ReferencedReferenceImageSequence' },
    '300C,0050' => { VR => 'SQ', Name => 'ReferencedDoseReferenceSequence' },
    '300C,0051' => { VR => 'IS', Name => 'ReferencedDoseReferenceNumber' },
    '300C,0055' => { VR => 'SQ', Name => 'BrachyReferencedDoseReferenceSeq' },
    '300C,0060' => { VR => 'SQ', Name => 'ReferencedStructureSetSequence' },
    '300C,006A' => { VR => 'IS', Name => 'ReferencedPatientSetupNumber' },
    '300C,0080' => { VR => 'SQ', Name => 'ReferencedDoseSequence' },
    '300C,00A0' => { VR => 'IS', Name => 'ReferencedToleranceTableNumber' },
    '300C,00B0' => { VR => 'SQ', Name => 'ReferencedBolusSequence' },
    '300C,00C0' => { VR => 'IS', Name => 'ReferencedWedgeNumber' },
    '300C,00D0' => { VR => 'IS', Name => 'ReferencedCompensatorNumber' },
    '300C,00E0' => { VR => 'IS', Name => 'ReferencedBlockNumber' },
    '300C,00F0' => { VR => 'IS', Name => 'ReferencedControlPointIndex' },
    '300C,00F2' => { VR => 'SQ', Name => 'ReferencedControlPointSequence' },
    '300C,00F4' => { VR => 'IS', Name => 'ReferencedStartControlPointIndex' },
    '300C,00F6' => { VR => 'IS', Name => 'ReferencedStopControlPointIndex' },
    '300C,0100' => { VR => 'IS', Name => 'ReferencedRangeShifterNumber' },
    '300C,0102' => { VR => 'IS', Name => 'ReferencedLateralSpreadingDevNum' },
    '300C,0104' => { VR => 'IS', Name => 'ReferencedRangeModulatorNumber' },
    '300C,0111' => { VR => 'SQ', Name => 'OmittedBeamTaskSequence' },
    '300C,0112' => { VR => 'CS', Name => 'ReasonForOmission' },
    '300C,0113' => { VR => 'LO', Name => 'ReasonForOmissionDescription' },
    '300C,0114' => { VR => 'SQ', Name => 'PrescriptionOverviewSequence' },
    '300C,0115' => { VR => 'FL', Name => 'TotalPrescriptionDose' },
    '300C,0116' => { VR => 'SQ', Name => 'PlanOverviewSequence' },
    '300C,0117' => { VR => 'US', Name => 'PlanOverviewIndex' },
    '300C,0118' => { VR => 'US', Name => 'ReferencedPlanOverviewIndex' },
    '300C,0119' => { VR => 'US', Name => 'NumberOfFractionsIncluded' },
    '300C,0120' => { VR => 'SQ', Name => 'DoseCalibrationConditionsSequence' },
    '300C,0121' => { VR => 'FD', Name => 'AbsorbedDoseToMetersetRatio' },
    '300C,0122' => { VR => 'FD', Name => 'DelineatedRadiationFieldSize' },
    '300C,0123' => { VR => 'CS', Name => 'DoseCalConditionsVerifiedFlag' },
    '300C,0124' => { VR => 'FD', Name => 'CalibrationReferencePointDepth' },
    '300C,0125' => { VR => 'SQ', Name => 'GatingBeamHoldTransitionSequence' },
    '300C,0126' => { VR => 'CS', Name => 'BeamHoldTransition' },
    '300C,0127' => { VR => 'DT', Name => 'BeamHoldTransitionDateTime' },
    '300C,0128' => { VR => 'SQ', Name => 'BeamHoldOriginatingDeviceSequence' },
    '300C,0129' => { VR => 'CS', Name => 'BeamHoldTransitionTriggerSource' },
    '300E,0002' => { VR => 'CS', Name => 'ApprovalStatus' },
    '300E,0004' => { VR => 'DA', Name => 'ReviewDate' },
    '300E,0005' => { VR => 'TM', Name => 'ReviewTime' },
    '300E,0008' => { VR => 'PN', Name => 'ReviewerName' },
    '3010,0001' => { VR => 'SQ', Name => 'RadiobiologicalDoseEffectSequence' },
    '3010,0002' => { VR => 'CS', Name => 'RadiobiologicalDoseEffectFlag' },
    '3010,0003' => { VR => 'SQ', Name => 'EffDoseCalcMethodCategoryCodeSeq' },
    '3010,0004' => { VR => 'SQ', Name => 'EffectiveDoseCalcMethodCodeSeq' },
    '3010,0005' => { VR => 'LO', Name => 'EffectiveDoseCalcMethodDescription' },
    '3010,0006' => { VR => 'UI', Name => 'ConceptualVolumeUID' },
    '3010,0007' => { VR => 'SQ', Name => 'OriginatingSOPInstanceReferenceSeq' },
    '3010,0008' => { VR => 'SQ', Name => 'ConceptualVolumeConstituentSeq' },
    '3010,0009' => { VR => 'SQ', Name => 'EquivConceptVolumeInstanceRefSeq' },
    '3010,000A' => { VR => 'SQ', Name => 'EquivalentConceptualVolumesSeq' },
    '3010,000B' => { VR => 'UI', Name => 'ReferencedConceptualVolumeUID' },
    '3010,000C' => { VR => 'UT', Name => 'ConceptualVolumeCombinationExpr' },
    '3010,000D' => { VR => 'US', Name => 'ConceptualVolumeConstituentIndex' },
    '3010,000E' => { VR => 'CS', Name => 'ConceptualVolumeCombinationFlag' },
    '3010,000F' => { VR => 'ST', Name => 'ConceptualVolumeCombinationDescr' },
    '3010,0010' => { VR => 'CS', Name => 'ConceptualVolumeSegmentDefinedFlag' },
    '3010,0011' => { VR => 'SQ', Name => 'ConceptualVolumeSegmentRefSeq' },
    '3010,0012' => { VR => 'SQ', Name => 'ConceptVolumeConstituentSegRefSeq' },
    '3010,0013' => { VR => 'UI', Name => 'ConstituentConceptualVolumeUID' },
    '3010,0014' => { VR => 'SQ', Name => 'DerivationConceptualVolumeSequence' },
    '3010,0015' => { VR => 'UI', Name => 'SourceConceptualVolumeUID' },
    '3010,0016' => { VR => 'SQ', Name => 'ConceptualVolumeDerivAlgorithmSeq' },
    '3010,0017' => { VR => 'ST', Name => 'ConceptualVolumeDescription' },
    '3010,0018' => { VR => 'SQ', Name => 'SourceConceptualVolumeSequence' },
    '3010,0019' => { VR => 'SQ', Name => 'AuthorIdentificationSequence' },
    '3010,001A' => { VR => 'LO', Name => 'ManufacturerModelVersion' },
    '3010,001B' => { VR => 'UC', Name => 'DeviceAlternateIdentifier' },
    '3010,001C' => { VR => 'CS', Name => 'DeviceAlternateIdentifierType' },
    '3010,001D' => { VR => 'LT', Name => 'DeviceAlternateIdentifierFormat' },
    '3010,001E' => { VR => 'LO', Name => 'SegmentationCreationTemplateLabel' },
    '3010,001F' => { VR => 'UI', Name => 'SegmentationTemplateUID' },
    '3010,0020' => { VR => 'US', Name => 'ReferencedSegmentReferenceIndex' },
    '3010,0021' => { VR => 'SQ', Name => 'SegmentReferenceSequence' },
    '3010,0022' => { VR => 'US', Name => 'SegmentReferenceIndex' },
    '3010,0023' => { VR => 'SQ', Name => 'DirectSegmentReferenceSequence' },
    '3010,0024' => { VR => 'SQ', Name => 'CombinationSegmentReferenceSeq' },
    '3010,0025' => { VR => 'SQ', Name => 'ConceptualVolumeSequence' },
    '3010,0026' => { VR => 'SQ', Name => 'SegmentedRTAccessoryDeviceSequence' },
    '3010,0027' => { VR => 'SQ', Name => 'SegmentCharacteristicsSequence' },
    '3010,0028' => { VR => 'SQ', Name => 'RelatedSegmentCharacteristicsSeq' },
    '3010,0029' => { VR => 'US', Name => 'SegmentCharacteristicsPrecedence' },
    '3010,002A' => { VR => 'SQ', Name => 'RTSegmentAnnotationSequence' },
    '3010,002B' => { VR => 'SQ', Name => 'SegmentAnnotationCategoryCodeSeq' },
    '3010,002C' => { VR => 'SQ', Name => 'SegmentAnnotationTypeCodeSequence' },
    '3010,002D' => { VR => 'LO', Name => 'DeviceLabel' },
    '3010,002E' => { VR => 'SQ', Name => 'DeviceTypeCodeSequence' },
    '3010,002F' => { VR => 'SQ', Name => 'SegmentAnnotTypeModifierCodeSeq' },
    '3010,0030' => { VR => 'SQ', Name => 'PatientEquipRelationshipCodeSeq' },
    '3010,0031' => { VR => 'UI', Name => 'ReferencedFiducialsUID' },
    '3010,0032' => { VR => 'SQ', Name => 'PatientTreatmentOrientationSeq' },
    '3010,0033' => { VR => 'SH', Name => 'UserContentLabel' },
    '3010,0034' => { VR => 'LO', Name => 'UserContentLongLabel' },
    '3010,0035' => { VR => 'SH', Name => 'EntityLabel' },
    '3010,0036' => { VR => 'LO', Name => 'EntityName' },
    '3010,0037' => { VR => 'ST', Name => 'EntityDescription' },
    '3010,0038' => { VR => 'LO', Name => 'EntityLongLabel' },
    '3010,0039' => { VR => 'US', Name => 'DeviceIndex' },
    '3010,003A' => { VR => 'US', Name => 'RTTreatmentPhaseIndex' },
    '3010,003B' => { VR => 'UI', Name => 'RTTreatmentPhaseUID' },
    '3010,003C' => { VR => 'US', Name => 'RTPrescriptionIndex' },
    '3010,003D' => { VR => 'US', Name => 'RTSegmentAnnotationIndex' },
    '3010,003E' => { VR => 'US', Name => 'BasisRTTreatmentPhaseIndex' },
    '3010,003F' => { VR => 'US', Name => 'RelatedRTTreatmentPhaseIndex' },
    '3010,0040' => { VR => 'US', Name => 'ReferencedRTTreatmentPhaseIndex' },
    '3010,0041' => { VR => 'US', Name => 'ReferencedRTPrescriptionIndex' },
    '3010,0042' => { VR => 'US', Name => 'ReferencedParentRTPrescriptionIdx' },
    '3010,0043' => { VR => 'ST', Name => 'ManufacturerDeviceIdentifier' },
    '3010,0044' => { VR => 'SQ', Name => 'InstanceLvlRefPerformedProcStepSeq' },
    '3010,0045' => { VR => 'CS', Name => 'RTTreatmentPhaseIntentPresenceFlag' },
    '3010,0046' => { VR => 'CS', Name => 'RadiotherapyTreatmentType' },
    '3010,0047' => { VR => 'CS', Name => 'TeletherapyRadiationType' },
    '3010,0048' => { VR => 'CS', Name => 'BrachytherapySourceType' },
    '3010,0049' => { VR => 'SQ', Name => 'ReferencedRTTreatmentPhaseSequence' },
    '3010,004A' => { VR => 'SQ', Name => 'ReferencedDirectSegmentInstanceSeq' },
    '3010,004B' => { VR => 'SQ', Name => 'IntendedRTTreatmentPhaseSequence' },
    '3010,004C' => { VR => 'DA', Name => 'IntendedPhaseStartDate' },
    '3010,004D' => { VR => 'DA', Name => 'IntendedPhaseEndDate' },
    '3010,004E' => { VR => 'SQ', Name => 'RTTreatmentPhaseIntervalSequence' },
    '3010,004F' => { VR => 'CS', Name => 'TemporalRelationshipIntervalAnchor' },
    '3010,0050' => { VR => 'FD', Name => 'MinimumNumberOfIntervalDays' },
    '3010,0051' => { VR => 'FD', Name => 'MaximumNumberOfIntervalDays' },
    '3010,0052' => { VR => 'UI', Name => 'PertinentSOPClassesInStudy' },
    '3010,0053' => { VR => 'UI', Name => 'PertinentSOPClassesInSeries' },
    '3010,0054' => { VR => 'LO', Name => 'RTPrescriptionLabel' },
    '3010,0055' => { VR => 'SQ', Name => 'RTPhysicianIntentPredecessorSeq' },
    '3010,0056' => { VR => 'LO', Name => 'RTTreatmentApproachLabel' },
    '3010,0057' => { VR => 'SQ', Name => 'RTPhysicianIntentSequence' },
    '3010,0058' => { VR => 'US', Name => 'RTPhysicianIntentIndex' },
    '3010,0059' => { VR => 'CS', Name => 'RTTreatmentIntentType' },
    '3010,005A' => { VR => 'UT', Name => 'RTPhysicianIntentNarrative' },
    '3010,005B' => { VR => 'SQ', Name => 'RTProtocolCodeSequence' },
    '3010,005C' => { VR => 'ST', Name => 'ReasonForSuperseding' },
    '3010,005D' => { VR => 'SQ', Name => 'RTDiagnosisCodeSequence' },
    '3010,005E' => { VR => 'US', Name => 'ReferencedRTPhysicianIntentIndex' },
    '3010,005F' => { VR => 'SQ', Name => 'RTPhysicianIntentInputInstanceSeq' },
    '3010,0060' => { VR => 'SQ', Name => 'RTAnatomicPrescriptionSequence' },
    '3010,0061' => { VR => 'UT', Name => 'PriorTreatmentDoseDescription' },
    '3010,0062' => { VR => 'SQ', Name => 'PriorTreatmentReferenceSequence' },
    '3010,0063' => { VR => 'CS', Name => 'DosimetricObjectiveEvaluationScope' },
    '3010,0064' => { VR => 'SQ', Name => 'TherapeuticRoleCategoryCodeSeq' },
    '3010,0065' => { VR => 'SQ', Name => 'TherapeuticRoleTypeCodeSequence' },
    '3010,0066' => { VR => 'US', Name => 'ConceptualVolumeOptPrecedence' },
    '3010,0067' => { VR => 'SQ', Name => 'ConceptualVolumeCategoryCodeSeq' },
    '3010,0068' => { VR => 'CS', Name => 'ConceptualVolumeBlockingConstraint' },
    '3010,0069' => { VR => 'SQ', Name => 'ConceptualVolumeTypeCodeSequence' },
    '3010,006A' => { VR => 'SQ', Name => 'ConceptualVolumeTypeModCodeSeq' },
    '3010,006B' => { VR => 'SQ', Name => 'RTPrescriptionSequence' },
    '3010,006C' => { VR => 'SQ', Name => 'DosimetricObjectiveSequence' },
    '3010,006D' => { VR => 'SQ', Name => 'DosimetricObjectiveTypeCodeSeq' },
    '3010,006E' => { VR => 'UI', Name => 'DosimetricObjectiveUID' },
    '3010,006F' => { VR => 'UI', Name => 'ReferencedDosimetricObjectiveUID' },
    '3010,0070' => { VR => 'SQ', Name => 'DosimetricObjectiveParameterSeq' },
    '3010,0071' => { VR => 'SQ', Name => 'ReferencedDosimetricObjectivesSeq' },
    '3010,0073' => { VR => 'CS', Name => 'AbsoluteDosimetricObjectiveFlag' },
    '3010,0074' => { VR => 'FD', Name => 'DosimetricObjectiveWeight' },
    '3010,0075' => { VR => 'CS', Name => 'DosimetricObjectivePurpose' },
    '3010,0076' => { VR => 'SQ', Name => 'PlanningInputInformationSequence' },
    '3010,0077' => { VR => 'LO', Name => 'TreatmentSite' },
    '3010,0078' => { VR => 'SQ', Name => 'TreatmentSiteCodeSequence' },
    '3010,0079' => { VR => 'SQ', Name => 'FractionPatternSequence' },
    '3010,007A' => { VR => 'UT', Name => 'TreatmentTechniqueNotes' },
    '3010,007B' => { VR => 'UT', Name => 'PrescriptionNotes' },
    '3010,007C' => { VR => 'IS', Name => 'NumberOfIntervalFractions' },
    '3010,007D' => { VR => 'US', Name => 'NumberOfFractions' },
    '3010,007E' => { VR => 'US', Name => 'IntendedDeliveryDuration' },
    '3010,007F' => { VR => 'UT', Name => 'FractionationNotes' },
    '3010,0080' => { VR => 'SQ', Name => 'RTTreatmentTechniqueCodeSequence' },
    '3010,0081' => { VR => 'SQ', Name => 'PrescriptionNotesSequence' },
    '3010,0082' => { VR => 'SQ', Name => 'FractionBasedRelationshipSequence' },
    '3010,0083' => { VR => 'CS', Name => 'FracBasedRelationIntervalAnchor' },
    '3010,0084' => { VR => 'FD', Name => 'MinimumHoursBetweenFractions' },
    '3010,0085' => { VR => 'TM', Name => 'IntendedFractionStartTime' },
    '3010,0086' => { VR => 'LT', Name => 'IntendedStartDayOfWeek' },
    '3010,0087' => { VR => 'SQ', Name => 'WeekdayFractionPatternSequence' },
    '3010,0088' => { VR => 'SQ', Name => 'DeliveryTimeStructureCodeSequence' },
    '3010,0089' => { VR => 'SQ', Name => 'TreatmentSiteModifierCodeSequence' },
    '3010,0090' => { VR => 'CS', Name => 'RoboticBaseLocationIndicator' },
    '3010,0091' => { VR => 'SQ', Name => 'RoboticPathNodeSetCodeSequence' },
    '3010,0092' => { VR => 'UL', Name => 'RoboticNodeIdentifier' },
    '3010,0093' => { VR => 'FD', Name => 'RTTreatmentSourceCoordinates' },
    '3010,0094' => { VR => 'FD', Name => 'RadiationSourceCoordYawAngle' },
    '3010,0095' => { VR => 'FD', Name => 'RadiationSourceCoordRollAngle' },
    '3010,0096' => { VR => 'FD', Name => 'RadiationSourceCoordPitchAngle' },
    '3010,0097' => { VR => 'SQ', Name => 'RoboticPathControlPointSequence' },
    '3010,0098' => { VR => 'SQ', Name => 'TomotherapeuticControlPointSeq' },
    '3010,0099' => { VR => 'FD', Name => 'TomotherapeuticLeafOpenDurations' },
    '3010,009A' => { VR => 'FD', Name => 'TomotherapeuticLeafInitClosedDur' },
    '3010,00A0' => { VR => 'SQ', Name => 'ConceptualVolumeIdentificationSeq' },
    # text group
    '4000,0000' => { VR => 'UL', Name => 'TextGroupLength' },
    '4000,0010' => { VR => 'LT', Name => 'Arbitrary' },
    '4000,4000' => { VR => 'LT', Name => 'TextComments' },
    # results group
    '4008,0040' => { VR => 'SH', Name => 'ResultsID' },
    '4008,0042' => { VR => 'LO', Name => 'ResultsIDIssuer' },
    '4008,0050' => { VR => 'SQ', Name => 'ReferencedInterpretationSequence' },
    '4008,00FF' => { VR => 'CS', Name => 'ReportProductionStatusTrial' },
    '4008,0100' => { VR => 'DA', Name => 'InterpretationRecordedDate' },
    '4008,0101' => { VR => 'TM', Name => 'InterpretationRecordedTime' },
    '4008,0102' => { VR => 'PN', Name => 'InterpretationRecorder' },
    '4008,0103' => { VR => 'LO', Name => 'ReferenceToRecordedSound' },
    '4008,0108' => { VR => 'DA', Name => 'InterpretationTranscriptionDate' },
    '4008,0109' => { VR => 'TM', Name => 'InterpretationTranscriptionTime' },
    '4008,010A' => { VR => 'PN', Name => 'InterpretationTranscriber' },
    '4008,010B' => { VR => 'ST', Name => 'InterpretationText' },
    '4008,010C' => { VR => 'PN', Name => 'InterpretationAuthor' },
    '4008,0111' => { VR => 'SQ', Name => 'InterpretationApproverSequence' },
    '4008,0112' => { VR => 'DA', Name => 'InterpretationApprovalDate' },
    '4008,0113' => { VR => 'TM', Name => 'InterpretationApprovalTime' },
    '4008,0114' => { VR => 'PN', Name => 'PhysicianApprovingInterpretation' },
    '4008,0115' => { VR => 'LT', Name => 'InterpretationDiagnosisDescription' },
    '4008,0117' => { VR => 'SQ', Name => 'InterpretationDiagnosisCodeSeq' },
    '4008,0118' => { VR => 'SQ', Name => 'ResultsDistributionListSequence' },
    '4008,0119' => { VR => 'PN', Name => 'DistributionName' },
    '4008,011A' => { VR => 'LO', Name => 'DistributionAddress' },
    '4008,0200' => { VR => 'SH', Name => 'InterpretationID' },
    '4008,0202' => { VR => 'LO', Name => 'InterpretationIDIssuer' },
    '4008,0210' => { VR => 'CS', Name => 'InterpretationTypeID' },
    '4008,0212' => { VR => 'CS', Name => 'InterpretationStatusID' },
    '4008,0300' => { VR => 'ST', Name => 'Impressions' },
    '4008,4000' => { VR => 'ST', Name => 'ResultsComments' },
    '4010,0001' => { VR => 'CS', Name => 'LowEnergyDetectors' },
    '4010,0002' => { VR => 'CS', Name => 'HighEnergyDetectors' },
    '4010,0004' => { VR => 'SQ', Name => 'DetectorGeometrySequence' },
    '4010,1001' => { VR => 'SQ', Name => 'ThreatROIVoxelSequence' },
    '4010,1004' => { VR => 'FL', Name => 'ThreatROIBase' },
    '4010,1005' => { VR => 'FL', Name => 'ThreatROIExtents' },
    '4010,1006' => { VR => 'OB', Name => 'ThreatROIBitmap' },
    '4010,1007' => { VR => 'SH', Name => 'RouteSegmentID' },
    '4010,1008' => { VR => 'CS', Name => 'GantryType' },
    '4010,1009' => { VR => 'CS', Name => 'OOIOwnerType' },
    '4010,100A' => { VR => 'SQ', Name => 'RouteSegmentSequence' },
    '4010,1010' => { VR => 'US', Name => 'PotentialThreatObjectID' },
    '4010,1011' => { VR => 'SQ', Name => 'ThreatSequence' },
    '4010,1012' => { VR => 'CS', Name => 'ThreatCategory' },
    '4010,1013' => { VR => 'LT', Name => 'ThreatCategoryDescription' },
    '4010,1014' => { VR => 'CS', Name => 'ATDAbilityAssessment' },
    '4010,1015' => { VR => 'CS', Name => 'ATDAssessmentFlag' },
    '4010,1016' => { VR => 'FL', Name => 'ATDAssessmentProbability' },
    '4010,1017' => { VR => 'FL', Name => 'Mass' },
    '4010,1018' => { VR => 'FL', Name => 'Density' },
    '4010,1019' => { VR => 'FL', Name => 'ZEffective' },
    '4010,101A' => { VR => 'SH', Name => 'BoardingPassID' },
    '4010,101B' => { VR => 'FL', Name => 'CenterOfMass' },
    '4010,101C' => { VR => 'FL', Name => 'CenterOfPTO' },
    '4010,101D' => { VR => 'FL', Name => 'BoundingPolygon' },
    '4010,101E' => { VR => 'SH', Name => 'RouteSegmentStartLocationID' },
    '4010,101F' => { VR => 'SH', Name => 'RouteSegmentEndLocationID' },
    '4010,1020' => { VR => 'CS', Name => 'RouteSegmentLocationIDType' },
    '4010,1021' => { VR => 'CS', Name => 'AbortReason' },
    '4010,1023' => { VR => 'FL', Name => 'VolumeOfPTO' },
    '4010,1024' => { VR => 'CS', Name => 'AbortFlag' },
    '4010,1025' => { VR => 'DT', Name => 'RouteSegmentStartTime' },
    '4010,1026' => { VR => 'DT', Name => 'RouteSegmentEndTime' },
    '4010,1027' => { VR => 'CS', Name => 'TDRType' },
    '4010,1028' => { VR => 'CS', Name => 'InternationalRouteSegment' },
    '4010,1029' => { VR => 'LO', Name => 'ThreatDetectionAlgorithmAndVersion' },
    '4010,102A' => { VR => 'SH', Name => 'AssignedLocation' },
    '4010,102B' => { VR => 'DT', Name => 'AlarmDecisionTime' },
    '4010,1031' => { VR => 'CS', Name => 'AlarmDecision' },
    '4010,1033' => { VR => 'US', Name => 'NumberOfTotalObjects' },
    '4010,1034' => { VR => 'US', Name => 'NumberOfAlarmObjects' },
    '4010,1037' => { VR => 'SQ', Name => 'PTORepresentationSequence' },
    '4010,1038' => { VR => 'SQ', Name => 'ATDAssessmentSequence' },
    '4010,1039' => { VR => 'CS', Name => 'TIPType' },
    '4010,103A' => { VR => 'CS', Name => 'DICOSVersion' },
    '4010,1041' => { VR => 'DT', Name => 'OOIOwnerCreationTime' },
    '4010,1042' => { VR => 'CS', Name => 'OOIType' },
    '4010,1043' => { VR => 'FL', Name => 'OOISize' },
    '4010,1044' => { VR => 'CS', Name => 'AcquisitionStatus' },
    '4010,1045' => { VR => 'SQ', Name => 'BasisMaterialsCodeSequence' },
    '4010,1046' => { VR => 'CS', Name => 'PhantomType' },
    '4010,1047' => { VR => 'SQ', Name => 'OOIOwnerSequence' },
    '4010,1048' => { VR => 'CS', Name => 'ScanType' },
    '4010,1051' => { VR => 'LO', Name => 'ItineraryID' },
    '4010,1052' => { VR => 'SH', Name => 'ItineraryIDType' },
    '4010,1053' => { VR => 'LO', Name => 'ItineraryIDAssigningAuthority' },
    '4010,1054' => { VR => 'SH', Name => 'RouteID' },
    '4010,1055' => { VR => 'SH', Name => 'RouteIDAssigningAuthority' },
    '4010,1056' => { VR => 'CS', Name => 'InboundArrivalType' },
    '4010,1058' => { VR => 'SH', Name => 'CarrierID' },
    '4010,1059' => { VR => 'CS', Name => 'CarrierIDAssigningAuthority' },
    '4010,1060' => { VR => 'FL', Name => 'SourceOrientation' },
    '4010,1061' => { VR => 'FL', Name => 'SourcePosition' },
    '4010,1062' => { VR => 'FL', Name => 'BeltHeight' },
    '4010,1064' => { VR => 'SQ', Name => 'AlgorithmRoutingCodeSequence' },
    '4010,1067' => { VR => 'CS', Name => 'TransportClassification' },
    '4010,1068' => { VR => 'LT', Name => 'OOITypeDescriptor' },
    '4010,1069' => { VR => 'FL', Name => 'TotalProcessingTime' },
    '4010,106C' => { VR => 'OB', Name => 'DetectorCalibrationData' },
    '4010,106D' => { VR => 'CS', Name => 'AdditionalScreeningPerformed' },
    '4010,106E' => { VR => 'CS', Name => 'AdditionalInspectionSelectCriteria' },
    '4010,106F' => { VR => 'SQ', Name => 'AdditionalInspectionMethodSequence' },
    '4010,1070' => { VR => 'CS', Name => 'AITDeviceType' },
    '4010,1071' => { VR => 'SQ', Name => 'QRMeasurementsSequence' },
    '4010,1072' => { VR => 'SQ', Name => 'TargetMaterialSequence' },
    '4010,1073' => { VR => 'FD', Name => 'SNRThreshold' },
    '4010,1075' => { VR => 'DS', Name => 'ImageScaleRepresentation' },
    '4010,1076' => { VR => 'SQ', Name => 'ReferencedPTOSequence' },
    '4010,1077' => { VR => 'SQ', Name => 'ReferencedTDRInstanceSequence' },
    '4010,1078' => { VR => 'ST', Name => 'PTOLocationDescription' },
    '4010,1079' => { VR => 'SQ', Name => 'AnomalyLocatorIndicatorSequence' },
    '4010,107A' => { VR => 'FL', Name => 'AnomalyLocatorIndicator' },
    '4010,107B' => { VR => 'SQ', Name => 'PTORegionSequence' },
    '4010,107C' => { VR => 'CS', Name => 'InspectionSelectionCriteria' },
    '4010,107D' => { VR => 'SQ', Name => 'SecondaryInspectionMethodSequence' },
    '4010,107E' => { VR => 'DS', Name => 'PRCSToRCSOrientation' },
    '4FFE,0001' => { VR => 'SQ', Name => 'MACParametersSequence' },
    # curve group
    '50xx,0005' => { VR => 'US', Name => 'CurveDimensions' },
    '50xx,0010' => { VR => 'US', Name => 'NumberOfPoints' },
    '50xx,0020' => { VR => 'CS', Name => 'TypeOfData' },
    '50xx,0022' => { VR => 'LO', Name => 'CurveDescription' },
    '50xx,0030' => { VR => 'SH', Name => 'AxisUnits' },
    '50xx,0040' => { VR => 'SH', Name => 'AxisLabels' },
    '50xx,0103' => { VR => 'US', Name => 'DataValueRepresentation' },
    '50xx,0104' => { VR => 'US', Name => 'MinimumCoordinateValue' },
    '50xx,0105' => { VR => 'US', Name => 'MaximumCoordinateValue' },
    '50xx,0106' => { VR => 'SH', Name => 'CurveRange' },
    '50xx,0110' => { VR => 'US', Name => 'CurveDataDescriptor' },
    '50xx,0112' => { VR => 'US', Name => 'CoordinateStartValue' },
    '50xx,0114' => { VR => 'US', Name => 'CoordinateStepValue' },
    '50xx,1001' => { VR => 'CS', Name => 'CurveActivationLayer' },
    '50xx,2000' => { VR => 'US', Name => 'AudioType' },
    '50xx,2002' => { VR => 'US', Name => 'AudioSampleFormat' },
    '50xx,2004' => { VR => 'US', Name => 'NumberOfChannels' },
    '50xx,2006' => { VR => 'UL', Name => 'NumberOfSamples' },
    '50xx,2008' => { VR => 'UL', Name => 'SampleRate' },
    '50xx,200A' => { VR => 'UL', Name => 'TotalTime' },
    '50xx,200C' => { VR => 'OW', Name => 'AudioSampleData' },
    '50xx,200E' => { VR => 'LT', Name => 'AudioComments' },
    '50xx,2500' => { VR => 'LO', Name => 'CurveLabel' },
    '50xx,2600' => { VR => 'SQ', Name => 'ReferencedOverlaySequence' },
    '50xx,2610' => { VR => 'US', Name => 'ReferencedOverlayGroup' },
    '50xx,3000' => { VR => 'OW', Name => 'CurveData' },
    '5200,9229' => { VR => 'SQ', Name => 'SharedFunctionalGroupsSequence' },
    '5200,9230' => { VR => 'SQ', Name => 'PerFrameFunctionalGroupsSequence' },
    '5400,0100' => { VR => 'SQ', Name => 'WaveformSequence' },
    '5400,0110' => { VR => 'OB', Name => 'ChannelMinimumValue' },
    '5400,0112' => { VR => 'OB', Name => 'ChannelMaximumValue' },
    '5400,1004' => { VR => 'US', Name => 'WaveformBitsAllocated' },
    '5400,1006' => { VR => 'CS', Name => 'WaveformSampleInterpretation' },
    '5400,100A' => { VR => 'OB', Name => 'WaveformPaddingValue' },
    '5400,1010' => { VR => 'OB', Name => 'WaveformData' },
    '5600,0010' => { VR => 'OF', Name => 'FirstOrderPhaseCorrectionAngle' },
    '5600,0020' => { VR => 'OF', Name => 'SpectroscopyData' },
    # overlay group
    '6000,0000' => { VR => 'UL', Name => 'OverlayGroupLength' },
    '60xx,0010' => { VR => 'US', Name => 'OverlayRows' },
    '60xx,0011' => { VR => 'US', Name => 'OverlayColumns' },
    '60xx,0012' => { VR => 'US', Name => 'OverlayPlanes' },
    '60xx,0015' => { VR => 'IS', Name => 'NumberOfFramesInOverlay' },
    '60xx,0022' => { VR => 'LO', Name => 'OverlayDescription' },
    '60xx,0040' => { VR => 'CS', Name => 'OverlayType' },
    '60xx,0045' => { VR => 'LO', Name => 'OverlaySubtype' },
    '60xx,0050' => { VR => 'SS', Name => 'OverlayOrigin' },
    '60xx,0051' => { VR => 'US', Name => 'ImageFrameOrigin' },
    '60xx,0052' => { VR => 'US', Name => 'OverlayPlaneOrigin' },
    '60xx,0060' => { VR => 'CS', Name => 'OverlayCompressionCode' },
    '60xx,0061' => { VR => 'SH', Name => 'OverlayCompressionOriginator' },
    '60xx,0062' => { VR => 'SH', Name => 'OverlayCompressionLabel' },
    '60xx,0063' => { VR => 'CS', Name => 'OverlayCompressionDescription' },
    '60xx,0066' => { VR => 'AT', Name => 'OverlayCompressionStepPointers' },
    '60xx,0068' => { VR => 'US', Name => 'OverlayRepeatInterval' },
    '60xx,0069' => { VR => 'US', Name => 'OverlayBitsGrouped' },
    '60xx,0100' => { VR => 'US', Name => 'OverlayBitsAllocated' },
    '60xx,0102' => { VR => 'US', Name => 'OverlayBitPosition' },
    '60xx,0110' => { VR => 'CS', Name => 'OverlayFormat' },
    '60xx,0200' => { VR => 'US', Name => 'OverlayLocation' },
    '60xx,0800' => { VR => 'CS', Name => 'OverlayCodeLabel' },
    '60xx,0802' => { VR => 'US', Name => 'OverlayNumberOfTables' },
    '60xx,0803' => { VR => 'AT', Name => 'OverlayCodeTableLocation' },
    '60xx,0804' => { VR => 'US', Name => 'OverlayBitsForCodeWord' },
    '60xx,1001' => { VR => 'CS', Name => 'OverlayActivationLayer' },
    '60xx,1100' => { VR => 'US', Name => 'OverlayDescriptorGray' },
    '60xx,1101' => { VR => 'US', Name => 'OverlayDescriptorRed' },
    '60xx,1102' => { VR => 'US', Name => 'OverlayDescriptorGreen' },
    '60xx,1103' => { VR => 'US', Name => 'OverlayDescriptorBlue' },
    '60xx,1200' => { VR => 'US', Name => 'OverlaysGray' },
    '60xx,1201' => { VR => 'US', Name => 'OverlaysRed' },
    '60xx,1202' => { VR => 'US', Name => 'OverlaysGreen' },
    '60xx,1203' => { VR => 'US', Name => 'OverlaysBlue' },
    '60xx,1301' => { VR => 'IS', Name => 'ROIArea' },
    '60xx,1302' => { VR => 'DS', Name => 'ROIMean' },
    '60xx,1303' => { VR => 'DS', Name => 'ROIStandardDeviation' },
    '60xx,1500' => { VR => 'LO', Name => 'OverlayLabel' },
    '60xx,3000' => { VR => 'OB', Name => 'OverlayData' },
    '60xx,4000' => { VR => 'LT', Name => 'OverlayComments' },
    '7FE0,0001' => { VR => 'OV', Name => 'ExtendedOffsetTable' },
    '7FE0,0002' => { VR => 'OV', Name => 'ExtendedOffsetTableLengths' },
    '7FE0,0003' => { VR => 'UV', Name => 'EncapsulatedPixelDataValueTotalLen' },
    '7FE0,0008' => { VR => 'OF', Name => 'FloatPixelData', Binary => 1 },
    '7FE0,0009' => { VR => 'OD', Name => 'DoubleFloatPixelData', Binary => 1 },
    # pixel data group
    '7Fxx,0000' => { VR => 'UL', Name => 'PixelDataGroupLength' },
    '7Fxx,0010' => { VR => 'OB', Name => 'PixelData', Binary => 1 },
    '7Fxx,0011' => { VR => 'US', Name => 'VariableNextDataGroup' },
    '7Fxx,0020' => { VR => 'OW', Name => 'VariableCoefficientsSDVN' },
    '7Fxx,0030' => { VR => 'OW', Name => 'VariableCoefficientsSDHN' },
    '7Fxx,0040' => { VR => 'OW', Name => 'VariableCoefficientsSDDN' },
    'FFFA,FFFA' => { VR => 'SQ', Name => 'DigitalSignaturesSequence' },
    'FFFC,FFFC' => { VR => 'OB', Name => 'DataSetTrailingPadding', Binary => 1 },
    # the sequence delimiters have no VR:
    'FFFE,E000' => 'StartOfItem',
    'FFFE,E00D' => 'EndOfItems',
    'FFFE,E0DD' => 'EndOfSequence',
);

# table to translate registered UID values to readable strings
# (the PrintConv is added dynamically when a 'UI' format tag is extracted)
%uid = (
    '1.2.840.10008.1.1' => 'Verification SOP Class',
    '1.2.840.10008.1.2' => 'Implicit VR Little Endian',
    '1.2.840.10008.1.2.1' => 'Explicit VR Little Endian',
    '1.2.840.10008.1.2.1.98' => 'Encapsulated Uncompressed Explicit VR Little Endian',
    '1.2.840.10008.1.2.1.99' => 'Deflated Explicit VR Little Endian',
    '1.2.840.10008.1.2.2' => 'Explicit VR Big Endian',
    '1.2.840.10008.1.2.4.50' => 'JPEG Baseline (Process 1)',
    '1.2.840.10008.1.2.4.51' => 'JPEG Extended (Process 2 & 4)',
    '1.2.840.10008.1.2.4.52' => 'JPEG Extended (Process 3 & 5)',
    '1.2.840.10008.1.2.4.53' => 'JPEG Spectral Selection, Non-Hierarchical (Process 6 & 8)',
    '1.2.840.10008.1.2.4.54' => 'JPEG Spectral Selection, Non-Hierarchical (Process 7 & 9)',
    '1.2.840.10008.1.2.4.55' => 'JPEG Full Progression, Non-Hierarchical (Process 10 & 12)',
    '1.2.840.10008.1.2.4.56' => 'JPEG Full Progression, Non-Hierarchical (Process 11 & 13)',
    '1.2.840.10008.1.2.4.57' => 'JPEG Lossless, Non-Hierarchical (Process 14)',
    '1.2.840.10008.1.2.4.58' => 'JPEG Lossless, Non-Hierarchical (Process 15) ',
    '1.2.840.10008.1.2.4.59' => 'JPEG Extended, Hierarchical (Process 16 & 18) ',
    '1.2.840.10008.1.2.4.60' => 'JPEG Extended, Hierarchical (Process 17 & 19) ',
    '1.2.840.10008.1.2.4.61' => 'JPEG Spectral Selection, Hierarchical (Process 20 & 22)',
    '1.2.840.10008.1.2.4.62' => 'JPEG Spectral Selection, Hierarchical (Process 21 & 23)',
    '1.2.840.10008.1.2.4.63' => 'JPEG Full Progression, Hierarchical (Process 24 & 26)',
    '1.2.840.10008.1.2.4.64' => 'JPEG Full Progression, Hierarchical (Process 25 & 27)',
    '1.2.840.10008.1.2.4.65' => 'JPEG Lossless, Hierarchical (Process 28) ',
    '1.2.840.10008.1.2.4.66' => 'JPEG Lossless, Hierarchical (Process 29) ',
    '1.2.840.10008.1.2.4.70' => 'JPEG Lossless, Non-Hierarchical, First-Order Prediction (Process 14-1)',
    '1.2.840.10008.1.2.4.80' => 'JPEG-LS Lossless Image Compression',
    '1.2.840.10008.1.2.4.81' => 'JPEG-LS Lossy (Near-Lossless) Image Compression',
    '1.2.840.10008.1.2.4.90' => 'JPEG 2000 Image Compression (Lossless Only)',
    '1.2.840.10008.1.2.4.91' => 'JPEG 2000 Image Compression',
    '1.2.840.10008.1.2.4.92' => 'JPEG 2000 Part 2 Multi-component Image Compression (Lossless Only)',
    '1.2.840.10008.1.2.4.93' => 'JPEG 2000 Part 2 Multi-component Image Compression',
    '1.2.840.10008.1.2.4.94' => 'JPIP Referenced',
    '1.2.840.10008.1.2.4.95' => 'JPIP Referenced Deflate',
    '1.2.840.10008.1.2.4.100' => 'MPEG2 Main Profile / Main Level',
    '1.2.840.10008.1.2.4.101' => 'MPEG2 Main Profile / High Level',
    '1.2.840.10008.1.2.4.100.1' => 'Fragmentable MPEG2 Main Profile / Main Level',
    '1.2.840.10008.1.2.4.102' => 'MPEG-4 AVC/H.264 High Profile / Level 4.1',
    '1.2.840.10008.1.2.4.102.1' => 'Fragmentable MPEG-4 AVC/H.264 High Profile / Level 4.1',
    '1.2.840.10008.1.2.4.103' => 'MPEG-4 AVC/H.264 BD-compatible High Profile / Level 4.1',
    '1.2.840.10008.1.2.4.103.1' => 'Fragmentable MPEG-4 AVC/H.264 BD-compatible High Profile / Level 4.1',
    '1.2.840.10008.1.2.4.104' => 'MPEG-4 AVC/H.264 High Profile / Level 4.2 For 2D Video',
    '1.2.840.10008.1.2.4.104.1' => 'Fragmentable MPEG-4 AVC/H.264 High Profile / Level 4.2 For 2D Video',
    '1.2.840.10008.1.2.4.105' => 'MPEG-4 AVC/H.264 High Profile / Level 4.2 For 3D Video',
    '1.2.840.10008.1.2.4.105.1' => 'Fragmentable MPEG-4 AVC/H.264 High Profile / Level 4.2 For 3D Video',
    '1.2.840.10008.1.2.4.106' => 'MPEG-4 AVC/H.264 Stereo High Profile / Level 4.2',
    '1.2.840.10008.1.2.4.106.1' => 'Fragmentable MPEG-4 AVC/H.264 Stereo High Profile / Level 4.2',
    '1.2.840.10008.1.2.4.107' => 'HEVC/H.265 Main Profile / Level 5.1',
    '1.2.840.10008.1.2.4.108' => 'HEVC/H.265 Main 10 Profile / Level 5.1',
    '1.2.840.10008.1.2.4.110' => 'JPEG XL Lossless',
    '1.2.840.10008.1.2.4.111' => 'JPEG XL JPEG Recompression',
    '1.2.840.10008.1.2.4.112' => 'JPEG XL',
    '1.2.840.10008.1.2.4.201' => 'High-Throughput JPEG 2000 Image Compression (Lossless Only)',
    '1.2.840.10008.1.2.4.202' => 'High-Throughput JPEG 2000 with RPCL Options Image Compression (Lossless Only)',
    '1.2.840.10008.1.2.4.203' => 'High-Throughput JPEG 2000 Image Compression',
    '1.2.840.10008.1.2.4.204' => 'JPIP HTJ2K Referenced',
    '1.2.840.10008.1.2.4.205' => 'JPIP HTJ2K Referenced Deflate',
    '1.2.840.10008.1.2.5' => 'RLE Lossless',
    '1.2.840.10008.1.2.6.1' => 'RFC 2557 MIME encapsulation',
    '1.2.840.10008.1.2.6.2' => 'XML Encoding',
    '1.2.840.10008.1.2.7.1' => 'SMPTE ST 2110-20 Uncompressed Progressive Active Video',
    '1.2.840.10008.1.2.7.2' => 'SMPTE ST 2110-20 Uncompressed Interlaced Active Video',
    '1.2.840.10008.1.2.7.3' => 'SMPTE ST 2110-30 PCM Digital Audio',
    '1.2.840.10008.1.2.8.1' => 'Deflated Image Frame Compression',
    '1.2.840.10008.1.3.10' => 'Media Storage Directory Storage',
    '1.2.840.10008.1.4.1.1' => 'Talairach Brain Atlas Frame of Reference',
    '1.2.840.10008.1.4.1.2' => 'SPM2 T1 Frame of Reference',
    '1.2.840.10008.1.4.1.3' => 'SPM2 T2 Frame of Reference',
    '1.2.840.10008.1.4.1.4' => 'SPM2 PD Frame of Reference',
    '1.2.840.10008.1.4.1.5' => 'SPM2 EPI Frame of Reference',
    '1.2.840.10008.1.4.1.6' => 'SPM2 FIL T1 Frame of Reference',
    '1.2.840.10008.1.4.1.7' => 'SPM2 PET Frame of Reference',
    '1.2.840.10008.1.4.1.8' => 'SPM2 TRANSM Frame of Reference',
    '1.2.840.10008.1.4.1.9' => 'SPM2 SPECT Frame of Reference',
    '1.2.840.10008.1.4.1.10' => 'SPM2 GRAY Frame of Reference',
    '1.2.840.10008.1.4.1.11' => 'SPM2 WHITE Frame of Reference',
    '1.2.840.10008.1.4.1.12' => 'SPM2 CSF Frame of Reference',
    '1.2.840.10008.1.4.1.13' => 'SPM2 BRAINMASK Frame of Reference',
    '1.2.840.10008.1.4.1.14' => 'SPM2 AVG305T1 Frame of Reference',
    '1.2.840.10008.1.4.1.15' => 'SPM2 AVG152T1 Frame of Reference',
    '1.2.840.10008.1.4.1.16' => 'SPM2 AVG152T2 Frame of Reference',
    '1.2.840.10008.1.4.1.17' => 'SPM2 AVG152PD Frame of Reference',
    '1.2.840.10008.1.4.1.18' => 'SPM2 SINGLESUBJT1 Frame of Reference',
    '1.2.840.10008.1.4.2.1' => 'ICBM 452 T1 Frame of Reference',
    '1.2.840.10008.1.4.2.2' => 'ICBM Single Subject MRI Frame of Reference',
    '1.2.840.10008.1.4.3.1' => 'IEC 61217 Fixed Coordinate System Frame of Reference',
    '1.2.840.10008.1.4.3.2' => 'Standard Robotic-Arm Coordinate System Frame of Reference',
    '1.2.840.10008.1.4.3.3' => 'IEC 61217 Table Top Coordinate System Frame of Reference',
    '1.2.840.10008.1.4.4.1' => 'SRI24 Frame of Reference',
    '1.2.840.10008.1.4.5.1' => 'Colin27 Frame of Reference',
    '1.2.840.10008.1.4.6.1' => 'LPBA40/AIR Frame of Reference',
    '1.2.840.10008.1.4.6.2' => 'LPBA40/FLIRT Frame of Reference',
    '1.2.840.10008.1.4.6.3' => 'LPBA40/SPM5 Frame of Reference',
    '1.2.840.10008.1.5.1' => 'Hot Iron Color Palette SOP Instance',
    '1.2.840.10008.1.5.2' => 'PET Color Palette SOP Instance',
    '1.2.840.10008.1.5.3' => 'Hot Metal Blue Color Palette SOP Instance',
    '1.2.840.10008.1.5.4' => 'PET 20 Step Color Palette SOP Instance',
    '1.2.840.10008.1.5.5' => 'Spring Color Palette SOP Instance',
    '1.2.840.10008.1.5.6' => 'Summer Color Palette SOP Instance',
    '1.2.840.10008.1.5.7' => 'Fall Color Palette SOP Instance',
    '1.2.840.10008.1.5.8' => 'Winter Color Palette SOP Instance',
    '1.2.840.10008.1.9' => 'Basic Study Content Notification SOP Class',
    '1.2.840.10008.1.20' => 'Papyrus 3 Implicit VR Little Endian (Retired)',
    '1.2.840.10008.1.20.1' => 'Storage Commitment Push Model SOP Class',
    '1.2.840.10008.1.20.1.1' => 'Storage Commitment Push Model SOP Instance',
    '1.2.840.10008.1.20.2' => 'Storage Commitment Pull Model SOP Class ',
    '1.2.840.10008.1.20.2.1' => 'Storage Commitment Pull Model SOP Instance ',
    '1.2.840.10008.1.40' => 'Procedural Event Logging SOP Class',
    '1.2.840.10008.1.40.1' => 'Procedural Event Logging SOP Instance',
    '1.2.840.10008.1.42' => 'Substance Administration Logging SOP Class',
    '1.2.840.10008.1.42.1' => 'Substance Administration Logging SOP Instance',
    '1.2.840.10008.2.6.1' => 'DICOM UID Registry',
    '1.2.840.10008.2.16.4' => 'DICOM Controlled Terminology',
    '1.2.840.10008.2.16.5' => 'Adult Mouse Anatomy Ontology',
    '1.2.840.10008.2.16.6' => 'Uberon Ontology',
    '1.2.840.10008.2.16.7' => 'Integrated Taxonomic Information System (ITIS) Taxonomic Serial Number (TSN)',
    '1.2.840.10008.2.16.8' => 'Mouse Genome Initiative (MGI)',
    '1.2.840.10008.2.16.9' => 'PubChem Compound CID',
    '1.2.840.10008.2.16.10' => 'Dublin Core',
    '1.2.840.10008.2.16.11' => 'New York University Melanoma Clinical Cooperative Group',
    '1.2.840.10008.2.16.12' => 'Mayo Clinic Non-radiological Images Specific Body Structure Anatomical Surface Region Guide',
    '1.2.840.10008.2.16.13' => 'Image Biomarker Standardisation Initiative',
    '1.2.840.10008.2.16.14' => 'Radiomics Ontology',
    '1.2.840.10008.2.16.15' => 'RadElement',
    '1.2.840.10008.2.16.16' => 'ICD-11',
    '1.2.840.10008.2.16.17' => 'Unified numbering system (UNS) for metals and alloys',
    '1.2.840.10008.2.16.18' => 'Research Resource Identification',
    '1.2.840.10008.3.1.1.1' => 'DICOM Application Context Name',
    '1.2.840.10008.3.1.2.1.1' => 'Detached Patient Management SOP Class',
    '1.2.840.10008.3.1.2.1.4' => 'Detached Patient Management Meta SOP Class',
    '1.2.840.10008.3.1.2.2.1' => 'Detached Visit Management SOP Class',
    '1.2.840.10008.3.1.2.3.1' => 'Detached Study Management SOP Class',
    '1.2.840.10008.3.1.2.3.2' => 'Study Component Management SOP Class',
    '1.2.840.10008.3.1.2.3.3' => 'Modality Performed Procedure Step SOP Class',
    '1.2.840.10008.3.1.2.3.4' => 'Modality Performed Procedure Step Retrieve SOP Class',
    '1.2.840.10008.3.1.2.3.5' => 'Modality Performed Procedure Step Notification SOP Class',
    '1.2.840.10008.3.1.2.5.1' => 'Detached Results Management SOP Class',
    '1.2.840.10008.3.1.2.5.4' => 'Detached Results Management Meta SOP Class',
    '1.2.840.10008.3.1.2.5.5' => 'Detached Study Management Meta SOP Class',
    '1.2.840.10008.3.1.2.6.1' => 'Detached Interpretation Management SOP Class',
    '1.2.840.10008.4.2' => 'Storage Service Class Service Class PS 3.4',
    '1.2.840.10008.5.1.1.1' => 'Basic Film Session SOP Class',
    '1.2.840.10008.5.1.1.2' => 'Basic Film Box SOP Class',
    '1.2.840.10008.5.1.1.4' => 'Basic Grayscale Image Box SOP Class',
    '1.2.840.10008.5.1.1.4.1' => 'Basic Color Image Box SOP Class',
    '1.2.840.10008.5.1.1.4.2' => 'Referenced Image Box SOP Class',
    '1.2.840.10008.5.1.1.9' => 'Basic Grayscale Print ManagementMeta SOP Class',
    '1.2.840.10008.5.1.1.9.1' => 'Referenced Grayscale Print Management Meta SOP Class',
    '1.2.840.10008.5.1.1.14' => 'Print Job SOP Class',
    '1.2.840.10008.5.1.1.15' => 'Basic Annotation Box SOP Class',
    '1.2.840.10008.5.1.1.16' => 'Printer SOP Class',
    '1.2.840.10008.5.1.1.16.376' => 'Printer Configuration Retrieval SOP Class',
    '1.2.840.10008.5.1.1.17' => 'Printer SOP Instance',
    '1.2.840.10008.5.1.1.17.376' => 'Printer Configuration RetrievalSOP Instance',
    '1.2.840.10008.5.1.1.18' => 'Basic Color Print Management Meta SOP Class',
    '1.2.840.10008.5.1.1.18.1' => 'Referenced Color Print Management Meta SOP Class',
    '1.2.840.10008.5.1.1.22' => 'VOI LUT Box SOP Class',
    '1.2.840.10008.5.1.1.23' => 'Presentation LUT SOP Class',
    '1.2.840.10008.5.1.1.24' => 'Image Overlay Box SOP Class',
    '1.2.840.10008.5.1.1.24.1' => 'Basic Print Image Overlay Box SOP Class',
    '1.2.840.10008.5.1.1.25' => 'Print Queue SOP Instance',
    '1.2.840.10008.5.1.1.26' => 'Print Queue Management SOP Class',
    '1.2.840.10008.5.1.1.27' => 'Stored Print Storage SOP Class',
    '1.2.840.10008.5.1.1.29' => 'Hardcopy Grayscale Image',
    '1.2.840.10008.5.1.1.30' => 'Hardcopy Color Image Storage SOP Class',
    '1.2.840.10008.5.1.1.31' => 'Pull Print Request SOP Class',
    '1.2.840.10008.5.1.1.32' => 'Pull Stored Print Management Meta SOP Class',
    '1.2.840.10008.5.1.1.33' => 'Media Creation Management SOP Class',
    '1.2.840.10008.5.1.1.40' => 'Display System SOP Class',
    '1.2.840.10008.5.1.1.40.1' => 'Display System SOP Instance',
    '1.2.840.10008.5.1.4.1.1.1' => 'Computed Radiography Image Storage',
    '1.2.840.10008.5.1.4.1.1.1.1' => 'Digital X-Ray Image Storage - For Presentation',
    '1.2.840.10008.5.1.4.1.1.1.1.1' => 'Digital X-Ray Image Storage - For Processing',
    '1.2.840.10008.5.1.4.1.1.1.2' => 'Digital Mammography X-Ray Image Storage - For Presentation',
    '1.2.840.10008.5.1.4.1.1.1.2.1' => 'Digital Mammography X-Ray Image Storage - For Processing',
    '1.2.840.10008.5.1.4.1.1.1.3' => 'Digital Intra-oral X-Ray Image Storage - For Presentation',
    '1.2.840.10008.5.1.4.1.1.1.3.1' => 'Digital Intra-oral X-Ray Image Storage - For Processing',
    '1.2.840.10008.5.1.4.1.1.2' => 'CT Image Storage',
    '1.2.840.10008.5.1.4.1.1.2.1' => 'Enhanced CT Image Storage',
    '1.2.840.10008.5.1.4.1.1.2.2' => 'Legacy Converted Enhanced CT Image Storage',
    '1.2.840.10008.5.1.4.1.1.3' => 'Ultrasound Multi-frame Image Storage ',
    '1.2.840.10008.5.1.4.1.1.3.1' => 'Ultrasound Multi-frame Image Storage',
    '1.2.840.10008.5.1.4.1.1.4' => 'MR Image Storage',
    '1.2.840.10008.5.1.4.1.1.4.1' => 'Enhanced MR Image Storage',
    '1.2.840.10008.5.1.4.1.1.4.2' => 'MR Spectroscopy Storage',
    '1.2.840.10008.5.1.4.1.1.4.3' => 'Enhanced MR Color Image Storage',
    '1.2.840.10008.5.1.4.1.1.4.4' => 'Legacy Converted Enhanced MR Image Storage',
    '1.2.840.10008.5.1.4.1.1.5' => 'Nuclear Medicine Image Storage',
    '1.2.840.10008.5.1.4.1.1.6' => 'Ultrasound Image Storage',
    '1.2.840.10008.5.1.4.1.1.6.1' => 'Ultrasound Image Storage',
    '1.2.840.10008.5.1.4.1.1.6.2' => 'Enhanced US Volume Storage',
    '1.2.840.10008.5.1.4.1.1.6.3' => 'Photoacoustic Image Storage',
    '1.2.840.10008.5.1.4.1.1.7' => 'Secondary Capture Image Storage',
    '1.2.840.10008.5.1.4.1.1.7.1' => 'Multi-frame Single Bit Secondary',
    '1.2.840.10008.5.1.4.1.1.7.2' => 'Multi-frame Grayscale Byte Secondary Capture Image Storage',
    '1.2.840.10008.5.1.4.1.1.7.3' => 'Multi-frame Grayscale Word Secondary Capture Image Storage',
    '1.2.840.10008.5.1.4.1.1.7.4' => 'Multi-frame True Color Secondary Capture Image Storage',
    '1.2.840.10008.5.1.4.1.1.8' => 'Standalone Overlay Storage',
    '1.2.840.10008.5.1.4.1.1.9' => 'Standalone Curve Storage',
    '1.2.840.10008.5.1.4.1.1.9.1' => 'Waveform Storage - Trial (Retired)',
    '1.2.840.10008.5.1.4.1.1.9.1.1' => '12-lead ECG Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.1.2' => 'General ECG Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.1.3' => 'Ambulatory ECG Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.1.4' => 'General 32-bit ECG Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.2.1' => 'Hemodynamic Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.3.1' => 'Cardiac Electrophysiology Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.4.1' => 'Basic Voice Audio Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.4.2' => 'General Audio Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.5.1' => 'Arterial Pulse Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.6.1' => 'Respiratory Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.6.2' => 'Multi-channel Respiratory Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.7.1' => 'Routine Scalp Electroencephalogram Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.7.2' => 'Electromyogram Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.7.3' => 'Electrooculogram Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.7.4' => 'Sleep Electroencephalogram Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.8.1' => 'Body Position Waveform Storage',
    '1.2.840.10008.5.1.4.1.1.9.100.1' => 'Waveform Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.9.100.2' => 'Waveform Acquisition Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.10' => 'Standalone Modality LUT Storage',
    '1.2.840.10008.5.1.4.1.1.11' => 'Standalone VOI LUT Storage',
    '1.2.840.10008.5.1.4.1.1.11.1' => 'Grayscale Softcopy Presentation State Storage SOP Class',
    '1.2.840.10008.5.1.4.1.1.11.2' => 'Color Softcopy Presentation State Storage SOP Class',
    '1.2.840.10008.5.1.4.1.1.11.3' => 'Pseudo-Color Softcopy Presentation State Storage SOP Class',
    '1.2.840.10008.5.1.4.1.1.11.4' => 'Blending Softcopy Presentation State Storage SOP Class',
    '1.2.840.10008.5.1.4.1.1.11.5' => 'XA/XRF Grayscale Softcopy Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.11.6' => 'Grayscale Planar MPR Volumetric Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.11.7' => 'Compositing Planar MPR Volumetric Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.11.8' => 'Advanced Blending Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.11.9' => 'Volume Rendering Volumetric Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.11.10' => 'Segmented Volume Rendering Volumetric Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.11.11' => 'Multiple Volume Rendering Volumetric Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.11.12' => 'Variable Modality LUT Softcopy Presentation State Storage',
    '1.2.840.10008.5.1.4.1.1.12.1' => 'X-Ray Angiographic Image Storage',
    '1.2.840.10008.5.1.4.1.1.12.1.1' => 'Enhanced XA Image Storage',
    '1.2.840.10008.5.1.4.1.1.12.2' => 'X-Ray Radiofluoroscopic Image Storage',
    '1.2.840.10008.5.1.4.1.1.12.2.1' => 'Enhanced XRF Image Storage',
    '1.2.840.10008.5.1.4.1.1.12.3' => 'X-Ray Angiographic Bi-Plane Image Storage ',
    '1.2.840.10008.5.1.4.1.1.13.1.1' => 'X-Ray 3D Angiographic Image Storage',
    '1.2.840.10008.5.1.4.1.1.13.1.2' => 'X-Ray 3D Craniofacial Image Storage',
    '1.2.840.10008.5.1.4.1.1.13.1.3' => 'Breast Tomosynthesis Image Storage',
    '1.2.840.10008.5.1.4.1.1.13.1.4' => 'Breast Projection X-Ray Image Storage - For Presentation',
    '1.2.840.10008.5.1.4.1.1.13.1.5' => 'Breast Projection X-Ray Image Storage - For Processing',
    '1.2.840.10008.5.1.4.1.1.14.1' => 'Intravascular Optical Coherence Tomography Image Storage - For Presentation',
    '1.2.840.10008.5.1.4.1.1.14.2' => 'Intravascular Optical Coherence Tomography Image Storage - For Processing',
    '1.2.840.10008.5.1.4.1.1.20' => 'Nuclear Medicine Image Storage',
    '1.2.840.10008.5.1.4.1.1.30' => 'Parametric Map Storage',
    '1.2.840.10008.5.1.4.1.1.66' => 'Raw Data Storage',
    '1.2.840.10008.5.1.4.1.1.66.1' => 'Spatial Registration Storage',
    '1.2.840.10008.5.1.4.1.1.66.2' => 'Spatial Fiducials Storage',
    '1.2.840.10008.5.1.4.1.1.66.3' => 'Deformable Spatial Registration Storage',
    '1.2.840.10008.5.1.4.1.1.66.4' => 'Segmentation Storage',
    '1.2.840.10008.5.1.4.1.1.66.5' => 'Surface Segmentation Storage',
    '1.2.840.10008.5.1.4.1.1.66.6' => 'Tractography Results Storage',
    '1.2.840.10008.5.1.4.1.1.66.7' => 'Label Map Segmentation Storage',
    '1.2.840.10008.5.1.4.1.1.66.8' => 'Height Map Segmentation Storage',
    '1.2.840.10008.5.1.4.1.1.66' => 'Raw Data Storage',
    '1.2.840.10008.5.1.4.1.1.67' => 'Real World Value Mapping Storage',
    '1.2.840.10008.5.1.4.1.1.68.1' => 'Surface Scan Mesh Storage',
    '1.2.840.10008.5.1.4.1.1.68.2' => 'Surface Scan Point Cloud Storage',
    '1.2.840.10008.5.1.4.1.1.77.1' => 'VL Image Storage ',
    '1.2.840.10008.5.1.4.1.1.77.1.1' => 'VL Endoscopic Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.1.1' => 'Video Endoscopic Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.2' => 'VL Microscopic Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.2.1' => 'Video Microscopic Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.3' => 'VL Slide-Coordinates Microscopic Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.4' => 'VL Photographic Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.4.1' => 'Video Photographic Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.1' => 'Ophthalmic Photography 8 Bit Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.2' => 'Ophthalmic Photography 16 Bit Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.3' => 'Stereometric Relationship Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.4' => 'Ophthalmic Tomography Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.5' => 'Wide Field Ophthalmic Photography Stereographic Projection Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.6' => 'Wide Field Ophthalmic Photography 3D Coordinates Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.7' => 'Ophthalmic Optical Coherence Tomography En Face Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.5.8' => 'Ophthalmic Optical Coherence Tomography B-scan Volume Analysis Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.6' => 'VL Whole Slide Microscopy Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.7' => 'Dermoscopic Photography Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.8' => 'Confocal Microscopy Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.1.9' => 'Confocal Microscopy Tiled Pyramidal Image Storage',
    '1.2.840.10008.5.1.4.1.1.77.2' => 'VL Multi-frame Image Storage',
    '1.2.840.10008.5.1.4.1.1.78.1' => 'Lensometry Measurements Storage',
    '1.2.840.10008.5.1.4.1.1.78.2' => 'Autorefraction Measurements Storage',
    '1.2.840.10008.5.1.4.1.1.78.3' => 'Keratometry Measurements Storage',
    '1.2.840.10008.5.1.4.1.1.78.4' => 'Subjective Refraction Measurements Storage',
    '1.2.840.10008.5.1.4.1.1.78.5' => 'Visual Acuity Measurements Storage',
    '1.2.840.10008.5.1.4.1.1.78.6' => 'Spectacle Prescription Report Storage',
    '1.2.840.10008.5.1.4.1.1.78.7' => 'Ophthalmic Axial Measurements Storage',
    '1.2.840.10008.5.1.4.1.1.78.8' => 'Intraocular Lens Calculations Storage',
    '1.2.840.10008.5.1.4.1.1.79.1' => 'Macular Grid Thickness and Volume Report Storage SOP Class',
    '1.2.840.10008.5.1.4.1.1.80.1' => 'Ophthalmic Visual Field Static Perimetry Measurements Storage',
    '1.2.840.10008.5.1.4.1.1.81.1' => 'Ophthalmic Thickness Map Storage',
    '1.2.840.10008.5.1.4.1.1.82.1' => 'Corneal Topography Map Storage',
    '1.2.840.10008.5.1.4.1.1.88.1' => 'Text SR Storage - Trial (Retired)',
    '1.2.840.10008.5.1.4.1.1.88.2' => 'Audio SR Storage - Trial (Retired)',
    '1.2.840.10008.5.1.4.1.1.88.3' => 'Detail SR Storage - Trial (Retired)',
    '1.2.840.10008.5.1.4.1.1.88.4' => 'Comprehensive SR Storage - Trial (Retired)',
    '1.2.840.10008.5.1.4.1.1.88.11' => 'Basic Text SR',
    '1.2.840.10008.5.1.4.1.1.88.22' => 'Enhanced SR',
    '1.2.840.10008.5.1.4.1.1.88.33' => 'Comprehensive SR',
    '1.2.840.10008.5.1.4.1.1.88.34' => 'Comprehensive 3D SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.35' => 'Extensible SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.40' => 'Procedure Log Storage',
    '1.2.840.10008.5.1.4.1.1.88.50' => 'Mammography CAD SR',
    '1.2.840.10008.5.1.4.1.1.88.59' => 'Key Object Selection Document',
    '1.2.840.10008.5.1.4.1.1.88.65' => 'Chest CAD SR',
    '1.2.840.10008.5.1.4.1.1.88.67' => 'X-Ray Radiation Dose SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.68' => 'Radiopharmaceutical Radiation Dose SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.69' => 'Colon CAD SR',
    '1.2.840.10008.5.1.4.1.1.88.70' => 'Implantation Plan SR Document Storage',
    '1.2.840.10008.5.1.4.1.1.88.71' => 'Acquisition Context SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.72' => 'Simplified Adult Echo SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.73' => 'Patient Radiation Dose SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.74' => 'Planned Imaging Agent Administration SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.75' => 'Performed Imaging Agent Administration SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.76' => 'Enhanced X-Ray Radiation Dose SR Storage',
    '1.2.840.10008.5.1.4.1.1.88.77' => 'Waveform Annotation SR Storage',
    '1.2.840.10008.5.1.4.1.1.90.1' => 'Content Assessment Results Storage',
    '1.2.840.10008.5.1.4.1.1.91.1' => 'Microscopy Bulk Simple Annotations Storage',
    '1.2.840.10008.5.1.4.1.1.104.1' => 'Encapsulated PDF Storage',
    '1.2.840.10008.5.1.4.1.1.104.2' => 'Encapsulated CDA Storage',
    '1.2.840.10008.5.1.4.1.1.104.3' => 'Encapsulated STL Storage',
    '1.2.840.10008.5.1.4.1.1.104.4' => 'Encapsulated OBJ Storage',
    '1.2.840.10008.5.1.4.1.1.104.5' => 'Encapsulated MTL Storage',
    '1.2.840.10008.5.1.4.1.1.128.1' => 'Legacy Converted Enhanced PET Image Storage',
    '1.2.840.10008.5.1.4.1.1.128' => 'Positron Emission Tomography Image Storage',
    '1.2.840.10008.5.1.4.1.1.129' => 'Standalone PET Curve Storage',
    '1.2.840.10008.5.1.4.1.1.130' => 'Enhanced PET Image Storage',
    '1.2.840.10008.5.1.4.1.1.131' => 'Basic Structured Display Storage',
    '1.2.840.10008.5.1.4.1.1.200.1' => 'CT Defined Procedure Protocol Storage',
    '1.2.840.10008.5.1.4.1.1.200.2' => 'CT Performed Procedure Protocol Storage',
    '1.2.840.10008.5.1.4.1.1.200.3' => 'Protocol Approval Storage',
    '1.2.840.10008.5.1.4.1.1.200.4' => 'Protocol Approval Information Model - FIND',
    '1.2.840.10008.5.1.4.1.1.200.5' => 'Protocol Approval Information Model - MOVE',
    '1.2.840.10008.5.1.4.1.1.200.6' => 'Protocol Approval Information Model - GET',
    '1.2.840.10008.5.1.4.1.1.200.7' => 'XA Defined Procedure Protocol Storage',
    '1.2.840.10008.5.1.4.1.1.200.8' => 'XA Performed Procedure Protocol Storage',
    '1.2.840.10008.5.1.4.1.1.201.1' => 'Inventory Storage',
    '1.2.840.10008.5.1.4.1.1.201.1.1' => 'Storage Management SOP Instance',
    '1.2.840.10008.5.1.4.1.1.201.2' => 'Inventory - FIND',
    '1.2.840.10008.5.1.4.1.1.201.3' => 'Inventory - MOVE',
    '1.2.840.10008.5.1.4.1.1.201.4' => 'Inventory - GET',
    '1.2.840.10008.5.1.4.1.1.201.5' => 'Inventory Creation',
    '1.2.840.10008.5.1.4.1.1.201.6' => 'Repository Query',
    '1.2.840.10008.5.1.4.1.1.481.1' => 'RT Image Storage',
    '1.2.840.10008.5.1.4.1.1.481.2' => 'RT Dose Storage',
    '1.2.840.10008.5.1.4.1.1.481.3' => 'RT Structure Set Storage',
    '1.2.840.10008.5.1.4.1.1.481.4' => 'RT Beams Treatment Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.5' => 'RT Plan Storage',
    '1.2.840.10008.5.1.4.1.1.481.6' => 'RT Brachy Treatment Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.7' => 'RT Treatment Summary Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.8' => 'RT Ion Plan Storage',
    '1.2.840.10008.5.1.4.1.1.481.9' => 'RT Ion Beams Treatment Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.10' => 'RT Physician Intent Storage',
    '1.2.840.10008.5.1.4.1.1.481.11' => 'RT Segment Annotation Storage',
    '1.2.840.10008.5.1.4.1.1.481.12' => 'RT Radiation Set Storage',
    '1.2.840.10008.5.1.4.1.1.481.13' => 'C-Arm Photon-Electron Radiation Storage',
    '1.2.840.10008.5.1.4.1.1.481.14' => 'Tomotherapeutic Radiation Storage',
    '1.2.840.10008.5.1.4.1.1.481.15' => 'Robotic-Arm Radiation Storage',
    '1.2.840.10008.5.1.4.1.1.481.16' => 'RT Radiation Record Set Storage',
    '1.2.840.10008.5.1.4.1.1.481.17' => 'RT Radiation Salvage Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.18' => 'Tomotherapeutic Radiation Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.19' => 'C-Arm Photon-Electron Radiation Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.20' => 'Robotic Radiation Record Storage',
    '1.2.840.10008.5.1.4.1.1.481.21' => 'RT Radiation Set Delivery Instruction Storage',
    '1.2.840.10008.5.1.4.1.1.481.22' => 'RT Treatment Preparation Storage',
    '1.2.840.10008.5.1.4.1.1.481.23' => 'Enhanced RT Image Storage',
    '1.2.840.10008.5.1.4.1.1.481.24' => 'Enhanced Continuous RT Image Storage',
    '1.2.840.10008.5.1.4.1.1.481.25' => 'RT Patient Position Acquisition Instruction Storage',
    '1.2.840.10008.5.1.4.1.1.501.1' => 'DICOS CT Image Storage',
    '1.2.840.10008.5.1.4.1.1.501.2.1' => 'DICOS Digital X-Ray Image Storage - For Presentation',
    '1.2.840.10008.5.1.4.1.1.501.2.2' => 'DICOS Digital X-Ray Image Storage - For Processing',
    '1.2.840.10008.5.1.4.1.1.501.3' => 'DICOS Threat Detection Report Storage',
    '1.2.840.10008.5.1.4.1.1.501.4' => 'DICOS 2D AIT Storage',
    '1.2.840.10008.5.1.4.1.1.501.5' => 'DICOS 3D AIT Storage',
    '1.2.840.10008.5.1.4.1.1.501.6' => 'DICOS Quadrupole Resonance (QR) Storage',
    '1.2.840.10008.5.1.4.1.1.601.1' => 'Eddy Current Image Storage',
    '1.2.840.10008.5.1.4.1.1.601.2' => 'Eddy Current Multi-frame Image Storage',
    '1.2.840.10008.5.1.4.1.1.601.3' => 'Thermography Image Storage',
    '1.2.840.10008.5.1.4.1.1.601.4' => 'Thermography Multi-frame Image Storage',
    '1.2.840.10008.5.1.4.1.1.601.5' => 'Ultrasound Waveform Storage',
    '1.2.840.10008.5.1.4.1.2.1.1' => 'Patient Root Query/Retrieve Information Model - FIND',
    '1.2.840.10008.5.1.4.1.2.1.2' => 'Patient Root Query/Retrieve Information Model - MOVE',
    '1.2.840.10008.5.1.4.1.2.1.3' => 'Patient Root Query/Retrieve Information Model - GET',
    '1.2.840.10008.5.1.4.1.2.2.1' => 'Study Root Query/Retrieve Information Model - FIND',
    '1.2.840.10008.5.1.4.1.2.2.2' => 'Study Root Query/Retrieve Information Model - MOVE',
    '1.2.840.10008.5.1.4.1.2.2.3' => 'Study Root Query/Retrieve Information Model - GET',
    '1.2.840.10008.5.1.4.1.2.3.1' => 'Patient/Study Only Query/Retrieve Information Model - FIND',
    '1.2.840.10008.5.1.4.1.2.3.2' => 'Patient/Study Only Query/Retrieve Information Model - MOVE',
    '1.2.840.10008.5.1.4.1.2.3.3' => 'Patient/Study Only Query/Retrieve Information Model - GET',
    '1.2.840.10008.5.1.4.1.2.4.2' => 'Composite Instance Root Retrieve - MOVE',
    '1.2.840.10008.5.1.4.1.2.4.3' => 'Composite Instance Root Retrieve - GET',
    '1.2.840.10008.5.1.4.1.2.5.3' => 'Composite Instance Retrieve Without Bulk Data - GET',
    '1.2.840.10008.5.1.4.20.1' => 'Defined Procedure Protocol Information Model - FIND',
    '1.2.840.10008.5.1.4.20.2' => 'Defined Procedure Protocol Information Model - MOVE',
    '1.2.840.10008.5.1.4.20.3' => 'Defined Procedure Protocol Information Model - GET',
    '1.2.840.10008.5.1.4.31' => 'Modality Worklist Information Model - FIND',
    '1.2.840.10008.5.1.4.32.1' => 'General Purpose Worklist Information Model - FIND',
    '1.2.840.10008.5.1.4.32.2' => 'General Purpose Scheduled Procedure Step SOP Class',
    '1.2.840.10008.5.1.4.32.3' => 'General Purpose Performed Procedure Step SOP Class',
    '1.2.840.10008.5.1.4.32' => 'General Purpose Worklist Management Meta SOP Class',
    '1.2.840.10008.5.1.4.33' => 'Instance Availability Notification SOP Class',
    '1.2.840.10008.5.1.4.34.1' => 'RT Beams Delivery Instruction Storage',
    '1.2.840.10008.5.1.4.34.2' => 'RT Conventional Machine Verification',
    '1.2.840.10008.5.1.4.34.3' => 'RT Ion Machine Verification',
    '1.2.840.10008.5.1.4.34.4' => 'Unified Worklist and Procedure Step Service Class',
    '1.2.840.10008.5.1.4.34.4.1' => 'Unified Procedure Step - Push SOP Class',
    '1.2.840.10008.5.1.4.34.4.2' => 'Unified Procedure Step - Watch SOP Class',
    '1.2.840.10008.5.1.4.34.4.3' => 'Unified Procedure Step - Pull SOP Class',
    '1.2.840.10008.5.1.4.34.4.4' => 'Unified Procedure Step - Event SOP Class',
    '1.2.840.10008.5.1.4.34.5' => 'Unified Worklist and Procedure Step SOP Instance',
    '1.2.840.10008.5.1.4.34.5.1' => 'UPS Filtered Global Subscription SOP Instance',
    '1.2.840.10008.5.1.4.34.6' => 'Unified Worklist and Procedure Step Service Class',
    '1.2.840.10008.5.1.4.34.6.1' => 'Unified Procedure Step - Push SOP Class',
    '1.2.840.10008.5.1.4.34.6.2' => 'Unified Procedure Step - Watch SOP Class',
    '1.2.840.10008.5.1.4.34.6.3' => 'Unified Procedure Step - Pull SOP Class',
    '1.2.840.10008.5.1.4.34.6.4' => 'Unified Procedure Step - Event SOP Class',
    '1.2.840.10008.5.1.4.34.6.5' => 'Unified Procedure Step - Query SOP Class',
    '1.2.840.10008.5.1.4.34.7' => 'RT Beams Delivery Instruction Storage',
    '1.2.840.10008.5.1.4.34.8' => 'RT Conventional Machine Verification',
    '1.2.840.10008.5.1.4.34.9' => 'RT Ion Machine Verification',
    '1.2.840.10008.5.1.4.34.10' => 'RT Brachy Application Setup Delivery Instruction Storage',
    '1.2.840.10008.5.1.4.37.1' => 'General Relevant Patient Information Query',
    '1.2.840.10008.5.1.4.37.2' => 'Breast Imaging Relevant Patient Information Query',
    '1.2.840.10008.5.1.4.37.3' => 'Cardiac Relevant Patient Information Query',
    '1.2.840.10008.5.1.4.38.1' => 'Hanging Protocol Storage',
    '1.2.840.10008.5.1.4.38.2' => 'Hanging Protocol Information Model - FIND',
    '1.2.840.10008.5.1.4.38.3' => 'Hanging Protocol Information Model - MOVE',
    '1.2.840.10008.5.1.4.38.4' => 'Hanging Protocol Information Model - GET',
    '1.2.840.10008.5.1.4.39.1' => 'Color Palette Storage',
    '1.2.840.10008.5.1.4.39.2' => 'Color Palette Information Model - FIND',
    '1.2.840.10008.5.1.4.39.3' => 'Color Palette Information Model - MOVE',
    '1.2.840.10008.5.1.4.39.4' => 'Color Palette Information Model - GET',
    '1.2.840.10008.5.1.4.41' => 'Product Characteristics Query SOP Class',
    '1.2.840.10008.5.1.4.42' => 'Substance Approval Query SOP Class',
    '1.2.840.10008.5.1.4.43.1' => 'Generic Implant Template Storage',
    '1.2.840.10008.5.1.4.43.2' => 'Generic Implant Template Information Model - FIND',
    '1.2.840.10008.5.1.4.43.3' => 'Generic Implant Template Information Model - MOVE',
    '1.2.840.10008.5.1.4.43.4' => 'Generic Implant Template Information Model - GET',
    '1.2.840.10008.5.1.4.44.1' => 'Implant Assembly Template Storage',
    '1.2.840.10008.5.1.4.44.2' => 'Implant Assembly Template Information Model - FIND',
    '1.2.840.10008.5.1.4.44.3' => 'Implant Assembly Template Information Model - MOVE',
    '1.2.840.10008.5.1.4.44.4' => 'Implant Assembly Template Information Model - GET',
    '1.2.840.10008.5.1.4.45.1' => 'Implant Template Group Storage',
    '1.2.840.10008.5.1.4.45.2' => 'Implant Template Group Information Model - FIND',
    '1.2.840.10008.5.1.4.45.3' => 'Implant Template Group Information Model - MOVE',
    '1.2.840.10008.5.1.4.45.4' => 'Implant Template Group Information Model - GET',
    '1.2.840.10008.6.1.1' => 'Anatomic Modifier',
    '1.2.840.10008.6.1.2' => 'Anatomic Region',
    '1.2.840.10008.6.1.3' => 'Transducer Approach',
    '1.2.840.10008.6.1.4' => 'Transducer Orientation',
    '1.2.840.10008.6.1.5' => 'Ultrasound Beam Path',
    '1.2.840.10008.6.1.6' => 'Angiographic Interventional Device',
    '1.2.840.10008.6.1.7' => 'Image Guided Therapeutic Procedure',
    '1.2.840.10008.6.1.8' => 'Interventional Drug',
    '1.2.840.10008.6.1.9' => 'Administration Route',
    '1.2.840.10008.6.1.10' => 'Imaging Contrast Agent',
    '1.2.840.10008.6.1.11' => 'Imaging Contrast Agent Ingredient',
    '1.2.840.10008.6.1.12' => 'Radiopharmaceutical Isotope',
    '1.2.840.10008.6.1.13' => 'Patient Orientation',
    '1.2.840.10008.6.1.14' => 'Patient Orientation Modifier',
    '1.2.840.10008.6.1.15' => 'Patient Equipment Relationship',
    '1.2.840.10008.6.1.16' => 'Cranio-Caudad Angulation',
    '1.2.840.10008.6.1.17' => 'Radiopharmaceutical',
    '1.2.840.10008.6.1.18' => 'Nuclear Medicine Projection',
    '1.2.840.10008.6.1.19' => 'Acquisition Modality',
    '1.2.840.10008.6.1.20' => 'DICOM Device',
    '1.2.840.10008.6.1.21' => 'Abstract Prior',
    '1.2.840.10008.6.1.22' => 'Numeric Value Qualifier',
    '1.2.840.10008.6.1.23' => 'Measurement Unit',
    '1.2.840.10008.6.1.24' => 'Real World Value Mapping Unit',
    '1.2.840.10008.6.1.25' => 'Significance Level',
    '1.2.840.10008.6.1.26' => 'Measurement Range Concept',
    '1.2.840.10008.6.1.27' => 'Normality',
    '1.2.840.10008.6.1.28' => 'Normal Range Value',
    '1.2.840.10008.6.1.29' => 'Selection Method',
    '1.2.840.10008.6.1.30' => 'Measurement Uncertainty Concept',
    '1.2.840.10008.6.1.31' => 'Population Statistical Descriptor',
    '1.2.840.10008.6.1.32' => 'Sample Statistical Descriptor',
    '1.2.840.10008.6.1.33' => 'Equation or Table',
    '1.2.840.10008.6.1.34' => 'Yes-No',
    '1.2.840.10008.6.1.35' => 'Present-Absent',
    '1.2.840.10008.6.1.36' => 'Normal-Abnormal',
    '1.2.840.10008.6.1.37' => 'Laterality',
    '1.2.840.10008.6.1.38' => 'Positive-Negative',
    '1.2.840.10008.6.1.39' => 'Complication Severity',
    '1.2.840.10008.6.1.40' => 'Observer Type',
    '1.2.840.10008.6.1.41' => 'Observation Subject Class',
    '1.2.840.10008.6.1.42' => 'Audio Channel Source',
    '1.2.840.10008.6.1.43' => 'ECG Lead',
    '1.2.840.10008.6.1.44' => 'Hemodynamic Waveform Source',
    '1.2.840.10008.6.1.45' => 'Cardiovascular Anatomic Structure',
    '1.2.840.10008.6.1.46' => 'Electrophysiology Anatomic Location',
    '1.2.840.10008.6.1.47' => 'Coronary Artery Segment',
    '1.2.840.10008.6.1.48' => 'Coronary Artery',
    '1.2.840.10008.6.1.49' => 'Cardiovascular Anatomic Structure Modifier',
    '1.2.840.10008.6.1.50' => 'CID 3082',
    '1.2.840.10008.6.1.51' => 'Time Synchronization Channel Type',
    '1.2.840.10008.6.1.52' => 'Cardiac Procedural State Value',
    '1.2.840.10008.6.1.53' => 'Electrophysiology Measurement Function/Technique',
    '1.2.840.10008.6.1.54' => 'Hemodynamic Measurement Technique',
    '1.2.840.10008.6.1.55' => 'Catheterization Procedure Phase',
    '1.2.840.10008.6.1.56' => 'Electrophysiology Procedure Phase',
    '1.2.840.10008.6.1.57' => 'Stress Protocol',
    '1.2.840.10008.6.1.58' => 'ECG Patient State Value',
    '1.2.840.10008.6.1.59' => 'Electrode Placement Value',
    '1.2.840.10008.6.1.60' => 'CID 3264',
    '1.2.840.10008.6.1.61' => 'Hemodynamic Physiological Challenge',
    '1.2.840.10008.6.1.62' => 'ECG Annotation',
    '1.2.840.10008.6.1.63' => 'Hemodynamic Annotation',
    '1.2.840.10008.6.1.64' => 'Electrophysiology Annotation',
    '1.2.840.10008.6.1.65' => 'Procedure Log Title',
    '1.2.840.10008.6.1.66' => 'Log Note Type',
    '1.2.840.10008.6.1.67' => 'Patient Status and Event',
    '1.2.840.10008.6.1.68' => 'Percutaneous Entry',
    '1.2.840.10008.6.1.69' => 'Staff Action',
    '1.2.840.10008.6.1.70' => 'Procedure Action Value',
    '1.2.840.10008.6.1.71' => 'Non-coronary Transcatheter Intervention',
    '1.2.840.10008.6.1.72' => 'Object Reference Purpose',
    '1.2.840.10008.6.1.73' => 'Consumable Action',
    '1.2.840.10008.6.1.74' => 'Drug/Contrast Administration',
    '1.2.840.10008.6.1.75' => 'Drug/Contrast Numeric Parameter',
    '1.2.840.10008.6.1.76' => 'Intracoronary Device',
    '1.2.840.10008.6.1.77' => 'Intervention Action/Status',
    '1.2.840.10008.6.1.78' => 'Adverse Outcome',
    '1.2.840.10008.6.1.79' => 'Procedure Urgency',
    '1.2.840.10008.6.1.80' => 'Cardiac Rhythm',
    '1.2.840.10008.6.1.81' => 'Respiration Rhythm',
    '1.2.840.10008.6.1.82' => 'Lesion Risk',
    '1.2.840.10008.6.1.83' => 'Finding Title',
    '1.2.840.10008.6.1.84' => 'Procedure Action',
    '1.2.840.10008.6.1.85' => 'Device Use Action',
    '1.2.840.10008.6.1.86' => 'Numeric Device Characteristic',
    '1.2.840.10008.6.1.87' => 'Intervention Parameter',
    '1.2.840.10008.6.1.88' => 'Consumables Parameter',
    '1.2.840.10008.6.1.89' => 'Equipment Event',
    '1.2.840.10008.6.1.90' => 'Cardiovascular Imaging Procedure',
    '1.2.840.10008.6.1.91' => 'Catheterization Device',
    '1.2.840.10008.6.1.92' => 'DateTime Qualifier',
    '1.2.840.10008.6.1.93' => 'Peripheral Pulse Location',
    '1.2.840.10008.6.1.94' => 'Patient Assessment',
    '1.2.840.10008.6.1.95' => 'Peripheral Pulse Method',
    '1.2.840.10008.6.1.96' => 'Skin Condition',
    '1.2.840.10008.6.1.97' => 'Airway Assessment',
    '1.2.840.10008.6.1.98' => 'Calibration Object',
    '1.2.840.10008.6.1.99' => 'Calibration Method',
    '1.2.840.10008.6.1.100' => 'Cardiac Volume Method',
    '1.2.840.10008.6.1.101' => 'Index Method',
    '1.2.840.10008.6.1.102' => 'Sub-segment Method',
    '1.2.840.10008.6.1.103' => 'Contour Realignment',
    '1.2.840.10008.6.1.104' => 'Circumferential Extent',
    '1.2.840.10008.6.1.105' => 'Regional Extent',
    '1.2.840.10008.6.1.106' => 'Chamber Identification',
    '1.2.840.10008.6.1.107' => 'QA Reference Method',
    '1.2.840.10008.6.1.108' => 'Plane Identification',
    '1.2.840.10008.6.1.109' => 'Ejection Fraction',
    '1.2.840.10008.6.1.110' => 'ED Volume',
    '1.2.840.10008.6.1.111' => 'ES Volume',
    '1.2.840.10008.6.1.112' => 'Vessel Lumen Cross-sectional Area Calculation Method',
    '1.2.840.10008.6.1.113' => 'Estimated Volume',
    '1.2.840.10008.6.1.114' => 'Cardiac Contraction Phase',
    '1.2.840.10008.6.1.115' => 'IVUS Procedure Phase',
    '1.2.840.10008.6.1.116' => 'IVUS Distance Measurement',
    '1.2.840.10008.6.1.117' => 'IVUS Area Measurement',
    '1.2.840.10008.6.1.118' => 'IVUS Longitudinal Measurement',
    '1.2.840.10008.6.1.119' => 'IVUS Index/Ratio',
    '1.2.840.10008.6.1.120' => 'IVUS Volume Measurement',
    '1.2.840.10008.6.1.121' => 'Vascular Measurement Site',
    '1.2.840.10008.6.1.122' => 'Intravascular Volumetric Region',
    '1.2.840.10008.6.1.123' => 'Min/Max/Mean',
    '1.2.840.10008.6.1.124' => 'Calcium Distribution',
    '1.2.840.10008.6.1.125' => 'IVUS Lesion Morphology',
    '1.2.840.10008.6.1.126' => 'Vascular Dissection Classification',
    '1.2.840.10008.6.1.127' => 'IVUS Relative Stenosis Severity',
    '1.2.840.10008.6.1.128' => 'IVUS Non Morphological Finding',
    '1.2.840.10008.6.1.129' => 'IVUS Plaque Composition',
    '1.2.840.10008.6.1.130' => 'IVUS Fiducial Point',
    '1.2.840.10008.6.1.131' => 'IVUS Arterial Morphology',
    '1.2.840.10008.6.1.132' => 'Pressure Unit',
    '1.2.840.10008.6.1.133' => 'Hemodynamic Resistance Unit',
    '1.2.840.10008.6.1.134' => 'Indexed Hemodynamic Resistance Unit',
    '1.2.840.10008.6.1.135' => 'Catheter Size Unit',
    '1.2.840.10008.6.1.136' => 'Specimen Collection',
    '1.2.840.10008.6.1.137' => 'Blood Source Type',
    '1.2.840.10008.6.1.138' => 'Blood Gas Pressure',
    '1.2.840.10008.6.1.139' => 'Blood Gas Content',
    '1.2.840.10008.6.1.140' => 'Blood Gas Saturation',
    '1.2.840.10008.6.1.141' => 'Blood Base Excess',
    '1.2.840.10008.6.1.142' => 'Blood pH',
    '1.2.840.10008.6.1.143' => 'Arterial / Venous Content',
    '1.2.840.10008.6.1.144' => 'Oxygen Administration Action',
    '1.2.840.10008.6.1.145' => 'Oxygen Administration',
    '1.2.840.10008.6.1.146' => 'Circulatory Support Action',
    '1.2.840.10008.6.1.147' => 'Ventilation Action',
    '1.2.840.10008.6.1.148' => 'Pacing Action',
    '1.2.840.10008.6.1.149' => 'Circulatory Support',
    '1.2.840.10008.6.1.150' => 'Ventilation',
    '1.2.840.10008.6.1.151' => 'Pacing',
    '1.2.840.10008.6.1.152' => 'Blood Pressure Method',
    '1.2.840.10008.6.1.153' => 'Relative Time',
    '1.2.840.10008.6.1.154' => 'Hemodynamic Patient State',
    '1.2.840.10008.6.1.155' => 'Arterial Lesion Location',
    '1.2.840.10008.6.1.156' => 'Arterial Source Location',
    '1.2.840.10008.6.1.157' => 'Venous Source Location',
    '1.2.840.10008.6.1.158' => 'Atrial Source Location',
    '1.2.840.10008.6.1.159' => 'Ventricular Source Location',
    '1.2.840.10008.6.1.160' => 'Gradient Source Location',
    '1.2.840.10008.6.1.161' => 'Pressure Measurement',
    '1.2.840.10008.6.1.162' => 'Blood Velocity Measurement',
    '1.2.840.10008.6.1.163' => 'Hemodynamic Time Measurement',
    '1.2.840.10008.6.1.164' => 'Non-mitral Valve Area',
    '1.2.840.10008.6.1.165' => 'Valve Area',
    '1.2.840.10008.6.1.166' => 'Hemodynamic Period Measurement',
    '1.2.840.10008.6.1.167' => 'Valve Flow',
    '1.2.840.10008.6.1.168' => 'Hemodynamic Flow',
    '1.2.840.10008.6.1.169' => 'Hemodynamic Resistance Measurement',
    '1.2.840.10008.6.1.170' => 'Hemodynamic Ratio',
    '1.2.840.10008.6.1.171' => 'Fractional Flow Reserve',
    '1.2.840.10008.6.1.172' => 'Measurement Type',
    '1.2.840.10008.6.1.173' => 'Cardiac Output Method',
    '1.2.840.10008.6.1.174' => 'Procedure Intent',
    '1.2.840.10008.6.1.175' => 'Cardiovascular Anatomic Location',
    '1.2.840.10008.6.1.176' => 'Hypertension',
    '1.2.840.10008.6.1.177' => 'Hemodynamic Assessment',
    '1.2.840.10008.6.1.178' => 'Degree Finding',
    '1.2.840.10008.6.1.179' => 'Hemodynamic Measurement Phase',
    '1.2.840.10008.6.1.180' => 'Body Surface Area Equation',
    '1.2.840.10008.6.1.181' => 'Oxygen Consumption Equation/Table',
    '1.2.840.10008.6.1.182' => 'P50 Equation',
    '1.2.840.10008.6.1.183' => 'Framingham Score',
    '1.2.840.10008.6.1.184' => 'Framingham Table',
    '1.2.840.10008.6.1.185' => 'ECG Procedure Type',
    '1.2.840.10008.6.1.186' => 'Reason for ECG Study',
    '1.2.840.10008.6.1.187' => 'Pacemaker',
    '1.2.840.10008.6.1.188' => 'CID 3673',
    '1.2.840.10008.6.1.189' => 'CID 3675',
    '1.2.840.10008.6.1.190' => 'Lead Measurement Technique',
    '1.2.840.10008.6.1.191' => 'Summary Codes ECG',
    '1.2.840.10008.6.1.192' => 'QT Correction Algorithm',
    '1.2.840.10008.6.1.193' => 'CID 3679',
    '1.2.840.10008.6.1.194' => 'ECG Lead Noise Description',
    '1.2.840.10008.6.1.195' => 'ECG Lead Noise Modifier (Retired)',
    '1.2.840.10008.6.1.196' => 'CID 3682',
    '1.2.840.10008.6.1.197' => 'CID 3683',
    '1.2.840.10008.6.1.198' => 'CID 3684',
    '1.2.840.10008.6.1.199' => 'CID 3685',
    '1.2.840.10008.6.1.200' => 'CID 3686',
    '1.2.840.10008.6.1.201' => 'Electrophysiology Waveform Duration',
    '1.2.840.10008.6.1.202' => 'Electrophysiology Waveform Voltage',
    '1.2.840.10008.6.1.203' => 'Cath Diagnosis',
    '1.2.840.10008.6.1.204' => 'Cardiac Valve/Tract',
    '1.2.840.10008.6.1.205' => 'Wall Motion',
    '1.2.840.10008.6.1.206' => 'Myocardium Wall Morphology Finding',
    '1.2.840.10008.6.1.207' => 'Chamber Size',
    '1.2.840.10008.6.1.208' => 'Overall Contractility',
    '1.2.840.10008.6.1.209' => 'VSD Description',
    '1.2.840.10008.6.1.210' => 'Aortic Root Description',
    '1.2.840.10008.6.1.211' => 'Coronary Dominance',
    '1.2.840.10008.6.1.212' => 'Valvular Abnormality',
    '1.2.840.10008.6.1.213' => 'Vessel Descriptor',
    '1.2.840.10008.6.1.214' => 'TIMI Flow Characteristic',
    '1.2.840.10008.6.1.215' => 'Thrombus',
    '1.2.840.10008.6.1.216' => 'Lesion Margin',
    '1.2.840.10008.6.1.217' => 'Severity',
    '1.2.840.10008.6.1.218' => 'Left Ventricle Myocardial Wall 17 Segment Model',
    '1.2.840.10008.6.1.219' => 'Myocardial Wall Segments in Projection',
    '1.2.840.10008.6.1.220' => 'Canadian Clinical Classification',
    '1.2.840.10008.6.1.221' => 'CID 3720',
    '1.2.840.10008.6.1.222' => 'Cardiovascular Surgery',
    '1.2.840.10008.6.1.223' => 'Diabetic Therapy',
    '1.2.840.10008.6.1.224' => 'MI Type',
    '1.2.840.10008.6.1.225' => 'Smoking History',
    '1.2.840.10008.6.1.226' => 'Coronary Intervention Indication',
    '1.2.840.10008.6.1.227' => 'Catheterization Indication',
    '1.2.840.10008.6.1.228' => 'Cath Finding',
    '1.2.840.10008.6.1.229' => 'Admission Status',
    '1.2.840.10008.6.1.230' => 'Insurance Payor',
    '1.2.840.10008.6.1.231' => 'Primary Cause of Death',
    '1.2.840.10008.6.1.232' => 'Acute Coronary Syndrome Time Period',
    '1.2.840.10008.6.1.233' => 'NYHA Classification',
    '1.2.840.10008.6.1.234' => 'Ischemia Non-invasive Test',
    '1.2.840.10008.6.1.235' => 'Pre-Cath Angina Type',
    '1.2.840.10008.6.1.236' => 'Cath Procedure Type',
    '1.2.840.10008.6.1.237' => 'Thrombolytic Administration',
    '1.2.840.10008.6.1.238' => 'Lab Visit Medication Administration',
    '1.2.840.10008.6.1.239' => 'PCI Medication Administration',
    '1.2.840.10008.6.1.240' => 'Clopidogrel/Ticlopidine Administration',
    '1.2.840.10008.6.1.241' => 'EF Testing Method',
    '1.2.840.10008.6.1.242' => 'Calculation Method',
    '1.2.840.10008.6.1.243' => 'Percutaneous Entry Site',
    '1.2.840.10008.6.1.244' => 'Percutaneous Closure',
    '1.2.840.10008.6.1.245' => 'Angiographic EF Testing Method',
    '1.2.840.10008.6.1.246' => 'PCI Procedure Result',
    '1.2.840.10008.6.1.247' => 'Previously Dilated Lesion',
    '1.2.840.10008.6.1.248' => 'Guidewire Crossing',
    '1.2.840.10008.6.1.249' => 'Vascular Complication',
    '1.2.840.10008.6.1.250' => 'Cath Complication',
    '1.2.840.10008.6.1.251' => 'Cardiac Patient Risk Factor',
    '1.2.840.10008.6.1.252' => 'Cardiac Diagnostic Procedure',
    '1.2.840.10008.6.1.253' => 'Cardiovascular Family History',
    '1.2.840.10008.6.1.254' => 'Hypertension Therapy',
    '1.2.840.10008.6.1.255' => 'Antilipemic Agent',
    '1.2.840.10008.6.1.256' => 'Antiarrhythmic Agent',
    '1.2.840.10008.6.1.257' => 'Myocardial Infarction Therapy',
    '1.2.840.10008.6.1.258' => 'Concern Type',
    '1.2.840.10008.6.1.259' => 'Problem Status',
    '1.2.840.10008.6.1.260' => 'Health Status',
    '1.2.840.10008.6.1.261' => 'Use Status',
    '1.2.840.10008.6.1.262' => 'Social History',
    '1.2.840.10008.6.1.263' => 'Cardiovascular Implant',
    '1.2.840.10008.6.1.264' => 'Plaque Structure',
    '1.2.840.10008.6.1.265' => 'Stenosis Measurement Method',
    '1.2.840.10008.6.1.266' => 'Stenosis Type',
    '1.2.840.10008.6.1.267' => 'Stenosis Shape',
    '1.2.840.10008.6.1.268' => 'Volume Measurement Method',
    '1.2.840.10008.6.1.269' => 'Aneurysm Type',
    '1.2.840.10008.6.1.270' => 'Associated Condition',
    '1.2.840.10008.6.1.271' => 'Vascular Morphology',
    '1.2.840.10008.6.1.272' => 'Stent Finding',
    '1.2.840.10008.6.1.273' => 'Stent Composition',
    '1.2.840.10008.6.1.274' => 'Source of Vascular Finding',
    '1.2.840.10008.6.1.275' => 'Vascular Sclerosis Type',
    '1.2.840.10008.6.1.276' => 'Non-invasive Vascular Procedure',
    '1.2.840.10008.6.1.277' => 'Papillary Muscle Included/Excluded',
    '1.2.840.10008.6.1.278' => 'Respiratory Status',
    '1.2.840.10008.6.1.279' => 'Heart Rhythm',
    '1.2.840.10008.6.1.280' => 'Vessel Segment',
    '1.2.840.10008.6.1.281' => 'Pulmonary Artery',
    '1.2.840.10008.6.1.282' => 'Stenosis Length',
    '1.2.840.10008.6.1.283' => 'Stenosis Grade',
    '1.2.840.10008.6.1.284' => 'Cardiac Ejection Fraction',
    '1.2.840.10008.6.1.285' => 'Cardiac Volume Measurement',
    '1.2.840.10008.6.1.286' => 'Time-based Perfusion Measurement',
    '1.2.840.10008.6.1.287' => 'Fiducial Feature',
    '1.2.840.10008.6.1.288' => 'Diameter Derivation',
    '1.2.840.10008.6.1.289' => 'Coronary Vein',
    '1.2.840.10008.6.1.290' => 'Pulmonary Vein',
    '1.2.840.10008.6.1.291' => 'Myocardial Subsegment',
    '1.2.840.10008.6.1.292' => 'Partial View Section for Mammography',
    '1.2.840.10008.6.1.293' => 'DX Anatomy Imaged',
    '1.2.840.10008.6.1.294' => 'DX View',
    '1.2.840.10008.6.1.295' => 'DX View Modifier',
    '1.2.840.10008.6.1.296' => 'Projection Eponymous Name',
    '1.2.840.10008.6.1.297' => 'Anatomic Region for Mammography',
    '1.2.840.10008.6.1.298' => 'View for Mammography',
    '1.2.840.10008.6.1.299' => 'View Modifier for Mammography',
    '1.2.840.10008.6.1.300' => 'Anatomic Region for Intra-oral Radiography',
    '1.2.840.10008.6.1.301' => 'Anatomic Region Modifier for Intra-oral Radiography',
    '1.2.840.10008.6.1.302' => 'Primary Anatomic Structure for Intra-oral Radiography (Permanent Dentition - Designation of Teeth)',
    '1.2.840.10008.6.1.303' => 'Primary Anatomic Structure for Intra-oral Radiography (Deciduous Dentition - Designation of Teeth)',
    '1.2.840.10008.6.1.304' => 'PET Radionuclide',
    '1.2.840.10008.6.1.305' => 'PET Radiopharmaceutical',
    '1.2.840.10008.6.1.306' => 'Craniofacial Anatomic Region',
    '1.2.840.10008.6.1.307' => 'CT, MR and PET Anatomy Imaged',
    '1.2.840.10008.6.1.308' => 'Common Anatomic Region',
    '1.2.840.10008.6.1.309' => 'MR Spectroscopy Metabolite',
    '1.2.840.10008.6.1.310' => 'MR Proton Spectroscopy Metabolite',
    '1.2.840.10008.6.1.311' => 'Endoscopy Anatomic Region',
    '1.2.840.10008.6.1.312' => 'XA/XRF Anatomy Imaged',
    '1.2.840.10008.6.1.313' => 'Drug or Contrast Agent Characteristic',
    '1.2.840.10008.6.1.314' => 'General Device',
    '1.2.840.10008.6.1.315' => 'Phantom Device',
    '1.2.840.10008.6.1.316' => 'Ophthalmic Imaging Agent',
    '1.2.840.10008.6.1.317' => 'Patient Eye Movement Command',
    '1.2.840.10008.6.1.318' => 'Ophthalmic Photography Acquisition Device',
    '1.2.840.10008.6.1.319' => 'Ophthalmic Photography Illumination',
    '1.2.840.10008.6.1.320' => 'Ophthalmic Filter',
    '1.2.840.10008.6.1.321' => 'Ophthalmic Lens',
    '1.2.840.10008.6.1.322' => 'Ophthalmic Channel Description',
    '1.2.840.10008.6.1.323' => 'Ophthalmic Image Position',
    '1.2.840.10008.6.1.324' => 'Mydriatic Agent',
    '1.2.840.10008.6.1.325' => 'Ophthalmic Anatomic Structure Imaged',
    '1.2.840.10008.6.1.326' => 'Ophthalmic Tomography Acquisition Device',
    '1.2.840.10008.6.1.327' => 'Ophthalmic OCT Anatomic Structure Imaged',
    '1.2.840.10008.6.1.328' => 'Language',
    '1.2.840.10008.6.1.329' => 'Country',
    '1.2.840.10008.6.1.330' => 'Overall Breast Composition',
    '1.2.840.10008.6.1.331' => 'Overall Breast Composition from BI-RADS',
    '1.2.840.10008.6.1.332' => 'Change Since Last Mammogram or Prior Surgery',
    '1.2.840.10008.6.1.333' => 'Change Since Last Mammogram or Prior Surgery from BI-RADS',
    '1.2.840.10008.6.1.334' => 'Mammography Shape Characteristic',
    '1.2.840.10008.6.1.335' => 'Shape Characteristic from BI-RADS',
    '1.2.840.10008.6.1.336' => 'Mammography Margin Characteristic',
    '1.2.840.10008.6.1.337' => 'Margin Characteristic from BI-RADS',
    '1.2.840.10008.6.1.338' => 'Density Modifier',
    '1.2.840.10008.6.1.339' => 'Density Modifier from BI-RADS',
    '1.2.840.10008.6.1.340' => 'Mammography Calcification Type',
    '1.2.840.10008.6.1.341' => 'Calcification Type from BI-RADS',
    '1.2.840.10008.6.1.342' => 'Calcification Distribution Modifier',
    '1.2.840.10008.6.1.343' => 'Calcification Distribution Modifier from BI-RADS',
    '1.2.840.10008.6.1.344' => 'Mammography Single Image Finding',
    '1.2.840.10008.6.1.345' => 'Single Image Finding from BI-RADS',
    '1.2.840.10008.6.1.346' => 'Mammography Composite Feature',
    '1.2.840.10008.6.1.347' => 'Composite Feature from BI-RADS',
    '1.2.840.10008.6.1.348' => 'Clockface Location or Region',
    '1.2.840.10008.6.1.349' => 'Clockface Location or Region from BI-RADS',
    '1.2.840.10008.6.1.350' => 'Quadrant Location',
    '1.2.840.10008.6.1.351' => 'Quadrant Location from BI-RADS',
    '1.2.840.10008.6.1.352' => 'Side',
    '1.2.840.10008.6.1.353' => 'Side from BI-RADS',
    '1.2.840.10008.6.1.354' => 'Depth',
    '1.2.840.10008.6.1.355' => 'Depth from BI-RADS',
    '1.2.840.10008.6.1.356' => 'Mammography Assessment',
    '1.2.840.10008.6.1.357' => 'Assessment from BI-RADS',
    '1.2.840.10008.6.1.358' => 'Mammography Recommended Follow-up',
    '1.2.840.10008.6.1.359' => 'Recommended Follow-up from BI-RADS',
    '1.2.840.10008.6.1.360' => 'Mammography Pathology Code',
    '1.2.840.10008.6.1.361' => 'Benign Pathology Code from BI-RADS',
    '1.2.840.10008.6.1.362' => 'High Risk Lesion Pathology Code from BI-RADS',
    '1.2.840.10008.6.1.363' => 'Malignant Pathology Code from BI-RADS',
    '1.2.840.10008.6.1.364' => 'CAD Output Intended Use',
    '1.2.840.10008.6.1.365' => 'Composite Feature Relation',
    '1.2.840.10008.6.1.366' => 'Feature Scope',
    '1.2.840.10008.6.1.367' => 'Mammography Quantitative Temporal Difference Type',
    '1.2.840.10008.6.1.368' => 'Mammography Qualitative Temporal Difference Type',
    '1.2.840.10008.6.1.369' => 'Nipple Characteristic',
    '1.2.840.10008.6.1.370' => 'Non-lesion Object Type',
    '1.2.840.10008.6.1.371' => 'Mammography Image Quality Finding',
    '1.2.840.10008.6.1.372' => 'Result Status',
    '1.2.840.10008.6.1.373' => 'Mammography CAD Analysis Type',
    '1.2.840.10008.6.1.374' => 'Image Quality Assessment Type',
    '1.2.840.10008.6.1.375' => 'Mammography Quality Control Standard Type',
    '1.2.840.10008.6.1.376' => 'Follow-up Interval Unit',
    '1.2.840.10008.6.1.377' => 'CAD Processing and Finding Summary',
    '1.2.840.10008.6.1.378' => 'CAD Operating Point Axis Label',
    '1.2.840.10008.6.1.379' => 'Breast Procedure Reported',
    '1.2.840.10008.6.1.380' => 'Breast Procedure Reason',
    '1.2.840.10008.6.1.381' => 'Breast Imaging Report Section Title',
    '1.2.840.10008.6.1.382' => 'Breast Imaging Report Element',
    '1.2.840.10008.6.1.383' => 'Breast Imaging Finding',
    '1.2.840.10008.6.1.384' => 'Breast Clinical Finding or Indicated Problem',
    '1.2.840.10008.6.1.385' => 'Associated Finding for Breast',
    '1.2.840.10008.6.1.386' => 'Ductography Finding for Breast',
    '1.2.840.10008.6.1.387' => 'Procedure Modifiers for Breast',
    '1.2.840.10008.6.1.388' => 'Breast Implant Type',
    '1.2.840.10008.6.1.389' => 'Breast Biopsy Technique',
    '1.2.840.10008.6.1.390' => 'Breast Imaging Procedure Modifier',
    '1.2.840.10008.6.1.391' => 'Interventional Procedure Complication',
    '1.2.840.10008.6.1.392' => 'Interventional Procedure Result',
    '1.2.840.10008.6.1.393' => 'Ultrasound Finding for Breast',
    '1.2.840.10008.6.1.394' => 'Instrument Approach',
    '1.2.840.10008.6.1.395' => 'Target Confirmation',
    '1.2.840.10008.6.1.396' => 'Fluid Color',
    '1.2.840.10008.6.1.397' => 'Tumor Stages From AJCC',
    '1.2.840.10008.6.1.398' => 'Nottingham Combined Histologic Grade',
    '1.2.840.10008.6.1.399' => 'Bloom-Richardson Histologic Grade',
    '1.2.840.10008.6.1.400' => 'Histologic Grading Method',
    '1.2.840.10008.6.1.401' => 'Breast Implant Finding',
    '1.2.840.10008.6.1.402' => 'Gynecological Hormone',
    '1.2.840.10008.6.1.403' => 'Breast Cancer Risk Factor',
    '1.2.840.10008.6.1.404' => 'Gynecological Procedure',
    '1.2.840.10008.6.1.405' => 'Procedures for Breast',
    '1.2.840.10008.6.1.406' => 'Mammoplasty Procedure',
    '1.2.840.10008.6.1.407' => 'Therapies for Breast',
    '1.2.840.10008.6.1.408' => 'Menopausal Phase',
    '1.2.840.10008.6.1.409' => 'General Risk Factor',
    '1.2.840.10008.6.1.410' => 'OB-GYN Maternal Risk Factor',
    '1.2.840.10008.6.1.411' => 'Substance',
    '1.2.840.10008.6.1.412' => 'Relative Usage/Exposure Amount',
    '1.2.840.10008.6.1.413' => 'Relative Frequency of Event Value',
    '1.2.840.10008.6.1.414' => 'Usage/Exposure Qualitative Concept',
    '1.2.840.10008.6.1.415' => 'Usage/Exposure/Amount Qualitative Concept',
    '1.2.840.10008.6.1.416' => 'Usage/Exposure/Frequency Qualitative Concept',
    '1.2.840.10008.6.1.417' => 'Procedure Numeric Property',
    '1.2.840.10008.6.1.418' => 'Pregnancy Status',
    '1.2.840.10008.6.1.419' => 'Side of Family',
    '1.2.840.10008.6.1.420' => 'Chest Component Category',
    '1.2.840.10008.6.1.421' => 'Chest Finding or Feature',
    '1.2.840.10008.6.1.422' => 'Chest Finding or Feature Modifier',
    '1.2.840.10008.6.1.423' => 'Abnormal Lines Finding or Feature',
    '1.2.840.10008.6.1.424' => 'Abnormal Opacity Finding or Feature',
    '1.2.840.10008.6.1.425' => 'Abnormal Lucency Finding or Feature',
    '1.2.840.10008.6.1.426' => 'Abnormal Texture Finding or Feature',
    '1.2.840.10008.6.1.427' => 'Width Descriptor',
    '1.2.840.10008.6.1.428' => 'Chest Anatomic Structure Abnormal Distribution',
    '1.2.840.10008.6.1.429' => 'Radiographic Anatomy Finding or Feature',
    '1.2.840.10008.6.1.430' => 'Lung Anatomy Finding or Feature',
    '1.2.840.10008.6.1.431' => 'Bronchovascular Anatomy Finding or Feature',
    '1.2.840.10008.6.1.432' => 'Pleura Anatomy Finding or Feature',
    '1.2.840.10008.6.1.433' => 'Mediastinum Anatomy Finding or Feature',
    '1.2.840.10008.6.1.434' => 'Osseous Anatomy Finding or Feature',
    '1.2.840.10008.6.1.435' => 'Osseous Anatomy Modifier',
    '1.2.840.10008.6.1.436' => 'Muscular Anatomy',
    '1.2.840.10008.6.1.437' => 'Vascular Anatomy',
    '1.2.840.10008.6.1.438' => 'Size Descriptor',
    '1.2.840.10008.6.1.439' => 'Chest Border Shape',
    '1.2.840.10008.6.1.440' => 'Chest Border Definition',
    '1.2.840.10008.6.1.441' => 'Chest Orientation Descriptor',
    '1.2.840.10008.6.1.442' => 'Chest Content Descriptor',
    '1.2.840.10008.6.1.443' => 'Chest Opacity Descriptor',
    '1.2.840.10008.6.1.444' => 'Location in Chest',
    '1.2.840.10008.6.1.445' => 'General Chest Location',
    '1.2.840.10008.6.1.446' => 'Location in Lung',
    '1.2.840.10008.6.1.447' => 'Segment Location in Lung',
    '1.2.840.10008.6.1.448' => 'Chest Distribution Descriptor',
    '1.2.840.10008.6.1.449' => 'Chest Site Involvement',
    '1.2.840.10008.6.1.450' => 'Severity Descriptor',
    '1.2.840.10008.6.1.451' => 'Chest Texture Descriptor',
    '1.2.840.10008.6.1.452' => 'Chest Calcification Descriptor',
    '1.2.840.10008.6.1.453' => 'Chest Quantitative Temporal Difference Type',
    '1.2.840.10008.6.1.454' => 'Chest Qualitative Temporal Difference Type',
    '1.2.840.10008.6.1.455' => 'Image Quality Finding',
    '1.2.840.10008.6.1.456' => 'Chest Types of Quality Control Standard',
    '1.2.840.10008.6.1.457' => 'CAD Analysis Type',
    '1.2.840.10008.6.1.458' => 'Chest Non-lesion Object Type',
    '1.2.840.10008.6.1.459' => 'Non-lesion Modifier',
    '1.2.840.10008.6.1.460' => 'Calculation Method',
    '1.2.840.10008.6.1.461' => 'Attenuation Coefficient Measurement',
    '1.2.840.10008.6.1.462' => 'Calculated Value',
    '1.2.840.10008.6.1.463' => 'Lesion Response',
    '1.2.840.10008.6.1.464' => 'RECIST Defined Lesion Response',
    '1.2.840.10008.6.1.465' => 'Baseline Category',
    '1.2.840.10008.6.1.466' => 'Background Echotexture',
    '1.2.840.10008.6.1.467' => 'Orientation',
    '1.2.840.10008.6.1.468' => 'Lesion Boundary',
    '1.2.840.10008.6.1.469' => 'Echo Pattern',
    '1.2.840.10008.6.1.470' => 'Posterior Acoustic Feature',
    '1.2.840.10008.6.1.471' => 'Vascularity',
    '1.2.840.10008.6.1.472' => 'Correlation to Other Finding',
    '1.2.840.10008.6.1.473' => 'Malignancy Type',
    '1.2.840.10008.6.1.474' => 'Breast Primary Tumor Assessment From AJCC',
    '1.2.840.10008.6.1.475' => 'Pathological Regional Lymph Node Assessment for Breast',
    '1.2.840.10008.6.1.476' => 'Assessment of Metastasis for Breast',
    '1.2.840.10008.6.1.477' => 'Menstrual Cycle Phase',
    '1.2.840.10008.6.1.478' => 'Time Interval',
    '1.2.840.10008.6.1.479' => 'Breast Linear Measurement',
    '1.2.840.10008.6.1.480' => 'CAD Geometry Secondary Graphical Representation',
    '1.2.840.10008.6.1.481' => 'Diagnostic Imaging Report Document Title',
    '1.2.840.10008.6.1.482' => 'Diagnostic Imaging Report Heading',
    '1.2.840.10008.6.1.483' => 'Diagnostic Imaging Report Element',
    '1.2.840.10008.6.1.484' => 'Diagnostic Imaging Report Purpose of Reference',
    '1.2.840.10008.6.1.485' => 'Waveform Purpose of Reference',
    '1.2.840.10008.6.1.486' => 'Contributing Equipment Purpose of Reference',
    '1.2.840.10008.6.1.487' => 'SR Document Purpose of Reference',
    '1.2.840.10008.6.1.488' => 'Signature Purpose',
    '1.2.840.10008.6.1.489' => 'Media Import',
    '1.2.840.10008.6.1.490' => 'Key Object Selection Document Title',
    '1.2.840.10008.6.1.491' => 'Rejected for Quality Reason',
    '1.2.840.10008.6.1.492' => 'Best in Set',
    '1.2.840.10008.6.1.493' => 'Document Title',
    '1.2.840.10008.6.1.494' => 'RCS Registration Method Type',
    '1.2.840.10008.6.1.495' => 'Brain Atlas Fiducial',
    '1.2.840.10008.6.1.496' => 'Segmentation Property Category',
    '1.2.840.10008.6.1.497' => 'Segmentation Property Type',
    '1.2.840.10008.6.1.498' => 'Cardiac Structure Segmentation Type',
    '1.2.840.10008.6.1.499' => 'CNS Segmentation Type',
    '1.2.840.10008.6.1.500' => 'Abdominal Segmentation Type',
    '1.2.840.10008.6.1.501' => 'Thoracic Segmentation Type',
    '1.2.840.10008.6.1.502' => 'Vascular Segmentation Type',
    '1.2.840.10008.6.1.503' => 'Device Segmentation Type',
    '1.2.840.10008.6.1.504' => 'Artifact Segmentation Type',
    '1.2.840.10008.6.1.505' => 'Lesion Segmentation Type',
    '1.2.840.10008.6.1.506' => 'Pelvic Organ Segmentation Type',
    '1.2.840.10008.6.1.507' => 'Physiology Segmentation Type',
    '1.2.840.10008.6.1.508' => 'Referenced Image Purpose of Reference',
    '1.2.840.10008.6.1.509' => 'Source Image Purpose of Reference',
    '1.2.840.10008.6.1.510' => 'Image Derivation',
    '1.2.840.10008.6.1.511' => 'Purpose of Reference to Alternate Representation',
    '1.2.840.10008.6.1.512' => 'Related Series Purpose of Reference',
    '1.2.840.10008.6.1.513' => 'Multi-Frame Subset Type',
    '1.2.840.10008.6.1.514' => 'Person Role',
    '1.2.840.10008.6.1.515' => 'Family Member',
    '1.2.840.10008.6.1.516' => 'Organizational Role',
    '1.2.840.10008.6.1.517' => 'Performing Role',
    '1.2.840.10008.6.1.518' => 'Animal Taxonomic Rank Value',
    '1.2.840.10008.6.1.519' => 'Sex',
    '1.2.840.10008.6.1.520' => 'Age Unit',
    '1.2.840.10008.6.1.521' => 'Linear Measurement Unit',
    '1.2.840.10008.6.1.522' => 'Area Measurement Unit',
    '1.2.840.10008.6.1.523' => 'Volume Measurement Unit',
    '1.2.840.10008.6.1.524' => 'Linear Measurement',
    '1.2.840.10008.6.1.525' => 'Area Measurement',
    '1.2.840.10008.6.1.526' => 'Volume Measurement',
    '1.2.840.10008.6.1.527' => 'General Area Calculation Method',
    '1.2.840.10008.6.1.528' => 'General Volume Calculation Method',
    '1.2.840.10008.6.1.529' => 'Breed',
    '1.2.840.10008.6.1.530' => 'Breed Registry',
    '1.2.840.10008.6.1.531' => 'Workitem Definition',
    '1.2.840.10008.6.1.532' => 'CID 9232',
    '1.2.840.10008.6.1.533' => 'Procedure Discontinuation Reason',
    '1.2.840.10008.6.1.534' => 'Scope of Accumulation',
    '1.2.840.10008.6.1.535' => 'UID Type',
    '1.2.840.10008.6.1.536' => 'Irradiation Event Type',
    '1.2.840.10008.6.1.537' => 'Equipment Plane Identification',
    '1.2.840.10008.6.1.538' => 'Fluoro Mode',
    '1.2.840.10008.6.1.539' => 'X-Ray Filter Material',
    '1.2.840.10008.6.1.540' => 'X-Ray Filter Type',
    '1.2.840.10008.6.1.541' => 'Dose Related Distance Measurement',
    '1.2.840.10008.6.1.542' => 'Measured/Calculated',
    '1.2.840.10008.6.1.543' => 'Dose Measurement Device',
    '1.2.840.10008.6.1.544' => 'Effective Dose Evaluation Method',
    '1.2.840.10008.6.1.545' => 'CT Acquisition Type',
    '1.2.840.10008.6.1.546' => 'CT IV Contrast Imaging Technique',
    '1.2.840.10008.6.1.547' => 'CT Dose Reference Authority',
    '1.2.840.10008.6.1.548' => 'Anode Target Material',
    '1.2.840.10008.6.1.549' => 'X-Ray Grid',
    '1.2.840.10008.6.1.550' => 'Ultrasound Protocol Type',
    '1.2.840.10008.6.1.551' => 'Ultrasound Protocol Stage Type',
    '1.2.840.10008.6.1.552' => 'OB-GYN Date',
    '1.2.840.10008.6.1.553' => 'Fetal Biometry Ratio',
    '1.2.840.10008.6.1.554' => 'Fetal Biometry Measurement',
    '1.2.840.10008.6.1.555' => 'Fetal Long Bones Biometry Measurement',
    '1.2.840.10008.6.1.556' => 'Fetal Cranium Measurement',
    '1.2.840.10008.6.1.557' => 'OB-GYN Amniotic Sac Measurement',
    '1.2.840.10008.6.1.558' => 'Early Gestation Biometry Measurement',
    '1.2.840.10008.6.1.559' => 'Ultrasound Pelvis and Uterus Measurement',
    '1.2.840.10008.6.1.560' => 'OB Equation/Table',
    '1.2.840.10008.6.1.561' => 'Gestational Age Equation/Table',
    '1.2.840.10008.6.1.562' => 'OB Fetal Body Weight Equation/Table',
    '1.2.840.10008.6.1.563' => 'Fetal Growth Equation/Table',
    '1.2.840.10008.6.1.564' => 'Estimated Fetal Weight Percentile Equation/Table',
    '1.2.840.10008.6.1.565' => 'Growth Distribution Rank',
    '1.2.840.10008.6.1.566' => 'OB-GYN Summary',
    '1.2.840.10008.6.1.567' => 'OB-GYN Fetus Summary',
    '1.2.840.10008.6.1.568' => 'Vascular Summary',
    '1.2.840.10008.6.1.569' => 'Temporal Period Relating to Procedure or Therapy',
    '1.2.840.10008.6.1.570' => 'Vascular Ultrasound Anatomic Location',
    '1.2.840.10008.6.1.571' => 'Extracranial Artery',
    '1.2.840.10008.6.1.572' => 'Intracranial Cerebral Vessel',
    '1.2.840.10008.6.1.573' => 'Intracranial Cerebral Vessel (Unilateral)',
    '1.2.840.10008.6.1.574' => 'Upper Extremity Artery',
    '1.2.840.10008.6.1.575' => 'Upper Extremity Vein',
    '1.2.840.10008.6.1.576' => 'Lower Extremity Artery',
    '1.2.840.10008.6.1.577' => 'Lower Extremity Vein',
    '1.2.840.10008.6.1.578' => 'Abdominopelvic Artery (Paired)',
    '1.2.840.10008.6.1.579' => 'Abdominopelvic Artery (Unpaired)',
    '1.2.840.10008.6.1.580' => 'Abdominopelvic Vein (Paired)',
    '1.2.840.10008.6.1.581' => 'Abdominopelvic Vein (Unpaired)',
    '1.2.840.10008.6.1.582' => 'Renal Vessel',
    '1.2.840.10008.6.1.583' => 'Vessel Segment Modifier',
    '1.2.840.10008.6.1.584' => 'Vessel Branch Modifier',
    '1.2.840.10008.6.1.585' => 'Vascular Ultrasound Property',
    '1.2.840.10008.6.1.586' => 'Ultrasound Blood Velocity Measurement',
    '1.2.840.10008.6.1.587' => 'Vascular Index/Ratio',
    '1.2.840.10008.6.1.588' => 'Other Vascular Property',
    '1.2.840.10008.6.1.589' => 'Carotid Ratio',
    '1.2.840.10008.6.1.590' => 'Renal Ratio',
    '1.2.840.10008.6.1.591' => 'Pelvic Vasculature Anatomical Location',
    '1.2.840.10008.6.1.592' => 'Fetal Vasculature Anatomical Location',
    '1.2.840.10008.6.1.593' => 'Echocardiography Left Ventricle Measurement',
    '1.2.840.10008.6.1.594' => 'Left Ventricle Linear Measurement',
    '1.2.840.10008.6.1.595' => 'Left Ventricle Volume Measurement',
    '1.2.840.10008.6.1.596' => 'Left Ventricle Other Measurement',
    '1.2.840.10008.6.1.597' => 'Echocardiography Right Ventricle Measurement',
    '1.2.840.10008.6.1.598' => 'Echocardiography Left Atrium Measurement',
    '1.2.840.10008.6.1.599' => 'Echocardiography Right Atrium Measurement',
    '1.2.840.10008.6.1.600' => 'Echocardiography Mitral Valve Measurement',
    '1.2.840.10008.6.1.601' => 'Echocardiography Tricuspid Valve Measurement',
    '1.2.840.10008.6.1.602' => 'Echocardiography Pulmonic Valve Measurement',
    '1.2.840.10008.6.1.603' => 'Echocardiography Pulmonary Artery Measurement',
    '1.2.840.10008.6.1.604' => 'Echocardiography Aortic Valve Measurement',
    '1.2.840.10008.6.1.605' => 'Echocardiography Aorta Measurement',
    '1.2.840.10008.6.1.606' => 'Echocardiography Pulmonary Vein Measurement',
    '1.2.840.10008.6.1.607' => 'Echocardiography Vena Cava Measurement',
    '1.2.840.10008.6.1.608' => 'Echocardiography Hepatic Vein Measurement',
    '1.2.840.10008.6.1.609' => 'Echocardiography Cardiac Shunt Measurement',
    '1.2.840.10008.6.1.610' => 'Echocardiography Congenital Anomaly Measurement',
    '1.2.840.10008.6.1.611' => 'Pulmonary Vein Modifier',
    '1.2.840.10008.6.1.612' => 'Echocardiography Common Measurement',
    '1.2.840.10008.6.1.613' => 'Flow Direction',
    '1.2.840.10008.6.1.614' => 'Orifice Flow Property',
    '1.2.840.10008.6.1.615' => 'Echocardiography Stroke Volume Origin',
    '1.2.840.10008.6.1.616' => 'Ultrasound Image Mode',
    '1.2.840.10008.6.1.617' => 'Echocardiography Image View',
    '1.2.840.10008.6.1.618' => 'Echocardiography Measurement Method',
    '1.2.840.10008.6.1.619' => 'Echocardiography Volume Method',
    '1.2.840.10008.6.1.620' => 'Echocardiography Area Method',
    '1.2.840.10008.6.1.621' => 'Gradient Method',
    '1.2.840.10008.6.1.622' => 'Volume Flow Method',
    '1.2.840.10008.6.1.623' => 'Myocardium Mass Method',
    '1.2.840.10008.6.1.624' => 'Cardiac Phase',
    '1.2.840.10008.6.1.625' => 'Respiration State',
    '1.2.840.10008.6.1.626' => 'Mitral Valve Anatomic Site',
    '1.2.840.10008.6.1.627' => 'Echocardiography Anatomic Site',
    '1.2.840.10008.6.1.628' => 'Echocardiography Anatomic Site Modifier',
    '1.2.840.10008.6.1.629' => 'Wall Motion Scoring Scheme',
    '1.2.840.10008.6.1.630' => 'Cardiac Output Property',
    '1.2.840.10008.6.1.631' => 'Left Ventricle Area Measurement',
    '1.2.840.10008.6.1.632' => 'Tricuspid Valve Finding Site',
    '1.2.840.10008.6.1.633' => 'Aortic Valve Finding Site',
    '1.2.840.10008.6.1.634' => 'Left Ventricle Finding Site',
    '1.2.840.10008.6.1.635' => 'Congenital Finding Site',
    '1.2.840.10008.6.1.636' => 'Surface Processing Algorithm Family',
    '1.2.840.10008.6.1.637' => 'Stress Test Procedure Phase',
    '1.2.840.10008.6.1.638' => 'Stage',
    '1.2.840.10008.6.1.735' => 'S-M-L Size Descriptor',
    '1.2.840.10008.6.1.736' => 'Major Coronary Artery',
    '1.2.840.10008.6.1.737' => 'Radioactivity Unit',
    '1.2.840.10008.6.1.738' => 'Rest/Stress State',
    '1.2.840.10008.6.1.739' => 'PET Cardiology Protocol',
    '1.2.840.10008.6.1.740' => 'PET Cardiology Radiopharmaceutical',
    '1.2.840.10008.6.1.741' => 'NM/PET Procedure',
    '1.2.840.10008.6.1.742' => 'Nuclear Cardiology Protocol',
    '1.2.840.10008.6.1.743' => 'Nuclear Cardiology Radiopharmaceutical',
    '1.2.840.10008.6.1.744' => 'Attenuation Correction',
    '1.2.840.10008.6.1.745' => 'Perfusion Defect Type',
    '1.2.840.10008.6.1.746' => 'Study Quality',
    '1.2.840.10008.6.1.747' => 'Stress Imaging Quality Issue',
    '1.2.840.10008.6.1.748' => 'NM Extracardiac Finding',
    '1.2.840.10008.6.1.749' => 'Attenuation Correction Method',
    '1.2.840.10008.6.1.750' => 'Level of Risk',
    '1.2.840.10008.6.1.751' => 'LV Function',
    '1.2.840.10008.6.1.752' => 'Perfusion Finding',
    '1.2.840.10008.6.1.753' => 'Perfusion Morphology',
    '1.2.840.10008.6.1.754' => 'Ventricular Enlargement',
    '1.2.840.10008.6.1.755' => 'Stress Test Procedure',
    '1.2.840.10008.6.1.756' => 'Indications for Stress Test',
    '1.2.840.10008.6.1.757' => 'Chest Pain',
    '1.2.840.10008.6.1.758' => 'Exerciser Device',
    '1.2.840.10008.6.1.759' => 'Stress Agent',
    '1.2.840.10008.6.1.760' => 'Indications for Pharmacological Stress Test',
    '1.2.840.10008.6.1.761' => 'Non-invasive Cardiac Imaging Procedure',
    '1.2.840.10008.6.1.763' => 'Exercise ECG Summary Code',
    '1.2.840.10008.6.1.764' => 'Stress Imaging Summary Code',
    '1.2.840.10008.6.1.765' => 'Speed of Response',
    '1.2.840.10008.6.1.766' => 'BP Response',
    '1.2.840.10008.6.1.767' => 'Treadmill Speed',
    '1.2.840.10008.6.1.768' => 'Stress Hemodynamic Finding',
    '1.2.840.10008.6.1.769' => 'Perfusion Finding Method',
    '1.2.840.10008.6.1.770' => 'Comparison Finding',
    '1.2.840.10008.6.1.771' => 'Stress Symptom',
    '1.2.840.10008.6.1.772' => 'Stress Test Termination Reason',
    '1.2.840.10008.6.1.773' => 'QTc Measurement',
    '1.2.840.10008.6.1.774' => 'ECG Timing Measurement',
    '1.2.840.10008.6.1.775' => 'ECG Axis Measurement',
    '1.2.840.10008.6.1.776' => 'ECG Finding',
    '1.2.840.10008.6.1.777' => 'ST Segment Finding',
    '1.2.840.10008.6.1.778' => 'ST Segment Location',
    '1.2.840.10008.6.1.779' => 'ST Segment Morphology',
    '1.2.840.10008.6.1.780' => 'Ectopic Beat Morphology',
    '1.2.840.10008.6.1.781' => 'Perfusion Comparison Finding',
    '1.2.840.10008.6.1.782' => 'Tolerance Comparison Finding',
    '1.2.840.10008.6.1.783' => 'Wall Motion Comparison Finding',
    '1.2.840.10008.6.1.784' => 'Stress Scoring Scale',
    '1.2.840.10008.6.1.785' => 'Perceived Exertion Scale',
    '1.2.840.10008.6.1.786' => 'Ventricle Identification',
    '1.2.840.10008.6.1.787' => 'Colon Overall Assessment',
    '1.2.840.10008.6.1.788' => 'Colon Finding or Feature',
    '1.2.840.10008.6.1.789' => 'Colon Finding or Feature Modifier',
    '1.2.840.10008.6.1.790' => 'Colon Non-lesion Object Type',
    '1.2.840.10008.6.1.791' => 'Anatomic Non-colon Finding',
    '1.2.840.10008.6.1.792' => 'Clockface Location for Colon',
    '1.2.840.10008.6.1.793' => 'Recumbent Patient Orientation for Colon',
    '1.2.840.10008.6.1.794' => 'Colon Quantitative Temporal Difference Type',
    '1.2.840.10008.6.1.795' => 'Colon Types of Quality Control Standard',
    '1.2.840.10008.6.1.796' => 'Colon Morphology Descriptor',
    '1.2.840.10008.6.1.797' => 'Location in Intestinal Tract',
    '1.2.840.10008.6.1.798' => 'Colon CAD Material Description',
    '1.2.840.10008.6.1.799' => 'Calculated Value for Colon Finding',
    '1.2.840.10008.6.1.800' => 'Ophthalmic Horizontal Direction',
    '1.2.840.10008.6.1.801' => 'Ophthalmic Vertical Direction',
    '1.2.840.10008.6.1.802' => 'Ophthalmic Visual Acuity Type',
    '1.2.840.10008.6.1.803' => 'Arterial Pulse Waveform',
    '1.2.840.10008.6.1.804' => 'Respiration Waveform',
    '1.2.840.10008.6.1.805' => 'Ultrasound Contrast/Bolus Agent',
    '1.2.840.10008.6.1.806' => 'Protocol Interval Event',
    '1.2.840.10008.6.1.807' => 'Transducer Scan Pattern',
    '1.2.840.10008.6.1.808' => 'Ultrasound Transducer Geometry',
    '1.2.840.10008.6.1.809' => 'Ultrasound Transducer Beam Steering',
    '1.2.840.10008.6.1.810' => 'Ultrasound Transducer Application',
    '1.2.840.10008.6.1.811' => 'Instance Availability Status',
    '1.2.840.10008.6.1.812' => 'Modality PPS Discontinuation Reason',
    '1.2.840.10008.6.1.813' => 'Media Import PPS Discontinuation Reason',
    '1.2.840.10008.6.1.814' => 'DX Anatomy Imaged for Animal',
    '1.2.840.10008.6.1.815' => 'Common Anatomic Regions for Animal',
    '1.2.840.10008.6.1.816' => 'DX View for Animal',
    '1.2.840.10008.6.1.817' => 'Institutional Department/Unit/Service',
    '1.2.840.10008.6.1.818' => 'Purpose of Reference to Predecessor Report',
    '1.2.840.10008.6.1.819' => 'Visual Fixation Quality During Acquisition',
    '1.2.840.10008.6.1.820' => 'Visual Fixation Quality Problem',
    '1.2.840.10008.6.1.821' => 'Ophthalmic Macular Grid Problem',
    '1.2.840.10008.6.1.822' => 'Organization',
    '1.2.840.10008.6.1.823' => 'Mixed Breed',
    '1.2.840.10008.6.1.824' => 'Broselow-Luten Pediatric Size Category',
    '1.2.840.10008.6.1.825' => 'CMDCTECC Calcium Scoring Patient Size Category',
    '1.2.840.10008.6.1.826' => 'Cardiac Ultrasound Report Title',
    '1.2.840.10008.6.1.827' => 'Cardiac Ultrasound Indication for Study',
    '1.2.840.10008.6.1.828' => 'Pediatric, Fetal and Congenital Cardiac Surgical Intervention',
    '1.2.840.10008.6.1.829' => 'Cardiac Ultrasound Summary Code',
    '1.2.840.10008.6.1.830' => 'Cardiac Ultrasound Fetal Summary Code',
    '1.2.840.10008.6.1.831' => 'Cardiac Ultrasound Common Linear Measurement',
    '1.2.840.10008.6.1.832' => 'Cardiac Ultrasound Linear Valve Measurement',
    '1.2.840.10008.6.1.833' => 'Cardiac Ultrasound Cardiac Function',
    '1.2.840.10008.6.1.834' => 'Cardiac Ultrasound Area Measurement',
    '1.2.840.10008.6.1.835' => 'Cardiac Ultrasound Hemodynamic Measurement',
    '1.2.840.10008.6.1.836' => 'CID 12255',
    '1.2.840.10008.6.1.838' => 'Cardiac Ultrasound Left Ventricle Measurement',
    '1.2.840.10008.6.1.839' => 'Cardiac Ultrasound Right Ventricle Measurement',
    '1.2.840.10008.6.1.840' => 'Cardiac Ultrasound Ventricles Measurement',
    '1.2.840.10008.6.1.841' => 'Cardiac Ultrasound Pulmonary Artery Measurement',
    '1.2.840.10008.6.1.842' => 'Cardiac Ultrasound Pulmonary Vein',
    '1.2.840.10008.6.1.843' => 'Cardiac Ultrasound Pulmonary Valve Measurement',
    '1.2.840.10008.6.1.844' => 'Cardiac Ultrasound Venous Return Pulmonary Measurement',
    '1.2.840.10008.6.1.845' => 'Cardiac Ultrasound Venous Return Systemic Measurement',
    '1.2.840.10008.6.1.846' => 'Cardiac Ultrasound Atria and Atrial Septum Measurement',
    '1.2.840.10008.6.1.847' => 'Cardiac Ultrasound Mitral Valve Measurement',
    '1.2.840.10008.6.1.848' => 'Cardiac Ultrasound Tricuspid Valve Measurement',
    '1.2.840.10008.6.1.849' => 'Cardiac Ultrasound Atrioventricular Valve Measurement',
    '1.2.840.10008.6.1.850' => 'Cardiac Ultrasound Interventricular Septum Measurement',
    '1.2.840.10008.6.1.851' => 'Cardiac Ultrasound Aortic Valve Measurement',
    '1.2.840.10008.6.1.852' => 'Cardiac Ultrasound Outflow Tract Measurement',
    '1.2.840.10008.6.1.853' => 'Cardiac Ultrasound Semilunar Valve, Annulate and Sinus Measurement',
    '1.2.840.10008.6.1.854' => 'Cardiac Ultrasound Aortic Sinotubular Junction Measurement',
    '1.2.840.10008.6.1.855' => 'Cardiac Ultrasound Aorta Measurement',
    '1.2.840.10008.6.1.856' => 'Cardiac Ultrasound Coronary Artery Measurement',
    '1.2.840.10008.6.1.857' => 'Cardiac Ultrasound Aorto Pulmonary Connection Measurement',
    '1.2.840.10008.6.1.858' => 'Cardiac Ultrasound Pericardium and Pleura Measurement',
    '1.2.840.10008.6.1.859' => 'Cardiac Ultrasound Fetal General Measurement',
    '1.2.840.10008.6.1.860' => 'Cardiac Ultrasound Target Site',
    '1.2.840.10008.6.1.861' => 'Cardiac Ultrasound Target Site Modifier',
    '1.2.840.10008.6.1.862' => 'Cardiac Ultrasound Venous Return Systemic Finding Site',
    '1.2.840.10008.6.1.863' => 'Cardiac Ultrasound Venous Return Pulmonary Finding Site',
    '1.2.840.10008.6.1.864' => 'Cardiac Ultrasound Atria and Atrial Septum Finding Site',
    '1.2.840.10008.6.1.865' => 'Cardiac Ultrasound Atrioventricular Valve Finding Site',
    '1.2.840.10008.6.1.866' => 'Cardiac Ultrasound Interventricular Septum Finding Site',
    '1.2.840.10008.6.1.867' => 'Cardiac Ultrasound Ventricle Finding Site',
    '1.2.840.10008.6.1.868' => 'Cardiac Ultrasound Outflow Tract Finding Site',
    '1.2.840.10008.6.1.869' => 'Cardiac Ultrasound Semilunar Valve, Annulus and Sinus Finding Site',
    '1.2.840.10008.6.1.870' => 'Cardiac Ultrasound Pulmonary Artery Finding Site',
    '1.2.840.10008.6.1.871' => 'Cardiac Ultrasound Aorta Finding Site',
    '1.2.840.10008.6.1.872' => 'Cardiac Ultrasound Coronary Artery Finding Site',
    '1.2.840.10008.6.1.873' => 'Cardiac Ultrasound Aortopulmonary Connection Finding Site',
    '1.2.840.10008.6.1.874' => 'Cardiac Ultrasound Pericardium and Pleura Finding Site',
    '1.2.840.10008.6.1.876' => 'Ophthalmic Ultrasound Axial Measurements Type',
    '1.2.840.10008.6.1.877' => 'Lens Status',
    '1.2.840.10008.6.1.878' => 'Vitreous Status',
    '1.2.840.10008.6.1.879' => 'Ophthalmic Axial Length Measurements Segment Name',
    '1.2.840.10008.6.1.880' => 'Refractive Surgery Type',
    '1.2.840.10008.6.1.881' => 'Keratometry Descriptor',
    '1.2.840.10008.6.1.882' => 'IOL Calculation Formula',
    '1.2.840.10008.6.1.883' => 'Lens Constant Type',
    '1.2.840.10008.6.1.884' => 'Refractive Error Type',
    '1.2.840.10008.6.1.885' => 'Anterior Chamber Depth Definition',
    '1.2.840.10008.6.1.886' => 'Ophthalmic Measurement or Calculation Data Source',
    '1.2.840.10008.6.1.887' => 'CID 4241',
    '1.2.840.10008.6.1.889' => 'Ophthalmic Quality Metric Type',
    '1.2.840.10008.6.1.890' => 'Ophthalmic Agent Concentration Unit',
    '1.2.840.10008.6.1.891' => 'Functional Condition Present During Acquisition',
    '1.2.840.10008.6.1.892' => 'Joint Position During Acquisition',
    '1.2.840.10008.6.1.893' => 'Joint Positioning Method',
    '1.2.840.10008.6.1.894' => 'Physical Force Applied During Acquisition',
    '1.2.840.10008.6.1.895' => 'ECG Control Numeric Variable',
    '1.2.840.10008.6.1.896' => 'ECG Control Text Variable',
    '1.2.840.10008.6.1.897' => 'Whole Slide Microscopy Image Referenced Image Purpose of Reference',
    '1.2.840.10008.6.1.898' => 'Microscopy Lens Type',
    '1.2.840.10008.6.1.899' => 'Microscopy Illuminator and Sensor Color',
    '1.2.840.10008.6.1.900' => 'Microscopy Illumination Method',
    '1.2.840.10008.6.1.901' => 'Microscopy Filter',
    '1.2.840.10008.6.1.902' => 'Microscopy Illuminator Type',
    '1.2.840.10008.6.1.903' => 'Audit Event ID',
    '1.2.840.10008.6.1.904' => 'Audit Event Type Code',
    '1.2.840.10008.6.1.905' => 'Audit Active Participant Role ID Code',
    '1.2.840.10008.6.1.906' => 'Security Alert Type Code',
    '1.2.840.10008.6.1.907' => 'Audit Participant Object ID Type Code',
    '1.2.840.10008.6.1.908' => 'Media Type Code',
    '1.2.840.10008.6.1.909' => 'Visual Field Static Perimetry Test Pattern',
    '1.2.840.10008.6.1.910' => 'Visual Field Static Perimetry Test Strategy',
    '1.2.840.10008.6.1.911' => 'Visual Field Static Perimetry Screening Test Mode',
    '1.2.840.10008.6.1.912' => 'Visual Field Static Perimetry Fixation Strategy',
    '1.2.840.10008.6.1.913' => 'Visual Field Static Perimetry Test Analysis Result',
    '1.2.840.10008.6.1.914' => 'Visual Field Illumination Color',
    '1.2.840.10008.6.1.915' => 'Visual Field Procedure Modifier',
    '1.2.840.10008.6.1.916' => 'Visual Field Global Index Name',
    '1.2.840.10008.6.1.917' => 'Abstract Multi-dimensional Image Model Component Semantic',
    '1.2.840.10008.6.1.918' => 'Abstract Multi-dimensional Image Model Component Unit',
    '1.2.840.10008.6.1.919' => 'Abstract Multi-dimensional Image Model Dimension Semantic',
    '1.2.840.10008.6.1.920' => 'Abstract Multi-dimensional Image Model Dimension Unit',
    '1.2.840.10008.6.1.921' => 'Abstract Multi-dimensional Image Model Axis Direction',
    '1.2.840.10008.6.1.922' => 'Abstract Multi-dimensional Image Model Axis Orientation',
    '1.2.840.10008.6.1.923' => 'Abstract Multi-dimensional Image Model Qualitative Dimension Sample Semantic',
    '1.2.840.10008.6.1.924' => 'Planning Method',
    '1.2.840.10008.6.1.925' => 'De-identification Method',
    '1.2.840.10008.6.1.926' => 'Measurement Orientation',
    '1.2.840.10008.6.1.927' => 'ECG Global Waveform Duration',
    '1.2.840.10008.6.1.930' => 'ICD',
    '1.2.840.10008.6.1.931' => 'Radiotherapy General Workitem Definition',
    '1.2.840.10008.6.1.932' => 'Radiotherapy Acquisition Workitem Definition',
    '1.2.840.10008.6.1.933' => 'Radiotherapy Registration Workitem Definition',
    '1.2.840.10008.6.1.934' => 'Contrast Bolus Substance',
    '1.2.840.10008.6.1.935' => 'Label Type',
    '1.2.840.10008.6.1.936' => 'Ophthalmic Mapping Unit for Real World Value Mapping',
    '1.2.840.10008.6.1.937' => 'Ophthalmic Mapping Acquisition Method',
    '1.2.840.10008.6.1.938' => 'Retinal Thickness Definition',
    '1.2.840.10008.6.1.939' => 'Ophthalmic Thickness Map Value Type',
    '1.2.840.10008.6.1.940' => 'Ophthalmic Map Purpose of Reference',
    '1.2.840.10008.6.1.941' => 'Ophthalmic Thickness Deviation Category',
    '1.2.840.10008.6.1.942' => 'Ophthalmic Anatomic Structure Reference Point',
    '1.2.840.10008.6.1.943' => 'Cardiac Synchronization Technique',
    '1.2.840.10008.6.1.944' => 'CID 8130',
    '1.2.840.10008.6.1.947' => 'Size Specific Dose Estimation Method for CT',
    '1.2.840.10008.6.1.948' => 'Pathology Imaging Protocol',
    '1.2.840.10008.6.1.949' => 'Magnification Selection',
    '1.2.840.10008.6.1.950' => 'Tissue Selection',
    '1.2.840.10008.6.1.951' => 'General Region of Interest Measurement Modifier',
    '1.2.840.10008.6.1.952' => 'Measurement Derived From Multiple ROI Measurements',
    '1.2.840.10008.6.1.953' => 'Surface Scan Acquisition Type',
    '1.2.840.10008.6.1.954' => 'Surface Scan Mode Type',
    '1.2.840.10008.6.1.956' => 'Surface Scan Registration Method Type',
    '1.2.840.10008.6.1.957' => 'Basic Cardiac View',
    '1.2.840.10008.6.1.958' => 'CT Reconstruction Algorithm',
    '1.2.840.10008.6.1.959' => 'Detector Type',
    '1.2.840.10008.6.1.960' => 'CR/DR Mechanical Configuration',
    '1.2.840.10008.6.1.961' => 'Projection X-Ray Acquisition Device Type',
    '1.2.840.10008.6.1.962' => 'Abstract Segmentation Type',
    '1.2.840.10008.6.1.963' => 'Common Tissue Segmentation Type',
    '1.2.840.10008.6.1.964' => 'Peripheral Nervous System Segmentation Type',
    '1.2.840.10008.6.1.965' => 'Corneal Topography Mapping Unit for Real World Value Mapping',
    '1.2.840.10008.6.1.966' => 'Corneal Topography Map Value Type',
    '1.2.840.10008.6.1.967' => 'Brain Structure for Volumetric Measurement',
    '1.2.840.10008.6.1.968' => 'RT Dose Derivation',
    '1.2.840.10008.6.1.969' => 'RT Dose Purpose of Reference',
    '1.2.840.10008.6.1.970' => 'Spectroscopy Purpose of Reference',
    '1.2.840.10008.6.1.971' => 'Scheduled Processing Parameter Concept Codes for RT Treatment',
    '1.2.840.10008.6.1.972' => 'Radiopharmaceutical Organ Dose Reference Authority',
    '1.2.840.10008.6.1.973' => 'Source of Radioisotope Activity Information',
    '1.2.840.10008.6.1.975' => 'Intravenous Extravasation Symptom',
    '1.2.840.10008.6.1.976' => 'Radiosensitive Organ',
    '1.2.840.10008.6.1.977' => 'Radiopharmaceutical Patient State',
    '1.2.840.10008.6.1.978' => 'GFR Measurement',
    '1.2.840.10008.6.1.979' => 'GFR Measurement Method',
    '1.2.840.10008.6.1.980' => 'Visual Evaluation Method',
    '1.2.840.10008.6.1.981' => 'Test Pattern Code',
    '1.2.840.10008.6.1.982' => 'Measurement Pattern Code',
    '1.2.840.10008.6.1.983' => 'Display Device Type',
    '1.2.840.10008.6.1.984' => 'SUV Unit',
    '1.2.840.10008.6.1.985' => 'T1 Measurement Method',
    '1.2.840.10008.6.1.986' => 'Tracer Kinetic Model',
    '1.2.840.10008.6.1.987' => 'Perfusion Measurement Method',
    '1.2.840.10008.6.1.988' => 'Arterial Input Function Measurement Method',
    '1.2.840.10008.6.1.989' => 'Bolus Arrival Time Derivation Method',
    '1.2.840.10008.6.1.990' => 'Perfusion Analysis Method',
    '1.2.840.10008.6.1.991' => 'Quantitative Method Used for Perfusion and Tracer Kinetic Model',
    '1.2.840.10008.6.1.992' => 'Tracer Kinetic Model Parameter',
    '1.2.840.10008.6.1.993' => 'Perfusion Model Parameter',
    '1.2.840.10008.6.1.994' => 'Model-Independent Dynamic Contrast Analysis Parameter',
    '1.2.840.10008.6.1.995' => 'Tracer Kinetic Modeling Covariate',
    '1.2.840.10008.6.1.996' => 'Contrast Characteristic',
    '1.2.840.10008.6.1.997' => 'Measurement Report Document Title',
    '1.2.840.10008.6.1.998' => 'Quantitative Diagnostic Imaging Procedure',
    '1.2.840.10008.6.1.999' => 'PET Region of Interest Measurement',
    '1.2.840.10008.6.1.1000' => 'Gray Level Co-occurrence Matrix Measurement',
    '1.2.840.10008.6.1.1001' => 'Texture Measurement',
    '1.2.840.10008.6.1.1002' => 'Time Point Type',
    '1.2.840.10008.6.1.1003' => 'Generic Intensity and Size Measurement',
    '1.2.840.10008.6.1.1004' => 'Response Criteria',
    '1.2.840.10008.6.1.1005' => 'Fetal Biometry Anatomic Site',
    '1.2.840.10008.6.1.1006' => 'Fetal Long Bone Anatomic Site',
    '1.2.840.10008.6.1.1007' => 'Fetal Cranium Anatomic Site',
    '1.2.840.10008.6.1.1008' => 'Pelvis and Uterus Anatomic Site',
    '1.2.840.10008.6.1.1009' => 'Parametric Map Derivation Image Purpose of Reference',
    '1.2.840.10008.6.1.1010' => 'Physical Quantity Descriptor',
    '1.2.840.10008.6.1.1011' => 'Lymph Node Anatomic Site',
    '1.2.840.10008.6.1.1012' => 'Head and Neck Cancer Anatomic Site',
    '1.2.840.10008.6.1.1013' => 'Fiber Tract In Brainstem',
    '1.2.840.10008.6.1.1014' => 'Projection and Thalamic Fiber',
    '1.2.840.10008.6.1.1015' => 'Association Fiber',
    '1.2.840.10008.6.1.1016' => 'Limbic System Tract',
    '1.2.840.10008.6.1.1017' => 'Commissural Fiber',
    '1.2.840.10008.6.1.1018' => 'Cranial Nerve',
    '1.2.840.10008.6.1.1019' => 'Spinal Cord Fiber',
    '1.2.840.10008.6.1.1020' => 'Tractography Anatomic Site',
    '1.2.840.10008.6.1.1021' => 'Primary Anatomic Structure for Intra-oral Radiography (Supernumerary Dentition - Designation of Teeth)',
    '1.2.840.10008.6.1.1022' => 'Primary Anatomic Structure for Intra-oral and Craniofacial Radiography - Teeth',
    '1.2.840.10008.6.1.1023' => 'IEC61217 Device Position Parameter',
    '1.2.840.10008.6.1.1024' => 'IEC61217 Gantry Position Parameter',
    '1.2.840.10008.6.1.1025' => 'IEC61217 Patient Support Position Parameter',
    '1.2.840.10008.6.1.1026' => 'Actionable Finding Classification',
    '1.2.840.10008.6.1.1027' => 'Image Quality Assessment',
    '1.2.840.10008.6.1.1028' => 'Summary Radiation Exposure Quantity',
    '1.2.840.10008.6.1.1029' => 'Wide Field Ophthalmic Photography Transformation Method',
    '1.2.840.10008.6.1.1030' => 'PET Unit',
    '1.2.840.10008.6.1.1031' => 'Implant Material',
    '1.2.840.10008.6.1.1032' => 'Intervention Type',
    '1.2.840.10008.6.1.1033' => 'Implant Template View Orientation',
    '1.2.840.10008.6.1.1034' => 'Implant Template Modified View Orientation',
    '1.2.840.10008.6.1.1035' => 'Implant Target Anatomy',
    '1.2.840.10008.6.1.1036' => 'Implant Planning Landmark',
    '1.2.840.10008.6.1.1037' => 'Human Hip Implant Planning Landmark',
    '1.2.840.10008.6.1.1038' => 'Implant Component Type',
    '1.2.840.10008.6.1.1039' => 'Human Hip Implant Component Type',
    '1.2.840.10008.6.1.1040' => 'Human Trauma Implant Component Type',
    '1.2.840.10008.6.1.1041' => 'Implant Fixation Method',
    '1.2.840.10008.6.1.1042' => 'Device Participating Role',
    '1.2.840.10008.6.1.1043' => 'Container Type',
    '1.2.840.10008.6.1.1044' => 'Container Component Type',
    '1.2.840.10008.6.1.1045' => 'Anatomic Pathology Specimen Type',
    '1.2.840.10008.6.1.1046' => 'Breast Tissue Specimen Type',
    '1.2.840.10008.6.1.1047' => 'Specimen Collection Procedure',
    '1.2.840.10008.6.1.1048' => 'Specimen Sampling Procedure',
    '1.2.840.10008.6.1.1049' => 'Specimen Preparation Procedure',
    '1.2.840.10008.6.1.1050' => 'Specimen Stain',
    '1.2.840.10008.6.1.1051' => 'Specimen Preparation Step',
    '1.2.840.10008.6.1.1052' => 'Specimen Fixative',
    '1.2.840.10008.6.1.1053' => 'Specimen Embedding Media',
    '1.2.840.10008.6.1.1054' => 'Source of Projection X-Ray Dose Information',
    '1.2.840.10008.6.1.1055' => 'Source of CT Dose Information',
    '1.2.840.10008.6.1.1056' => 'Radiation Dose Reference Point',
    '1.2.840.10008.6.1.1057' => 'Volumetric View Description',
    '1.2.840.10008.6.1.1058' => 'Volumetric View Modifier',
    '1.2.840.10008.6.1.1059' => 'Diffusion Acquisition Value Type',
    '1.2.840.10008.6.1.1060' => 'Diffusion Model Value Type',
    '1.2.840.10008.6.1.1061' => 'Diffusion Tractography Algorithm Family',
    '1.2.840.10008.6.1.1062' => 'Diffusion Tractography Measurement Type',
    '1.2.840.10008.6.1.1063' => 'Research Animal Source Registry',
    '1.2.840.10008.6.1.1064' => 'Yes-No Only',
    '1.2.840.10008.6.1.1065' => 'Biosafety Level',
    '1.2.840.10008.6.1.1066' => 'Biosafety Control Reason',
    '1.2.840.10008.6.1.1067' => 'Sex - Male Female or Both',
    '1.2.840.10008.6.1.1068' => 'Animal Room Type',
    '1.2.840.10008.6.1.1069' => 'Device Reuse',
    '1.2.840.10008.6.1.1070' => 'Animal Bedding Material',
    '1.2.840.10008.6.1.1071' => 'Animal Shelter Type',
    '1.2.840.10008.6.1.1072' => 'Animal Feed Type',
    '1.2.840.10008.6.1.1073' => 'Animal Feed Source',
    '1.2.840.10008.6.1.1074' => 'Animal Feeding Method',
    '1.2.840.10008.6.1.1075' => 'Water Type',
    '1.2.840.10008.6.1.1076' => 'Anesthesia Category Code Type for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1077' => 'Anesthesia Category Code Type from Anesthesia Quality Initiative',
    '1.2.840.10008.6.1.1078' => 'Anesthesia Induction Code Type for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1079' => 'Anesthesia Induction Code Type from Anesthesia Quality Initiative',
    '1.2.840.10008.6.1.1080' => 'Anesthesia Maintenance Code Type for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1081' => 'Anesthesia Maintenance Code Type from Anesthesia Quality Initiative',
    '1.2.840.10008.6.1.1082' => 'Airway Management Method Code Type for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1083' => 'Airway Management Method Code Type from Anesthesia Quality Initiative',
    '1.2.840.10008.6.1.1084' => 'Airway Management Sub-Method Code Type for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1085' => 'Airway Management Sub-Method Code Type from Anesthesia Quality Initiative',
    '1.2.840.10008.6.1.1086' => 'Medication Type for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1087' => 'Medication Type Code Type from Anesthesia Quality Initiative',
    '1.2.840.10008.6.1.1088' => 'Medication for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1089' => 'Inhalational Anesthesia Agent for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1090' => 'Injectable Anesthesia Agent for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1091' => 'Premedication Agent for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1092' => 'Neuromuscular Blocking Agent for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1093' => 'Ancillary Medications for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1094' => 'Carrier Gases for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1095' => 'Local Anesthetics for Small Animal Anesthesia',
    '1.2.840.10008.6.1.1096' => 'Procedure Phase Requiring Anesthesia',
    '1.2.840.10008.6.1.1097' => 'Surgical Procedure Phase Requiring Anesthesia',
    '1.2.840.10008.6.1.1098' => 'CID 633',
    '1.2.840.10008.6.1.1099' => 'Animal Handling Phase',
    '1.2.840.10008.6.1.1100' => 'Heating Method',
    '1.2.840.10008.6.1.1101' => 'Temperature Sensor Device Component Type for Small Animal Procedure',
    '1.2.840.10008.6.1.1102' => 'Exogenous Substance Type',
    '1.2.840.10008.6.1.1103' => 'Exogenous Substance',
    '1.2.840.10008.6.1.1104' => 'Tumor Graft Histologic Type',
    '1.2.840.10008.6.1.1105' => 'Fibril',
    '1.2.840.10008.6.1.1106' => 'Virus',
    '1.2.840.10008.6.1.1107' => 'Cytokine',
    '1.2.840.10008.6.1.1108' => 'Toxin',
    '1.2.840.10008.6.1.1109' => 'Exogenous Substance Administration Site',
    '1.2.840.10008.6.1.1110' => 'Exogenous Substance Origin Tissue',
    '1.2.840.10008.6.1.1111' => 'Preclinical Small Animal Imaging Procedure',
    '1.2.840.10008.6.1.1112' => 'Position Reference Indicator for Frame of Reference',
    '1.2.840.10008.6.1.1113' => 'Present-Absent Only',
    '1.2.840.10008.6.1.1114' => 'Water Equivalent Diameter Method',
    '1.2.840.10008.6.1.1115' => 'Radiotherapy Purpose of Reference',
    '1.2.840.10008.6.1.1116' => 'Content Assessment Type',
    '1.2.840.10008.6.1.1117' => 'RT Content Assessment Type',
    '1.2.840.10008.6.1.1118' => 'Assessment Basis',
    '1.2.840.10008.6.1.1119' => 'Reader Specialty',
    '1.2.840.10008.6.1.1120' => 'Requested Report Type',
    '1.2.840.10008.6.1.1121' => 'CT Transverse Plane Reference Basis',
    '1.2.840.10008.6.1.1122' => 'Anatomical Reference Basis',
    '1.2.840.10008.6.1.1123' => 'Anatomical Reference Basis - Head',
    '1.2.840.10008.6.1.1124' => 'Anatomical Reference Basis - Spine',
    '1.2.840.10008.6.1.1125' => 'Anatomical Reference Basis - Chest',
    '1.2.840.10008.6.1.1126' => 'Anatomical Reference Basis - Abdomen/Pelvis',
    '1.2.840.10008.6.1.1127' => 'Anatomical Reference Basis - Extremity',
    '1.2.840.10008.6.1.1128' => 'Reference Geometry - Plane',
    '1.2.840.10008.6.1.1129' => 'Reference Geometry - Point',
    '1.2.840.10008.6.1.1130' => 'Patient Alignment Method',
    '1.2.840.10008.6.1.1131' => 'Contraindications For CT Imaging',
    '1.2.840.10008.6.1.1132' => 'Fiducial Category',
    '1.2.840.10008.6.1.1133' => 'Fiducial',
    '1.2.840.10008.6.1.1134' => 'Non-Image Source Instance Purpose of Reference',
    '1.2.840.10008.6.1.1135' => 'RT Process Output',
    '1.2.840.10008.6.1.1136' => 'RT Process Input',
    '1.2.840.10008.6.1.1137' => 'RT Process Input Used',
    '1.2.840.10008.6.1.1138' => 'Prostate Anatomy',
    '1.2.840.10008.6.1.1139' => 'Prostate Sector Anatomy from PI-RADS v2',
    '1.2.840.10008.6.1.1140' => 'Prostate Sector Anatomy from European Concensus 16 Sector (Minimal) Model',
    '1.2.840.10008.6.1.1141' => 'Prostate Sector Anatomy from European Concensus 27 Sector (Optimal) Model',
    '1.2.840.10008.6.1.1142' => 'Measurement Selection Reason',
    '1.2.840.10008.6.1.1143' => 'Echo Finding Observation Type',
    '1.2.840.10008.6.1.1144' => 'Echo Measurement Type',
    '1.2.840.10008.6.1.1145' => 'Cardiovascular Measured Property',
    '1.2.840.10008.6.1.1146' => 'Basic Echo Anatomic Site',
    '1.2.840.10008.6.1.1147' => 'Echo Flow Direction',
    '1.2.840.10008.6.1.1148' => 'Cardiac Phase and Time Point',
    '1.2.840.10008.6.1.1149' => 'Core Echo Measurement',
    '1.2.840.10008.6.1.1150' => 'OCT-A Processing Algorithm Family',
    '1.2.840.10008.6.1.1151' => 'En Face Image Type',
    '1.2.840.10008.6.1.1152' => 'OPT Scan Pattern Type',
    '1.2.840.10008.6.1.1153' => 'Retinal Segmentation Surface',
    '1.2.840.10008.6.1.1154' => 'Organ for Radiation Dose Estimate',
    '1.2.840.10008.6.1.1155' => 'Absorbed Radiation Dose Type',
    '1.2.840.10008.6.1.1156' => 'Equivalent Radiation Dose Type',
    '1.2.840.10008.6.1.1157' => 'Radiation Dose Estimate Distribution Representation',
    '1.2.840.10008.6.1.1158' => 'Patient Model Type',
    '1.2.840.10008.6.1.1159' => 'Radiation Transport Model Type',
    '1.2.840.10008.6.1.1160' => 'Attenuator Category',
    '1.2.840.10008.6.1.1161' => 'Radiation Attenuator Material',
    '1.2.840.10008.6.1.1162' => 'Estimate Method Type',
    '1.2.840.10008.6.1.1163' => 'Radiation Dose Estimate Parameter',
    '1.2.840.10008.6.1.1164' => 'Radiation Dose Type',
    '1.2.840.10008.6.1.1165' => 'MR Diffusion Component Semantic',
    '1.2.840.10008.6.1.1166' => 'MR Diffusion Anisotropy Index',
    '1.2.840.10008.6.1.1167' => 'MR Diffusion Model Parameter',
    '1.2.840.10008.6.1.1168' => 'MR Diffusion Model',
    '1.2.840.10008.6.1.1169' => 'MR Diffusion Model Fitting Method',
    '1.2.840.10008.6.1.1170' => 'MR Diffusion Model Specific Method',
    '1.2.840.10008.6.1.1171' => 'MR Diffusion Model Input',
    '1.2.840.10008.6.1.1172' => 'Diffusion Rate Area Over Time Unit',
    '1.2.840.10008.6.1.1173' => 'Pediatric Size Category',
    '1.2.840.10008.6.1.1174' => 'Calcium Scoring Patient Size Category',
    '1.2.840.10008.6.1.1175' => 'Reason for Repeating Acquisition',
    '1.2.840.10008.6.1.1176' => 'Protocol Assertion',
    '1.2.840.10008.6.1.1177' => 'Radiotherapeutic Dose Measurement Device',
    '1.2.840.10008.6.1.1178' => 'Export Additional Information Document Title',
    '1.2.840.10008.6.1.1179' => 'Export Delay Reason',
    '1.2.840.10008.6.1.1180' => 'Level of Difficulty',
    '1.2.840.10008.6.1.1181' => 'Category of Teaching Material - Imaging',
    '1.2.840.10008.6.1.1182' => 'Miscellaneous Document Title',
    '1.2.840.10008.6.1.1183' => 'Segmentation Non-Image Source Purpose of Reference',
    '1.2.840.10008.6.1.1184' => 'Longitudinal Temporal Event Type',
    '1.2.840.10008.6.1.1185' => 'Non-lesion Object Type - Physical Object',
    '1.2.840.10008.6.1.1186' => 'Non-lesion Object Type - Substance',
    '1.2.840.10008.6.1.1187' => 'Non-lesion Object Type - Tissue',
    '1.2.840.10008.6.1.1188' => 'Chest Non-lesion Object Type - Physical Object',
    '1.2.840.10008.6.1.1189' => 'Chest Non-lesion Object Type - Tissue',
    '1.2.840.10008.6.1.1190' => 'Tissue Segmentation Property Type',
    '1.2.840.10008.6.1.1191' => 'Anatomical Structure Segmentation Property Type',
    '1.2.840.10008.6.1.1192' => 'Physical Object Segmentation Property Type',
    '1.2.840.10008.6.1.1193' => 'Morphologically Abnormal Structure Segmentation Property Type',
    '1.2.840.10008.6.1.1194' => 'Function Segmentation Property Type',
    '1.2.840.10008.6.1.1195' => 'Spatial and Relational Concept Segmentation Property Type',
    '1.2.840.10008.6.1.1196' => 'Body Substance Segmentation Property Type',
    '1.2.840.10008.6.1.1197' => 'Substance Segmentation Property Type',
    '1.2.840.10008.6.1.1198' => 'Interpretation Request Discontinuation Reason',
    '1.2.840.10008.6.1.1199' => 'Gray Level Run Length Based Feature',
    '1.2.840.10008.6.1.1200' => 'Gray Level Size Zone Based Feature',
    '1.2.840.10008.6.1.1201' => 'Encapsulated Document Source Purpose of Reference',
    '1.2.840.10008.6.1.1202' => 'Model Document Title',
    '1.2.840.10008.6.1.1203' => 'Purpose of Reference to Predecessor 3D Model',
    '1.2.840.10008.6.1.1204' => 'Model Scale Unit',
    '1.2.840.10008.6.1.1205' => 'Model Usage',
    '1.2.840.10008.6.1.1206' => 'Radiation Dose Unit',
    '1.2.840.10008.6.1.1207' => 'Radiotherapy Fiducial',
    '1.2.840.10008.6.1.1208' => 'Multi-energy Relevant Material',
    '1.2.840.10008.6.1.1209' => 'Multi-energy Material Unit',
    '1.2.840.10008.6.1.1210' => 'Dosimetric Objective Type',
    '1.2.840.10008.6.1.1211' => 'Prescription Anatomy Category',
    '1.2.840.10008.6.1.1212' => 'RT Segment Annotation Category',
    '1.2.840.10008.6.1.1213' => 'Radiotherapy Therapeutic Role Category',
    '1.2.840.10008.6.1.1214' => 'RT Geometric Information',
    '1.2.840.10008.6.1.1215' => 'Fixation or Positioning Device',
    '1.2.840.10008.6.1.1216' => 'Brachytherapy Device',
    '1.2.840.10008.6.1.1217' => 'External Body Model',
    '1.2.840.10008.6.1.1218' => 'Non-specific Volume',
    '1.2.840.10008.6.1.1219' => 'Purpose of Reference For RT Physician Intent Input',
    '1.2.840.10008.6.1.1220' => 'Purpose of Reference For RT Treatment Planning Input',
    '1.2.840.10008.6.1.1221' => 'General External Radiotherapy Procedure Technique',
    '1.2.840.10008.6.1.1222' => 'Tomotherapeutic Radiotherapy Procedure Technique',
    '1.2.840.10008.6.1.1223' => 'Fixation Device',
    '1.2.840.10008.6.1.1224' => 'Anatomical Structure For Radiotherapy',
    '1.2.840.10008.6.1.1225' => 'RT Patient Support Device',
    '1.2.840.10008.6.1.1226' => 'Radiotherapy Bolus Device Type',
    '1.2.840.10008.6.1.1227' => 'Radiotherapy Block Device Type',
    '1.2.840.10008.6.1.1228' => 'Radiotherapy Accessory No-slot Holder Device Type',
    '1.2.840.10008.6.1.1229' => 'Radiotherapy Accessory Slot Holder Device Type',
    '1.2.840.10008.6.1.1230' => 'Segmented RT Accessory Device',
    '1.2.840.10008.6.1.1231' => 'Radiotherapy Treatment Energy Unit',
    '1.2.840.10008.6.1.1232' => 'Multi-source Radiotherapy Procedure Technique',
    '1.2.840.10008.6.1.1233' => 'Robotic Radiotherapy Procedure Technique',
    '1.2.840.10008.6.1.1234' => 'Radiotherapy Procedure Technique',
    '1.2.840.10008.6.1.1235' => 'Radiation Therapy Particle',
    '1.2.840.10008.6.1.1236' => 'Ion Therapy Particle',
    '1.2.840.10008.6.1.1237' => 'Teletherapy Isotope',
    '1.2.840.10008.6.1.1238' => 'Brachytherapy Isotope',
    '1.2.840.10008.6.1.1239' => 'Single Dose Dosimetric Objective',
    '1.2.840.10008.6.1.1240' => 'Percentage and Dose Dosimetric Objective',
    '1.2.840.10008.6.1.1241' => 'Volume and Dose Dosimetric Objective',
    '1.2.840.10008.6.1.1242' => 'No-Parameter Dosimetric Objective',
    '1.2.840.10008.6.1.1243' => 'Delivery Time Structure',
    '1.2.840.10008.6.1.1244' => 'Radiotherapy Target',
    '1.2.840.10008.6.1.1245' => 'Radiotherapy Dose Calculation Role',
    '1.2.840.10008.6.1.1246' => 'Radiotherapy Prescribing and Segmenting Person Role',
    '1.2.840.10008.6.1.1247' => 'Effective Dose Calculation Method Category',
    '1.2.840.10008.6.1.1248' => 'Radiation Transport-based Effective Dose Method Modifier',
    '1.2.840.10008.6.1.1249' => 'Fractionation-based Effective Dose Method Modifier',
    '1.2.840.10008.6.1.1250' => 'Imaging Agent Administration Adverse Event',
    '1.2.840.10008.6.1.1251' => 'CID 61',
    '1.2.840.10008.6.1.1252' => 'Imaging Agent Administration Phase Type',
    '1.2.840.10008.6.1.1253' => 'Imaging Agent Administration Mode',
    '1.2.840.10008.6.1.1254' => 'Imaging Agent Administration Patient State',
    '1.2.840.10008.6.1.1255' => 'Imaging Agent Administration Premedication',
    '1.2.840.10008.6.1.1256' => 'Imaging Agent Administration Medication',
    '1.2.840.10008.6.1.1257' => 'Imaging Agent Administration Completion Status',
    '1.2.840.10008.6.1.1258' => 'Imaging Agent Administration Pharmaceutical Presentation Unit',
    '1.2.840.10008.6.1.1259' => 'Imaging Agent Administration Consumable',
    '1.2.840.10008.6.1.1260' => 'Flush',
    '1.2.840.10008.6.1.1261' => 'Imaging Agent Administration Injector Event Type',
    '1.2.840.10008.6.1.1262' => 'Imaging Agent Administration Step Type',
    '1.2.840.10008.6.1.1263' => 'Bolus Shaping Curve',
    '1.2.840.10008.6.1.1264' => 'Imaging Agent Administration Consumable Catheter Type',
    '1.2.840.10008.6.1.1265' => 'Low High or Equal',
    '1.2.840.10008.6.1.1266' => 'Premedication Type',
    '1.2.840.10008.6.1.1267' => 'Laterality with Median',
    '1.2.840.10008.6.1.1268' => 'Dermatology Anatomic Site',
    '1.2.840.10008.6.1.1269' => 'Quantitative Image Feature',
    '1.2.840.10008.6.1.1270' => 'Global Shape Descriptor',
    '1.2.840.10008.6.1.1271' => 'Intensity Histogram Feature',
    '1.2.840.10008.6.1.1272' => 'Grey Level Distance Zone Based Feature',
    '1.2.840.10008.6.1.1273' => 'Neighbourhood Grey Tone Difference Based Feature',
    '1.2.840.10008.6.1.1274' => 'Neighbouring Grey Level Dependence Based Feature',
    '1.2.840.10008.6.1.1275' => 'Cornea Measurement Method Descriptor',
    '1.2.840.10008.6.1.1276' => 'Segmented Radiotherapeutic Dose Measurement Device',
    '1.2.840.10008.6.1.1277' => 'Clinical Course of Disease',
    '1.2.840.10008.6.1.1278' => 'Racial Group',
    '1.2.840.10008.6.1.1279' => 'Relative Laterality',
    '1.2.840.10008.6.1.1280' => 'Brain Lesion Segmentation Type With Necrosis',
    '1.2.840.10008.6.1.1281' => 'Brain Lesion Segmentation Type Without Necrosis',
    '1.2.840.10008.6.1.1282' => 'Non-Acquisition Modality',
    '1.2.840.10008.6.1.1283' => 'Modality',
    '1.2.840.10008.6.1.1284' => 'Laterality Left-Right Only',
    '1.2.840.10008.6.1.1285' => 'Qualitative Evaluation Modifier Type',
    '1.2.840.10008.6.1.1286' => 'Qualitative Evaluation Modifier Value',
    '1.2.840.10008.6.1.1287' => 'Generic Anatomic Location Modifier',
    '1.2.840.10008.6.1.1288' => 'Beam Limiting Device Type',
    '1.2.840.10008.6.1.1289' => 'Compensator Device Type',
    '1.2.840.10008.6.1.1290' => 'Radiotherapy Treatment Machine Mode',
    '1.2.840.10008.6.1.1291' => 'Radiotherapy Distance Reference Location',
    '1.2.840.10008.6.1.1292' => 'Fixed Beam Limiting Device Type',
    '1.2.840.10008.6.1.1293' => 'Radiotherapy Wedge Type',
    '1.2.840.10008.6.1.1294' => 'RT Beam Limiting Device Orientation Label',
    '1.2.840.10008.6.1.1295' => 'General Accessory Device Type',
    '1.2.840.10008.6.1.1296' => 'Radiation Generation Mode Type',
    '1.2.840.10008.6.1.1297' => 'C-Arm Photon-Electron Delivery Rate Unit',
    '1.2.840.10008.6.1.1298' => 'Treatment Delivery Device Type',
    '1.2.840.10008.6.1.1299' => 'C-Arm Photon-Electron Dosimeter Unit',
    '1.2.840.10008.6.1.1300' => 'Treatment Point',
    '1.2.840.10008.6.1.1301' => 'Equipment Reference Point',
    '1.2.840.10008.6.1.1302' => 'Radiotherapy Treatment Planning Person Role',
    '1.2.840.10008.6.1.1303' => 'Real Time Video Rendition Title',
    '1.2.840.10008.6.1.1304' => 'Geometry Graphical Representation',
    '1.2.840.10008.6.1.1305' => 'Visual Explanation',
    '1.2.840.10008.6.1.1306' => 'Prostate Sector Anatomy from PI-RADS v2.1',
    '1.2.840.10008.6.1.1307' => 'Radiotherapy Robotic Node Set',
    '1.2.840.10008.6.1.1308' => 'Tomotherapeutic Dosimeter Unit',
    '1.2.840.10008.6.1.1309' => 'Tomotherapeutic Dose Rate Unit',
    '1.2.840.10008.6.1.1310' => 'Robotic Delivery Device Dosimeter Unit',
    '1.2.840.10008.6.1.1311' => 'Robotic Delivery Device Dose Rate Unit',
    '1.2.840.10008.6.1.1312' => 'Anatomic Structure',
    '1.2.840.10008.6.1.1313' => 'Mediastinum Finding or Feature',
    '1.2.840.10008.6.1.1314' => 'Mediastinum Anatomy',
    '1.2.840.10008.6.1.1315' => 'Vascular Ultrasound Report Document Title',
    '1.2.840.10008.6.1.1316' => 'Organ Part (Non-Lateralized)',
    '1.2.840.10008.6.1.1317' => 'Organ Part (Lateralized)',
    '1.2.840.10008.6.1.1318' => 'Treatment Termination Reason',
    '1.2.840.10008.6.1.1319' => 'Radiotherapy Treatment Delivery Person Role',
    '1.2.840.10008.6.1.1320' => 'Radiotherapy Interlock Resolution',
    '1.2.840.10008.6.1.1321' => 'Treatment Session Confirmation Assertion',
    '1.2.840.10008.6.1.1322' => 'Treatment Tolerance Violation Cause',
    '1.2.840.10008.6.1.1323' => 'Clinical Tolerance Violation Type',
    '1.2.840.10008.6.1.1324' => 'Machine Tolerance Violation Type',
    '1.2.840.10008.6.1.1325' => 'Radiotherapy Treatment Interlock',
    '1.2.840.10008.6.1.1326' => 'Isocentric Patient Support Position Parameter',
    '1.2.840.10008.6.1.1327' => 'RT Overridden Treatment Parameter',
    '1.2.840.10008.6.1.1328' => 'EEG Lead',
    '1.2.840.10008.6.1.1329' => 'Lead Location Near or in Muscle',
    '1.2.840.10008.6.1.1330' => 'Lead Location Near Peripheral Nerve',
    '1.2.840.10008.6.1.1331' => 'EOG Lead',
    '1.2.840.10008.6.1.1332' => 'Body Position Channel',
    '1.2.840.10008.6.1.1333' => 'EEG Annotation - Neurophysiologic Enumeration',
    '1.2.840.10008.6.1.1334' => 'EMG Annotation - Neurophysiological Enumeration',
    '1.2.840.10008.6.1.1335' => 'EOG Annotation - Neurophysiological Enumeration',
    '1.2.840.10008.6.1.1336' => 'Pattern Event',
    '1.2.840.10008.6.1.1337' => 'Device-related and Environment-related Event',
    '1.2.840.10008.6.1.1338' => 'EEG Annotation - Neurological Monitoring Measurement',
    '1.2.840.10008.6.1.1339' => 'OB-GYN Ultrasound Report Document Title',
    '1.2.840.10008.6.1.1340' => 'Automation of Measurement',
    '1.2.840.10008.6.1.1341' => 'OB-GYN Ultrasound Beam Path',
    '1.2.840.10008.6.1.1342' => 'Angle Measurement',
    '1.2.840.10008.6.1.1343' => 'Generic Purpose of Reference to Images and Coordinates in Measurement',
    '1.2.840.10008.6.1.1344' => 'Generic Purpose of Reference to Images in Measurement',
    '1.2.840.10008.6.1.1345' => 'Generic Purpose of Reference to Coordinates in Measurement',
    '1.2.840.10008.6.1.1346' => 'Fitzpatrick Skin Type',
    '1.2.840.10008.6.1.1347' => 'History of Malignant Melanoma',
    '1.2.840.10008.6.1.1348' => 'History of Melanoma in Situ',
    '1.2.840.10008.6.1.1349' => 'History of Non-Melanoma Skin Cancer',
    '1.2.840.10008.6.1.1350' => 'Skin Disorder',
    '1.2.840.10008.6.1.1351' => 'Patient Reported Lesion Characteristic',
    '1.2.840.10008.6.1.1352' => 'Lesion Palpation Finding',
    '1.2.840.10008.6.1.1353' => 'Lesion Visual Finding',
    '1.2.840.10008.6.1.1354' => 'Skin Procedure',
    '1.2.840.10008.6.1.1355' => 'Abdominopelvic Vessel',
    '1.2.840.10008.6.1.1356' => 'Numeric Value Failure Qualifier',
    '1.2.840.10008.6.1.1357' => 'Numeric Value Unknown Qualifier',
    '1.2.840.10008.6.1.1358' => 'Couinaud Liver Segment',
    '1.2.840.10008.6.1.1359' => 'Liver Segmentation Type',
    '1.2.840.10008.6.1.1360' => 'Contraindications For XA Imaging',
    '1.2.840.10008.6.1.1361' => 'Neurophysiologic Stimulation Mode',
    '1.2.840.10008.6.1.1362' => 'Reported Value Type',
    '1.2.840.10008.6.1.1363' => 'Value Timing',
    '1.2.840.10008.6.1.1364' => 'RDSR Frame of Reference Origin',
    '1.2.840.10008.6.1.1365' => 'Microscopy Annotation Property Type',
    '1.2.840.10008.6.1.1366' => 'Microscopy Measurement Type',
    '1.2.840.10008.6.1.1367' => 'Prostate Reporting System',
    '1.2.840.10008.6.1.1368' => 'MR Signal Intensity',
    '1.2.840.10008.6.1.1369' => 'Cross-sectional Scan Plane Orientation',
    '1.2.840.10008.6.1.1370' => 'History of Prostate Disease',
    '1.2.840.10008.6.1.1371' => 'Prostate MRI Study Quality Finding',
    '1.2.840.10008.6.1.1372' => 'Prostate MRI Series Quality Finding',
    '1.2.840.10008.6.1.1373' => 'MR Imaging Artifact',
    '1.2.840.10008.6.1.1374' => 'Prostate DCE MRI Quality Finding',
    '1.2.840.10008.6.1.1375' => 'Prostate DWI MRI Quality Finding',
    '1.2.840.10008.6.1.1376' => 'Abdominal Intervention Type',
    '1.2.840.10008.6.1.1377' => 'Abdominopelvic Intervention',
    '1.2.840.10008.6.1.1378' => 'Prostate Cancer Diagnostic Procedure',
    '1.2.840.10008.6.1.1379' => 'Prostate Cancer Family History',
    '1.2.840.10008.6.1.1380' => 'Prostate Cancer Therapy',
    '1.2.840.10008.6.1.1381' => 'Prostate MRI Assessment',
    '1.2.840.10008.6.1.1382' => 'Overall Assessment from PI-RADS',
    '1.2.840.10008.6.1.1383' => 'Image Quality Control Standard',
    '1.2.840.10008.6.1.1384' => 'Prostate Imaging Indication',
    '1.2.840.10008.6.1.1385' => 'PI-RADS v2 Lesion Assessment Category',
    '1.2.840.10008.6.1.1386' => 'PI-RADS v2 T2WI PZ Lesion Assessment Category',
    '1.2.840.10008.6.1.1387' => 'PI-RADS v2 T2WI TZ Lesion Assessment Category',
    '1.2.840.10008.6.1.1388' => 'PI-RADS v2 DWI Lesion Assessment Category',
    '1.2.840.10008.6.1.1389' => 'PI-RADS v2 DCE Lesion Assessment Category',
    '1.2.840.10008.6.1.1390' => 'mpMRI Assessment Type',
    '1.2.840.10008.6.1.1391' => 'mpMRI Assessment Type from PI-RADS',
    '1.2.840.10008.6.1.1392' => 'mpMRI Assessment Value',
    '1.2.840.10008.6.1.1393' => 'MRI Abnormality',
    '1.2.840.10008.6.1.1394' => 'mpMRI Prostate Abnormality from PI-RADS',
    '1.2.840.10008.6.1.1395' => 'mpMRI Benign Prostate Abnormality from PI-RADS',
    '1.2.840.10008.6.1.1396' => 'MRI Shape Characteristic',
    '1.2.840.10008.6.1.1397' => 'Prostate MRI Shape Characteristic from PI-RADS',
    '1.2.840.10008.6.1.1398' => 'MRI Margin Characteristic',
    '1.2.840.10008.6.1.1399' => 'Prostate MRI Margin Characteristic from PI-RADS',
    '1.2.840.10008.6.1.1400' => 'MRI Signal Characteristic',
    '1.2.840.10008.6.1.1401' => 'Prostate MRI Signal Characteristic from PI-RADS',
    '1.2.840.10008.6.1.1402' => 'MRI Enhancement Pattern',
    '1.2.840.10008.6.1.1403' => 'Prostate MRI Enhancement Pattern from PI-RADS',
    '1.2.840.10008.6.1.1404' => 'Prostate MRI Extra-prostatic Finding',
    '1.2.840.10008.6.1.1405' => 'Prostate MRI Assessment of Extra-prostatic Anatomic Site',
    '1.2.840.10008.6.1.1406' => 'MR Coil Type',
    '1.2.840.10008.6.1.1407' => 'Endorectal Coil Fill Substance',
    '1.2.840.10008.6.1.1408' => 'Prostate Relational Measurement',
    '1.2.840.10008.6.1.1409' => 'Prostate Cancer Diagnostic Blood Lab Measurement',
    '1.2.840.10008.6.1.1410' => 'Prostate Imaging Types of Quality Control Standard',
    '1.2.840.10008.6.1.1411' => 'Ultrasound Shear Wave Measurement',
    '1.2.840.10008.6.1.1412' => 'CID 3780',
    '1.2.840.10008.6.1.1413' => 'Left Ventricle Myocardial Wall 18 Segment Model',
    '1.2.840.10008.6.1.1414' => 'Left Ventricle Basal Wall 6 Segments',
    '1.2.840.10008.6.1.1415' => 'Left Ventricle Midlevel Wall 6 Segments',
    '1.2.840.10008.6.1.1416' => 'Left Ventricle Apical Wall 4 Segments',
    '1.2.840.10008.6.1.1417' => 'Left Ventricle Apical Wall 6 Segments',
    '1.2.840.10008.6.1.1418' => 'Patient Treatment Preparation Method',
    '1.2.840.10008.6.1.1419' => 'Patient Shielding Device',
    '1.2.840.10008.6.1.1420' => 'Patient Treatment Preparation Device',
    '1.2.840.10008.6.1.1421' => 'Patient Position Displacement Reference Point',
    '1.2.840.10008.6.1.1422' => 'Patient Alignment Device',
    '1.2.840.10008.6.1.1423' => 'Reasons for RT Radiation Treatment Omission',
    '1.2.840.10008.6.1.1424' => 'Patient Treatment Preparation Procedure',
    '1.2.840.10008.6.1.1425' => 'Motion Management Setup Device',
    '1.2.840.10008.6.1.1426' => 'Core Echo Strain Measurement',
    '1.2.840.10008.6.1.1427' => 'Myocardial Strain Method',
    '1.2.840.10008.6.1.1428' => 'Echo Measured Strain Property',
    '1.2.840.10008.6.1.1429' => 'Assessment from CAD-RADS',
    '1.2.840.10008.6.1.1430' => 'CAD-RADS Stenosis Assessment Modifier',
    '1.2.840.10008.6.1.1431' => 'CAD-RADS Assessment Modifier',
    '1.2.840.10008.6.1.1432' => 'RT Segment Material',
    '1.2.840.10008.6.1.1433' => 'Vertebral Anatomic Structure',
    '1.2.840.10008.6.1.1434' => 'Vertebra',
    '1.2.840.10008.6.1.1435' => 'Intervertebral Disc',
    '1.2.840.10008.6.1.1436' => 'Imaging Procedure',
    '1.2.840.10008.6.1.1437' => 'NICIP Short Code Imaging Procedure',
    '1.2.840.10008.6.1.1438' => 'NICIP SNOMED Imaging Procedure',
    '1.2.840.10008.6.1.1439' => 'ICD-10-PCS Imaging Procedure',
    '1.2.840.10008.6.1.1440' => 'ICD-10-PCS Nuclear Medicine Procedure',
    '1.2.840.10008.6.1.1441' => 'ICD-10-PCS Radiation Therapy Procedure',
    '1.2.840.10008.6.1.1442' => 'RT Segmentation Property Category',
    '1.2.840.10008.6.1.1443' => 'Radiotherapy Registration Mark',
    '1.2.840.10008.6.1.1444' => 'Radiotherapy Dose Region',
    '1.2.840.10008.6.1.1445' => 'Anatomically Localized Lesion Segmentation Type',
    '1.2.840.10008.6.1.1446' => 'Reason for Removal from Operational Use',
    '1.2.840.10008.6.1.1447' => 'General Ultrasound Report Document Title',
    '1.2.840.10008.6.1.1448' => 'Elastography Site',
    '1.2.840.10008.6.1.1449' => 'Elastography Measurement Site',
    '1.2.840.10008.6.1.1450' => 'Ultrasound Relevant Patient Condition',
    '1.2.840.10008.6.1.1451' => 'Shear Wave Detection Method',
    '1.2.840.10008.6.1.1452' => 'Liver Ultrasound Study Indication',
    '1.2.840.10008.6.1.1453' => 'Analog Waveform Filter',
    '1.2.840.10008.6.1.1454' => 'Digital Waveform Filter',
    '1.2.840.10008.6.1.1455' => 'Waveform Filter Lookup Table Input Frequency Unit',
    '1.2.840.10008.6.1.1456' => 'Waveform Filter Lookup Table Output Magnitude Unit',
    '1.2.840.10008.6.1.1457' => 'Specific Observation Subject Class',
    '1.2.840.10008.6.1.1458' => 'Movable Beam Limiting Device Type',
    '1.2.840.10008.6.1.1459' => 'Radiotherapy Acquisition WorkItem Subtasks',
    '1.2.840.10008.6.1.1460' => 'Patient Position Acquisition Radiation Source Locations',
    '1.2.840.10008.6.1.1461' => 'Energy Derivation Types',
    '1.2.840.10008.6.1.1462' => 'KV Imaging Acquisition Techniques',
    '1.2.840.10008.6.1.1463' => 'MV Imaging Acquisition Techniques',
    '1.2.840.10008.6.1.1464' => 'Patient Position Acquisition - Projection Techniques',
    '1.2.840.10008.6.1.1465' => 'Patient Position Acquisition - CT Techniques',
    '1.2.840.10008.6.1.1466' => 'Patient Positioning Related Object Purposes',
    '1.2.840.10008.6.1.1467' => 'Patient Position Acquisition Devices',
    '1.2.840.10008.6.1.1468' => 'RT Radiation Meterset Units',
    '1.2.840.10008.6.1.1469' => 'Acquisition Initiation Types',
    '1.2.840.10008.6.1.1470' => 'RT Image Patient Position Acquisition Devices',
    '1.2.840.10008.6.1.1471' => 'Photoacoustic Illumination Method',
    '1.2.840.10008.6.1.1472' => 'Acoustic Coupling Medium',
    '1.2.840.10008.6.1.1473' => 'Ultrasound Transducer Technology',
    '1.2.840.10008.6.1.1474' => 'Speed of Sound Correction Mechanisms',
    '1.2.840.10008.6.1.1475' => 'Photoacoustic Reconstruction Algorithm Family',
    '1.2.840.10008.6.1.1476' => 'Photoacoustic Imaged Property',
    '1.2.840.10008.6.1.1477' => 'X-Ray Radiation Dose Procedure Type Reported',
    '1.2.840.10008.6.1.1478' => 'Topical Treatment',
    '1.2.840.10008.6.1.1479' => 'Lesion Color',
    '1.2.840.10008.6.1.1480' => 'Specimen Stain for Confocal Microscopy',
    '1.2.840.10008.6.1.1481' => 'RT ROI Image Acquisition Context',
    '1.2.840.10008.6.1.1482' => 'Lobe of Lung',
    '1.2.840.10008.6.1.1483' => 'Zone of Lung',
    '1.2.840.10008.6.1.1484' => 'Sleep Stage',
    '1.2.840.10008.6.1.1485' => 'Patient Position Acquisition - MR Techniques',
    '1.2.840.10008.6.1.1486' => 'RT Plan Radiotherapy Procedure Technique',
    '1.2.840.10008.6.1.1487' => 'Waveform Annotation Classification',
    '1.2.840.10008.6.1.1488' => 'Waveform Annotations Document Title',
    '1.2.840.10008.6.1.1489' => 'EEG Procedure',
    '1.2.840.10008.6.1.1490' => 'Patient Consciousness',
    '1.2.840.10008.6.1.1491' => 'Follicle Type',
    '1.2.840.10008.6.1.1492' => 'Breast Tissue Segmentation Type',
    '1.2.840.10008.6.1.1493' => 'Implanted Device',
    '1.2.840.10008.6.1.1494' => 'Similarity Measure',
    '1.2.840.10008.6.1.1495' => 'Waveform Acquisition Modality',
    '1.2.840.10008.6.1.1496' => 'En Face Processing Algorithm Family',
    '1.2.840.10008.6.1.1497' => 'Anterior Eye Segmentation Surface',
    '1.2.840.10008.6.1.1498' => 'Fetal Echocardiography Image View',
    '1.2.840.10008.6.1.1499' => 'Cardiac Ultrasound Fetal Arrhythmia Measurements',
    '1.2.840.10008.6.1.1500' => 'Common Fetal Echocardiography Measurements',
    '1.2.840.10008.6.1.1501' => 'Head and Neck Primary Anatomic Structure',
    '1.2.840.10008.6.1.1502' => 'VL View',
    '1.2.840.10008.6.1.1503' => 'VL Dental View',
    '1.2.840.10008.6.1.1504' => 'VL View Modifier',
    '1.2.840.10008.6.1.1505' => 'VL Dental View Modifier',
    '1.2.840.10008.6.1.1506' => 'Orthognathic Functional Condition',
    '1.2.840.10008.6.1.1507' => 'Orthodontic Finding by Inspection',
    '1.2.840.10008.6.1.1508' => 'Orthodontic Observable Entity',
    '1.2.840.10008.6.1.1509' => 'Dental Occlusion',
    '1.2.840.10008.6.1.1510' => 'Orthodontic Treatment Progress',
    '1.2.840.10008.6.1.1511' => 'General Photography Device',
    '1.2.840.10008.6.1.1512' => 'Devices for the Purpose of Dental Photography',
    '1.2.840.10008.6.1.1513' => 'CTDI Phantom Device',
    '1.2.840.10008.6.1.1514' => 'Diagnostic Imaging Procedure without IV Contrast',
    '1.2.840.10008.6.1.1515' => 'Diagnostic Imaging Procedure with IV Contrast',
    '1.2.840.10008.6.1.1516' => 'Structural Heart Procedure',
    '1.2.840.10008.6.1.1517' => 'Structural Heart Device',
    '1.2.840.10008.6.1.1518' => 'Structural Heart Measurement',
    '1.2.840.10008.6.1.1519' => 'Aortic Valve Structural Measurement',
    '1.2.840.10008.6.1.1520' => 'Mitral Valve Structural Measurement',
    '1.2.840.10008.6.1.1521' => 'Tricuspid Valve Structural Measurement',
    '1.2.840.10008.6.1.1522' => 'Structural Heart Echo Measurement',
    '1.2.840.10008.6.1.1523' => 'Left Atrial Appendage Closure Measurement',
    '1.2.840.10008.6.1.1524' => 'Structural Heart Procedure Anatomic Site',
    '1.2.840.10008.6.1.1525' => 'Indication for Structural Heart Procedure',
    '1.2.840.10008.6.1.1526' => 'Bradycardiac Agent',
    '1.2.840.10008.6.1.1527' => 'Transesophageal Echocardiography Scan Plane',
    '1.2.840.10008.6.1.1528' => 'Structural Heart Measurement Report Document Title',
    '1.2.840.10008.6.1.1529' => 'Person Gender Identity',
    '1.2.840.10008.6.1.1530' => 'Category of Sex Parameters for Clinical Use',
    '1.2.840.10008.6.1.1531' => 'Third Person Pronoun Set',
    '1.2.840.10008.6.1.1532' => 'Cardiac Structure Calcification Qualitative Evaluation',
    '1.2.840.10008.6.1.1533' => 'Visual Field Measurements',
    '1.2.840.10008.6.1.1534' => 'Optic Disc Key Measurements',
    '1.2.840.10008.6.1.1535' => 'Retinal Sector Methods',
    '1.2.840.10008.6.1.1536' => 'RNFL Sector Measurements',
    '1.2.840.10008.6.1.1537' => 'RNFL Clockface Measurements',
    '1.2.840.10008.6.1.1538' => 'Macular Thickness Key Measurements',
    '1.2.840.10008.6.1.1539' => 'Ganglion Cell Measurement Extent',
    '1.2.840.10008.6.1.1540' => 'Ganglion Cell Key Measurements',
    '1.2.840.10008.6.1.1541' => 'Ganglion Cell Sector Measurements',
    '1.2.840.10008.6.1.1542' => 'Ganglion Cell Sector Methods',
    '1.2.840.10008.6.1.1543' => 'Endothelial Cell Count Measurements',
    '1.2.840.10008.6.1.1544' => 'Ophthalmic Image ROI Measurements',
    '1.2.840.10008.6.1.1545' => 'RT Plan Approval Assertion',
    '1.2.840.10008.6.1.1546' => 'Estimated Delivery Date Method',
    '1.2.840.10008.6.1.1547' => 'RT Dose Calculation Algorithm Family',
    '1.2.840.10008.6.1.1548' => 'Dose Index for Dose Calibration',
    '1.2.840.10008.6.1.1549' => 'Ultrasound Attenuation Imaging Site',
    '1.2.840.10008.6.1.1550' => 'Fetal Anatomy Survey Assessment',
    '1.2.840.10008.6.1.1551' => 'Fetal Anatomy Survey Assessment - Head',
    '1.2.840.10008.6.1.1552' => 'Fetal Anatomy Survey Assessment - Face and Neck',
    '1.2.840.10008.6.1.1553' => 'Fetal Anatomy Survey Assessment - Chest',
    '1.2.840.10008.6.1.1554' => 'Fetal Anatomy Survey Assessment - Heart',
    '1.2.840.10008.6.1.1555' => 'Fetal Anatomy Survey Assessment - Abdomen and Pelvis',
    '1.2.840.10008.6.1.1556' => 'Fetal Anatomy Survey Assessment - Spine',
    '1.2.840.10008.6.1.1557' => 'Fetal Anatomy Survey Assessment - Extremities',
    '1.2.840.10008.6.1.1558' => 'Fetal Anatomy Survey Assessment - Maternal',
    '1.2.840.10008.6.1.1559' => 'Fetal Anatomy Survey Practice Guideline',
    '1.2.840.10008.6.1.1560' => 'Sensitive Content Category',
    '1.2.840.10008.6.1.1561' => 'Sensitive Content Detail',
    '1.2.840.10008.6.1.1562' => 'Application Type Code',
    '1.2.840.10008.6.1.1563' => 'X-Ray Modulation Type',
    '1.2.840.10008.6.1.1564' => 'Radiotherapy Dose Real World Units',
    '1.2.840.10008.6.1.1565' => 'Radiotherapy Dose Interpreted Type Codes',
    '1.2.840.10008.6.1.1566' => 'Radiotherapy Dose Interpreted Type Modifier Codes',
    '1.2.840.10008.6.1.1567' => 'Radiotherapy Dose Intent Codes',
    '1.2.840.10008.6.1.1568' => 'Quality Segmentation Property Type',
    '1.2.840.10008.6.1.1569' => 'Ultrasound Z-Score Population Index',
    '1.2.840.10008.6.1.1570' => 'Fetal Ultrasound Z-Score Reference Authority',
    '1.2.840.10008.6.1.1571' => 'Metal Artifact Reduction Algorithm Family',
    '1.2.840.10008.7.1.1' => 'Native DICOM Model',
    '1.2.840.10008.7.1.2' => 'Abstract Multi-Dimensional Image Model',
    '1.2.840.10008.8.1.1' => 'DICOM Content Mapping Resource',
    '1.2.840.10008.9.1' => 'Imaging Report',
    '1.2.840.10008.9.2' => 'Clinical Information',
    '1.2.840.10008.9.3' => 'Imaging Procedure Description',
    '1.2.840.10008.9.4' => 'Comparison Study',
    '1.2.840.10008.9.5' => 'Impression',
    '1.2.840.10008.9.6' => 'Addendum',
    '1.2.840.10008.9.7' => 'Request',
    '1.2.840.10008.9.8' => 'Radiation Exposure and Protection Information',
    '1.2.840.10008.9.9' => 'Fetus Findings',
    '1.2.840.10008.9.10' => 'Labeled Subsection',
    '1.2.840.10008.9.11' => 'Communication of Actionable Findings',
    '1.2.840.10008.9.12' => 'Recommendation',
    '1.2.840.10008.9.13' => 'Procedural Medication',
    '1.2.840.10008.9.14' => 'Procedure Technique',
    '1.2.840.10008.9.15' => 'Image Quality',
    '1.2.840.10008.9.16' => 'Study Act',
    '1.2.840.10008.9.17' => 'Series Act',
    '1.2.840.10008.9.18' => 'SOP Instance Observation',
    '1.2.840.10008.9.19' => 'Section Text',
    '1.2.840.10008.9.20' => 'General Header',
    '1.2.840.10008.9.21' => 'Imaging Header',
    '1.2.840.10008.9.22' => 'Parent Document',
    '1.2.840.10008.9.23' => 'General Section Entries',
    '1.2.840.10008.9.24' => 'Imaging Addendum Report',
    '1.2.840.10008.10.1' => 'Video Endoscopic Image Real-Time Communication',
    '1.2.840.10008.10.2' => 'Video Photographic Image Real-Time Communication',
    '1.2.840.10008.10.3' => 'Audio Waveform Real-Time Communication',
    '1.2.840.10008.10.4' => 'Rendition Selection Document Real-Time Communication',
    '1.2.840.10008.15.0.3.1' => 'dicomDeviceName',
    '1.2.840.10008.15.0.3.2' => 'dicomDescription',
    '1.2.840.10008.15.0.3.3' => 'dicomManufacturer',
    '1.2.840.10008.15.0.3.4' => 'dicomManufacturerModelName',
    '1.2.840.10008.15.0.3.5' => 'dicomSoftwareVersion',
    '1.2.840.10008.15.0.3.6' => 'dicomVendorData',
    '1.2.840.10008.15.0.3.7' => 'dicomAETitle',
    '1.2.840.10008.15.0.3.8' => 'dicomNetworkConnectionReference',
    '1.2.840.10008.15.0.3.9' => 'dicomApplicationCluster',
    '1.2.840.10008.15.0.3.10' => 'dicomAssociationInitiator',
    '1.2.840.10008.15.0.3.11' => 'dicomAssociationAcceptor',
    '1.2.840.10008.15.0.3.12' => 'dicomHostname',
    '1.2.840.10008.15.0.3.13' => 'dicomPort',
    '1.2.840.10008.15.0.3.14' => 'dicomSOPClass',
    '1.2.840.10008.15.0.3.15' => 'dicomTransferRole',
    '1.2.840.10008.15.0.3.16' => 'dicomTransferSyntax',
    '1.2.840.10008.15.0.3.17' => 'dicomPrimaryDeviceType',
    '1.2.840.10008.15.0.3.18' => 'dicomRelatedDeviceReference',
    '1.2.840.10008.15.0.3.19' => 'dicomPreferredCalledAETitle',
    '1.2.840.10008.15.0.3.20' => 'dicomTLSCyphersuite',
    '1.2.840.10008.15.0.3.21' => 'dicomAuthorizedNodeCertificateReference',
    '1.2.840.10008.15.0.3.22' => 'dicomThisNodeCertificateReference',
    '1.2.840.10008.15.0.3.23' => 'dicomInstalled',
    '1.2.840.10008.15.0.3.24' => 'dicomStationName',
    '1.2.840.10008.15.0.3.25' => 'dicomDeviceSerialNumber',
    '1.2.840.10008.15.0.3.26' => 'dicomInstitutionName',
    '1.2.840.10008.15.0.3.27' => 'dicomInstitutionAddress',
    '1.2.840.10008.15.0.3.28' => 'dicomInstitutionDepartmentName',
    '1.2.840.10008.15.0.3.29' => 'dicomIssuerOfPatientID',
    '1.2.840.10008.15.0.3.30' => 'dicomPreferredCallingAETitle',
    '1.2.840.10008.15.0.3.31' => 'dicomSupportedCharacterSet',
    '1.2.840.10008.15.0.4.1' => 'dicomConfigurationRoot',
    '1.2.840.10008.15.0.4.2' => 'dicomDevicesRoot',
    '1.2.840.10008.15.0.4.3' => 'dicomUniqueAETitlesRegistryRoot',
    '1.2.840.10008.15.0.4.4' => 'dicomDevice',
    '1.2.840.10008.15.0.4.5' => 'dicomNetworkAE',
    '1.2.840.10008.15.0.4.6' => 'dicomNetworkConnection',
    '1.2.840.10008.15.0.4.7' => 'dicomUniqueAETitle',
    '1.2.840.10008.15.0.4.8' => 'dicomTransferCapability',
);

#------------------------------------------------------------------------------
# Extract information from a DICOM (DCM) image
# Inputs: 0) ExifTool object reference, 1) DirInfo reference
# Returns: 1 on success, 0 if this wasn't a valid DICOM file
sub ProcessDICOM($$)
{
    my ($et, $dirInfo) = @_;
    my $raf = $$dirInfo{RAF};
    my $unknown = $et->Options('Unknown');
    my $verbose = $et->Options('Verbose');
    my ($hdr, $buff, $implicit, $vr, $len);
#
# identify the DICOM or ACR-NEMA file
#
    $raf->Read($hdr, 12) == 12 or return 0; # save for ACR identification later
    $raf->Seek(128, 0) or return 0;         # skip to end of DICM header
    $raf->Read($buff, 4) == 4 or return 0;  # read signature
    if ($buff eq 'DICM') {
        # file meta information transfer syntax is explicit little endian
        SetByteOrder('II');
        $et->SetFileType('DICOM');
    } else {
        # test for a RAW DCM image (ACR-NEMA format, ie. no header)
        foreach ('II','MM','') {
            return 0 unless $_; # no luck identifying the syntax
            SetByteOrder($_);
            my $g = Get16u(\$hdr, 0);
            # expect group number to be small and even
            next if $g < 2 or $g > 8 or $g & 0x01;
            my $e = Get16u(\$hdr, 2);
            next if $e > 0x20;          # expect a low element number at start
            $vr = substr($hdr, 4, 2);   # look for explicit VR
            if ($vr =~ /^[A-Z]{2}$/) {
                $implicit = 0;
                if ($vr32{$vr}) {
                    next unless Get16u(\$hdr, 6) == 0;  # must be 2 zero bytes
                    $len = Get32u(\$hdr, 8);
                } else {
                    next if $e == 0 and $vr ne 'UL';    # group length must be UL
                    $len = Get16u(\$hdr, 6);
                }
            } else {
                $implicit = 1;
                $len = Get32u(\$hdr, 4);
            }
            next if $e == 0 and $len != 4;  # group length value must be 4 bytes
            next if $len > 64;      # first element shouldn't be too long
            last;   # success!
        }
        $raf->Seek(0, 0) or return 0;   # rewind to start of file
        $et->SetFileType('ACR');
    }
#
# process the meta information
#
    my $tagTablePtr = GetTagTable('Image::ExifTool::DICOM::Main');
    my $pos = $raf->Tell();
    my $err = 1;
    my ($transferSyntax, $group2end);
    for (;;) {
        $raf->Read($buff, 8) == 8 or $err = 0, last;
        $pos += 8;
        my $group = Get16u(\$buff, 0);
        # implement the transfer syntax at the end of the group 2 data
        if ($transferSyntax and ($group != 0x0002 or
            ($group2end and $pos > $group2end)))
        {
            # 1.2.840.10008.1.2   = implicit VR little endian
            # 1.2.840.10008.1.2.2 = explicit VR big endian
            # 1.2.840.10008.1.2.x = explicit VR little endian
            # 1.2.840.10008.1.2.1.99 = deflated
            unless ($transferSyntax =~ /^1\.2\.840\.10008\.1\.2(\.\d+)?(\.\d+)?/) {
                $et->Warn("Unrecognized transfer syntax $transferSyntax");
                last;
            }
            if (not $1) {
                $implicit = 1;
            } elsif ($1 eq '.2') {
                SetByteOrder('MM');
                $group = Get16u(\$buff, 0); # must get group again
            } elsif ($1 eq '.1' and $2 and $2 eq '.99') {
                # inflate compressed data stream
                if (eval { require Compress::Zlib }) {
                    # must use undocumented zlib feature to disable zlib header information
                    # because DICOM deflated data doesn't have the zlib header (ref 3)
                    my $wbits = -Compress::Zlib::MAX_WBITS();
                    my $inflate = Compress::Zlib::inflateInit(-WindowBits => $wbits);
                    if ($inflate) {
                        $raf->Seek(-8, 1) or last;
                        my $data = '';
                        while ($raf->Read($buff, 65536)) {
                            my ($buf, $stat) = $inflate->inflate($buff);
                            if ($stat == Compress::Zlib::Z_OK() or
                                $stat == Compress::Zlib::Z_STREAM_END())
                            {
                                $data .= $buf;
                                last if $stat == Compress::Zlib::Z_STREAM_END();
                            } else {
                                $et->Warn('Error inflating compressed data stream');
                                return 1;
                            }
                        }
                        last if length $data < 8;
                        # create new RAF object from inflated data stream
                        $raf = File::RandomAccess->new(\$data);
                        # re-read start of stream (now decompressed)
                        $raf->Read($buff, 8) == 8 or last;
                        $group = Get16u(\$buff, 0);
                    } else {
                        $et->Warn('Error initializing inflation');
                        return 1;
                    }
                } else {
                    $et->Warn('Install Compress::Zlib to decode compressed data stream');
                    return 1;
                }
            }
            undef $transferSyntax;
        }
        my $element = Get16u(\$buff,2);
        my $tag = sprintf('%.4X,%.4X', $group, $element);

        if ($implicit or $implicitVR{$tag}) {
            # treat everything as string if implicit VR because it
            # isn't worth it to generate the necessary VR lookup tables
            # for the thousands of defined data elements
            $vr = '';       # no VR (treat everything as string)
            $len = Get32u(\$buff, 4);
        } else {
            $vr = substr($buff,4,2);
            last unless $vr =~ /^[A-Z]{2}$/;
            if ($vr32{$vr}) {
                $raf->Read($buff, 4) == 4 or last;
                $pos += 4;
                $len = Get32u(\$buff, 0);
                $len = 0 if $vr eq 'SQ';    # just recurse into sequences
            } else {
                $len = Get16u(\$buff, 6);
            }
        }
        if ($len == 0xffffffff) {
            $len = 0;   # don't read value if undefined length
            if ($verbose) {
                # start list of items in verbose output
                $et->VPrint(0, "$$et{INDENT}+ [List of items]\n");
                $$et{INDENT} .= '| ';
            }
        }
        # read the element value
        if ($len) {
            $raf->Read($buff, $len) == $len or last;
            $pos += $len;
        } else {
            $buff = '';
        }

        # handle tags not found in the table
        my $tagInfo = $$tagTablePtr{$tag};
        unless ($tagInfo) {
            # accept tag ID's with "x" for a wildcard in the following patterns:
            # '60xx,1203', '0020,31xx', '0028,04x2', '1000,xxx0', '1010,xxxx'
            my $xx;
            if ((($xx = $tag) =~ s/^(..)../$1xx/  and $$tagTablePtr{$xx}) or
                (($xx = $tag) =~ s/..$/xx/        and $$tagTablePtr{$xx}) or
                (($xx = $tag) =~ s/.(.)$/x$1/     and $$tagTablePtr{$xx}) or
                (($xx = $tag) =~ s/...(.)$/xxx$1/ and $$tagTablePtr{$xx}) or
                (($xx = $tag) =~ s/....$/xxxx/    and $$tagTablePtr{$xx}))
            {
                $tag = $xx;
                $tagInfo = $$tagTablePtr{$xx};
            } elsif ($unknown) {
                # create tag info hash for unknown elements
                if ($element == 0) {    # element zero is group length
                    $tagInfo = {
                        Name => sprintf("Group%.4X_Length", $group),
                        Description => sprintf("Group %.4X Length", $group),
                    };
                } else {
                    $tagInfo = {
                        Name => sprintf("DICOM_%.4X_%.4X", $group, $element),
                        Description => sprintf("DICOM %.4X,%.4X", $group, $element),
                    };
                }
                $$tagInfo{Unknown} = 1;
                AddTagToTable($tagTablePtr, $tag, $tagInfo);
            }
        }
        # get VR from our tag information if implicit
        $vr = $$tagInfo{VR} || '  ' if $tagInfo and not $vr;

        if ($element == 0) {
            $vr = 'UL'; # group length element is always unsigned long
        }
        my $val;
        my $format = $dicomFormat{$vr};
        # remove trailing space used to pad to an even number of characters
        $buff =~ s/ $// unless $format or length($buff) & 0x01;
        if ($len > 1024) {
            # treat large data elements as binary data
            my $binData;
            my $lcTag = $tagInfo ? lc($$tagInfo{Name}) : 'unknown';
            if ($$et{REQ_TAG_LOOKUP}{$lcTag} or
                ($$et{OPTIONS}{Binary} and not $$et{EXCL_TAG_LOOKUP}{$lcTag}))
            {
                $binData = $buff;   # must make a copy
            } else {
                $binData = "Binary data $len bytes";
            }
            $val = \$binData;
        } elsif ($format) {
            $val = ReadValue(\$buff, 0, $format, undef, $len);
        } else {
            $val = $buff;
            $format = 'string';
            if ($vr eq 'DA') {
                # format date values
                $val =~ s/^ *(\d{4})(\d{2})(\d{2})/$1:$2:$3/;
            } elsif ($vr eq 'TM') {
                # format time values
                $val =~ s/^ *(\d{2})(\d{2})(\d{2}[^ ]*)/$1:$2:$3/;
            } elsif ($vr eq 'DT') {
                # format date/time values
                $val =~ s/^ *(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2}[^ ]*)/$1:$2:$3 $4:$5:$6/;
            } elsif ($vr eq 'AT' and $len == 4) {
                # convert attribute tag ID to hex format
                my ($g, $e) = (Get16u(\$buff,0), Get16u(\$buff,2));
                $val = sprintf('%.4X,%.4X', $g, $e);
            } elsif ($vr eq 'UI') {
                # add PrintConv to translate registered UID's
                $val =~ s/\0.*//s; # truncate at null
                if ($tagInfo) {
                    $$tagInfo{PrintConv} = \%uid if $uid{$val};
                    $$tagInfo{Hidden} = '' unless defined $$tagInfo{Hidden}; # (don't add PrintConv to -listx output)
                }
            } elsif ($vr =~ /^(AE|CS|DS|IS|LO|PN|SH)$/) {
                $val =~ s/ +$//;    # leading/trailing spaces not significant
                $val =~ s/^ +//;
            } elsif ($vr =~ /^(LT|ST|UT)$/) {
                $val =~ s/ +$//;    # trailing spaces not significant
            }
        }
        # save the group 2 end position and transfer syntax
        if ($group == 0x0002) {
            $element == 0x0000 and $group2end = $pos + $val;
            $element == 0x0010 and $transferSyntax = $val;
        }

        # handle the new tag information
        $et->HandleTag($tagTablePtr, $tag, $val,
            DataPt => \$buff,
            DataPos => $pos - $len,
            Format => $format,
            Start => 0,
            Size => $len,
            Extra => " ($vr)",
        );

        # stop indenting for list if we reached EndOfItems tag
        $$et{INDENT} =~ s/..$// if $verbose and $tag eq 'FFFE,E00D';
    }
    $err and $et->Warn('Error reading DICOM file (corrupted?)');
    return 1;
}

1;  # end

__END__

=head1 NAME

Image::ExifTool::DICOM - Read DICOM and ACR-NEMA medical images

=head1 SYNOPSIS

This module is used by Image::ExifTool

=head1 DESCRIPTION

This module contains routines required by Image::ExifTool to extract meta
information from DICOM (Digital Imaging and Communications in Medicine) DCM
and ACR-NEMA (American College of Radiology - National Electrical
Manufacturer's Association) ACR medical images.

=head1 NOTES

Images compressed using the DICOM deflated transfer syntax will be decoded
if Compress::Zlib is installed.

No translation of special characters sets is done.

=head1 AUTHOR

Copyright 2003-2026, Phil Harvey (philharvey66 at gmail.com)

This library is free software; you can redistribute it and/or modify it
under the same terms as Perl itself.

=head1 REFERENCES

=over 4

=item L<http://medical.nema.org/>

=item L<http://www.sph.sc.edu/comd/rorden/dicom.html>

=item L<http://www.dclunie.com/>

=item L<http://www.gehealthcare.com/usen/interoperability/dicom/docs/2258357r3.pdf>

=back

=head1 SEE ALSO

L<Image::ExifTool::TagNames/DICOM Tags>,
L<Image::ExifTool(3pm)|Image::ExifTool>

=cut

